"""
Consistency check of batch configs.

Felix Metzner 2020
"""

import os
import yaml
import rdstar
import collections
import analysistools.batch_tools.job_handler as handler

from textops import find
from typing import Optional, List, Dict, Any

from rdstar.utilities import PathType

__all__ = [
    "config_check",
]

pkg_dir = os.path.dirname(rdstar.__file__)  # type: PathType
configs_dir_path = os.path.join(pkg_dir, "configs")  # type: PathType

accepted_none_confs = [
    "batch_config.json",
    "test_batch_config.json",
]  # type: List[str]


def config_check(
    production_tag: str,
    configs_sub_dir_name: Optional[str] = None,
) -> None:
    print(f"Checking configs with production tag {production_tag}")

    error_msgs = []  # type: List[str]

    if configs_sub_dir_name is None:
        this_configs_dir_path = configs_dir_path  # type: PathType
    else:
        this_configs_dir_path = os.path.join(configs_dir_path, configs_sub_dir_name)

    for config_dir in this_configs_dir_path | find(f"{production_tag}*"):
        print(f"Checking config directory {os.path.basename(config_dir)}\n\tIncludes the configs:")
        configs_dict = {}  # type: Dict[PathType, Dict[Any, Any]]

        for config_path in config_dir | find("*.json"):
            print(f"\t\t- {os.path.basename(config_path)}")
            configs_dict.update({config_path: handler.load_json_config(path_to_file=config_path)})

        configs = list(configs_dict.values())  # type: List[Dict[Any, Any]]

        data_type_keys = [k for c in configs for k in c["data"].keys()]
        if not len(data_type_keys) == len(set(data_type_keys)):
            duplicate_keys = [k for k, count in collections.Counter(data_type_keys).items() if count > 1]
            err_msg = f"ERROR: The keys {duplicate_keys} appear in multiple configs of {os.path.basename(config_dir)}!"
            error_msgs.append(err_msg)
            print(err_msg)

        streams = []  # type: List[Optional[int]]

        for p, c in configs_dict.items():
            c_streams = [
                s
                for sl in [[v["streams"]] if v["streams"] is None else v["streams"] for v in c["data"].values()]
                for s in sl
            ]

            if not (len(set(c_streams)) == 1 or (len(set(c_streams)) == 2 and None in c_streams)):
                err_msg = (
                    f"ERROR: Multiple streams used in config "
                    f"{os.path.basename(config_dir)}/{os.path.basename(p)}: "
                    f"The found streams are {list(set(c_streams))}!"
                )
                error_msgs.append(err_msg)
                print(err_msg)

            if None in c_streams and not (os.path.basename(p) in accepted_none_confs or len(set(c_streams)) == 1):
                err_msg = (
                    f"ERROR: Found None stream used in config "
                    f"{os.path.basename(config_dir)}/{os.path.basename(p)} "
                    f"which should not contain any None values for streams! Streams: {c_streams}"
                )
                error_msgs.append(err_msg)
                print(err_msg)

            streams.extend(c_streams)

        if not (len(set(streams)) == 1 or (len(set(streams)) == 2 and None in streams)):
            err_msg = (
                f"ERROR: Multiple streams used in config directory {os.path.basename(config_dir)}: "
                f"The found streams are {list(set(streams))}!"
            )
            error_msgs.append(err_msg)
            print(err_msg)

        print(streams)

        offline_config_paths = list(config_dir | find("*.yaml"))
        if not len(offline_config_paths) == 1:
            err_msg = (
                f"ERROR: Found none or multiple offline configs in config dir {os.path.basename(config_dir)}: "
                f"The found offline configs are {[os.path.basename(p) for p in offline_config_paths]}!"
            )
            error_msgs.append(err_msg)
            print(err_msg)
        offline_config_path = offline_config_paths[0]

        with open(offline_config_path) as f:
            offline_config = yaml.load(f, Loader=yaml.FullLoader)

        expected_batch_confs = offline_config["batch_configs"]
        available_batch_confs = [
            os.path.basename(p) for p in configs_dict.keys() if os.path.basename(p) not in accepted_none_confs
        ]

        assert len(expected_batch_confs) == len(set(expected_batch_confs)), (
            len(expected_batch_confs),
            len(set(expected_batch_confs)),
        )
        if not set(expected_batch_confs) == set(available_batch_confs):
            missing_confs = [c for c in expected_batch_confs if c not in available_batch_confs]
            missing_in_offline_conf = [c for c in available_batch_confs if c not in expected_batch_confs]
            err_msg = (
                f"ERROR: Configs missing for config dir {os.path.basename(config_dir)}! "
                f"Missing batch configs: {missing_confs}; "
                f"batch configs missing in offline conf: {missing_in_offline_conf}"
            )
            error_msgs.append(err_msg)
            print(err_msg)

    if len(error_msgs) > 0:
        msg = "\nThe following ERRORS occurred during the check:\n" + "\n\t- ".join(error_msgs)  # type: str
        print(msg)
        raise RuntimeError(msg)

    print("\nNo ERRORS occurred during the check :)\n")


if __name__ == "__main__":
    tag = "prod_02ndMay2021"  # type: str
    config_check(production_tag=tag)
