"""
Utilities for branching fraction calculations

Felix Metzner 2021
"""

import numpy as np

from uncertainties import ufloat, correlated_values, UFloat
from typing import Union, Optional, List, Tuple, NamedTuple, Sequence, Dict

from rdstar.utilities import PathType
from rdstar.latex_labels.latex_output_utilities import create_latex_command_file, escape_special_chars


__all__ = [
    "BFEntryType",
    "BFSDType",
    "BelleBBFEntry",
    "update_branching_fraction",
    "BFMeasurement",
    "IsoSpinVarianceWeightedAvgCalculator",
    "SimpleIsoSpinVarianceWeightedAvgCalculator",
    "assume_100_percent_uncertainty",
    "split_up_by_bfs",
    "info_str",
    "create_latex_value_file",
]


class BFEntryType(NamedTuple):
    bf_value: UFloat
    charged_b: bool
    decay_descr: str

    @property
    def bf(self) -> float:
        return self.bf_value.nominal_value

    @property
    def bf_std(self) -> float:
        return self.bf_value.std_dev


BFSDType = Dict[str, BFEntryType]


class BelleBBFEntry(NamedTuple):
    decay: str
    model: str
    bf: float

    @property
    def clean_decay(self) -> str:
        return self.get_clean_decay(decay_str=self.decay)

    @staticmethod
    def get_clean_decay(decay_str: str) -> str:
        if decay_str.count("  ") == 0:
            return decay_str
        return BelleBBFEntry.get_clean_decay(decay_str=decay_str.replace("  ", " "))


def update_branching_fraction(bf_dict: BFSDType, bf_factor: float) -> BFSDType:
    assert all(isinstance(v.bf, float) for v in bf_dict.values()), [
        (k, v.bf, type(v.bf)) for k, v in bf_dict.items() if not isinstance(v.bf, float)
    ]

    new_dict = {
        k: BFEntryType(
            bf_value=v.bf_value * bf_factor,
            charged_b=v.charged_b,
            decay_descr=v.decay_descr,
        )
        for k, v in bf_dict.items()
    }  # type: BFSDType

    assert len(new_dict.keys()) == len(bf_dict.keys()), (len(new_dict.keys()), len(bf_dict.keys()))
    assert len(new_dict.keys()) == len(bf_dict.keys()), (len(new_dict.keys()), len(bf_dict.keys()))

    assert all(new_dict[k].charged_b == v.charged_b for k, v in bf_dict.items())
    assert all(new_dict[k].decay_descr == v.decay_descr for k, v in bf_dict.items())

    return new_dict


class BFMeasurement(NamedTuple):
    central_value: float
    stat_uncert: float
    sys_uncert: float

    @property
    def combined_uncert_sq(self) -> float:
        return self.stat_uncert * self.stat_uncert + self.sys_uncert * self.sys_uncert

    @property
    def combined_uncert(self) -> float:
        return np.sqrt(self.combined_uncert_sq)

    @property
    def ufloat(self) -> UFloat:
        return ufloat(self.central_value, self.combined_uncert, tag="combined")

    @property
    def ufloat_stat(self) -> UFloat:
        return ufloat(self.central_value, self.stat_uncert, tag="stat")

    @property
    def ufloat_sys(self) -> UFloat:
        return ufloat(self.central_value, self.sys_uncert, tag="sys")


class IsoSpinVarianceWeightedAvgCalculator:
    # Isospin variance weighted average calculation assuming 0 correlation between measurements
    def __init__(
        self,
        neutral_b_bf: BFMeasurement,
        charged_b_bf: BFMeasurement,
        lifetime_ratio_tau_p0: UFloat,
        correlation_rho_sys: Optional[float] = None,
    ) -> None:

        self.measured_b0_bf = neutral_b_bf  # type: BFMeasurement
        self.measured_bp_bf = charged_b_bf  # type: BFMeasurement
        self.tau_p0 = lifetime_ratio_tau_p0  # type: UFloat
        self._rho_sys = correlation_rho_sys  # type: Optional[float]

        self._cov_mat = None  # type: Optional[np.matrix]

        self._avg_b0_and_bp_bfs = None  # type: Optional[Tuple[UFloat, UFloat]]

    @property
    def rho_sys(self) -> float:
        if self._rho_sys is None:
            return 0.0
        return self._rho_sys

    @property
    def cov_mat(self) -> np.matrix:
        if self._cov_mat is None:
            self._cov_mat = self.calc_cov_mat()
        return self._cov_mat

    @property
    def inv_cov_mat(self) -> np.matrix:
        return np.linalg.inv(self.cov_mat)

    @property
    def average_neutral_b_bf(self) -> UFloat:
        return self._b0_and_bp_bf_averages[0]

    @property
    def average_charged_b_bf(self) -> UFloat:
        return self._b0_and_bp_bf_averages[1]

    @property
    def _b0_and_bp_bf_averages(self) -> Tuple[UFloat, UFloat]:
        if self._avg_b0_and_bp_bfs is None:
            self._avg_b0_and_bp_bfs = self._calc_b0_and_bp_bf_averages()
        return self._avg_b0_and_bp_bfs

    def calc_cov_mat(self) -> np.matrix:
        b_to_d_cov_00 = (self.measured_b0_bf.ufloat_stat * self.tau_p0).std_dev ** 2 + (
            self.measured_b0_bf.ufloat_sys * self.tau_p0
        ).std_dev ** 2  # type: float
        b_to_d_cov_01 = (
            self.rho_sys * (self.measured_b0_bf.ufloat_sys * self.tau_p0).std_dev * self.measured_bp_bf.ufloat_sys.std_dev
        )  # type: float
        b_to_d_cov_10 = b_to_d_cov_01  # type: float
        b_to_d_cov_11 = (
            self.measured_bp_bf.ufloat_stat.std_dev ** 2 + self.measured_bp_bf.ufloat_sys.std_dev ** 2
        )  # type: float

        cov_mat = np.matrix([[b_to_d_cov_00, b_to_d_cov_01], [b_to_d_cov_10, b_to_d_cov_11]])  # type: np.matrix
        return cov_mat

    def calc_variance_weights(self) -> Tuple[float, float]:
        inv_cov_mat = self.inv_cov_mat  # type: np.matrix
        w_b0 = (inv_cov_mat[0, 0] + inv_cov_mat[0, 1]) / inv_cov_mat.flatten().sum()  # type: float
        w_bp = (inv_cov_mat[1, 0] + inv_cov_mat[1, 1]) / inv_cov_mat.flatten().sum()  # type: float
        return w_b0, w_bp

    def _calc_b0_and_bp_bf_averages(self) -> Tuple[UFloat, UFloat]:
        _nom_vals = [
            (self.measured_b0_bf.ufloat_stat * self.tau_p0).nominal_value,
            self.measured_bp_bf.central_value,
        ]  # type: List[float]

        corr_b0_bf, corr_bp_bf = correlated_values(
            nom_values=np.array(_nom_vals),
            covariance_mat=np.array(self.cov_mat),
        )  # type: UFloat, UFloat

        w_b0, w_bp = self.calc_variance_weights()  # type: float, float

        _bf_bp_avg = w_b0 * corr_b0_bf + w_bp * corr_bp_bf  # type: UFloat
        bf_b0_avg = _bf_bp_avg / self.tau_p0  # type: UFloat
        bf_bp_avg = _bf_bp_avg  # type: UFloat
        return bf_b0_avg, bf_bp_avg

    def info_str(
        self,
        decay_id_str: str,
    ) -> str:
        _info_str = (
            f"Averaged BF neutral {decay_id_str}: {self.average_neutral_b_bf:2.6f}\n"
            f"Averaged BF charged {decay_id_str}: {self.average_charged_b_bf:2.6f}\n"
            f"{decay_id_str} -- w_B0, w_B+ = {self.calc_variance_weights()}"
        )  # type: str
        return _info_str


class SimpleIsoSpinVarianceWeightedAvgCalculator:
    def __init__(
        self,
        neutral_b_bf: BFMeasurement,
        charged_b_bf: BFMeasurement,
        two_body_iso_spin_ratio: Union[float, UFloat],
        lifetime_ratio_tau_p0: UFloat,
    ) -> None:

        self._measured_b0_bf = neutral_b_bf  # type: BFMeasurement
        self._measured_bp_bf = charged_b_bf  # type: BFMeasurement
        self.two_body_iso_spin_ratio = two_body_iso_spin_ratio  # type: Union[float, UFloat]
        self.tau_p0 = lifetime_ratio_tau_p0  # type: UFloat

        self._wb0_wbp = None  # type: Optional[Tuple[float, float]]

        self._avg_b0_and_bp_bfs = None  # type: Optional[Tuple[UFloat, UFloat]]

    @property
    def measured_b0_bf_ufloat(self) -> UFloat:
        return self._measured_b0_bf.ufloat * self.two_body_iso_spin_ratio / self.tau_p0

    @property
    def measured_bp_bf_ufloat(self) -> UFloat:
        return self._measured_bp_bf.ufloat * self.two_body_iso_spin_ratio

    @property
    def average_neutral_b_bf(self) -> UFloat:
        return self._b0_and_bp_bf_averages[0]

    @property
    def average_charged_b_bf(self) -> UFloat:
        return self._b0_and_bp_bf_averages[1]

    @property
    def _b0_and_bp_bf_averages(self) -> Tuple[UFloat, UFloat]:
        if self._avg_b0_and_bp_bfs is None:
            self._avg_b0_and_bp_bfs = self._calc_b0_and_bp_bf_averages()
        return self._avg_b0_and_bp_bfs

    @property
    def wb0_and_wbp(self) -> Tuple[float, float]:
        if self._wb0_wbp is None:
            self._wb0_wbp = self.calc_variance_weights()
        return self._wb0_wbp

    def calc_variance_weights(self) -> Tuple[float, float]:
        neutral_stat_inv_sq = self.measured_b0_bf_ufloat.std_dev ** -2  # type: float
        charged_stat_inv_sq = self.measured_bp_bf_ufloat.std_dev ** -2  # type: float

        w_b0 = neutral_stat_inv_sq / (charged_stat_inv_sq + neutral_stat_inv_sq)  # type: float
        w_bp = charged_stat_inv_sq / (charged_stat_inv_sq + neutral_stat_inv_sq)  # type: float

        return w_b0, w_bp

    def _calc_b0_and_bp_bf_averages(self) -> Tuple[UFloat, UFloat]:
        w_b0, w_bp = self.wb0_and_wbp  # type: float, float

        _bf_bp_avg = w_b0 * self.measured_b0_bf_ufloat + w_bp * self.measured_bp_bf_ufloat  # type: UFloat
        bf_b0_avg = _bf_bp_avg / self.tau_p0  # type: UFloat
        bf_bp_avg = _bf_bp_avg  # type: UFloat
        return bf_b0_avg, bf_bp_avg

    def info_str(
        self,
        decay_id_str: str,
    ) -> str:
        _info_str = (
            f"Simple Averaged BF neutral {decay_id_str}: {self.average_neutral_b_bf:2.6f}\n"
            f"Simple Averaged BF charged {decay_id_str}: {self.average_charged_b_bf:2.6f}\n"
            f"{decay_id_str} -- w_B0, w_B+ = {self.wb0_and_wbp}"
        )  # type: str
        return _info_str


def assume_100_percent_uncertainty(
    bf_value: UFloat,
) -> UFloat:
    return ufloat(nominal_value=bf_value.nominal_value, std_dev=bf_value.nominal_value)


def split_up_by_bfs(
    bf_value: UFloat,
    base_bfs: Sequence[UFloat],
) -> Tuple[UFloat, ...]:
    sum_base = sum(base_bfs)
    return tuple([bf_value * base_bf / sum_base for base_bf in base_bfs])


def info_str(decay_id_str: str, bf_value: UFloat) -> str:
    return f"Our BF for {decay_id_str}: {bf_value:2.6f}"


def create_latex_value_file(
    path: PathType,
    bf_dict: Dict[str, UFloat],
    in_percent: bool = False,
    overwrite: bool = False,
) -> None:
    factor = 100.0 if in_percent else 1.0  # type: float

    round_to = 3  # type: int
    simple = False  # type: bool

    if simple:
        command_mapping = {
            escape_special_chars(s=bf_key): f"{factor * bf_value:0.{round_to}L}" for bf_key, bf_value in bf_dict.items()
        }  # type: Dict[str, str]
    else:
        command_mapping = {
            escape_special_chars(
                s=bf_key
            ): f"{factor * bf_value.nominal_value:0.{round_to}f} \\pm {factor * bf_value.std_dev:0.{round_to}f}"
            for bf_key, bf_value in bf_dict.items()
        }

    create_latex_command_file(
        output_path=path,
        command_mapping_dict=command_mapping,
        overwrite=overwrite,
    )
