"""
Providing common latex strings.

Felix Metzner 2021
"""

from typing import List, Dict

from rdstar.latex_labels.latex_particle_names import LatexParticleNames as LPNs
from rdstar.latex_labels.latex_string_methods import LatexObsNameMethods as LONM


__all__ = [
    "LatexCommonStrings",
]


class LatexCommonStrings:
    # fmt: off

    # R(D) and R(D*)
    rd_str = r"$\mathcal{R}(" + LONM.get_bare_var_str(var_str=LPNs.d_str) + ")$"  # type: str
    rd_star_str = r"$\mathcal{R}(" + LONM.get_bare_var_str(var_str=LPNs.d_st_str) + ")$"  # type: str
    rd_wwo_star_str = r"$\mathcal{R}(" + LONM.get_bare_var_str(var_str=LPNs.d_w_wo_st_str) + ")$"  # type: str

    # fmt: on

    @staticmethod
    def get_unfiltered_attributes_mapping() -> Dict[str, str]:
        _method_names = [
            "get_unfiltered_attributes_mapping",
            "get_attributes_mapping",
        ]  # type: List[str]

        return {
            attr_name: val for attr_name, val in LatexCommonStrings.__dict__.items() if not (attr_name in _method_names)
        }

    @staticmethod
    def get_attributes_mapping() -> Dict[str, str]:
        return {
            attr_name: val
            for attr_name, val in LatexCommonStrings.get_unfiltered_attributes_mapping().items()
            if not (attr_name.endswith("__") or attr_name.startswith("_"))
        }


if __name__ == "__main__":
    from collections import Counter
    from matplotlib import pyplot as plt

    all_latex_commons_keys = tuple(
        [c_key for c_key in LatexCommonStrings.get_unfiltered_attributes_mapping().keys() if not (c_key.endswith("__"))]
    )
    all_commons = tuple([LatexCommonStrings.__dict__[common_key] for common_key in all_latex_commons_keys])

    assert len(set(all_commons)) == len(all_commons), [v for v, c in Counter(all_commons).items() if c > 1]

    test_text = ""  # type: str
    for common_key in all_latex_commons_keys:
        common_str = LatexCommonStrings.__dict__[common_key]
        print("\t- " + common_key + "  ->  " + common_str)
        assert common_str.count("{") == common_str.count("}"), (common_key, common_str)
        assert common_str.count("$") % 2 == 0, (common_key, common_str, common_str.count("$"))

        test_text += f"{common_str}\n"

        fig, ax = plt.subplots(figsize=(6, 4), tight_layout=True)
        try:
            ax.annotate(text=common_str, xy=(0.5, 0.5), xytext=(0.6, 0.6))
            plt.show()
        except:  # noqa: E722  # Ignore this flake8 error for this test.
            plt.close(fig)
            fig.clf()
            raise ValueError("Error in " + common_key + "  ->  " + common_str)

        plt.close(fig)
        fig.clf()

    fig, ax = plt.subplots(figsize=(6, 4), tight_layout=True)
    ax.annotate(text=test_text, xy=(0.5, 0.5), xytext=(0.6, 0.6))
    plt.show()
