"""
Evaluate issues in the mc production.

Felix Metzner 2021
"""

import os
import json

from collections import Counter
from typing import Optional, Tuple, List, Dict, Any, NamedTuple

from rdstar.utilities import PathType
from rdstar.mc_production.check_mc.mc_production_issue_validator import MCIssueValidatorMaster
from rdstar.mc_production.check_mc.mc_production_issue_finder import MCProductionEvaluator, FoundFile

__all__ = [
    "MCIssueEvaluator",
]


class MCFileCountInfo(NamedTuple):
    log_file_path: PathType
    file_info: FoundFile
    category: str
    basf_gsim_input: int
    basf_gsim_output: int
    basf2_input: int
    basf2_converted: int

    def get_input_loss(self) -> float:
        loss = 1.0 - float(self.basf2_input) / float(self.basf_gsim_input)
        assert 0.0 <= loss <= 1.0, (loss, self.basf_gsim_input, self.basf2_input, self.log_file_path)
        return loss

    def get_input_diff(self) -> int:
        diff = self.basf_gsim_input - self.basf2_input
        assert diff >= 0, (diff, self.basf_gsim_input, self.basf2_input, self.log_file_path)
        return diff

    def get_output_loss(self) -> float:
        loss = 1.0 - float(self.basf2_converted) / float(self.basf_gsim_output)
        assert 0.0 <= loss <= 1.0, (loss, self.basf_gsim_output, self.basf2_converted, self.log_file_path)
        return loss

    def get_output_diff(self) -> int:
        diff = self.basf_gsim_output - self.basf2_converted
        assert diff >= 0, (diff, self.basf_gsim_output, self.basf2_converted, self.log_file_path)
        return diff

    def get_initial_diff_estimate(self, efficiency_estimate: Optional[float] = None) -> int:
        if efficiency_estimate is None:
            return int(self.get_output_loss() * self.basf_gsim_input)
        else:
            eff = 1.0 if self.get_output_loss() == 1.0 else efficiency_estimate
            return int(self.get_output_loss() / eff * self.basf_gsim_input)


class UpdateInfo(NamedTuple):
    diff_initial_count: int
    diff_usable_count: int

    def as_dict(self) -> Dict[str, int]:
        return dict(self._asdict())


class MCCountCorrectionInfo:
    # rough_estimation_of_skim_efficiency = 594271. / 600000.  # type: float  # Based on Charged Events
    rough_estimation_of_skim_efficiency = 492873.0 / 500000.0  # type: float  # Based on Dstst Events

    def __init__(self) -> None:
        self.file_count_infos = []  # type: List[MCFileCountInfo]

    def add_mc_file_count_info(self, file_count_info: MCFileCountInfo) -> None:
        assert file_count_info.log_file_path not in [
            i.log_file_path for i in self.file_count_infos
        ], file_count_info.log_file_path
        self.file_count_infos.append(file_count_info)

    def get_update_info_by_data_type(self, output_loss_threshold: float) -> Dict[str, UpdateInfo]:
        data_types = set([i.file_info.data_type for i in self.file_count_infos])
        eff = None  # self.rough_estimation_of_skim_efficiency
        update_dict = {}  # type: Dict[str, UpdateInfo]

        for data_type in data_types:
            relevant_infos = [i for i in self.file_count_infos if i.file_info.data_type == data_type]
            file_counts_over_threshold = [i for i in relevant_infos if i.get_output_loss() > output_loss_threshold]
            initial_check = sum([i.get_input_diff() for i in file_counts_over_threshold])
            assert initial_check == 0, initial_check

            initial_diff = sum([i.get_initial_diff_estimate(efficiency_estimate=eff) for i in file_counts_over_threshold])
            usable_diff = sum([i.get_output_diff() for i in file_counts_over_threshold])
            assert initial_diff >= usable_diff, (initial_diff, usable_diff)

            assert data_type not in update_dict.keys(), data_type

            if file_counts_over_threshold:
                update_dict.update(
                    {data_type: UpdateInfo(diff_initial_count=initial_diff, diff_usable_count=usable_diff)}
                )

        return update_dict


class MCIssueEvaluator:
    # rough_estimate_of_skim_loss = (1. - MCCountCorrectionInfo.rough_estimation_of_skim_efficiency) * 3.  # type: float
    rough_estimate_of_skim_loss = (1.0 - MCCountCorrectionInfo.rough_estimation_of_skim_efficiency) * 2.0  # type: float

    default_output_file_name = "event_count_update_info.json"  # type: str

    def __init__(
        self,
        do_basf2_evaluation: bool = True,
        info_file_path: Optional[PathType] = None,
    ) -> None:
        self.with_basf2_evaluation = do_basf2_evaluation
        self.info_file_path = MCProductionEvaluator.get_log_file_path() if info_file_path is None else info_file_path

        self.mc_check_dict = self.load_dict()
        self.gsim_log_dict = {
            p: info for logs in self.mc_check_dict.values() for p, info in logs.items()
        }  # type: Dict[str, FoundFile]
        self.gsim_log_to_category_map = {
            p: cat for cat, logs in self.mc_check_dict.items() for p in logs.keys()
        }  # type: Dict[str, str]

        assert len(self.gsim_log_dict.keys()) == len(self.gsim_log_to_category_map), (
            len(self.gsim_log_dict.keys()),
            len(self.gsim_log_to_category_map),
        )

        self.basf2_from_log_dict, self.basf2_init_from_log_dict, self.basf2_from_tuple_dict = self.load_basf2_dicts()

        if self.with_basf2_evaluation:
            assert len(self.gsim_log_dict.keys()) == len(self.basf2_from_log_dict.keys()), (
                len(self.gsim_log_dict.keys()),
                len(self.basf2_from_log_dict.keys()),
            )

        self.print_info_dict_stats()

        self.file_count_info = self.evaluate_info_dict_stats()  # type: MCCountCorrectionInfo

    def evaluate_info_dict_stats(self) -> MCCountCorrectionInfo:
        corr_info = MCCountCorrectionInfo()

        for mc_log_path, gsim_info in self.gsim_log_dict.items():
            assert gsim_info.events is not None

            basf2_init_log_info = self.basf2_init_from_log_dict[mc_log_path]
            basf2_log_info = self.basf2_from_log_dict[mc_log_path]
            basf2_tuple_log_info = self.basf2_from_tuple_dict[mc_log_path]
            assert basf2_log_info == basf2_tuple_log_info, (basf2_log_info, basf2_tuple_log_info)

            corr_info.add_mc_file_count_info(
                MCFileCountInfo(
                    log_file_path=mc_log_path,
                    file_info=gsim_info,
                    category=self.gsim_log_to_category_map[mc_log_path],
                    basf_gsim_input=gsim_info.events.input_events,
                    basf_gsim_output=gsim_info.events.output_events,
                    basf2_input=basf2_init_log_info,
                    basf2_converted=basf2_log_info,
                )
            )

        return corr_info

    def get_update_info(self, output_loss_threshold: Optional[float] = None) -> Dict[str, UpdateInfo]:
        threshold = self.rough_estimate_of_skim_loss if output_loss_threshold is None else output_loss_threshold
        return self.file_count_info.get_update_info_by_data_type(output_loss_threshold=threshold)

    def write_update_infos_to_file(
        self,
        file_path: Optional[PathType] = None,
        output_loss_threshold: Optional[float] = None,
        overwrite: bool = False,
    ):
        if file_path is not None:
            _output_path = file_path  # type: PathType
        else:
            _output_path = os.path.join(MCProductionEvaluator.get_log_file_dir_path(), self.default_output_file_name)

        if os.path.isfile(_output_path) and not overwrite:
            raise RuntimeError(f"Output file {_output_path} already exists!")

        _update_info = self.get_update_info(output_loss_threshold=output_loss_threshold)
        update_info = {dt: info.as_dict() for dt, info in _update_info.items()}
        with open(_output_path, "w") as out_file:
            json.dump(update_info, out_file, indent=2)

    def print_info_dict_stats(self) -> None:
        count_in_categories = {c: len(d) for c, d in self.mc_check_dict.items()}
        print(count_in_categories)
        print(f"Sum of files = {sum(list(count_in_categories.values()))}")

        data_type_in_categories = {c: [i[1].data_type for i in d.items()] for c, d in self.mc_check_dict.items()}
        count_in_data_type = {c: Counter(dts) for c, dts in data_type_in_categories.items()}
        print(count_in_data_type)
        print(f"Sum of files = {sum([sum(list(cv.values())) for cv in count_in_data_type.values()])} (Just Checking)")

        print(f"Sum of files in basf2 n-tuple check = {len(self.basf2_from_tuple_dict)}")
        print(f"Sum of files in basf2 log check = {len(self.basf2_from_log_dict)}")
        print(f"Sum of files in basf2 log check = {len(self.basf2_init_from_log_dict)}")

        assert set(self.basf2_from_tuple_dict.keys()) == set(self.basf2_from_log_dict.keys())

        print(
            f"Files with 0 events in basf2 n-tuple check = {len([i for i in self.basf2_from_tuple_dict.values() if i == 0])}"
        )
        print(f"Files with 0 events in basf2 log check = {len([i for i in self.basf2_from_log_dict.values() if i == 0])}")

        logs_frank_found = [c.to_path(category="Dststlnu") for c in MCProductionEvaluator.get_files_frank_found()]
        print(
            "Files with 0 events and missing in what Frank found:\n\t- "
            + "\n\t- ".join([p for p, i in self.basf2_from_log_dict.items() if p not in logs_frank_found and i == 0])
        )

        print(
            "Checking if initial event numbers match between basf and basf2:\n\t- "
            + "\n\t- ".join(
                [
                    f"{p}: basf = {self.gsim_log_dict[p].events.input_events} != i = basf2"  # type: ignore
                    for p, i in self.basf2_init_from_log_dict.items()
                    if self.gsim_log_dict[p].events.input_events != i  # type: ignore
                ]
            )
        )

        # print(":\n\t- " + "\n\t- ".join([p for p, i in self.basf2_from_log_dict.items() if i == 0]))

        print(
            "\n\nChecking if event numbers are consistent in basf2 check:\n\t- "
            + "\n\t- ".join([p for p, tc in self.basf2_from_tuple_dict.items() if tc != self.basf2_from_log_dict[p]])
        )

        print(
            "\n\nChecking if basf2 event counts match basf gsim log count:\n\t- "
            + "\n\t- ".join(
                [
                    f"basf2={c:7d} in {p}\n\t   basf={self.gsim_log_dict[p].events.output_events:7d}  "  # type: ignore
                    f"=> {self.gsim_log_dict[p].events.output_events - c:7d} less in basf2"  # type: ignore
                    for p, c in self.basf2_from_log_dict.items()
                    if c != self.gsim_log_dict[p].events.output_events  # type: ignore
                ]
            )
        )

        print(
            "\n\nFiles with significant lower event count in basf2:\n\t- "
            + "\n\t- ".join(
                [
                    f"in {p}\n\t\tbasf2={c:7d}\n\t\t basf={self.gsim_log_dict[p].events.output_events:7d}  "  # type: ignore
                    f"=> {self.gsim_log_dict[p].events.output_events - c:7d} less in basf2 "  # type: ignore
                    f"(loss = {1. - (float(c) / self.gsim_log_dict[p].events.output_events):.3f};\t\t"  # type: ignore
                    f"initial basf = {self.gsim_log_dict[p].events.input_events} , basf2 = {self.basf2_init_from_log_dict[p]})"  # type: ignore
                    for p, c in self.basf2_from_log_dict.items()
                    if 1.0 - (float(c) / self.gsim_log_dict[p].events.output_events) > self.rough_estimate_of_skim_loss  # type: ignore
                ]
            )
        )

        print(
            "\n\nFiles which are partially affected:\n\t- "
            + "\n\t- ".join(
                [
                    f"in {p}\n\t\tbasf2={c:7d}\n\t\t basf={self.gsim_log_dict[p].events.output_events:7d}  "  # type: ignore
                    f"=> {self.gsim_log_dict[p].events.output_events - c:7d} less in basf2 "  # type: ignore
                    f"(loss = {1. - (float(c) / self.gsim_log_dict[p].events.output_events):.3f};\t\t"  # type: ignore
                    f"initial basf = {self.gsim_log_dict[p].events.input_events} , basf2 = {self.basf2_init_from_log_dict[p]})"  # type: ignore
                    for p, c in self.basf2_from_log_dict.items()
                    if 1.0
                    > 1.0 - (float(c) / self.gsim_log_dict[p].events.output_events)  # type: ignore
                    > self.rough_estimate_of_skim_loss
                ]
            )
        )

        print(self.rough_estimate_of_skim_loss)

    def load_dict(self) -> Dict[str, Any]:
        log_dict, _ = MCProductionEvaluator.read_info_from_file(input_path=self.info_file_path)
        return log_dict

    def load_basf2_dicts(self) -> Tuple[Dict[str, int], Dict[str, int], Dict[str, int]]:
        if not self.with_basf2_evaluation:
            return {}, {}, {}

        mc_check_path = MCProductionEvaluator.get_log_file_dir_path()
        basf2_log_dict_path = os.path.join(mc_check_path, MCIssueValidatorMaster.log_event_count_file_name)
        basf2_init_log_dict_path = os.path.join(mc_check_path, MCIssueValidatorMaster.log_initial_event_count_file_name)
        basf2_n_tuple_dict_path = os.path.join(mc_check_path, MCIssueValidatorMaster.n_tuple_event_count_file_name)

        basf2_log_dict = self._load_basf2_dict(file_path=basf2_log_dict_path)
        basf2_init_log_dict = self._load_basf2_dict(file_path=basf2_init_log_dict_path)
        basf2_n_tuple_dict = self._load_basf2_dict(file_path=basf2_n_tuple_dict_path)

        return basf2_log_dict, basf2_init_log_dict, basf2_n_tuple_dict

    @staticmethod
    def _load_basf2_dict(file_path: PathType) -> Dict[str, int]:
        assert os.path.isfile(file_path), file_path
        with open(file_path, "r") as in_file:
            file_dict = json.load(fp=in_file)
        return {str(MCIssueValidatorMaster.mdst_path_to_log_path(mdst_path=k)): v for k, v in file_dict.items()}


if __name__ == "__main__":
    issue_evaluator = MCIssueEvaluator()

    issue_evaluator.write_update_infos_to_file(overwrite=True)

    for dk, dv in issue_evaluator.get_update_info().items():
        print(dk, dv.as_dict())
