"""
Management of specific data sample definitions.

Felix Metzner 2021
"""

from dataclasses import dataclass
from typing import Tuple, List, Sequence, Dict, ClassVar

from rdstar.offline_analysis import rdstar_samples as rdstar_samples_info


__all__ = [
    "SampleSetupInfo",
    "SampleSetupHandler",
]


@dataclass(frozen=True)
class SampleSetupInfo:
    name: str
    data_sample_key: str
    mc_sample_keys: Tuple[str, ...]


class SampleSetupHandler:
    default_gap_sample_name: ClassVar[str] = "default_gap_setup"
    alternative_gap_sample_name: ClassVar[str] = "alternative_gap_setup"
    off_resonance_sample_name: ClassVar[str] = "off_resonance_setup"

    on_resonance_sample_setup_names: ClassVar[Tuple[str, ...]] = (
        default_gap_sample_name,
        alternative_gap_sample_name,
    )

    sample_setup_mapping: ClassVar[Dict[str, SampleSetupInfo]] = {
        default_gap_sample_name: SampleSetupInfo(
            name=default_gap_sample_name,
            data_sample_key=rdstar_samples_info.y4s_data_key,
            mc_sample_keys=tuple(rdstar_samples_info.main_sample_keys),
        ),
        alternative_gap_sample_name: SampleSetupInfo(
            name=alternative_gap_sample_name,
            data_sample_key=rdstar_samples_info.y4s_data_key,
            mc_sample_keys=tuple(rdstar_samples_info.alternative_sample_keys),
        ),
        off_resonance_sample_name: SampleSetupInfo(
            name=off_resonance_sample_name,
            data_sample_key=rdstar_samples_info.off_res_data_key,
            mc_sample_keys=tuple(rdstar_samples_info.off_res_sample_keys),
        ),
    }

    @staticmethod
    def identify_sample_setup_key(mc_sample_keys: Sequence[str], data_sample_key: str) -> str:
        matching_sample_keys = list()  # type: List[str]
        for sample_key, sample_info in SampleSetupHandler.sample_setup_mapping.items():
            if set(mc_sample_keys) == set(sample_info.mc_sample_keys) and data_sample_key == sample_info.data_sample_key:
                matching_sample_keys.append(sample_key)
        assert len(matching_sample_keys) == 1, matching_sample_keys
        return matching_sample_keys[0]

    @staticmethod
    def get_sample(sample_name: str) -> SampleSetupInfo:
        if sample_name not in SampleSetupHandler.sample_setup_mapping:
            raise KeyError(
                f"Unknown sample setup name {sample_name}!"
                f"Must be one of {list(SampleSetupHandler.sample_setup_mapping.keys())}"
            )
        return SampleSetupHandler.sample_setup_mapping[sample_name]

    @staticmethod
    def get_mc_sample_keys(sample_name: str) -> Tuple[str, ...]:
        return SampleSetupHandler.get_sample(sample_name=sample_name).mc_sample_keys

    @staticmethod
    def get_data_sample_key(sample_name: str) -> str:
        return SampleSetupHandler.get_sample(sample_name=sample_name).data_sample_key

    @staticmethod
    def get_on_resonance_sample_key_for(use_alternative_gap: bool) -> str:
        if use_alternative_gap:
            return SampleSetupHandler.alternative_gap_sample_name
        else:
            return SampleSetupHandler.default_gap_sample_name
