"""
Management tools for loading and handling of data samples post selection.

Felix Metzner 2021
"""

import os
import copy
import pandas as pd

from typing import Optional, Tuple, List, Sequence, Dict, ClassVar, NamedTuple

from rdstar.utilities import PathType

from rdstar.offline_analysis.data_handling.data_sample_tag_names import DataSampleTags
from rdstar.offline_analysis.data_handling.sample_setup_definitions import SampleSetupHandler


__all__ = [
    "DataSampleLoader",
]


class DataLoaderReturnTuple(NamedTuple):
    data: pd.DataFrame
    mc: pd.DataFrame


class DataSampleLoader:
    base_identifier: ClassVar[str] = "BASE"
    base_file_tag: ClassVar[str] = ""

    identifier_separator_str: ClassVar[str] = "__AND__"

    input_level_mapping: ClassVar[Dict[int, str]] = {
        0: "",
        1: f"{DataSampleTags.weight_finalized_sample_tag}",
        2: f"{DataSampleTags.weight_finalized_sample_tag}_{DataSampleTags.resolution_corrected_sample_tag}",
    }

    def __init__(
        self,
        input_dict: Dict[str, PathType],
        sample_setup_key: str,
        cs_tag: str,
        selection_name: str,
        input_level: int,
        sideband_identifiers: Sequence[str],
    ) -> None:
        if input_level not in self.input_level_mapping:
            raise KeyError(
                f"Invalid input level {input_level} provided!"
                f"Must be one of {tuple(self.input_level_mapping.keys())} to match the categories\n\t"
                + "\n\t".join([f"{k}: {v}" for k, v in self.input_level_mapping.items()])
            )

        self._input_dict = copy.copy(input_dict)  # type: Dict[str, PathType]

        self._sample_setup_key = sample_setup_key  # type: str

        self._cs_tag = cs_tag  # type: str
        self._selection_name = selection_name  # type: str
        self._input_level = input_level  # type: int
        self._input_file_type_appendix = self.input_level_mapping[self._input_level]  # type: str

        self._sideband_identifiers = tuple(sideband_identifiers)  # type: Tuple[str, ...]

        self._potential_mc_paths = None  # type: Optional[Tuple[PathType, ...]]

    @property
    def data_sample_key(self) -> str:
        return SampleSetupHandler.get_data_sample_key(sample_name=self._sample_setup_key)

    @property
    def mc_sample_keys(self) -> Tuple[str, ...]:
        return SampleSetupHandler.get_mc_sample_keys(sample_name=self._sample_setup_key)

    @property
    def input_dict(self) -> Dict[str, PathType]:
        return self._input_dict

    @property
    def input_file_type_appendix(self) -> str:
        return f"_{self._input_file_type_appendix}" if self._input_file_type_appendix else ""

    def get_file_tag_for(self, sideband_identifier: str) -> str:
        return self.base_file_tag if sideband_identifier == self.base_identifier else sideband_identifier

    def get_selection_and_sideband_identifier_for(self, sideband_identifier: str) -> str:
        _prefix = f"_{self._selection_name}_with_CS_{self._cs_tag}"  # type: str
        __sideband_file_tag = self.get_file_tag_for(sideband_identifier=sideband_identifier)  # type: str
        _sideband_file_tag = f"_{__sideband_file_tag}" if __sideband_file_tag else __sideband_file_tag  # type: str

        return f"{_prefix}{_sideband_file_tag}"

    @staticmethod
    def is_hdf_file(fp: PathType) -> bool:
        return os.path.splitext(fp)[1] == ".h5"

    def _get_data_paths(self) -> Tuple[PathType, ...]:
        data_file_paths = []  # type: List[PathType]
        for sideband_id in self._sideband_identifiers:
            key_start = f"{self.data_sample_key}{self.input_file_type_appendix}_"  # type: str
            key_end = self.get_selection_and_sideband_identifier_for(sideband_identifier=sideband_id)  # type: str

            data_paths = [
                p
                for k, p in self.input_dict.items()
                if k.startswith(key_start) and k.endswith(key_end) and self.is_hdf_file(fp=p)
            ]  # type: List[PathType]
            assert len(data_paths) == 1, data_paths
            assert all(d_p not in data_file_paths for d_p in data_paths), (data_paths, data_file_paths)
            data_file_paths.extend(data_paths)

        assert len(data_file_paths) == len(self._sideband_identifiers), (data_file_paths, self._sideband_identifiers)
        return tuple(data_file_paths)

    def _get_potential_mc_paths(self) -> Tuple[PathType, ...]:
        mc_file_paths = []  # type: List[PathType]
        for sideband_id in self._sideband_identifiers:
            key_end = self.get_selection_and_sideband_identifier_for(sideband_identifier=sideband_id)  # type: str
            for mc_key in self.mc_sample_keys:
                key_start = f"{mc_key}{self.input_file_type_appendix}_"  # type: str
                mc_paths = [
                    p
                    for k, p in self.input_dict.items()
                    if k.startswith(key_start) and k.endswith(key_end) and self.is_hdf_file(fp=p)
                ]  # type: List[PathType]
                assert len(mc_paths) <= 1, mc_paths
                assert all(mc_p not in mc_file_paths for mc_p in mc_paths), (mc_paths, mc_file_paths)
                mc_file_paths.extend(mc_paths)
        return tuple(mc_file_paths)

    @property
    def potential_mc_paths(self) -> Tuple[PathType, ...]:
        if self._potential_mc_paths is None:
            self._potential_mc_paths = self._get_potential_mc_paths()
        return self._potential_mc_paths

    def _get_mc_paths(self) -> Tuple[PathType, ...]:
        mc_file_paths = self.potential_mc_paths  # type: Tuple[PathType, ...]
        assert len(mc_file_paths) == len(self.mc_sample_keys) * len(self._sideband_identifiers), (
            (len(mc_file_paths), len(self.mc_sample_keys), len(self._sideband_identifiers)),
            mc_file_paths,
            self.mc_sample_keys,
            self._sideband_identifiers,
        )

        return tuple(mc_file_paths)

    @property
    def files_are_available(self) -> bool:
        if len(self.potential_mc_paths) == 0:
            return False
        else:
            assert len(self.potential_mc_paths) == len(self._get_mc_paths()), (
                self.potential_mc_paths,
                self._get_mc_paths(),
            )
            return True

    def load_data_and_mc(self) -> DataLoaderReturnTuple:
        assert self.files_are_available

        data_file_paths = self._get_data_paths()  # type: Tuple[PathType, ...]
        assert all(os.path.isfile(p) for p in data_file_paths), [p for p in data_file_paths if not os.path.isfile(p)]
        if len(data_file_paths) == 1:
            _data_df = pd.read_hdf(data_file_paths[0])
            assert isinstance(_data_df, pd.DataFrame), type(_data_df)
            data_df = _data_df  # type: pd.DataFrame
        elif len(data_file_paths) > 1:
            data_df = pd.concat([pd.read_hdf(fp) for fp in data_file_paths], ignore_index=True, sort=False)
        else:
            raise RuntimeError("No data file paths provided!")

        mc_file_paths = self._get_mc_paths()  # type: Tuple[PathType, ...]
        assert all(os.path.isfile(p) for p in mc_file_paths), [p for p in mc_file_paths if not os.path.isfile(p)]
        mc_df = pd.concat([pd.read_hdf(mc_path) for mc_path in mc_file_paths], ignore_index=True, sort=False)

        return DataLoaderReturnTuple(
            data=data_df,
            mc=mc_df,
        )

    @staticmethod
    def get_identifiers_from_sideband_key(sideband_key: str) -> Tuple[str, ...]:
        return tuple(sideband_key.split(DataSampleLoader.identifier_separator_str))

    @staticmethod
    def get_sideband_key_for(sideband_keys: Sequence[str]) -> str:
        assert len(sideband_keys) > 1, sideband_keys
        _sideband_keys = [
            DataSampleLoader.base_identifier if k == DataSampleLoader.base_file_tag else k for k in sideband_keys
        ]  # type: List[str]
        return DataSampleLoader.identifier_separator_str.join(_sideband_keys)
