import json
import numpy as np
from typing import Dict, Tuple, Union, Optional

import pandas as pd
from templatefitter.minimizer import MinimizeResult

from rdstar.utilities import PathType
from rdstar.offline_analysis.dss_fitting.setup.bf_fitting_setup import BFFittingSetup
from rdstar.offline_analysis.dss_fitting.setup.dss_bf_fit_cases import BFFitCaseDict
from rdstar.offline_analysis.efficiency_calculation.simple_efficiency_calculator import EfficiencyNumbersContainer


__all__ = [
    "BFCalculator",
]


YieldDictType = Dict[str, Dict[str, Union[Tuple[float, float], int]]]


class BFCalculator:
    def __init__(
        self,
        fitting_setup: BFFittingSetup,
        fit_result: MinimizeResult,
        efficiency_dict: EfficiencyNumbersContainer,
    ) -> None:
        self._fitting_setup = fitting_setup
        self._fit_case = BFFitCaseDict.get_fit_case(fit_case_name=self._fitting_setup.fit_case_name)
        self._yield_parameters = self._fitting_setup.yield_parameters
        self._norm_parameter = self._fitting_setup.norm_parameter
        self._process_latex_labels = self._fit_case.plot_labels

        self._yield_dict = None  # type: Optional[YieldDictType]
        self._yields = None  # type: Optional[Tuple[np.ndarray, np.ndarray]]
        self._ratios = None  # type: Optional[Tuple[np.ndarray, np.ndarray]]

        self._bf_dict = dict()  # type: Dict[str, Dict[str, float]]
        self._eff_dict = dict()  # type: Dict[str, Dict[str, float]]

        self._fit_result = fit_result
        self._efficiency_dict = efficiency_dict  # type: EfficiencyNumbersContainer

        self.all_mode_ids = [d_id for d_ids in [d_ids for _, d_ids in self._fit_case.reco_mode_infos] for d_id in d_ids]

    def get_yield_dict(self) -> YieldDictType:
        yield_dict = dict()  # type: YieldDictType
        for yield_parameter in self._yield_parameters:
            yield_value = self._fit_result.params.get_param_value(f"yield_{yield_parameter}")
            yield_error = self._fit_result.params.get_param_error(f"yield_{yield_parameter}")

            yield_dict[yield_parameter] = {
                "yield": (yield_value, yield_error),
                "id": self._fit_result.params.param_id_to_index(f"yield_{yield_parameter}"),
            }

        norm_yield = self._fit_result.params.get_param_value(f"yield_{self._norm_parameter}")
        norm_error = self._fit_result.params.get_param_error(f"yield_{self._norm_parameter}")
        yield_dict[self._norm_parameter] = {
            "yield": (norm_yield, norm_error),
            "id": self._fit_result.params.param_id_to_index(f"yield_{self._norm_parameter}"),
        }

        return yield_dict

    def get_all_yields(self) -> Tuple[np.ndarray, np.ndarray]:
        shape = self._fit_result.params.covariance.shape[0]
        assert all(y > 0 for y in self._fit_result.params.values[:shape]), (
            self._fit_case.name,
            self._fit_result.params.values[:shape],
        )
        return self._fit_result.params.values[:shape], self._fit_result.params.errors[:shape]

    def calculate_efficiency_for_parameter(
        self,
        parameter: str,
    ) -> Tuple[float, float, float]:
        sig_ids = self._fit_case.get_sig_id_for_component(component_name=parameter)
        efficiency_info = self._efficiency_dict.get_efficiency(
            sig_ids=sig_ids,
            decay_mode_ids=tuple(self.all_mode_ids),
        )

        assert parameter not in self._eff_dict.keys(), "You are trying to overwrite an efficiency entry!"
        self._eff_dict[parameter] = {
            "efficiency": efficiency_info.nominal_value,
            "upper_error": efficiency_info.std_dev_up,
            "lower_error": efficiency_info.std_dev_down,
        }

        return efficiency_info.nominal_value, efficiency_info.std_dev_up, efficiency_info.std_dev_down

    def get_efficiency_from_dict(self, parameter: str) -> Tuple[float, float, float]:
        assert parameter in self._eff_dict.keys()
        efficiency = self._eff_dict[parameter]["efficiency"]
        upper_error = self._eff_dict[parameter]["upper_error"]
        lower_error = self._eff_dict[parameter]["lower_error"]

        return efficiency, upper_error, lower_error

    def get_covariance_from_fit(self) -> np.ndarray:
        if self._fit_case.fix_nuisance_parameters:
            return self._fit_result.params.covariance
        else:
            raise RuntimeError("Not implemented for unfixed nuisance parameters!")

    def generate_jacobi_matrix(self) -> np.ndarray:
        assert self._yields is not None, "Yields were not calculated!"
        yields, _ = self._yields
        dim = len(yields)
        assert self._yield_dict is not None, "Yield Dict was not set!"
        norm_param_id = self._yield_dict[self._norm_parameter]["id"]
        jacobi = np.zeros((dim, dim))
        for i in range(0, dim):
            for j in range(0, dim):
                if i == j and i != norm_param_id:
                    jacobi[i][j] = 1 / yields[norm_param_id]
                elif i != j and j == norm_param_id:
                    jacobi[i][j] = -yields[i] / (yields[norm_param_id] ** 2)
                else:
                    jacobi[i][j] = 0
        return jacobi

    def calculate_ratio_errors(self) -> np.ndarray:
        assert self._yields is not None, "Yields were not calculated!"
        yields, _ = self._yields

        cov = self.get_covariance_from_fit()
        assert len(yields) == cov.shape[0], (len(yields), cov.shape[0])
        dim = len(yields)
        jacobi = self.generate_jacobi_matrix()
        dov = np.matmul(np.matmul(jacobi, cov), np.transpose(jacobi))

        ratio_errors = [np.sqrt(dov[i][i]) for i in range(0, dim)]
        return np.array(ratio_errors)

    def calculate_ratios(self) -> Tuple[np.ndarray, np.ndarray]:
        assert self._yields is not None, "Yields were not calculated!"
        yields, _ = self._yields
        dim = len(yields)
        assert self._yield_dict is not None, "Yield Dict was not set!"
        norm_id = self._yield_dict[self._norm_parameter]["id"]

        ratios = np.array([(yields[i] / yields[norm_id]) for i in range(0, dim)])
        ratio_errors = self.calculate_ratio_errors()

        return ratios, ratio_errors

    def calculate_bf_for_parameter(
        self,
        yield_parameter: str,
    ) -> Tuple[float, float, float]:
        assert self._yield_dict is not None, "Yield Dict was not set!"
        param_id = self._yield_dict[yield_parameter]["id"]
        assert isinstance(param_id, int)
        norm_id = self._yield_dict[self._norm_parameter]["id"]
        assert isinstance(norm_id, int)

        assert self._ratios is not None, "Ratios were not calculated!"
        ratios, ratio_errors = self._ratios
        ratio_for_param = ratios[param_id]

        if yield_parameter not in self._eff_dict.keys():
            eff, upper_eff, lower_eff = self.calculate_efficiency_for_parameter(
                parameter=yield_parameter,
            )
        else:
            eff, upper_eff, lower_eff = self.get_efficiency_from_dict(parameter=yield_parameter)

        if self._norm_parameter not in self._eff_dict.keys():
            norm_eff, upper_norm, lower_norm = self.calculate_efficiency_for_parameter(
                parameter=self._norm_parameter,
            )
        else:
            norm_eff, upper_norm, lower_norm = self.get_efficiency_from_dict(parameter=self._norm_parameter)

        norm_bf = self._fitting_setup.norm_bf
        norm_bf_error = self._fitting_setup.norm_bf_std_dev
        charged_mixed_factor = self._fitting_setup.charged_mixed_factor
        bf = ratio_for_param * (norm_eff / eff) * norm_bf * charged_mixed_factor

        # calculate the errors on this
        upper_error = np.sqrt(
            ((norm_eff / eff) * norm_bf * ratio_errors[param_id]) ** 2
            + ((1 / eff) * ratio_for_param * norm_bf * upper_norm) ** 2
            + ((norm_eff / eff ** 2) * ratio_for_param * norm_bf * lower_eff) ** 2
            + ((norm_eff / eff) * ratio_for_param * norm_bf_error) ** 2
        )
        lower_error = np.sqrt(
            ((norm_eff / eff) * norm_bf * ratio_errors[param_id]) ** 2
            + ((1 / eff) * ratio_for_param * norm_bf * lower_norm) ** 2
            + ((norm_eff / eff ** 2) * ratio_for_param * norm_bf * upper_eff) ** 2
            + ((norm_eff / eff) * ratio_for_param * norm_bf_error) ** 2
        )

        return bf, upper_error, lower_error

    def calculate(self) -> None:
        self._yield_dict = self.get_yield_dict()
        assert self._yield_dict is not None

        self._yields = self.get_all_yields()

        self._ratios = self.calculate_ratios()

        for parameter in self._yield_parameters:
            bf, upper_error, lower_error = self.calculate_bf_for_parameter(yield_parameter=parameter)
            self._bf_dict[parameter] = {
                "branching_fraction": bf,
                "upper_bf_error": upper_error,
                "lower_bf_error": lower_error,
            }

    def reset(self) -> None:
        self._yield_dict = None
        self._yields = None
        self._ratios = None

        self._bf_dict = {}
        self._eff_dict = {}

    def save_yields_to_file(
        self,
        file_path: PathType,
    ) -> None:
        with open(file_path, "w") as fp:
            json.dump(self._yield_dict, fp, indent=2)

    def save_yields_to_latex(
        self,
        file_path: PathType,
        output_order: Tuple[str, ...],
    ) -> None:
        df = pd.DataFrame(columns=["process", "yields"])
        for process in output_order:
            process_name = self._process_latex_labels[process] if self._process_latex_labels else process
            yield_entry = self.get_yield_dict()[process]["yield"]
            assert isinstance(yield_entry, tuple)
            process_yield, error = yield_entry

            yield_value = "{:0.2f}".format(process_yield)
            uncert = "{:0.2f}".format(error)
            yield_str = rf"${yield_value} \pm {uncert}$"

            df = df.append(
                {"process": process_name, "yields": yield_str},
                ignore_index=True,
            )

        with open(file_path, "w") as f:
            with pd.option_context("max_colwidth", 1000):
                f.write(
                    df.to_latex(
                        columns=["process", "yields"],
                        index=False,
                        escape=False,
                        header=["Process", r"Yields"],
                        float_format="%.2f",
                        column_format="l|r",
                    )
                )

    def save_efficiencies_to_file(
        self,
        file_path: PathType,
    ) -> None:
        with open(file_path, "w") as fp:
            json.dump(self._eff_dict, fp, indent=2)

    def save_efficiencies_to_latex(
        self,
        file_path: PathType,
        output_order: Tuple[str, ...],
        scaling: float = 1e6,
    ) -> None:
        df = pd.DataFrame(columns=["process", "efficiency"])
        for process in output_order:
            process_name = self._process_latex_labels[process] if self._process_latex_labels else process
            efficiency, upper_error, lower_error = self.get_efficiency_from_dict(parameter=process)

            eff = "{:0.2f}".format(efficiency * scaling)
            uncert = "{:0.2f}".format(max(upper_error, lower_error) * scaling)
            eff_str = rf"${eff} \pm {uncert}$"

            df = df.append(
                {"process": process_name, "efficiency": eff_str},
                ignore_index=True,
            )

        with open(file_path, "w") as f:
            with pd.option_context("max_colwidth", 1000):
                f.write(
                    df.to_latex(
                        columns=["process", "efficiency"],
                        index=False,
                        escape=False,
                        header=["Process", r"Efficiency $\times 10^{-6}$"],
                        float_format="%.2f",
                        column_format="l|r",
                    )
                )

    def save_branching_fractions_to_file(
        self,
        file_path: PathType,
    ) -> None:
        with open(file_path, "w") as fp:
            json.dump(self._bf_dict, fp, indent=2)

    def get_bf_for_parameter(
        self,
        parameter: str,
    ) -> Tuple[float, float, float]:
        assert parameter in self._bf_dict.keys()
        bf = self._bf_dict[parameter]["branching_fraction"]
        upper_error = self._bf_dict[parameter]["upper_bf_error"]
        lower_error = self._bf_dict[parameter]["lower_bf_error"]

        return bf, upper_error, lower_error

    def eff_dict(self) -> Dict[str, Dict[str, float]]:
        return self._eff_dict

    def bf_dict(self) -> Dict[str, Dict[str, float]]:
        return self._bf_dict
