import os
import json
import copy
import pandas as pd
import b2luigi as luigi
import matplotlib.pyplot as plt

from luigi.util import inherits
from typing import Dict, Generator, Tuple, List, Optional, Union

from templatefitter.fitter import TemplateFitter
from templatefitter.fit_model.model_builder import FitModel
from templatefitter.plotter.histogram_variable import HistVariable
from templatefitter.plotter.fit_result_plots import FitResultPlotter

from rdstar.utilities.logging import log_to_default
from rdstar.utilities import PathType, merge_plot_pdf_files

from rdstar.offline_analysis.tasks import utilities
from rdstar.offline_analysis.tasks.logistics_tasks import DataCopyTask
from rdstar.offline_analysis.observables.common_observables import Cols
from rdstar.offline_analysis.mc_matching.sig_id_costomizer_tools import SigIDAndGenIDCombiner

from rdstar.offline_analysis.fitting.fit_case import FitCase
from rdstar.offline_analysis.fitting.fit_model_creator import FitModelCreator
from rdstar.offline_analysis.efficiency_calculation.simple_efficiency_calculator import SimpleEfficiencyCalculator

from rdstar.offline_analysis.dss_fitting.setup.dss_bf_fit_cases import BFFitCaseDict
from rdstar.offline_analysis.dss_fitting.utilities import utilities as dss_utils
from rdstar.offline_analysis.dss_fitting.tasks.preparation_task import FittingPreparationTask, CopyToSSDTask


__all__ = [
    "BFFittingTask",
]


logger = log_to_default(name=__name__)


@inherits(FittingPreparationTask)
class BFFittingTask(utilities.BaseOfflineTask):
    fit_case_name: str = luigi.Parameter(positional=False)
    plot_output: bool = luigi.BoolParameter(positional=False)
    mva_cut_info: Optional[float] = luigi.Parameter(positional=False)

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

        self.fit_case: FitCase = BFFitCaseDict.get_fit_case(fit_case_name=self.fit_case_name)

        if not self.mva_cut_info:
            self.base_dir: PathType = os.path.join(self.working_dir, "bf_fit_results", self.fit_case_name)
        else:
            self.base_dir = os.path.join(
                self.working_dir, "bf_fit_results", self.fit_case_name, f"mva_cut_{self.mva_cut_info}"
            )

        self.base_input_path: PathType = f"/ceph/fmetzner/rdstar/{self.production_key}_stream0/"

        self.file_dict_dir_path: PathType = DataCopyTask.get_local_file_dict_path_from_main_dir(
            main_work_dir_path=self.base_input_path
        )

        self.fit_output_tag = self.fit_case.output_tag

        self.mc_key = "combined_mc_prepared"

        self._input_dict: Optional[Dict[str, PathType]] = None

    @property
    def input_dict(self) -> Dict[str, PathType]:
        if self._input_dict is None:
            self._input_dict = {k: v.path for d in self.input() if isinstance(d, dict) for k, v in d.items()}
        return self._input_dict

    def get_data_path(
        self,
        data_key: str,
        sideband: Optional[str] = None,
    ) -> PathType:
        if not sideband:
            input_key = f"{data_key}_{self.selection}"
        else:
            input_key = f"{data_key}_{self.selection}_{sideband}"
        return self.input_dict[input_key]

    def get_fit_output_key_file_name_pair(self, file_tag: str) -> Tuple[str, str]:
        if not self.mva_cut_info:
            key = f"{self.fit_output_tag}_{file_tag}_{self.fit_case_name}"
        else:
            key = f"{self.fit_output_tag}_{file_tag}_{self.fit_case_name}_mva_cut_{self.mva_cut_info}"
        return key, f"{key}.json"

    def get_fit_plot_output_key_file_name_pair(self, file_tag: str) -> Tuple[str, str]:
        if not self.mva_cut_info:
            key = f"{self.fit_output_tag}_{file_tag}_{self.fit_case_name}"
        else:
            key = f"{self.fit_output_tag}_{file_tag}_{self.fit_case_name}_mva_cut_{self.mva_cut_info}"
        return key, f"{key}.pdf"

    def initialize_model_creator(self) -> Tuple[FitModelCreator, List[str]]:
        FitModelCreator.fit_cases = {self.fit_case.name: self.fit_case}

        model_creator = FitModelCreator(
            fit_case=self.fit_case.name,
        )
        all_yield_components = [
            model_creator.get_yield_parameter_name(component_name=comp) for comp in model_creator.all_components
        ]

        return model_creator, all_yield_components

    def apply_scope_cuts(self, df: pd.DataFrame) -> pd.DataFrame:
        query_str_list: List[str] = [self.fit_case.get_cut_query(reco_mode_column_str=Cols.decay_mode_column_name)]

        mva_cut_info = self.mva_cut_info
        if mva_cut_info is not None:
            assert isinstance(mva_cut_info, float), type(mva_cut_info)
            mva_variable = self.fit_case.get_additional_setting(setting_key="mva_variable_name")
            mva_cut_str = f"{mva_variable} >= {mva_cut_info}"
            query_str_list.append(mva_cut_str)

        query_str: str = " & ".join(query_str_list)
        return df.query(query_str, inplace=False).reset_index(drop=True, inplace=False)

    def generate_plot_output_dict(self) -> Tuple[Dict[str, PathType], str, str]:
        plot_output_dict: Dict[str, PathType] = dict()
        pre_fit_plot_key, pre_fit_plot_file_name = self.get_fit_plot_output_key_file_name_pair(file_tag="PreFitPlot")
        pre_fit_plot_path = os.path.join(self.base_dir, pre_fit_plot_file_name)
        plot_output_dict.update({pre_fit_plot_key: pre_fit_plot_path})

        post_fit_plot_key, post_fit_plot_file_name = self.get_fit_plot_output_key_file_name_pair(file_tag="PostFitPlot")
        post_fit_plot_path = os.path.join(self.base_dir, post_fit_plot_file_name)
        plot_output_dict.update({post_fit_plot_key: post_fit_plot_path})

        return plot_output_dict, pre_fit_plot_file_name, post_fit_plot_file_name

    def do_plots(
        self,
        fit_variables: Union[Dict[str, Tuple[HistVariable, ...]], Tuple[HistVariable, ...]],
        fit_model: FitModel,
        plot_file_name: str,
        use_initial_values: bool,
    ) -> None:
        plt.switch_backend("Agg")
        fit_plot_path = os.path.join(self.base_dir, plot_file_name)

        fit_result_plotter = FitResultPlotter(
            variables_by_channel=fit_variables,
            fit_model=fit_model,
            reference_dimension=0,
            channel_label_dict=self.fit_case.reco_mode_labels_dict,
            data_label=self.fit_case.plot_info_text,
        )

        if not self.mva_cut_info:
            output_name_tag = f"{self.fit_output_tag}_{self.fit_case_name}"
        else:
            output_name_tag = f"{self.fit_output_tag}_{self.fit_case_name}_mva_cut_{self.mva_cut_info}"
        pre_fit_plot_paths = fit_result_plotter.plot_fit_result(
            use_initial_values=use_initial_values,
            output_dir_path=self.base_dir,
            output_name_tag=output_name_tag,
        )
        assert len(pre_fit_plot_paths["pdf"]) == len(pre_fit_plot_paths["png"]), (
            len(pre_fit_plot_paths["pdf"]),
            len(pre_fit_plot_paths["png"]),
        )
        temp_fit_plot_path = os.path.join(self.base_dir, f"temp_{plot_file_name}")
        merge_plot_pdf_files(input_paths=pre_fit_plot_paths["pdf"], output_path=temp_fit_plot_path)
        os.rename(src=temp_fit_plot_path, dst=fit_plot_path)

    def get_toy_key_path_pair(self) -> Tuple[str, PathType]:
        if not self.mva_cut_info:
            toy_key = f"toy_yields_{self.fit_case_name}"
        else:
            toy_key = f"toy_yields_{self.fit_case_name}_mva_cut_{self.mva_cut_info}"
        return toy_key, os.path.join(self.base_dir, f"{toy_key}.json")

    def do_toy_study(
        self,
        df_mc: pd.DataFrame,
        iterations: int = 20,
    ) -> None:
        logger.debug("Performing Toy Study!")
        model_creator, all_yield_components = self.initialize_model_creator()
        model_creator.setup_initial_templates(mc_df=df_mc)

        yield_dict: Dict[str, Tuple[List, List]] = dict()
        for yield_parameter in all_yield_components:
            yield_dict.update({yield_parameter: ([], [])})

        toy_key, toy_path = self.get_toy_key_path_pair()
        intermediate_toy_path = os.path.join(self.base_dir, f"intermediate_{toy_key}.json")
        tmp_intermediate_toy_path = os.path.join(self.base_dir, f"tmp_intermediate_{toy_key}.json")
        tmp_toy_path = os.path.join(self.base_dir, f"tmp_{toy_key}.json")

        for i in range(iterations):
            logger.debug(f"Iteration Number {i+1} of {iterations} for fit case {self.fit_case_name}")
            new_model_creator = copy.deepcopy(model_creator)
            new_model_creator.add_data_to_fit_model_and_finalize(
                mc_df=df_mc,
                data_df=None,
                use_toy_data=True,
                round_toy_data_bin_counts=True,
            )
            fit_model = copy.deepcopy(new_model_creator.fit_model)

            fitter = TemplateFitter(fit_model=fit_model, minimizer_id="iminuit")
            for yield_parameter in all_yield_components:
                fitter.set_parameter_bounds(param_id=yield_parameter, bounds=(0, None))
            fit_result = fitter.do_fit(
                update_templates=True,
                get_hesse=True,
                verbose=False,
                fix_nui_params=self.fit_case.fix_nuisance_parameters,
            )
            for yield_parameter in all_yield_components:
                yield_value = fit_result.params.get_param_value(yield_parameter)
                yield_error = fit_result.params.get_param_error(yield_parameter)
                yield_dict[yield_parameter][0].append(yield_value)
                yield_dict[yield_parameter][1].append(yield_error)

            with open(tmp_intermediate_toy_path, "w") as fp:
                json.dump(yield_dict, fp, indent=2)
            os.rename(src=tmp_intermediate_toy_path, dst=intermediate_toy_path)

        with open(tmp_toy_path, "w") as fp:
            json.dump(yield_dict, fp, indent=2)
        os.rename(src=tmp_toy_path, dst=toy_path)

    def run_fit(self, dry_run: bool = False) -> Dict[str, PathType]:
        output_dict: Dict[str, PathType] = dict()

        fit_result_key, fit_result_file_name = self.get_fit_output_key_file_name_pair(file_tag="FitResult")
        temp_fit_result_path = os.path.join(self.base_dir, f"temp_{fit_result_file_name}")
        fit_result_path = os.path.join(self.base_dir, fit_result_file_name)
        output_dict.update({fit_result_key: fit_result_path})

        pre_fit_plot_file_name = None
        post_fit_plot_file_name = None
        if self.plot_output:
            plot_output_dict, pre_fit_plot_file_name, post_fit_plot_file_name = self.generate_plot_output_dict()
            output_dict.update(plot_output_dict)

        if not self.mva_cut_info:
            significance_key = f"significance_{self.fit_case_name}"
        else:
            significance_key = f"significance_{self.fit_case_name}__mva_cut_{self.mva_cut_info}"
        significance_path = os.path.join(self.base_dir, f"{significance_key}.json")
        if self.fit_case.get_additional_setting_with_default(setting_key="do_significance", default_setting=False):
            output_dict.update({significance_key: significance_path})

        efficiency_key = f"efficiency_{self.fit_case_name}"
        efficiency_path = os.path.join(self.base_dir, f"{efficiency_key}.json")
        output_dict.update({efficiency_key: efficiency_path})

        if self.fit_case.get_additional_setting_with_default(setting_key="do_toys", default_setting=False):
            toy_key, toy_path = self.get_toy_key_path_pair()
            output_dict.update({toy_key: toy_path})

        if not dry_run:
            sideband = self.fit_case.get_additional_setting_with_default(setting_key="sideband", default_setting=None)

            df_mc_full = pd.read_hdf(path_or_buf=self.get_data_path(data_key=self.mc_key, sideband=sideband))
            assert isinstance(df_mc_full, pd.DataFrame)
            df_mc_full.reset_index(drop=True, inplace=True)
            df_mc = self.apply_scope_cuts(df=df_mc_full)

            sig_id_customizing_dict = self.fit_case.get_additional_setting_with_default(
                setting_key="sig_id_customizing_dict",
                default_setting=None,
            )
            if sig_id_customizing_dict is not None:
                sig_id_customizer = SigIDAndGenIDCombiner(
                    customizing_dict=sig_id_customizing_dict,
                    sig_id_column_name=self.fit_case.sig_id_column_name,
                    decay_mode_dict=dss_utils.all_decay_modes,
                )
                sig_id_customizer.customize_sig_id(df=df_mc)

            assert not (self.fit_case.asimov_data and self.fit_case.sample_key)
            df_data = (
                pd.read_hdf(path_or_buf=self.get_data_path(data_key=self.fit_case.sample_key, sideband=sideband))
                if self.fit_case.sample_key and not self.fit_case.asimov_data
                else None
            )
            if df_data is not None:
                assert isinstance(df_data, pd.DataFrame)
                df_data.reset_index(drop=True, inplace=True)

            logger.debug("Setting up Fit Model...")
            model_creator, all_yield_components = self.initialize_model_creator()
            model_creator.setup_model(
                mc_df=df_mc,
                data_df=df_data,
                use_asimov_data=self.fit_case.asimov_data,
                round_asimov_data_bin_counts=self.fit_case.round_asimov_data,
            )
            fit_model = model_creator.fit_model
            logger.debug("Fit Model set up!")

            if self.plot_output:
                assert pre_fit_plot_file_name is not None
                self.do_plots(
                    fit_variables=model_creator.hist_variables_by_mode_dict,
                    fit_model=fit_model,
                    plot_file_name=pre_fit_plot_file_name,
                    use_initial_values=True,
                )

            fitter = TemplateFitter(fit_model=fit_model, minimizer_id="iminuit")
            for yield_parameter in all_yield_components:
                fitter.set_parameter_bounds(param_id=yield_parameter, bounds=(0, None))
            fit_result = fitter.do_fit(
                update_templates=True,
                get_hesse=True,
                verbose=True,
                fix_nui_params=self.fit_case.fix_nuisance_parameters,
            )
            fit_result.save_to(path=temp_fit_result_path, overwrite=True)
            os.rename(src=temp_fit_result_path, dst=fit_result_path)
            logger.debug("Fit performed!")

            if self.plot_output:
                assert post_fit_plot_file_name is not None
                self.do_plots(
                    fit_variables=model_creator.hist_variables_by_mode_dict,
                    fit_model=fit_model,
                    plot_file_name=post_fit_plot_file_name,
                    use_initial_values=False,
                )

            if self.fit_case.get_additional_setting_with_default(setting_key="do_significance", default_setting=False):
                logger.debug("Calculate Significance...")
                significance_dict: Dict[str, float] = dict()
                for yield_parameter in self.fit_case.all_components:
                    if yield_parameter in self.fit_case.fix_component_yields:
                        continue
                    try:
                        significance: float = fitter.get_significance(
                            yield_parameter=f"yield_{yield_parameter}",
                            fix_nui_params=True,
                            catch_exception=True,
                        )
                    except IndexError as ie:
                        logger.warning(
                            f"Catching IndexError in significance calculation for yield parameter "
                            f"'{yield_parameter}'. The error was:\n{str(ie)}"
                        )
                        significance = -1.0
                    significance_dict.update({yield_parameter: significance})

                tmp_significance_path = os.path.join(self.base_dir, f"tmp_{significance_key}.json")
                with open(tmp_significance_path, "w") as fp:
                    json.dump(significance_dict, fp, indent=2)
                os.rename(src=tmp_significance_path, dst=significance_path)

            if self.fit_case.get_additional_setting_with_default(setting_key="do_toys", default_setting=False):
                self.do_toy_study(
                    df_mc=df_mc,
                    iterations=500,
                )

            efficiency_calculator = SimpleEfficiencyCalculator(
                df=df_mc,
                sig_id_data_type_mapping=self.fit_case.sig_id_to_data_type_mapping,
                sig_id_column_name=self.fit_case.sig_id_column_name,
                file_dict_base_path=self.file_dict_dir_path,
            )
            efficiency_calculator.write_eff_dict_to_file(file_path=efficiency_path)

        return output_dict

    def run(self) -> None:
        utilities.host_check(hosts=utilities.GlobalParameters().work_host_list)
        self.run_fit()

    def output(self) -> Dict[str, luigi.LocalTarget]:
        output_dict = dict()
        for output_key, target_path in self.run_fit(dry_run=True).items():
            output_dict.update({output_key: luigi.LocalTarget(target_path)})
        return output_dict

    def requires(self) -> Generator[utilities.TaskType, None, None]:
        yield self.clone(CopyToSSDTask)
