import os
import json
import numpy as np
import b2luigi as luigi
import matplotlib.pyplot as plt

from luigi.util import inherits
from typing import Dict, Generator, Optional, Tuple, List

from templatefitter.plotter import plot_style
from templatefitter.plotter.plot_utilities import export

from rdstar.utilities import PathType
from rdstar.offline_analysis.tasks import utilities

from rdstar.offline_analysis.dss_fitting.tasks.bf_fitting_task import BFFittingTask
from rdstar.offline_analysis.dss_fitting.tasks.preparation_task import FittingPreparationTask
from rdstar.offline_analysis.dss_fitting.setup.dss_bf_fit_cases import BFFitCaseDict


__all__ = [
    "BFToyEvaluationTask",
]


@inherits(FittingPreparationTask)
class BFToyEvaluationTask(utilities.BaseOfflineTask):
    fit_case_name = luigi.Parameter(positional=False)  # type: str
    plot_output = luigi.BoolParameter(positional=False)  # type: bool
    mva_cut_info = luigi.Parameter(positional=False)  # type: Optional[float]

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

        if not self.mva_cut_info:
            self.base_dir = os.path.join(self.working_dir, "bf_toy_results", self.fit_case_name)  # type: PathType
        else:
            self.base_dir = os.path.join(
                self.working_dir, "bf_toy_results", self.fit_case_name, f"mva_cut_{self.mva_cut_info}"
            )

        self.fit_case = BFFitCaseDict.get_fit_case(fit_case_name=self.fit_case_name)
        self.fit_output_tag = self.fit_case.output_tag

        self.input_dict = {k: v.path for d in self.input() if isinstance(d, dict) for k, v in d.items()}

    def read_yields(self) -> Dict[str, Tuple[List, List]]:
        if not self.mva_cut_info:
            toy_key = f"toy_yields_{self.fit_case_name}"
        else:
            toy_key = f"toy_yields_{self.fit_case_name}_mva_cut_{self.mva_cut_info}"
        with open(self.input_dict[toy_key]) as f:
            yield_dict = json.load(f)
        return yield_dict

    def get_output_dict(self) -> Dict[str, PathType]:
        output_dict = dict()  # type: Dict[str, PathType]
        for component in self.fit_case.all_components:
            yield_key = f"toy_evaluation_yields_{component}"
            yield_path = os.path.join(self.base_dir, f"{yield_key}.pdf")
            output_dict.update({yield_key: yield_path})
        return output_dict

    @staticmethod
    def gauss(x: np.ndarray, mean: float, sigma: float) -> np.ndarray:
        return 1 / (np.sqrt(2 * np.pi) * sigma) * np.exp(-0.5 * (x - mean) ** 2 / sigma ** 2)

    def do_plots(
        self,
        dry_run: bool = False,
    ) -> Dict[str, PathType]:
        output_dict = self.get_output_dict()  # type: Dict[str, PathType]

        if not dry_run:
            yield_dict = self.read_yields()

            for component in self.fit_case.all_components:
                yield_values, yield_errors = yield_dict[f"yield_{component}"]

                plot_style.set_matplotlibrc_params()
                fig, ax = plt.subplots(figsize=(6.5, 5))

                minimum = np.min(yield_values)
                maximum = np.max(yield_values)
                mean = np.mean(yield_values)
                std = np.std(yield_values)
                std_fit = np.mean(yield_errors)

                x = np.linspace(minimum, maximum, num=1000)

                color = self.fit_case.plot_colors[component]
                if color == plot_style.KITColors.white:
                    color = plot_style.KITColors.grey
                ax.hist(
                    np.array(yield_values),
                    bins=50,
                    color=color,
                    density=True,
                    range=(minimum, maximum),
                    alpha=0.6,
                )

                ax.plot(
                    x,
                    self.gauss(x, mean=float(mean), sigma=float(std)),
                    color=color,
                    label="Std from fit results",
                    lw=2,
                )
                ax.plot(
                    x,
                    self.gauss(x, mean=float(mean), sigma=float(std_fit)),
                    color=plot_style.KITColors.kit_red,
                    label="Mean of fitted stds",
                    lw=2,
                )

                ax.set_xlabel(r"Yield", plot_style.xlabel_pos)
                fig.set_tight_layout(True)
                # Plot the figure once, then append the offset text to the x-axis label
                # Hide the offset text as we no longer need it, then redraw the figure
                fig.savefig(os.path.join(self.base_dir, f"toy_evaluation_yields_{component}.pdf"))
                offset = ax.get_xaxis().get_offset_text()
                ax.set_xlabel("{0} {1}".format(ax.get_xlabel(), offset.get_text()))
                ax.set_ylabel(r"Normalized in arb. units", plot_style.ylabel_pos)
                ax.set_title(self.fit_case.plot_labels[component], loc="right")
                offset.set_visible(False)

                ax.text(
                    0.95,
                    0.95,
                    "Mean = {:0.1f}".format(mean),
                    color=plot_style.KITColors.kit_black,
                    va="center",
                    ha="right",
                    transform=ax.transAxes,
                )
                ax.text(
                    0.95,
                    0.90,
                    "Std = {:0.1f}".format(std),
                    color=plot_style.KITColors.kit_black,
                    va="center",
                    ha="right",
                    transform=ax.transAxes,
                )
                ax.text(
                    0.95,
                    0.85,
                    "Mean of fit stds = {:0.1f}".format(std_fit),
                    color=plot_style.KITColors.kit_black,
                    va="center",
                    ha="right",
                    transform=ax.transAxes,
                )
                plt.legend(loc=2)

                export(fig=fig, filename=f"toy_evaluation_yields_{component}", target_dir=self.base_dir)

        return output_dict

    def run_toy_evaluation(self, dry_run: bool = False) -> Dict[str, PathType]:

        output_dict = dict()  # type: Dict[str, PathType]

        plot_output_dict = self.do_plots(
            dry_run=dry_run,
        )
        output_dict.update(plot_output_dict)

        return output_dict

    def run(self) -> None:
        utilities.host_check(hosts=utilities.GlobalParameters().work_host_list)
        self.run_toy_evaluation()

    def output(self) -> Dict[str, luigi.LocalTarget]:
        output_dict = dict()
        for output_key, target_path in self.run_toy_evaluation(dry_run=True).items():
            output_dict.update({output_key: luigi.LocalTarget(target_path)})
        return output_dict

    def requires(self) -> Generator[utilities.TaskType, None, None]:
        yield self.clone(
            BFFittingTask,
            fit_case_name=self.fit_case_name,
            plot_output=self.plot_output,
            mva_cut_info=self.mva_cut_info,
        )
