import os
import numpy as np
import pandas as pd
import b2luigi as luigi
import matplotlib.pyplot as plt

from luigi.util import inherits
from typing import Optional, Dict, Generator

from templatefitter.plotter import plot_style
from templatefitter.plotter.plot_utilities import export
from templatefitter.plotter.plot_style import set_matplotlibrc_params

from rdstar.utilities import PathType
from rdstar.utilities.logging import log_to_default
from rdstar.offline_analysis.tasks import utilities

from rdstar.offline_analysis.fitting.fit_case import TemplateSetup
from rdstar.offline_analysis.observables.q_squared_residual import QSquaredResidualVariable

from rdstar.offline_analysis.dss_fitting.tasks.preparation_task import FittingPreparationTask, CopyToSSDTask
from rdstar.offline_analysis.dss_fitting.setup.fitting_setup import FittingSetupCollection


__all__ = [
    "MigrationTask",
]


logger = log_to_default(name=__name__)


@inherits(FittingPreparationTask)
class MigrationTask(utilities.BaseOfflineTask):
    fitting_setup_name = luigi.Parameter(positional=False)  # type: str

    decay_mode_column_name = "daughter__bo1__cm__spextraInfo__bodecayModeID__bc__bc"  # type: str

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

        self.base_dir = os.path.join(self.working_dir, "migration", self.fitting_setup_name)

        self.fitting_setup = FittingSetupCollection.available_setups[self.fitting_setup_name]
        self.fit_case = self.fitting_setup.fit_case_dict.pivot_fit_case

        self.variable = self.fitting_setup.fit_in_bins_of

        assert self.fit_case.plot_labels is not None
        self.yield_parameter_labels = self.fit_case.plot_labels

        self.mc_key = "combined_mc"
        self.q_squared_truth_variable = QSquaredResidualVariable.q_squared_truth_col_name

        self.all_mode_ids = [d_id for d_ids in [d_ids for _, d_ids in self.fit_case.reco_mode_infos] for d_id in d_ids]

        self._input_dict = None  # type: Optional[Dict[str, PathType]]

    @property
    def input_dict(self) -> Dict[str, PathType]:
        if self._input_dict is None:
            self._input_dict = {k: v.path for d in self.input() if isinstance(d, dict) for k, v in d.items()}
        return self._input_dict

    def generate_output_dict(self) -> Dict[str, PathType]:
        output_dict = dict()  # type: Dict[str, PathType]
        for yield_param in self.fitting_setup.ratio_yield_parameters:
            output_key = f"{self.fitting_setup_name}_{yield_param}_migration_matrix"
            output_path = os.path.join(self.base_dir, f"{output_key}.pdf")
            output_dict.update({output_key: output_path})
        return output_dict

    def get_mc_yields(self, df: pd.DataFrame, yield_parameter: str) -> np.ndarray:
        assert isinstance(self.fit_case.template_setup, TemplateSetup)
        template_setup = self.fit_case.template_setup.template_setup
        if isinstance(template_setup, Dict):
            sig_ids = tuple([self.fit_case.sig_queries[comp].id for comp in template_setup[yield_parameter]])
        else:
            sig_ids = (self.fit_case.sig_queries[yield_parameter].id,)
        sig_id_condition = df[self.fit_case.sig_id_column_name].isin(sig_ids)
        df_mc = df.loc[sig_id_condition]

        scopes = self.fitting_setup.fit_in_bins_of_scopes

        dim = self.variable.n_bins
        mc_yields = np.zeros((dim, dim))
        for i in range(0, dim):
            df_cut = df_mc.query(f"{scopes[i]} <= {self.variable.df_label} < {scopes[i + 1]}")
            for j in range(0, dim):
                df_truth_cut = df_cut.query(f"{scopes[j]} <= {self.q_squared_truth_variable} < {scopes[j + 1]}")
                mc_yields[i][j] = np.sum(df_truth_cut[self.fit_case.weight_column_name])
            mc_yields[i] /= np.sum(mc_yields[i])

        return mc_yields

    def run_migration_calculation(self, dry_run: bool = False) -> Dict[str, PathType]:
        output_dict = self.generate_output_dict()

        if not dry_run:
            path = self.input_dict[f"{self.mc_key}_prepared_{self.selection}"]
            df = pd.read_hdf(path_or_buf=path)
            assert isinstance(df, pd.DataFrame)

            df_mc = df.loc[df[self.decay_mode_column_name].isin(self.all_mode_ids)]

            for yield_param in self.fitting_setup.ratio_yield_parameters:
                mc_yields = self.get_mc_yields(df=df_mc, yield_parameter=yield_param)
                logger.debug(mc_yields)

                set_matplotlibrc_params()
                fig, ax = plt.subplots()
                ax.imshow(mc_yields, cmap="GnBu")

                for i in range(self.variable.n_bins):
                    for j in range(self.variable.n_bins):
                        ax.text(
                            j,
                            i,
                            round(mc_yields[i, j], 2),
                            ha="center",
                            va="center",
                            color=plot_style.KITColors.kit_black,
                        )

                ax.set_title(self.yield_parameter_labels[yield_param], loc="right")
                ax.set_ylabel(r"$\mathrm{q}^2_\mathrm{Reco}$ Bin", plot_style.ylabel_pos)
                ax.set_xlabel(r"$\mathrm{q}^2_\mathrm{Truth}$ Bin", plot_style.xlabel_pos)

                export(
                    fig=fig,
                    filename=f"{self.fitting_setup_name}_{yield_param}_migration_matrix",
                    target_dir=self.base_dir,
                )

        return output_dict

    def run(self) -> None:
        utilities.host_check(hosts=utilities.GlobalParameters().work_host_list)
        self.run_migration_calculation()

    def output(self) -> Dict[str, luigi.LocalTarget]:
        output_dict = dict()
        for output_key, target_path in self.run_migration_calculation(dry_run=True).items():
            output_dict.update({output_key: luigi.LocalTarget(target_path)})
        return output_dict

    def requires(self) -> Generator[utilities.TaskType, None, None]:
        yield self.clone(CopyToSSDTask)
