import os
import json
import b2luigi as luigi

from luigi.util import inherits
from typing import Optional, Dict, Generator, Tuple

from rdstar.utilities import PathType
from rdstar.offline_analysis.tasks import utilities

from rdstar.offline_analysis.dss_fitting.tasks.preparation_task import FittingPreparationTask
from rdstar.offline_analysis.dss_fitting.tasks.bf_calculation_task import BFCalculationTask
from rdstar.offline_analysis.dss_fitting.tasks.bf_fitting_task import BFFittingTask
from rdstar.offline_analysis.dss_fitting.setup.bf_fitting_setup import BFFittingSetupCollection
from rdstar.offline_analysis.dss_fitting.mva_optimizer import MVABFDictType, MVAOptimizer


__all__ = [
    "MVAOptimizationTask",
]


@inherits(FittingPreparationTask)
class MVAOptimizationTask(utilities.BaseOfflineTask):
    fitting_setup_name = luigi.Parameter(positional=False)  # type: str

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

        self.base_dir = os.path.join(self.working_dir, "mva_optimization", self.fitting_setup_name)  # type: PathType

        self.fitting_setup = BFFittingSetupCollection.available_setups[self.fitting_setup_name]

        self.fit_case = self.fitting_setup.fit_case

        self.fit_output_tag = self.fit_case.output_tag

        self.mva_cut_infos = (
            0.1,
            0.2,
            0.25,
            0.3,
            0.35,
            0.4,
            0.45,
            0.5,
            0.55,
            0.6,
            0.65,
            0.7,
            0.75,
            0.8,
        )  # type: Tuple[float, ...]

        self.yield_parameters = self.fitting_setup.yield_parameters

        self._input_dict = None  # type: Optional[Dict[str, PathType]]

    @property
    def input_dict(self) -> Dict[str, PathType]:
        if self._input_dict is None:
            self._input_dict = {k: v.path for d in self.input() if isinstance(d, dict) for k, v in d.items()}
        return self._input_dict

    def create_bf_path(self, yield_parameter: str) -> Tuple[str, PathType]:
        bf_key = f"mva_branching_fractions_{yield_parameter}"
        bf_path = os.path.join(self.base_dir, f"{bf_key}.pdf")
        return bf_key, bf_path

    def create_bf_error_path(self, yield_parameter: str) -> Tuple[str, PathType]:
        bf_err_key = f"mva_branching_fractions_errors_{yield_parameter}"
        bf_err_path = os.path.join(self.base_dir, f"{bf_err_key}.pdf")
        return bf_err_key, bf_err_path

    def create_significance_path(self, yield_parameter: str) -> Tuple[str, PathType]:
        sig_key = f"mva_significance_{yield_parameter}"
        sig_path = os.path.join(self.base_dir, f"{sig_key}.pdf")
        return sig_key, sig_path

    def create_output_dict(self) -> Dict[str, PathType]:
        output_dict = dict()  # type: Dict[str, PathType]
        for yield_parameter in self.yield_parameters:
            bf_key, bf_path = self.create_bf_path(yield_parameter=yield_parameter)
            output_dict.update({bf_key: bf_path})
            bf_err_key, bf_err_path = self.create_bf_error_path(yield_parameter=yield_parameter)
            output_dict.update({bf_err_key: bf_err_path})
            sig_key, sig_path = self.create_significance_path(yield_parameter=yield_parameter)
            output_dict.update({sig_key: sig_path})
        return output_dict

    def run_mva_optimization(self, dry_run: bool = False) -> Dict[str, PathType]:
        output_dict = self.create_output_dict()

        bf_dict = dict()  # type: MVABFDictType
        significance_dict = dict()  # type: Dict[float, Dict[str, float]]
        if not dry_run:
            for mva_cut_info in self.mva_cut_infos:
                bf_path = self.input_dict[f"branching_fractions_{self.fitting_setup_name}_mva_cut_{mva_cut_info}"]
                with open(bf_path, "r") as read_file:
                    branching_fractions = json.load(read_file)
                    bf_dict.update({mva_cut_info: branching_fractions})
                sig_path = self.input_dict[f"significance_{self.fitting_setup_name}__mva_cut_{mva_cut_info}"]
                with open(sig_path, "r") as read_file:
                    significance = json.load(read_file)
                    significance_dict.update({mva_cut_info: significance})

            mva_optimizer = MVAOptimizer(
                mva_cut_infos=self.mva_cut_infos,
                bf_dict=bf_dict,
                significance_dict=significance_dict,
                fitting_setup=self.fitting_setup,
            )
            for yield_parameter in self.yield_parameters:
                bf_key, _ = self.create_bf_path(yield_parameter=yield_parameter)
                mva_optimizer.plot_branching_fractions(
                    yield_parameter=yield_parameter,
                    filename=bf_key,
                    output_dir=self.base_dir,
                )
                bf_err_key, _ = self.create_bf_error_path(yield_parameter=yield_parameter)
                mva_optimizer.plot_branching_fraction_errors(
                    yield_parameter=yield_parameter,
                    filename=bf_err_key,
                    output_dir=self.base_dir,
                )
                sig_key, _ = self.create_significance_path(yield_parameter=yield_parameter)
                mva_optimizer.plot_significance(
                    yield_parameter=yield_parameter,
                    filename=sig_key,
                    output_dir=self.base_dir,
                )

        return output_dict

    def run(self) -> None:
        utilities.host_check(hosts=utilities.GlobalParameters().work_host_list)
        self.run_mva_optimization()

    def output(self) -> Dict[str, luigi.LocalTarget]:
        output_dict = dict()
        for output_key, target_path in self.run_mva_optimization(dry_run=True).items():
            output_dict.update({output_key: luigi.LocalTarget(target_path)})
        return output_dict

    def requires(self) -> Generator[utilities.TaskType, None, None]:
        for mva_cut_info in self.mva_cut_infos:
            yield self.clone(BFCalculationTask, plot_output=False, mva_cut_info=mva_cut_info)
            yield self.clone(BFFittingTask, plot_output=False, mva_cut_info=mva_cut_info)
