import os
import itertools
import pandas as pd
import b2luigi as luigi

from luigi.util import inherits
from typing import Optional, Dict, Generator, Tuple

from rdstar.utilities import PathType
from rdstar.offline_analysis.tasks import utilities

from rdstar.offline_analysis.dss_fitting.setup.fitting_setup import FittingSetupCollection
from rdstar.offline_analysis.dss_fitting.tasks.preparation_task import FittingPreparationTask, CopyToSSDTask


__all__ = [
    "SampleSplitterTask",
]


@inherits(FittingPreparationTask)
class SampleSplitterTask(utilities.BaseOfflineTask):
    fitting_setup_name = luigi.Parameter(positional=False)  # type: str

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

        self.base_dir = os.path.join(self.working_dir, "samples", self.fitting_setup_name)

        self.fitting_setup = FittingSetupCollection.available_setups[self.fitting_setup_name]

        self.mc_key = "combined_mc"
        self.data_key = "data"

        self._input_dict = None  # type: Optional[Dict[str, PathType]]

    @property
    def input_dict(self) -> Dict[str, PathType]:
        if self._input_dict is None:
            self._input_dict = {k: v.path for d in self.input() if isinstance(d, dict) for k, v in d.items()}
        return self._input_dict

    def generate_split_dict(self) -> Dict[int, str]:
        split_dict = {}
        scopes = self.fitting_setup.fit_in_bins_of_scopes
        variable = self.fitting_setup.fit_in_bins_of.df_label
        for i in range(0, self.fitting_setup.fit_in_bins_of.n_bins):
            split_str = f"{scopes[i]} <= {variable} < {scopes[i + 1]}"
            split_dict.update({i: split_str})
        assert len(split_dict) == self.fitting_setup.fit_in_bins_of.n_bins
        return split_dict

    def generate_output_key_path_pair(self, bin_number: int, data_key: str) -> Tuple[str, PathType]:
        output_key = f"splitted_{data_key}_{self.selection}_{self.fitting_setup_name}_bin_{bin_number}"
        output_path = os.path.join(self.base_dir, f"{output_key}.h5")
        return output_key, output_path

    def perform_split(self, dry_run: bool = False) -> Dict[str, PathType]:
        split_dict = self.generate_split_dict()
        output_dict = {
            k: p
            for k, p in [
                self.generate_output_key_path_pair(bin_number=i, data_key=key)
                for i, key in itertools.product(split_dict.keys(), [self.data_key, self.mc_key])
            ]
        }

        if not dry_run:
            for key in [self.mc_key, self.data_key]:
                path = self.input_dict[f"{key}_prepared_{self.selection}"]
                df = pd.read_hdf(path_or_buf=path)
                assert isinstance(df, pd.DataFrame)

                for bin_number in split_dict.keys():
                    df_split = df.query(split_dict[bin_number])
                    df_split.reset_index(drop=True, inplace=True)

                    output_key, output_path = self.generate_output_key_path_pair(bin_number=bin_number, data_key=key)
                    tmp_output_path = os.path.join(self.base_dir, f"tmp_{output_key}.h5")

                    df_split.to_hdf(path_or_buf=tmp_output_path, key="fitting")
                    os.rename(src=tmp_output_path, dst=output_path)

        return output_dict

    def run(self) -> None:
        utilities.host_check(hosts=utilities.GlobalParameters().work_host_list)
        self.perform_split()

    def output(self) -> Dict[str, luigi.LocalTarget]:
        output_dict = dict()
        for output_key, target_path in self.perform_split(dry_run=True).items():
            output_dict.update({output_key: luigi.LocalTarget(target_path)})
        return output_dict

    def requires(self) -> Generator[utilities.TaskType, None, None]:
        yield self.clone(CopyToSSDTask)
