import pandas as pd

from typing import List, Dict, Tuple

from templatefitter.plotter.plot_style import KITColors
from templatefitter.plotter.histogram_variable import HistVariable

from rdstar.latex_labels.latex_units import LatexUnits
from rdstar.latex_labels.latex_observable_names import LatexObsNames

from rdstar.offline_analysis.mc_matching.mc_matching_utilities import MatchQueryInfo, MCSigIDNameToQueryMappingType

from rdstar.offline_analysis.fitting.fit_case import FitVariable
from rdstar.offline_analysis.fitting.fit_case import TemplateSetup
from rdstar.offline_analysis.mms_resolution_correction.resolution_correction import MissingMassResolutionCorrector


advanced_dss_components: MCSigIDNameToQueryMappingType = {
    "DPiZ": MatchQueryInfo(id_info=101, query=lambda df: pd.Series(data=True, index=df.index), latex_label=str(101)),
    "DSPiZ": MatchQueryInfo(id_info=102, query=lambda df: pd.Series(data=True, index=df.index), latex_label=str(102)),
    "DPiP": MatchQueryInfo(id_info=103, query=lambda df: pd.Series(data=True, index=df.index), latex_label=str(103)),
    "DSPiP": MatchQueryInfo(id_info=104, query=lambda df: pd.Series(data=True, index=df.index), latex_label=str(104)),
    "Dl": MatchQueryInfo(id_info=105, query=lambda df: pd.Series(data=True, index=df.index), latex_label=str(105)),
    "DSl": MatchQueryInfo(id_info=106, query=lambda df: pd.Series(data=True, index=df.index), latex_label=str(106)),
    "BKG": MatchQueryInfo(id_info=107, query=lambda df: pd.Series(data=True, index=df.index), latex_label=str(107)),
}

reduced_dss_labels: Dict[str, str] = {
    "BKG": "BKG",
    "Dl": r"$D^{(*)} \ell \nu$",
    "DSSPiP": r"$D^{**}(\rightarrow D^{(*)} \pi^\pm) \ell \nu$",
    "DSPiZ": r"$D^{**}(\rightarrow D^{*} \pi^0) \ell \nu$",
    "DPiZ": r"$D^{**}(\rightarrow D \pi^0) \ell \nu$",
    "DSSPiZ": r"$D^{**}(\rightarrow D^{(*)} \pi^0) \ell \nu$",
    "DSPiP": r"$D^{**}(\rightarrow D^{*} \pi^\pm) \ell \nu$",
    "DPiP": r"$D^{**}(\rightarrow D \pi^\pm) \ell \nu$",
}

reduced_dss_colors: Dict[str, str] = {
    "BKG": KITColors.light_grey,
    "Dl": KITColors.kit_purple,
    "DSSPiP": KITColors.kit_orange,
    "DSPiZ": KITColors.kit_maygreen,
    "DPiZ": KITColors.kit_green,
    "DSSPiZ": KITColors.kit_red,
    "DSPiP": KITColors.kit_cyan,
    "DPiP": KITColors.kit_blue,
}

reduced_dss_mapping: Dict[str, List[str]] = {
    "BKG": ["BKG"],
    "Dl": ["Dl", "DSl"],
    "DSSPiP": ["DPiP", "DSPiP"],
    "DSPiZ": ["DSPiZ"],
    "DPiZ": ["DPiZ"],
    "DSSPiZ": ["DPiZ", "DSPiZ"],
    "DSPiP": ["DSPiP"],
    "DPiP": ["DPiP"],
}

reduced_advanced_dss_plot_order_for_DPiZ: Dict[str, List[str]] = {
    "BKG": ["BKG"],
    "Dl": ["Dl", "DSl"],
    "DSSPiP": ["DPiP", "DSPiP"],
    "DSPiZ": ["DSPiZ"],
    "DPiZ": ["DPiZ"],
}
dss_ff_template_setup_for_DPiZ = TemplateSetup(template_setup=reduced_advanced_dss_plot_order_for_DPiZ)

reduced_advanced_dss_plot_order_for_DSPiZ: Dict[str, List[str]] = {
    "BKG": ["BKG"],
    "Dl": ["Dl", "DSl"],
    "DSSPiP": ["DPiP", "DSPiP"],
    "DPiZ": ["DPiZ"],
    "DSPiZ": ["DSPiZ"],
}
dss_ff_template_setup_for_DSPiZ = TemplateSetup(template_setup=reduced_advanced_dss_plot_order_for_DSPiZ)

reduced_advanced_dss_plot_order_for_DPiP: Dict[str, List[str]] = {
    "BKG": ["BKG"],
    "Dl": ["Dl", "DSl"],
    "DSSPiZ": ["DPiZ", "DSPiZ"],
    "DSPiP": ["DSPiP"],
    "DPiP": ["DPiP"],
}
dss_ff_template_setup_for_DPiP = TemplateSetup(template_setup=reduced_advanced_dss_plot_order_for_DPiP)

reduced_advanced_dss_plot_order_for_DSPiP: Dict[str, List[str]] = {
    "BKG": ["BKG"],
    "Dl": ["Dl", "DSl"],
    "DSSPiZ": ["DPiZ", "DSPiZ"],
    "DPiP": ["DPiP"],
    "DSPiP": ["DSPiP"],
}
dss_ff_template_setup_for_DSPiP = TemplateSetup(template_setup=reduced_advanced_dss_plot_order_for_DSPiP)

missing_mass_for_ff_fitting = HistVariable(
    df_label=MissingMassResolutionCorrector.get_main_single_stage_correction_column_name(
        use_laplace_smeared_version=False
    ),
    n_bins=24,
    scope=(-0.5, 4.0),
    var_name=LatexObsNames.fitting_mm2,
    unit=LatexUnits.gev_sq,
)

delta_M_for_ff_fitting_hvar = HistVariable(
    df_label="daughter__bo1__cm__spformula__bodaughterInvM__bo0__cm__sp2__bc__sp__mi__spdaughter__bo0__cm__spInvM__bc__bc__bc",
    n_bins=30,
    scope=(0.2, 0.9),
    var_name=LatexObsNames.d_st_st_delta_m,
    unit=LatexUnits.gev,
)

ff_study_fit_variables_dict: Dict[str, Tuple[FitVariable, ...]] = {
    "DZ": (FitVariable(variable=missing_mass_for_ff_fitting, alternative_binning=30, alternative_scope=None),),
    "DM": (FitVariable(variable=missing_mass_for_ff_fitting, alternative_binning=30, alternative_scope=None),),
    "DSZ": (FitVariable(variable=missing_mass_for_ff_fitting, alternative_binning=30, alternative_scope=None),),
    "DSM": (FitVariable(variable=missing_mass_for_ff_fitting, alternative_binning=30, alternative_scope=None),),
    "DZPiZ": (FitVariable(variable=missing_mass_for_ff_fitting, alternative_binning=25, alternative_scope=None),),
    "DMPiZ": (FitVariable(variable=missing_mass_for_ff_fitting, alternative_binning=20, alternative_scope=None),),
    "DSZPiZ": (FitVariable(variable=missing_mass_for_ff_fitting, alternative_binning=20, alternative_scope=None),),
    "DSMPiZ": (FitVariable(variable=missing_mass_for_ff_fitting, alternative_binning=20, alternative_scope=None),),
    "DZPiM": (FitVariable(variable=missing_mass_for_ff_fitting, alternative_binning=30, alternative_scope=None),),
    "DPPiM": (FitVariable(variable=missing_mass_for_ff_fitting, alternative_binning=30, alternative_scope=None),),
    "DSZPiM": (FitVariable(variable=missing_mass_for_ff_fitting, alternative_binning=25, alternative_scope=None),),
    "DSPPiM": (FitVariable(variable=missing_mass_for_ff_fitting, alternative_binning=20, alternative_scope=None),),
}

ff_study_fit_variables_dict_opt_1: Dict[str, Tuple[FitVariable, ...]] = {
    "DZ": (FitVariable(variable=missing_mass_for_ff_fitting, alternative_binning=30, alternative_scope=(-2.0, 2.0)),),
    "DM": (FitVariable(variable=missing_mass_for_ff_fitting, alternative_binning=30, alternative_scope=(-2.0, 2.0)),),
    "DSZ": (FitVariable(variable=missing_mass_for_ff_fitting, alternative_binning=30, alternative_scope=(-2.0, 2.0)),),
    "DSM": (FitVariable(variable=missing_mass_for_ff_fitting, alternative_binning=30, alternative_scope=(-2.0, 2.0)),),
    "DZPiZ": (FitVariable(variable=missing_mass_for_ff_fitting, alternative_binning=25, alternative_scope=(-2.0, 2.0)),),
    "DMPiZ": (FitVariable(variable=missing_mass_for_ff_fitting, alternative_binning=20, alternative_scope=(-2.0, 2.0)),),
    "DSZPiZ": (FitVariable(variable=missing_mass_for_ff_fitting, alternative_binning=20, alternative_scope=(-2.0, 2.0)),),
    "DSMPiZ": (FitVariable(variable=missing_mass_for_ff_fitting, alternative_binning=20, alternative_scope=(-2.0, 2.0)),),
    "DZPiM": (FitVariable(variable=missing_mass_for_ff_fitting, alternative_binning=30, alternative_scope=(-2.0, 2.0)),),
    "DPPiM": (FitVariable(variable=missing_mass_for_ff_fitting, alternative_binning=30, alternative_scope=(-2.0, 2.0)),),
    "DSZPiM": (FitVariable(variable=missing_mass_for_ff_fitting, alternative_binning=25, alternative_scope=(-2.0, 2.0)),),
    "DSPPiM": (FitVariable(variable=missing_mass_for_ff_fitting, alternative_binning=20, alternative_scope=(-2.0, 2.0)),),
}

ff_study_fit_variables_dict_opt_2: Dict[str, Tuple[FitVariable, ...]] = {
    "DZ": (FitVariable(variable=missing_mass_for_ff_fitting, alternative_binning=30, alternative_scope=(-2.0, 1.0)),),
    "DM": (FitVariable(variable=missing_mass_for_ff_fitting, alternative_binning=30, alternative_scope=(-2.0, 1.0)),),
    "DSZ": (FitVariable(variable=missing_mass_for_ff_fitting, alternative_binning=30, alternative_scope=(-2.0, 1.0)),),
    "DSM": (FitVariable(variable=missing_mass_for_ff_fitting, alternative_binning=30, alternative_scope=(-2.0, 1.0)),),
    "DZPiZ": (FitVariable(variable=missing_mass_for_ff_fitting, alternative_binning=25, alternative_scope=(-2.0, 1.0)),),
    "DMPiZ": (FitVariable(variable=missing_mass_for_ff_fitting, alternative_binning=20, alternative_scope=(-2.0, 1.0)),),
    "DSZPiZ": (FitVariable(variable=missing_mass_for_ff_fitting, alternative_binning=20, alternative_scope=(-2.0, 1.0)),),
    "DSMPiZ": (FitVariable(variable=missing_mass_for_ff_fitting, alternative_binning=20, alternative_scope=(-2.0, 1.0)),),
    "DZPiM": (FitVariable(variable=missing_mass_for_ff_fitting, alternative_binning=30, alternative_scope=(-2.0, 1.0)),),
    "DPPiM": (FitVariable(variable=missing_mass_for_ff_fitting, alternative_binning=30, alternative_scope=(-2.0, 1.0)),),
    "DSZPiM": (FitVariable(variable=missing_mass_for_ff_fitting, alternative_binning=25, alternative_scope=(-2.0, 1.0)),),
    "DSPPiM": (FitVariable(variable=missing_mass_for_ff_fitting, alternative_binning=20, alternative_scope=(-2.0, 1.0)),),
}

alternative_ff_study_fit_variables_dict: Dict[str, Tuple[FitVariable, ...]] = {
    "DZ": (
        FitVariable(variable=delta_M_for_ff_fitting_hvar, alternative_binning=None, alternative_scope=None),
        FitVariable(variable=missing_mass_for_ff_fitting, alternative_binning=2, alternative_scope=(-0.4, 0.4)),
    ),
    "DM": (
        FitVariable(variable=delta_M_for_ff_fitting_hvar, alternative_binning=None, alternative_scope=None),
        FitVariable(variable=missing_mass_for_ff_fitting, alternative_binning=2, alternative_scope=(-0.4, 0.4)),
    ),
    "DSZ": (
        FitVariable(variable=delta_M_for_ff_fitting_hvar, alternative_binning=None, alternative_scope=None),
        FitVariable(variable=missing_mass_for_ff_fitting, alternative_binning=2, alternative_scope=(-0.4, 0.4)),
    ),
    "DSM": (
        FitVariable(variable=delta_M_for_ff_fitting_hvar, alternative_binning=None, alternative_scope=None),
        FitVariable(variable=missing_mass_for_ff_fitting, alternative_binning=2, alternative_scope=(-0.4, 0.4)),
    ),
    "DZPiZ": (
        FitVariable(variable=delta_M_for_ff_fitting_hvar, alternative_binning=None, alternative_scope=None),
        # FitVariable(variable=missing_mass_for_ff_fitting, alternative_binning=2, alternative_scope=(-0.4, 0.4)),
    ),
    "DMPiZ": (
        FitVariable(variable=delta_M_for_ff_fitting_hvar, alternative_binning=None, alternative_scope=None),
        # FitVariable(variable=missing_mass_for_ff_fitting, alternative_binning=2, alternative_scope=(-0.4, 0.4)),
    ),
    "DSZPiZ": (
        FitVariable(variable=delta_M_for_ff_fitting_hvar, alternative_binning=None, alternative_scope=None),
        # FitVariable(variable=missing_mass_for_ff_fitting, alternative_binning=2, alternative_scope=(-0.4, 0.4)),
    ),
    "DSMPiZ": (
        FitVariable(variable=delta_M_for_ff_fitting_hvar, alternative_binning=None, alternative_scope=None),
        # FitVariable(variable=missing_mass_for_ff_fitting, alternative_binning=2, alternative_scope=(-0.4, 0.4)),
    ),
    "DZPiM": (
        FitVariable(variable=delta_M_for_ff_fitting_hvar, alternative_binning=None, alternative_scope=None),
        FitVariable(variable=missing_mass_for_ff_fitting, alternative_binning=2, alternative_scope=(-0.4, 0.4)),
    ),
    "DPPiM": (
        FitVariable(variable=delta_M_for_ff_fitting_hvar, alternative_binning=None, alternative_scope=None),
        FitVariable(variable=missing_mass_for_ff_fitting, alternative_binning=2, alternative_scope=(-0.4, 0.4)),
    ),
    "DSZPiM": (
        FitVariable(variable=delta_M_for_ff_fitting_hvar, alternative_binning=None, alternative_scope=None),
        FitVariable(variable=missing_mass_for_ff_fitting, alternative_binning=2, alternative_scope=(-0.4, 0.4)),
    ),
    "DSPPiM": (
        FitVariable(variable=delta_M_for_ff_fitting_hvar, alternative_binning=None, alternative_scope=None),
        # FitVariable(variable=missing_mass_for_ff_fitting, alternative_binning=2, alternative_scope=(-0.4, 0.4)),
    ),
}

alternative_ff_study_pStarL_fit_variables_dict: Dict[str, Tuple[FitVariable, ...]] = {
    "DZ": (FitVariable(variable=delta_M_for_ff_fitting_hvar, alternative_binning=None, alternative_scope=None),),
    "DM": (FitVariable(variable=delta_M_for_ff_fitting_hvar, alternative_binning=None, alternative_scope=None),),
    "DSZ": (FitVariable(variable=delta_M_for_ff_fitting_hvar, alternative_binning=None, alternative_scope=None),),
    "DSM": (FitVariable(variable=delta_M_for_ff_fitting_hvar, alternative_binning=None, alternative_scope=None),),
    "DZPiZ": (FitVariable(variable=delta_M_for_ff_fitting_hvar, alternative_binning=None, alternative_scope=None),),
    "DMPiZ": (FitVariable(variable=delta_M_for_ff_fitting_hvar, alternative_binning=None, alternative_scope=None),),
    "DSZPiZ": (FitVariable(variable=delta_M_for_ff_fitting_hvar, alternative_binning=None, alternative_scope=None),),
    "DSMPiZ": (FitVariable(variable=delta_M_for_ff_fitting_hvar, alternative_binning=None, alternative_scope=None),),
    "DZPiM": (FitVariable(variable=delta_M_for_ff_fitting_hvar, alternative_binning=None, alternative_scope=None),),
    "DPPiM": (FitVariable(variable=delta_M_for_ff_fitting_hvar, alternative_binning=None, alternative_scope=None),),
    "DSZPiM": (FitVariable(variable=delta_M_for_ff_fitting_hvar, alternative_binning=None, alternative_scope=None),),
    "DSPPiM": (FitVariable(variable=delta_M_for_ff_fitting_hvar, alternative_binning=None, alternative_scope=None),),
}

q_square_advanced_binning_for_fitting_var = HistVariable(
    df_label="qSquaredReco",
    n_bins=6,
    scope=(0, 10),
    var_name=LatexObsNames.fitting_q2,
    unit=LatexUnits.gev_sq,
)

pStarL_advanced_binning_for_fitting_var = HistVariable(
    df_label="pStarL",
    n_bins=8,
    scope=(0.1, 2.2),
    var_name=LatexObsNames.fitting_p_l_star,
    unit=LatexUnits.gev,
)

q_square_advanced_binning_scopes = (0.0, 1.0, 2.0, 3.0, 4.5, 7.0, 10.0)

pStarL_advanced_binning_scopes = (0.1, 0.5, 0.8, 1.0, 1.2, 1.4, 1.6, 1.8, 2.2)
