import json
import os

import numpy as np
import matplotlib.pyplot as plt

from templatefitter.plotter.plot_utilities import export
from templatefitter.plotter.plot_style import KITColors, xlabel_pos, ylabel_pos

from rdstar.offline_analysis.dstst_analysis.template_definitions import channel_dec_modes

import maxtemplatefitter as tf


def chisq(obs, exp, exp_unc):
    if len(obs.shape) > 1:
        return np.sum(np.nan_to_num((obs - exp) ** 2 / exp_unc), axis=1)
    else:
        return np.sum(np.nan_to_num((obs - exp) ** 2 / exp_unc))


def mc_chisq_test(obs, exp, exp_unc):
    exp_ge_zero = exp > 0
    obs = obs[exp_ge_zero]
    exp = exp[exp_ge_zero]
    exp_unc = exp_unc[exp_ge_zero]
    print(exp_unc)

    obs_chisq = chisq(obs, exp, exp_unc)

    toys = np.random.poisson(exp, size=(1000000, len(exp)))

    toy_chisq = chisq(toys, exp, exp_unc)

    return obs_chisq, toy_chisq


def get_mode(df, mode_ids):
    return df.loc[df.daughter__bo1__cm__spextraInfo__bodecayModeID__bc__bc.isin(mode_ids), :]


def chisq_missingmass(mc, data, mode, mode_id, sig_var, weight_col, sig_var_label, lin_space, export_path):
    mc_mode = get_mode(mc, mode_id)
    data_mode = get_mode(data, mode_id)

    nbins = len(lin_space) - 1

    mc_hist = np.histogram(mc_mode[sig_var], bins=lin_space, weights=mc_mode[weight_col])[0]
    mc_hist_unc = np.sqrt(np.histogram(mc_mode[sig_var], bins=lin_space, weights=mc_mode[weight_col].values ** 2)[0])

    data_hist = np.histogram(data_mode[sig_var], bins=lin_space)[0]

    obs, toys = mc_chisq_test(data_hist, mc_hist, mc_hist_unc)

    fig, ax = plt.subplots(1, 1, figsize=(4, 4), dpi=200)

    bc, be, _ = ax.hist(toys, bins=100, density=True, histtype="stepfilled", edgecolor=KITColors.kit_black,
                        label=r"$\chi^2$ Toys")

    bm = (be[1:] + be[:-1]) / 2
    bw = (be[1:] - be[:-1])
    p_val = np.sum(bc[bm > obs] * bw[bm > obs])

    ax.axvline(obs, color=KITColors.kit_orange, label=r"Observed $\chi^2$" + "\n" + r"$p_{\chi^2}=%.2f$" % p_val)
    ax.text(0.45, 0.6, r"%s for $\mathrm{M}_\mathrm{bc}<5.27$ GeV" % sig_var_label, size="small",
            transform=ax.transAxes)
    ax.text(0.45, 0.55, r"$N_\mathrm{bins}=%d$" % nbins, size="small", transform=ax.transAxes)

    ax.legend(frameon=True)
    ax.set_title(channel_dec_modes[mode], loc="right")

    ax.ticklabel_format(axis="y", style="sci", scilimits=(-3, 3))
    ax.set_xlabel(r"$\chi^2$", **xlabel_pos)
    ax.set_ylabel("Normalized in arb. units", **ylabel_pos)
    ax.set_ylim((0, 1.2 * np.amax(bc)))

    plt.tight_layout()
    if export_path:
        fig_name = f"chi_sq_test_{sig_var}_{len(mc_hist)}_{mode}"
        export(fig, fig_name, export_path)


def mct_gof(mct, result_dir, tag, title=None, text=None):
    data = np.concatenate([channel.data.bin_counts for channel in mct.channels.values()])

    mc = np.concatenate(
        [sum(template.values for template in channel.templates.values()) for channel in mct.channels.values()]
    )
    mc_errors_sq = np.concatenate(
        [sum(template.errors ** 2 for template in channel.templates.values()) for channel in mct.channels.values()]
    )

    cowan = tf.stats.cowan_binned_likelihood_gof(data, mc, 4)
    chi2 = tf.stats.pearson_chi2_test(data, mc, len(data) - 4)

    with open(os.path.join(result_dir, f"{tag}.json"), "w") as outfile:
        json.dump({"cowan": cowan, "chis2": chi2}, outfile, indent=2)

    obs, toys = mc_chisq_test(data, mc, np.sqrt(mc_errors_sq))

    fig, ax = plt.subplots(1, 1, figsize=(3.5, 3.5), dpi=200)

    bc, be, _ = ax.hist(toys, bins=100, density=True, histtype="stepfilled", edgecolor=KITColors.kit_black,
                        label=r"$\chi^2$ Toys")

    bm = (be[1:] + be[:-1]) / 2
    bw = (be[1:] - be[:-1])
    p_val = np.sum(bc[bm > obs] * bw[bm > obs])

    ax.axvline(obs, color=KITColors.kit_orange, label=r"Observed $\chi^2$" + "\n" + r"$p_{\chi^2}=%.2f$" % p_val)

    ax.legend(frameon=True)
    ax.ticklabel_format(axis="y", style="sci", scilimits=(-3, 3))
    ax.set_xlabel(r"$\chi^2$", **xlabel_pos)
    ax.set_ylabel("Normalized in arb. units", **ylabel_pos)
    ax.set_ylim((0, 1.2 * np.amax(bc)))

    if text:
        for t in text:
            ax.text(t[0], t[1], t[2], size="small", transform=ax.transAxes)

    # ax.text(0.45, 0.6, r"%s for $\mathrm{M}_\mathrm{bc}<5.27$ GeV" %sig_var_label , size="small",
    #         transform=ax.transAxes)
    # ax.text(0.45, 0.55, r"$N_\mathrm{bins}=%d$" % nbins , size="small", transform=ax.transAxes)

    if title:
        ax.set_title(title, loc="right")

    export(fig, tag, result_dir)
