"""
Definitions for calculation of rdstar analysis reconstruction efficiencies.

Felix Metzner 2021
"""

from enum import Enum
from typing import Tuple, Dict

from rdstar.branching_fractions import belle_neutral_b_meson_decay_info as belle_mixed_b_bfs
from rdstar.branching_fractions import belle_charged_b_meson_decay_info as belle_charged_b_bfs
from rdstar.branching_fractions import belle_special_b_meson_decay_info as belle_special_b_bfs

from rdstar.offline_analysis.rdstar_samples import data_type_combinations
from rdstar.offline_analysis.mc_matching.old_mc_matching_info import OldMCMatchingIDsQueryInfo
from rdstar.offline_analysis.mc_matching.mc_matching_utilities import MCSigIDNameToQueryMappingType
from rdstar.offline_analysis.efficiency_calculation.simple_efficiency_calculator import DataTypeFraction


__all__ = [
    "SigIDToDTMappingType",
    "get_two_sided_event_fraction",
    "get_one_sided_event_fraction",
    "get_two_sided_event_fraction_with_two_bfs",
    "get_rdstar_sig_id_data_type_mapping",
    "get_rdstar_true_charge_sig_id_data_type_mapping",
]


SigIDToDTMappingType = Dict[int, Tuple[DataTypeFraction, ...]]


# Sig ID category keys
_st_dl: str = "Dl"
_st_dsl: str = "DSl"
_st_dtau: str = "Dtau"
_st_dstau: str = "DStau"


# Data type keys
_mixed_dts: Tuple[str, ...] = tuple(["Mixed"])
_charged_dts: Tuple[str, ...] = tuple(["Charged"])

_mixed_special_dts: Tuple[str, ...] = tuple([c for c in data_type_combinations["Special"] if c.endswith("Mixed")])
_charged_special_dts: Tuple[str, ...] = tuple([c for c in data_type_combinations["Special"] if c.endswith("Charged")])

_mixed_dstst_dts: Tuple[str, ...] = tuple(
    [
        *[c for c in data_type_combinations["Dss"] if c.startswith("BZ")],
        *[c for c in data_type_combinations["DssTau"] if c.startswith("TBZ")],
    ]
)
_charged_dstst_dts: Tuple[str, ...] = tuple(
    [
        *[c for c in data_type_combinations["Dss"] if c.startswith("BM")],
        *[c for c in data_type_combinations["DssTau"] if c.startswith("TBM")],
    ]
)

_mixed_gap_done_dts: Tuple[str, ...] = tuple([c for c in data_type_combinations["GapD1"] if c.endswith("Mixed")])
_charged_gap_done_dts: Tuple[str, ...] = tuple([c for c in data_type_combinations["GapD1"] if c.endswith("Charged")])

_mixed_default_gap_dts: Tuple[str, ...] = tuple(
    [
        *[c for c in data_type_combinations["GapDEtaOpt1"] if c.endswith("Mixed")],
        *[c for c in data_type_combinations["GapDsEtaOpt1"] if c.endswith("Mixed")],
        *[c for c in data_type_combinations["GapDPPOpt1"] if c.endswith("Mixed")],
        *[c for c in data_type_combinations["GapDsPPOpt1"] if c.endswith("Mixed")],
    ]
)
_charged_default_gap_dts: Tuple[str, ...] = tuple(
    [
        *[c for c in data_type_combinations["GapDEtaOpt1"] if c.endswith("Charged")],
        *[c for c in data_type_combinations["GapDsEtaOpt1"] if c.endswith("Charged")],
        *[c for c in data_type_combinations["GapDPPOpt1"] if c.endswith("Charged")],
        *[c for c in data_type_combinations["GapDsPPOpt1"] if c.endswith("Charged")],
    ]
)

_mixed_alt_gap_dts: Tuple[str, ...] = tuple(
    [
        *[c for c in data_type_combinations["GapDEtaOpt2"] if c.endswith("Mixed")],
        *[c for c in data_type_combinations["GapDsEtaOpt2"] if c.endswith("Mixed")],
        *[c for c in data_type_combinations["GapDPPOpt2"] if c.endswith("Mixed")],
        *[c for c in data_type_combinations["GapDsPPOpt2"] if c.endswith("Mixed")],
    ]
)
_charged_alt_gap_dts: Tuple[str, ...] = tuple(
    [
        *[c for c in data_type_combinations["GapDEtaOpt2"] if c.endswith("Charged")],
        *[c for c in data_type_combinations["GapDsEtaOpt2"] if c.endswith("Charged")],
        *[c for c in data_type_combinations["GapDPPOpt2"] if c.endswith("Charged")],
        *[c for c in data_type_combinations["GapDsPPOpt2"] if c.endswith("Charged")],
    ]
)


def get_two_sided_event_fraction(bf: float, bf_removed: float) -> float:
    # return 2 * bf * (1.0 - bf_removed) - bf ** 2  # Event based version
    return 2 * bf * (1.0 - bf_removed)  # Candidate based version


def get_one_sided_event_fraction(bf: float, bf_removed: float) -> float:
    return bf * (1.0 - bf_removed)


def get_two_sided_event_fraction_with_two_bfs(bf_1: float, bf_2: float, bf_removed: float) -> float:
    # return bf_1 * (1.0 - bf_removed) + bf_2  # * (1.0 - bf_1)  # Event based version
    return bf_1 * (1.0 - bf_removed) + bf_2  # Candidate based version


bf_b0_dlnu: float = 2 * belle_mixed_b_bfs.belle_bf_b0_to_d_l_nu
bf_bp_dlnu: float = 2 * belle_charged_b_bfs.belle_bf_bp_to_d_l_nu
bf_b0_dstlnu: float = 2 * belle_mixed_b_bfs.belle_bf_b0_to_d_st_l_nu
bf_bp_dstlnu: float = 2 * belle_charged_b_bfs.belle_bf_bp_to_d_st_l_nu
bf_b0_dtaunu: float = belle_mixed_b_bfs.belle_bf_b0_to_d_tau_nu
bf_bp_dtaunu: float = belle_charged_b_bfs.belle_bf_bp_to_d_tau_nu
bf_b0_dsttaunu: float = belle_mixed_b_bfs.belle_bf_b0_to_d_st_tau_nu
bf_bp_dsttaunu: float = belle_charged_b_bfs.belle_bf_bp_to_d_st_tau_nu


mixed_removed: float = (
    2 * belle_mixed_b_bfs.sum_of_belle_removed_b0_bfs
    + belle_mixed_b_bfs.sum_of_belle_b0_to_dss_l_nu_bfs
    + belle_special_b_bfs.sum_bf_b0_rare
    + belle_special_b_bfs.sum_bf_b0_ulnu
)

charged_removed: float = (
    2 * belle_charged_b_bfs.sum_of_belle_removed_bp_bfs
    + belle_charged_b_bfs.sum_of_belle_bp_to_dss_l_nu_bfs
    + belle_special_b_bfs.sum_bf_bp_rare
    + belle_special_b_bfs.sum_bf_bp_ulnu
)

mixed_special_removed: float = 0.0
charged_special_removed: float = 0.0
mixed_dstst_removed: float = 0.0
charged_dstst_removed: float = 0.0
mixed_gap_done_removed: float = 0.0
charged_gap_done_removed: float = 0.0
mixed_gap_removed: float = 0.0
charged_gap_removed: float = 0.0


def get_rdstar_sig_id_data_type_mapping(
    use_alternative_gap: bool = False,
) -> SigIDToDTMappingType:
    _mixed_gap_dts: Tuple[str, ...] = _mixed_alt_gap_dts if use_alternative_gap else _mixed_default_gap_dts
    _charged_gap_dts: Tuple[str, ...] = _charged_alt_gap_dts if use_alternative_gap else _charged_default_gap_dts

    sig_id_mapping: MCSigIDNameToQueryMappingType = OldMCMatchingIDsQueryInfo.get_sig_id_mapping()

    rdstar_sig_id_data_type_mapping: SigIDToDTMappingType = {
        sig_id_mapping[_st_dl].id: (
            DataTypeFraction(
                data_types=_mixed_dts,
                fraction=get_two_sided_event_fraction(bf=bf_b0_dlnu, bf_removed=mixed_removed),
            ),
            DataTypeFraction(
                data_types=_charged_dts,
                fraction=get_two_sided_event_fraction(bf=bf_bp_dlnu, bf_removed=charged_removed),
            ),
            DataTypeFraction(
                data_types=_mixed_special_dts,
                fraction=get_one_sided_event_fraction(bf=bf_b0_dlnu, bf_removed=mixed_special_removed),
            ),
            DataTypeFraction(
                data_types=_charged_special_dts,
                fraction=get_one_sided_event_fraction(bf=bf_bp_dlnu, bf_removed=charged_special_removed),
            ),
            DataTypeFraction(
                data_types=_mixed_dstst_dts,
                fraction=get_one_sided_event_fraction(bf=bf_b0_dlnu, bf_removed=mixed_dstst_removed),
            ),
            DataTypeFraction(
                data_types=_charged_dstst_dts,
                fraction=get_one_sided_event_fraction(bf=bf_bp_dlnu, bf_removed=charged_dstst_removed),
            ),
            DataTypeFraction(
                data_types=_mixed_gap_done_dts,
                fraction=get_one_sided_event_fraction(bf=bf_b0_dlnu, bf_removed=mixed_gap_done_removed),
            ),
            DataTypeFraction(
                data_types=_charged_gap_done_dts,
                fraction=get_one_sided_event_fraction(bf=bf_bp_dlnu, bf_removed=charged_gap_done_removed),
            ),
            DataTypeFraction(
                data_types=_mixed_gap_dts,
                fraction=get_one_sided_event_fraction(bf=bf_b0_dlnu, bf_removed=mixed_gap_removed),
            ),
            DataTypeFraction(
                data_types=_charged_gap_dts,
                fraction=get_one_sided_event_fraction(bf=bf_bp_dlnu, bf_removed=charged_gap_removed),
            ),
        ),
        sig_id_mapping[_st_dsl].id: (
            DataTypeFraction(
                data_types=_mixed_dts,
                fraction=get_two_sided_event_fraction(bf=bf_b0_dstlnu, bf_removed=mixed_removed),
            ),
            DataTypeFraction(
                data_types=_charged_dts,
                fraction=get_two_sided_event_fraction(bf=bf_bp_dstlnu, bf_removed=charged_removed),
            ),
            DataTypeFraction(
                data_types=_mixed_special_dts,
                fraction=get_one_sided_event_fraction(bf=bf_b0_dstlnu, bf_removed=mixed_special_removed),
            ),
            DataTypeFraction(
                data_types=_charged_special_dts,
                fraction=get_one_sided_event_fraction(bf=bf_bp_dstlnu, bf_removed=charged_special_removed),
            ),
            DataTypeFraction(
                data_types=_mixed_dstst_dts,
                fraction=get_one_sided_event_fraction(bf=bf_b0_dstlnu, bf_removed=mixed_dstst_removed),
            ),
            DataTypeFraction(
                data_types=_charged_dstst_dts,
                fraction=get_one_sided_event_fraction(bf=bf_bp_dstlnu, bf_removed=charged_dstst_removed),
            ),
            DataTypeFraction(
                data_types=_mixed_gap_done_dts,
                fraction=get_one_sided_event_fraction(bf=bf_b0_dstlnu, bf_removed=mixed_gap_done_removed),
            ),
            DataTypeFraction(
                data_types=_charged_gap_done_dts,
                fraction=get_one_sided_event_fraction(bf=bf_bp_dstlnu, bf_removed=charged_gap_done_removed),
            ),
            DataTypeFraction(
                data_types=_mixed_gap_dts,
                fraction=get_one_sided_event_fraction(bf=bf_b0_dstlnu, bf_removed=mixed_gap_removed),
            ),
            DataTypeFraction(
                data_types=_charged_gap_dts,
                fraction=get_one_sided_event_fraction(bf=bf_bp_dstlnu, bf_removed=charged_gap_removed),
            ),
        ),
        sig_id_mapping[_st_dtau].id: (
            DataTypeFraction(
                data_types=_mixed_dts,
                fraction=get_two_sided_event_fraction(bf=bf_b0_dtaunu, bf_removed=mixed_removed),
            ),
            DataTypeFraction(
                data_types=_charged_dts,
                fraction=get_two_sided_event_fraction(bf=bf_bp_dtaunu, bf_removed=charged_removed),
            ),
            DataTypeFraction(
                data_types=_mixed_special_dts,
                fraction=get_one_sided_event_fraction(bf=bf_b0_dtaunu, bf_removed=mixed_special_removed),
            ),
            DataTypeFraction(
                data_types=_charged_special_dts,
                fraction=get_one_sided_event_fraction(bf=bf_bp_dtaunu, bf_removed=charged_special_removed),
            ),
            DataTypeFraction(
                data_types=_mixed_dstst_dts,
                fraction=get_one_sided_event_fraction(bf=bf_b0_dtaunu, bf_removed=mixed_dstst_removed),
            ),
            DataTypeFraction(
                data_types=_charged_dstst_dts,
                fraction=get_one_sided_event_fraction(bf=bf_bp_dtaunu, bf_removed=charged_dstst_removed),
            ),
            DataTypeFraction(
                data_types=_mixed_gap_done_dts,
                fraction=get_one_sided_event_fraction(bf=bf_b0_dtaunu, bf_removed=mixed_gap_done_removed),
            ),
            DataTypeFraction(
                data_types=_charged_gap_done_dts,
                fraction=get_one_sided_event_fraction(bf=bf_bp_dtaunu, bf_removed=charged_gap_done_removed),
            ),
            DataTypeFraction(
                data_types=_mixed_gap_dts,
                fraction=get_one_sided_event_fraction(bf=bf_b0_dtaunu, bf_removed=mixed_gap_removed),
            ),
            DataTypeFraction(
                data_types=_charged_gap_dts,
                fraction=get_one_sided_event_fraction(bf=bf_bp_dtaunu, bf_removed=charged_gap_removed),
            ),
        ),
        sig_id_mapping[_st_dstau].id: (
            DataTypeFraction(
                data_types=_mixed_dts,
                fraction=get_two_sided_event_fraction(bf=bf_b0_dsttaunu, bf_removed=mixed_removed),
            ),
            DataTypeFraction(
                data_types=_charged_dts,
                fraction=get_two_sided_event_fraction(bf=bf_bp_dsttaunu, bf_removed=charged_removed),
            ),
            DataTypeFraction(
                data_types=_mixed_special_dts,
                fraction=get_one_sided_event_fraction(bf=bf_b0_dsttaunu, bf_removed=mixed_special_removed),
            ),
            DataTypeFraction(
                data_types=_charged_special_dts,
                fraction=get_one_sided_event_fraction(bf=bf_bp_dsttaunu, bf_removed=charged_special_removed),
            ),
            DataTypeFraction(
                data_types=_mixed_dstst_dts,
                fraction=get_one_sided_event_fraction(bf=bf_b0_dsttaunu, bf_removed=mixed_dstst_removed),
            ),
            DataTypeFraction(
                data_types=_charged_dstst_dts,
                fraction=get_one_sided_event_fraction(bf=bf_bp_dsttaunu, bf_removed=charged_dstst_removed),
            ),
            DataTypeFraction(
                data_types=_mixed_gap_done_dts,
                fraction=get_one_sided_event_fraction(bf=bf_b0_dsttaunu, bf_removed=mixed_gap_done_removed),
            ),
            DataTypeFraction(
                data_types=_charged_gap_done_dts,
                fraction=get_one_sided_event_fraction(bf=bf_bp_dsttaunu, bf_removed=charged_gap_done_removed),
            ),
            DataTypeFraction(
                data_types=_mixed_gap_dts,
                fraction=get_one_sided_event_fraction(bf=bf_b0_dsttaunu, bf_removed=mixed_gap_removed),
            ),
            DataTypeFraction(
                data_types=_charged_gap_dts,
                fraction=get_one_sided_event_fraction(bf=bf_bp_dsttaunu, bf_removed=charged_gap_removed),
            ),
        ),
    }

    return rdstar_sig_id_data_type_mapping


class TrueChargeRDStarSigIDs(Enum):
    BpDztau: int = 1
    BzDmtau: int = 1001
    BpDzStau: int = 2
    BzDmStau: int = 1002
    BpDzl: int = 3
    BzDml: int = 1003
    BpDzSl: int = 4
    BzDmSl: int = 1004


def get_rdstar_true_charge_sig_id_data_type_mapping(
    use_alternative_gap: bool = False,
) -> SigIDToDTMappingType:
    _mixed_gap_dts: Tuple[str, ...] = _mixed_alt_gap_dts if use_alternative_gap else _mixed_default_gap_dts
    _charged_gap_dts: Tuple[str, ...] = _charged_alt_gap_dts if use_alternative_gap else _charged_default_gap_dts

    rdstar_true_charge_sig_id_data_type_mapping: SigIDToDTMappingType = {
        TrueChargeRDStarSigIDs.BpDzl.value: (
            DataTypeFraction(
                data_types=_mixed_dts,
                fraction=get_two_sided_event_fraction(bf=0.0, bf_removed=mixed_removed),
            ),
            DataTypeFraction(
                data_types=_charged_dts,
                fraction=get_two_sided_event_fraction(bf=bf_bp_dlnu, bf_removed=charged_removed),
            ),
            DataTypeFraction(
                data_types=_mixed_special_dts,
                fraction=get_one_sided_event_fraction(bf=0.0, bf_removed=mixed_special_removed),
            ),
            DataTypeFraction(
                data_types=_charged_special_dts,
                fraction=get_one_sided_event_fraction(bf=bf_bp_dlnu, bf_removed=charged_special_removed),
            ),
            DataTypeFraction(
                data_types=_mixed_dstst_dts,
                fraction=get_one_sided_event_fraction(bf=0.0, bf_removed=mixed_dstst_removed),
            ),
            DataTypeFraction(
                data_types=_charged_dstst_dts,
                fraction=get_one_sided_event_fraction(bf=bf_bp_dlnu, bf_removed=charged_dstst_removed),
            ),
            DataTypeFraction(
                data_types=_mixed_gap_done_dts,
                fraction=get_one_sided_event_fraction(bf=0.0, bf_removed=mixed_gap_done_removed),
            ),
            DataTypeFraction(
                data_types=_charged_gap_done_dts,
                fraction=get_one_sided_event_fraction(bf=bf_bp_dlnu, bf_removed=charged_gap_done_removed),
            ),
            DataTypeFraction(
                data_types=_mixed_gap_dts,
                fraction=get_one_sided_event_fraction(bf=0.0, bf_removed=mixed_gap_removed),
            ),
            DataTypeFraction(
                data_types=_charged_gap_dts,
                fraction=get_one_sided_event_fraction(bf=bf_bp_dlnu, bf_removed=charged_gap_removed),
            ),
        ),
        TrueChargeRDStarSigIDs.BzDml.value: (
            DataTypeFraction(
                data_types=_mixed_dts,
                fraction=get_two_sided_event_fraction(bf=bf_b0_dlnu, bf_removed=mixed_removed),
            ),
            DataTypeFraction(
                data_types=_charged_dts,
                fraction=get_two_sided_event_fraction(bf=0.0, bf_removed=charged_removed),
            ),
            DataTypeFraction(
                data_types=_mixed_special_dts,
                fraction=get_one_sided_event_fraction(bf=bf_b0_dlnu, bf_removed=mixed_special_removed),
            ),
            DataTypeFraction(
                data_types=_charged_special_dts,
                fraction=get_one_sided_event_fraction(bf=0.0, bf_removed=charged_special_removed),
            ),
            DataTypeFraction(
                data_types=_mixed_dstst_dts,
                fraction=get_one_sided_event_fraction(bf=bf_b0_dlnu, bf_removed=mixed_dstst_removed),
            ),
            DataTypeFraction(
                data_types=_charged_dstst_dts,
                fraction=get_one_sided_event_fraction(bf=0.0, bf_removed=charged_dstst_removed),
            ),
            DataTypeFraction(
                data_types=_mixed_gap_done_dts,
                fraction=get_one_sided_event_fraction(bf=bf_b0_dlnu, bf_removed=mixed_gap_done_removed),
            ),
            DataTypeFraction(
                data_types=_charged_gap_done_dts,
                fraction=get_one_sided_event_fraction(bf=0.0, bf_removed=charged_gap_done_removed),
            ),
            DataTypeFraction(
                data_types=_mixed_gap_dts,
                fraction=get_one_sided_event_fraction(bf=bf_b0_dlnu, bf_removed=mixed_gap_removed),
            ),
            DataTypeFraction(
                data_types=_charged_gap_dts,
                fraction=get_one_sided_event_fraction(bf=0.0, bf_removed=charged_gap_removed),
            ),
        ),
        TrueChargeRDStarSigIDs.BpDzSl.value: (
            DataTypeFraction(
                data_types=_mixed_dts,
                fraction=get_two_sided_event_fraction(bf=0.0, bf_removed=mixed_removed),
            ),
            DataTypeFraction(
                data_types=_charged_dts,
                fraction=get_two_sided_event_fraction(bf=bf_bp_dstlnu, bf_removed=charged_removed),
            ),
            DataTypeFraction(
                data_types=_mixed_special_dts,
                fraction=get_one_sided_event_fraction(bf=0.0, bf_removed=mixed_special_removed),
            ),
            DataTypeFraction(
                data_types=_charged_special_dts,
                fraction=get_one_sided_event_fraction(bf=bf_bp_dstlnu, bf_removed=charged_special_removed),
            ),
            DataTypeFraction(
                data_types=_mixed_dstst_dts,
                fraction=get_one_sided_event_fraction(bf=0.0, bf_removed=mixed_dstst_removed),
            ),
            DataTypeFraction(
                data_types=_charged_dstst_dts,
                fraction=get_one_sided_event_fraction(bf=bf_bp_dstlnu, bf_removed=charged_dstst_removed),
            ),
            DataTypeFraction(
                data_types=_mixed_gap_done_dts,
                fraction=get_one_sided_event_fraction(bf=0.0, bf_removed=mixed_gap_done_removed),
            ),
            DataTypeFraction(
                data_types=_charged_gap_done_dts,
                fraction=get_one_sided_event_fraction(bf=bf_bp_dstlnu, bf_removed=charged_gap_done_removed),
            ),
            DataTypeFraction(
                data_types=_mixed_gap_dts,
                fraction=get_one_sided_event_fraction(bf=0.0, bf_removed=mixed_gap_removed),
            ),
            DataTypeFraction(
                data_types=_charged_gap_dts,
                fraction=get_one_sided_event_fraction(bf=bf_bp_dstlnu, bf_removed=charged_gap_removed),
            ),
        ),
        TrueChargeRDStarSigIDs.BzDmSl.value: (
            DataTypeFraction(
                data_types=_mixed_dts,
                fraction=get_two_sided_event_fraction(bf=bf_b0_dstlnu, bf_removed=mixed_removed),
            ),
            DataTypeFraction(
                data_types=_charged_dts,
                fraction=get_two_sided_event_fraction(bf=0.0, bf_removed=charged_removed),
            ),
            DataTypeFraction(
                data_types=_mixed_special_dts,
                fraction=get_one_sided_event_fraction(bf=bf_b0_dstlnu, bf_removed=mixed_special_removed),
            ),
            DataTypeFraction(
                data_types=_charged_special_dts,
                fraction=get_one_sided_event_fraction(bf=0.0, bf_removed=charged_special_removed),
            ),
            DataTypeFraction(
                data_types=_mixed_dstst_dts,
                fraction=get_one_sided_event_fraction(bf=bf_b0_dstlnu, bf_removed=mixed_dstst_removed),
            ),
            DataTypeFraction(
                data_types=_charged_dstst_dts,
                fraction=get_one_sided_event_fraction(bf=0.0, bf_removed=charged_dstst_removed),
            ),
            DataTypeFraction(
                data_types=_mixed_gap_done_dts,
                fraction=get_one_sided_event_fraction(bf=bf_b0_dstlnu, bf_removed=mixed_gap_done_removed),
            ),
            DataTypeFraction(
                data_types=_charged_gap_done_dts,
                fraction=get_one_sided_event_fraction(bf=0.0, bf_removed=charged_gap_done_removed),
            ),
            DataTypeFraction(
                data_types=_mixed_gap_dts,
                fraction=get_one_sided_event_fraction(bf=bf_b0_dstlnu, bf_removed=mixed_gap_removed),
            ),
            DataTypeFraction(
                data_types=_charged_gap_dts,
                fraction=get_one_sided_event_fraction(bf=0.0, bf_removed=charged_gap_removed),
            ),
        ),
        TrueChargeRDStarSigIDs.BpDztau.value: (
            DataTypeFraction(
                data_types=_mixed_dts,
                fraction=get_two_sided_event_fraction(bf=0.0, bf_removed=mixed_removed),
            ),
            DataTypeFraction(
                data_types=_charged_dts,
                fraction=get_two_sided_event_fraction(bf=bf_bp_dtaunu, bf_removed=charged_removed),
            ),
            DataTypeFraction(
                data_types=_mixed_special_dts,
                fraction=get_one_sided_event_fraction(bf=0.0, bf_removed=mixed_special_removed),
            ),
            DataTypeFraction(
                data_types=_charged_special_dts,
                fraction=get_one_sided_event_fraction(bf=bf_bp_dtaunu, bf_removed=charged_special_removed),
            ),
            DataTypeFraction(
                data_types=_mixed_dstst_dts,
                fraction=get_one_sided_event_fraction(bf=0.0, bf_removed=mixed_dstst_removed),
            ),
            DataTypeFraction(
                data_types=_charged_dstst_dts,
                fraction=get_one_sided_event_fraction(bf=bf_bp_dtaunu, bf_removed=charged_dstst_removed),
            ),
            DataTypeFraction(
                data_types=_mixed_gap_done_dts,
                fraction=get_one_sided_event_fraction(bf=0.0, bf_removed=mixed_gap_done_removed),
            ),
            DataTypeFraction(
                data_types=_charged_gap_done_dts,
                fraction=get_one_sided_event_fraction(bf=bf_bp_dtaunu, bf_removed=charged_gap_done_removed),
            ),
            DataTypeFraction(
                data_types=_mixed_gap_dts,
                fraction=get_one_sided_event_fraction(bf=0.0, bf_removed=mixed_gap_removed),
            ),
            DataTypeFraction(
                data_types=_charged_gap_dts,
                fraction=get_one_sided_event_fraction(bf=bf_bp_dtaunu, bf_removed=charged_gap_removed),
            ),
        ),
        TrueChargeRDStarSigIDs.BzDmtau.value: (
            DataTypeFraction(
                data_types=_mixed_dts,
                fraction=get_two_sided_event_fraction(bf=bf_b0_dtaunu, bf_removed=mixed_removed),
            ),
            DataTypeFraction(
                data_types=_charged_dts,
                fraction=get_two_sided_event_fraction(bf=0.0, bf_removed=charged_removed),
            ),
            DataTypeFraction(
                data_types=_mixed_special_dts,
                fraction=get_one_sided_event_fraction(bf=bf_b0_dtaunu, bf_removed=mixed_special_removed),
            ),
            DataTypeFraction(
                data_types=_charged_special_dts,
                fraction=get_one_sided_event_fraction(bf=0.0, bf_removed=charged_special_removed),
            ),
            DataTypeFraction(
                data_types=_mixed_dstst_dts,
                fraction=get_one_sided_event_fraction(bf=bf_b0_dtaunu, bf_removed=mixed_dstst_removed),
            ),
            DataTypeFraction(
                data_types=_charged_dstst_dts,
                fraction=get_one_sided_event_fraction(bf=0.0, bf_removed=charged_dstst_removed),
            ),
            DataTypeFraction(
                data_types=_mixed_gap_done_dts,
                fraction=get_one_sided_event_fraction(bf=bf_b0_dtaunu, bf_removed=mixed_gap_done_removed),
            ),
            DataTypeFraction(
                data_types=_charged_gap_done_dts,
                fraction=get_one_sided_event_fraction(bf=0.0, bf_removed=charged_gap_done_removed),
            ),
            DataTypeFraction(
                data_types=_mixed_gap_dts,
                fraction=get_one_sided_event_fraction(bf=bf_b0_dtaunu, bf_removed=mixed_gap_removed),
            ),
            DataTypeFraction(
                data_types=_charged_gap_dts,
                fraction=get_one_sided_event_fraction(bf=0.0, bf_removed=charged_gap_removed),
            ),
        ),
        TrueChargeRDStarSigIDs.BpDzStau.value: (
            DataTypeFraction(
                data_types=_mixed_dts,
                fraction=get_two_sided_event_fraction(bf=0.0, bf_removed=mixed_removed),
            ),
            DataTypeFraction(
                data_types=_charged_dts,
                fraction=get_two_sided_event_fraction(bf=bf_bp_dsttaunu, bf_removed=charged_removed),
            ),
            DataTypeFraction(
                data_types=_mixed_special_dts,
                fraction=get_one_sided_event_fraction(bf=0.0, bf_removed=mixed_special_removed),
            ),
            DataTypeFraction(
                data_types=_charged_special_dts,
                fraction=get_one_sided_event_fraction(bf=bf_bp_dsttaunu, bf_removed=charged_special_removed),
            ),
            DataTypeFraction(
                data_types=_mixed_dstst_dts,
                fraction=get_one_sided_event_fraction(bf=0.0, bf_removed=mixed_dstst_removed),
            ),
            DataTypeFraction(
                data_types=_charged_dstst_dts,
                fraction=get_one_sided_event_fraction(bf=bf_bp_dsttaunu, bf_removed=charged_dstst_removed),
            ),
            DataTypeFraction(
                data_types=_mixed_gap_done_dts,
                fraction=get_one_sided_event_fraction(bf=0.0, bf_removed=mixed_gap_done_removed),
            ),
            DataTypeFraction(
                data_types=_charged_gap_done_dts,
                fraction=get_one_sided_event_fraction(bf=bf_bp_dsttaunu, bf_removed=charged_gap_done_removed),
            ),
            DataTypeFraction(
                data_types=_mixed_gap_dts,
                fraction=get_one_sided_event_fraction(bf=0.0, bf_removed=mixed_gap_removed),
            ),
            DataTypeFraction(
                data_types=_charged_gap_dts,
                fraction=get_one_sided_event_fraction(bf=bf_bp_dsttaunu, bf_removed=charged_gap_removed),
            ),
        ),
        TrueChargeRDStarSigIDs.BzDmStau.value: (
            DataTypeFraction(
                data_types=_mixed_dts,
                fraction=get_two_sided_event_fraction(bf=bf_b0_dsttaunu, bf_removed=mixed_removed),
            ),
            DataTypeFraction(
                data_types=_charged_dts,
                fraction=get_two_sided_event_fraction(bf=0.0, bf_removed=charged_removed),
            ),
            DataTypeFraction(
                data_types=_mixed_special_dts,
                fraction=get_one_sided_event_fraction(bf=bf_b0_dsttaunu, bf_removed=mixed_special_removed),
            ),
            DataTypeFraction(
                data_types=_charged_special_dts,
                fraction=get_one_sided_event_fraction(bf=0.0, bf_removed=charged_special_removed),
            ),
            DataTypeFraction(
                data_types=_mixed_dstst_dts,
                fraction=get_one_sided_event_fraction(bf=bf_b0_dsttaunu, bf_removed=mixed_dstst_removed),
            ),
            DataTypeFraction(
                data_types=_charged_dstst_dts,
                fraction=get_one_sided_event_fraction(bf=0.0, bf_removed=charged_dstst_removed),
            ),
            DataTypeFraction(
                data_types=_mixed_gap_done_dts,
                fraction=get_one_sided_event_fraction(bf=bf_b0_dsttaunu, bf_removed=mixed_gap_done_removed),
            ),
            DataTypeFraction(
                data_types=_charged_gap_done_dts,
                fraction=get_one_sided_event_fraction(bf=0.0, bf_removed=charged_gap_done_removed),
            ),
            DataTypeFraction(
                data_types=_mixed_gap_dts,
                fraction=get_one_sided_event_fraction(bf=bf_b0_dsttaunu, bf_removed=mixed_gap_removed),
            ),
            DataTypeFraction(
                data_types=_charged_gap_dts,
                fraction=get_one_sided_event_fraction(bf=0.0, bf_removed=charged_gap_removed),
            ),
        ),
    }

    return rdstar_true_charge_sig_id_data_type_mapping
