"""
Fitter class for extraction of Continuum calibration factors.

Felix Metzner 2023
"""

import copy
import numpy as np
import pandas as pd

from typing import Optional, Tuple

from rdstar.utilities.logging import log_to_default

from rdstar.offline_analysis.observables.common_observables import Cols
from rdstar.offline_analysis.reco_modes.reco_mode_infos import TagRecoModeInfos, RecoModeInfos

from rdstar.offline_analysis.fitting.dedicated_fit_approach.fit_info_container import FitSetupInfoContainer
from rdstar.offline_analysis.fitting.dedicated_fit_approach.cont_cal_fitters.cont_cal_fitter_base import (
    ContCalFitterBase,
    ContCalMappingType,
    SimpleContCalMappingType,
)

from rdstar.offline_analysis.fitting.dedicated_fit_approach.fit_setups import FitSetupCollection
from rdstar.offline_analysis.fitting.dedicated_fit_approach.fit_setups.fit_setup_utilities import FitSetupTriplet


__all__ = [
    "ContCalFitter",
]


logger = log_to_default(name=__name__)


class ContCalFitter(ContCalFitterBase):

    this_fitter_setup: FitSetupTriplet = FitSetupCollection.cont_cal_in_b_tag_modes__fs

    secondary_reco_id_col: Optional[str] = Cols.decay_mode_column_name

    ordered_parameter_names: Tuple[str, ...] = (
        "continuum_cb_tag_yield",
        "continuum_nb_tag_yield",
        "cb_tag_b_bkg_yield",
        "nb_tag_b_bkg_yield",
    )

    continuum_cb_tag_yield_index: int = 0
    continuum_nb_tag_yield_index: int = 1
    cb_tag_b_bkg_yield_index: int = 2
    nb_tag_b_bkg_yield_index: int = 3

    end_yield_params_index: int = nb_tag_b_bkg_yield_index

    main_floating_params_slice = np.s_[continuum_cb_tag_yield_index : end_yield_params_index + 1 : 1]

    number_of_hardcoded_floating_params: int = end_yield_params_index + 1

    fixed_params_indices: Tuple[int, ...] = tuple([])
    manual_nuisances_params_indices: Tuple[int, ...] = tuple([])

    def __init__(
        self,
        *ignore,
        fit_setup_info: FitSetupInfoContainer,
        mc_df: pd.DataFrame,
        data_df: Optional[pd.DataFrame],
        **kwargs,
    ) -> None:
        super().__init__(
            *ignore,
            fit_setup_info=fit_setup_info,
            mc_df=mc_df,
            data_df=data_df,
            file_dict_dir_path=None,
            **kwargs,
        )

        assert len(self.fit_setup_info.components) == self.end_yield_params_index + 1, (
            [(c.comp_index, c.name) for c in self.fit_setup_info.components],
            len(self.fit_setup_info.components),
            self.end_yield_params_index + 1,
        )

        self.set_initial_parameters(values=self._calculate_initial_parameter_vector())

    @staticmethod
    def event_selection_function(df: pd.DataFrame) -> pd.DataFrame:
        _df = copy.deepcopy(df)
        return _df[_df[Cols.decay_mode_column_name].isin(RecoModeInfos.main_reco_modes.ids)].reset_index(drop=True)

    # region Cost function
    def cost_fcn(self, params: np.ndarray) -> float:
        return self.chi2_fcn(params=params) + self.gaussian_constrains(params=params)

    def chi2_fcn(self, params: np.ndarray) -> float:
        _chi2 = np.sum(
            (self.expected_bin_counts(params=params) - self.data_bin_counts()) ** 2
            / (self.data_bin_stat_errors_squared + self.mc_bin_stat_errors_squared),
            axis=None,
        )
        assert isinstance(_chi2, float)
        return _chi2

    def gaussian_constrains(self, params: np.ndarray) -> float:
        # TODO!!!!!!
        return 0.0

    # endregion

    # region Expected bin count
    def expected_bin_counts(self, params: np.ndarray) -> np.ndarray:
        return np.sum(
            self.get_yields(params=params) * self.local_efficiencies * self.template_shapes,
            axis=1,
        )

    # endregion

    # region Yields relations
    def get_yields(self, params: np.ndarray) -> np.ndarray:
        return params[self.main_floating_params_slice]

    # endregion

    # region Parameter information
    @property
    def end_of_floating_params(self) -> int:
        _number_of_automated_nuisances: int = 0  # TODO!
        return self.number_of_hardcoded_floating_params + _number_of_automated_nuisances

    def _calculate_initial_parameter_vector(self) -> np.ndarray:

        # Yields
        _normalization_factors = self._calculate_normalization_factor_per_component(
            templates_per_ch=self.base_mc_bin_counts,
        )

        # Nuisances:
        automated_nuisances_initial_values: Tuple[float, ...] = tuple([])

        _initial_params = np.array(
            [
                _normalization_factors[self.continuum_cb_tag_yield_index],
                _normalization_factors[self.continuum_nb_tag_yield_index],
                #
                _normalization_factors[self.cb_tag_b_bkg_yield_index],
                _normalization_factors[self.nb_tag_b_bkg_yield_index],
                #
                *automated_nuisances_initial_values,
            ]
        )

        assert len(_initial_params) == self.end_of_floating_params, (
            len(_initial_params),
            self.end_of_floating_params,
            _initial_params,
        )

        return _initial_params

    # endregion

    # region Correction application info
    @classmethod
    def get_correction_mapping(cls) -> ContCalMappingType:
        sec_ids: Tuple[int, ...] = RecoModeInfos.main_reco_modes.ids

        _result: ContCalMappingType = {
            (TagRecoModeInfos.charged_tag_reco_modes.ids, sec_ids): cls.continuum_cb_tag_yield_index,
            (TagRecoModeInfos.neutral_tag_reco_modes.ids, sec_ids): cls.continuum_nb_tag_yield_index,
        }
        return _result

    @classmethod
    def get_simple_correction_mapping(cls) -> SimpleContCalMappingType:
        _result: SimpleContCalMappingType = {
            TagRecoModeInfos.charged_tag_reco_modes.ids: cls.continuum_cb_tag_yield_index,
            TagRecoModeInfos.neutral_tag_reco_modes.ids: cls.continuum_nb_tag_yield_index,
        }
        return _result

    # endregion
