"""
Base fitter class for extraction of B -> D** l nu BF and/or efficiency correction factors.

Felix Metzner 2023
"""

import copy
import numpy as np
import pandas as pd

from abc import ABC, abstractmethod
from typing import Optional, Tuple, Dict

from rdstar.utilities import PathType
from rdstar.utilities.logging import log_to_default

from rdstar.offline_analysis.observables.common_observables import Cols

from rdstar.offline_analysis.fitting.dedicated_fit_approach.fitter_base import FitterBase
from rdstar.offline_analysis.fitting.dedicated_fit_approach.fit_info_container import FitSetupInfoContainer

from rdstar.offline_analysis.fitting.dedicated_fit_approach.fit_setups.fit_setup_utilities import FitSetupTriplet


__all__ = [
    "DssBFCalibrationMappingType",
    "DssBFCorrFitterBase",
]


logger = log_to_default(name=__name__)


DssBFCalibrationMappingType = Dict[Tuple[Tuple[int, ...], Tuple[int, ...]], int]


class DssBFCorrFitterBase(FitterBase, ABC):

    this_fitter_setup: Optional[FitSetupTriplet] = None

    def __init__(
        self,
        *ignore,
        fit_setup_info: FitSetupInfoContainer,
        mc_df: pd.DataFrame,
        data_df: Optional[pd.DataFrame],
        file_dict_dir_path: Optional[PathType] = None,
        **kwargs,
    ) -> None:
        super().__init__(
            *ignore,
            fit_setup_info=fit_setup_info,
            mc_df=mc_df,
            data_df=data_df,
            file_dict_dir_path=file_dict_dir_path,
            **kwargs,
        )

    # region Cost function
    @abstractmethod
    def cost_fcn(self, params: np.ndarray) -> float:
        raise NotImplementedError("The 'cost_fcn' must be implemented in the derived class.")

    @abstractmethod
    def chi2_fcn(self, params: np.ndarray) -> float:
        raise NotImplementedError("The 'chi2_fcn' must be implemented in the derived class.")

    @abstractmethod
    def gaussian_constrains(self, params: np.ndarray) -> float:
        raise NotImplementedError("The 'gaussian_constrains_fcn' must be implemented in the derived class.")

    # endregion

    # region Expected bin count
    def expected_bin_counts(self, params: np.ndarray) -> np.ndarray:
        return np.sum(
            self.get_yields(params=params) * self.local_efficiencies * self.template_shapes,
            axis=1,
        )

    # endregion

    # region Yields relations
    @abstractmethod
    def get_yields(self, params: np.ndarray) -> np.ndarray:
        raise NotImplementedError("The 'get_yields' must be implemented in the derived class.")

    # endregion

    # region Correction application info
    @classmethod
    @abstractmethod
    def get_correction_mapping(cls) -> DssBFCalibrationMappingType:
        raise NotImplementedError("The 'get_correction_mapping' must be implemented in the derived class.")

    @classmethod
    def get_calibrated_df(
        cls,
        df: pd.DataFrame,
        ratio_result_values: np.ndarray,
    ) -> pd.DataFrame:
        _sample = copy.deepcopy(df)

        assert cls.this_fitter_setup is not None
        base_sig_id_col: str = Cols.base_sig_id_col
        component_id_col: str = cls.this_fitter_setup.asimov.component_id_column_name
        weight_col: str = cls.this_fitter_setup.asimov.weight_info.col_name

        if component_id_col not in _sample.columns:
            cls.this_fitter_setup.asimov.prepare_mc_df(df=_sample)

        test_mask: np.ndarray = np.full(shape=len(_sample), fill_value=False)

        for (sig_id, comp_id), ratio_index in cls.get_correction_mapping().items():

            _sig_id_mask: np.ndarray = _sample[base_sig_id_col].isin(sig_id)
            _comp_id_mask: np.ndarray = _sample[component_id_col].isin(comp_id)
            evt_mask: np.ndarray = _sig_id_mask & _comp_id_mask

            scale_factor: float = ratio_result_values[ratio_index]
            _sample.loc[evt_mask, weight_col] *= scale_factor

            assert not np.any(test_mask & evt_mask), np.sum(test_mask & evt_mask)
            test_mask |= evt_mask

        return _sample

    # endregion
