"""
Plots for fit evaluation.

Felix Metzner 2024
"""


import os
import numpy as np
import matplotlib.pyplot as plt

from dataclasses import dataclass
from typing import Tuple, List, Sequence, Optional, ClassVar, Generator

from templatefitter.plotter.plot_utilities import export, AxesType, FigureType
from templatefitter.plotter.plot_style import KITColors, TangoColors, set_matplotlibrc_params, xlabel_pos, ylabel_pos

from rdstar.utilities import PathType
from rdstar.utilities.logging import log_to_default

from rdstar.offline_analysis.fitting.dedicated_fit_approach.dedicated_sys_handler import SystematicsInfo
from rdstar.offline_analysis.fitting.dedicated_fit_approach.fit_result_container import RDStarFitResultContainer
from rdstar.offline_analysis.fitting.dedicated_fit_approach.fit_setups.fit_setup_utilities import FitSetupTriplet
from rdstar.offline_analysis.fitting.dedicated_fit_approach.systematics_manager.shape_sys_evaluater import (
    FitSetupBinningPerRecoChInfo,
)
from rdstar.offline_analysis.fitting.dedicated_fit_approach.fit_info_container import (
    ComponentInfo,
    FitSetupInfoContainer,
    FitObservableInfo,
    RecoChannelInfo,
)


__all__ = [
    "NuisancePullPlotter",
    "ShapePlotInfoContainer",
    "SystematicsShapePlotter",
    "SpecificShapePlotInfoContainer",
]


logger = log_to_default(name=__name__)


# region nuisance pull plotter


@dataclass(frozen=True)
class NuisancePullInfo:
    name: str
    latex_str: str
    param_ids: np.ndarray
    param_values: np.ndarray
    param_errors: np.ndarray

    def __post_init__(self) -> None:
        assert self.param_ids.shape == self.param_values.shape, (self.param_ids.shape, self.param_values.shape)
        assert self.param_ids.shape == self.param_errors.shape, (self.param_ids.shape, self.param_errors.shape)

    @property
    def number_of_nuisances(self) -> int:
        return len(self.param_ids)

    @staticmethod
    def create_info_from(
        sys_info: SystematicsInfo,
        fit_result: RDStarFitResultContainer,
    ) -> "NuisancePullInfo":
        assert sys_info.is_active

        _nu_param_ids: np.ndarray = np.array(sys_info.nuisance_param_ids)
        _nu_param_values: np.ndarray = fit_result.param_values[_nu_param_ids]
        _nu_param_errors: np.ndarray = fit_result.errors[_nu_param_ids]

        return NuisancePullInfo(
            name=sys_info.systematics_key,
            latex_str=sys_info.latex_str,
            param_ids=_nu_param_ids,
            param_values=_nu_param_values,
            param_errors=_nu_param_errors,
        )


class NuisancePullPlotter:

    output_dir_name: ClassVar[str] = "NuisancePulls"
    plot_name_prefix: ClassVar[str] = "nuisance_pulls_for"

    def __init__(
        self,
        base_output_dir_path: PathType,
    ) -> None:

        assert os.path.isdir(base_output_dir_path), base_output_dir_path

        self.output_dir_path: PathType = os.path.join(base_output_dir_path, self.output_dir_name)

    def create_nuisance_pull_plot_from(
        self,
        sys_info: SystematicsInfo,
        fit_result: RDStarFitResultContainer,
        fig_size: Tuple[float, float] = (8.0, 5.0),
    ) -> None:
        nuisance_pull_info: NuisancePullInfo = NuisancePullInfo.create_info_from(
            sys_info=sys_info,
            fit_result=fit_result,
        )
        self.plot_nuisance_pull(
            infos=nuisance_pull_info,
            fig_size=fig_size,
        )

    def plot_nuisance_pull(
        self,
        infos: NuisancePullInfo,
        fig_size: Tuple[float, float] = (8.0, 5.0),
    ) -> None:
        set_matplotlibrc_params()

        os.makedirs(self.output_dir_path, exist_ok=True)

        plot_file_name: str = f"{self.plot_name_prefix}_{infos.name}"

        fig, ax = plt.subplots(figsize=fig_size, dpi=300)  # type: FigureType, AxesType

        ax.set_xlim(-0.4, infos.number_of_nuisances - 0.6)
        ax.set_ylim(-2.2, 2.2)

        ls_min: int = -1
        ls_max: int = infos.number_of_nuisances
        ls_num: int = infos.number_of_nuisances + 2

        ax.fill_between(x=np.linspace(ls_min, ls_max, ls_num), y1=-1.0, y2=+1.0, color=TangoColors.chameleon2)

        ax.fill_between(x=np.linspace(ls_min, ls_max, ls_num), y1=-2.0, y2=-1.0, color=TangoColors.butter2)
        ax.fill_between(x=np.linspace(ls_min, ls_max, ls_num), y1=+1.0, y2=+2.0, color=TangoColors.butter2)

        ax.errorbar(
            x=np.arange(infos.number_of_nuisances),
            y=[nu_v for nu_v in infos.param_values],
            yerr=[nu_e for nu_e in infos.param_errors],
            marker=".",
            color=KITColors.kit_black,
            linestyle="",
            capsize=6,
        )

        plt.title(r"$\mathrm{Nuisance\;Parameter\;Pulls}$", fontsize=22)

        plt.xlabel(infos.latex_str + r" $\mathrm{Nuisance\;Parameters}$", fontsize=18, **xlabel_pos)
        plt.ylabel(r"$\mathrm{Standard\;Deviations}$", fontsize=18, **ylabel_pos)

        export(
            fig=fig,
            filename=plot_file_name,
            target_dir=self.output_dir_path,
            close_figure=False,
        )


# endregion


# region systematics shape plotter


@dataclass(frozen=True)
class ShapePlotInfoContainer:
    name: str
    latex_str: str
    normed_base_shape: np.ndarray
    relative_shape_error: np.ndarray
    pure_bin_counts: np.ndarray
    scale_factor: Optional[float] = None
    components_to_consider: Optional[Tuple[str, ...]] = None
    plot_subsets: bool = True

    def __post_init__(self) -> None:
        assert len(self.normed_base_shape.shape) == 2, (
            len(self.normed_base_shape.shape),
            self.normed_base_shape.shape,
        )
        assert len(self.relative_shape_error.shape) == 3, (
            len(self.relative_shape_error.shape),
            self.relative_shape_error.shape,
        )
        assert self.normed_base_shape.shape == self.relative_shape_error.shape[:2], (
            self.normed_base_shape.shape,
            self.relative_shape_error.shape,
        )
        assert self.normed_base_shape.shape == self.pure_bin_counts.shape, (
            self.normed_base_shape.shape,
            self.pure_bin_counts.shape,
        )

    @property
    def relative_stat_error(self) -> np.ndarray:
        norm: np.ndarray = np.where(
            self.pure_bin_counts > 0,
            self.pure_bin_counts,
            np.ones_like(self.pure_bin_counts),
        )
        return np.sqrt(self.pure_bin_counts) / norm


@dataclass(frozen=True)
class SpecificShapePlotInfoContainer:
    name: str
    latex_str: str
    subset_index: int
    normed_base_shape: np.ndarray
    shape_error: np.ndarray
    stat_error: np.ndarray
    observable: FitObservableInfo
    reco_ch_info: RecoChannelInfo
    component_info: ComponentInfo
    scale_factor: Optional[float] = None
    add_statistical_uncertainty: bool = True

    def __post_init__(self) -> None:
        assert len(self.normed_base_shape.shape) == 1, (len(self.normed_base_shape.shape), self.normed_base_shape.shape)

        assert self.normed_base_shape.shape == self.shape_error.shape, (
            self.normed_base_shape.shape,
            self.shape_error.shape,
        )


class SystematicsShapePlotter:

    output_dir_name: ClassVar[str] = "SystematicsShape"
    plot_name_prefix: ClassVar[str] = "sys_shape_effect_for"

    def __init__(
        self,
        base_output_dir_path: PathType,
        fit_setup: FitSetupTriplet,
        fig_size: Tuple[float, float] = (6.0, 5.0),
        height_ratio: Tuple[float, float] = (3.5, 1.0),
    ) -> None:

        assert os.path.isdir(base_output_dir_path), base_output_dir_path

        self.output_dir_path: PathType = os.path.join(base_output_dir_path, self.output_dir_name)
        self.fig_size: Tuple[float, float] = fig_size
        self.height_ratio: Tuple[float, float] = height_ratio

        self._asimov_fit_setup: FitSetupInfoContainer = fit_setup.asimov
        self._fit_binning_info: FitSetupBinningPerRecoChInfo = FitSetupBinningPerRecoChInfo.init_from(
            fit_setup=self._asimov_fit_setup,
        )

    @property
    def fit_components(self) -> Tuple[ComponentInfo, ...]:
        return self._asimov_fit_setup.components

    @property
    def fit_reco_channels(self) -> Tuple[RecoChannelInfo, ...]:
        return self._asimov_fit_setup.reco_channels

    @property
    def fit_observable_infos(self) -> Tuple[FitObservableInfo, ...]:
        return self._asimov_fit_setup.observable_infos

    @property
    def expected_obs_df_labels(self) -> Tuple[str, ...]:
        return tuple(o.col_name for o in self.fit_observable_infos)

    def calculate_specific_shape_info_for(
        self,
        sys_shape_info: ShapePlotInfoContainer,
        selected_normed_base_shape: np.ndarray,
        selected_relative_shape_error: np.ndarray,
        selected_relative_stat_error: np.ndarray,
        reco_ch_info: RecoChannelInfo,
        component_info: ComponentInfo,
        add_statistical_uncertainty: bool,
    ) -> Generator[SpecificShapePlotInfoContainer, None, None]:

        assert len(selected_normed_base_shape.shape) == 1, selected_normed_base_shape.shape
        assert len(selected_relative_shape_error.shape) == 2, selected_relative_shape_error.shape

        _rel_shape_error_splits: List[np.ndarray] = [
            np.squeeze(v, axis=1)
            for v in np.split(selected_relative_shape_error, selected_relative_shape_error.shape[1], axis=1)
        ]
        assert len(_rel_shape_error_splits) == sys_shape_info.relative_shape_error.shape[2], (
            len(_rel_shape_error_splits),
            sys_shape_info.relative_shape_error.shape[2],
            sys_shape_info.relative_shape_error.shape,
        )

        binning_info: Tuple[int, ...] = tuple(oi.n_bins for oi in reco_ch_info.observable_infos)
        assert np.product(binning_info) == selected_normed_base_shape.shape[0], (
            (np.product(binning_info), selected_normed_base_shape.shape[0]),
            binning_info,
            selected_normed_base_shape.shape,
        )

        # Recreate 2D binning by reshaping 1st dimension
        _2d_normed_base_shape: np.ndarray = selected_normed_base_shape.reshape(binning_info)
        _2d_relative_shape_error_splits: List[np.ndarray] = [v.reshape(binning_info) for v in _rel_shape_error_splits]

        assert len(_2d_normed_base_shape.shape) == 2, _2d_normed_base_shape.shape
        for _this_shape_err_id, _2d_rel_shape_err in enumerate(_2d_relative_shape_error_splits):
            assert _2d_rel_shape_err.shape == _2d_normed_base_shape.shape, (
                _this_shape_err_id,
                _2d_rel_shape_err.shape,
                _2d_normed_base_shape.shape,
                [v.shape for v in _2d_relative_shape_error_splits],
            )

        # Calculate projections for nominal template
        projected_normed_base_shape_list: Tuple[np.ndarray, np.ndarray] = (
            _2d_normed_base_shape.sum(axis=1),
            _2d_normed_base_shape.sum(axis=0),
        )

        # Calculate projections for uncertainties
        projected_shape_error_splits: Tuple[Tuple[np.ndarray, np.ndarray], ...] = tuple(
            (
                (rel_err * _2d_normed_base_shape).sum(axis=1),
                (rel_err * _2d_normed_base_shape).sum(axis=0),
            )
            for rel_err in _2d_relative_shape_error_splits
        )

        sum_of_2d_relative_shape_error_splits: np.ndarray = np.sum(_2d_relative_shape_error_splits, axis=0)
        projected_shape_error_combination: Tuple[np.ndarray, np.ndarray] = (
            (sum_of_2d_relative_shape_error_splits * _2d_normed_base_shape).sum(axis=1),
            (sum_of_2d_relative_shape_error_splits * _2d_normed_base_shape).sum(axis=0),
        )

        # Stat error handling
        _2d_relative_stat_error: np.ndarray = selected_relative_stat_error.reshape(binning_info)
        projected_stat_error: Tuple[np.ndarray, np.ndarray] = (
            np.sqrt((np.square(_2d_relative_stat_error * _2d_normed_base_shape)).sum(axis=1)),
            np.sqrt((np.square(_2d_relative_stat_error * _2d_normed_base_shape)).sum(axis=0)),
        )

        for obs_index, obs_info in enumerate(reco_ch_info.observable_infos):
            assert obs_info.col_name in self.expected_obs_df_labels, (obs_info.col_name, self.expected_obs_df_labels)

            projected_normed_base_shape: np.ndarray = projected_normed_base_shape_list[obs_index]
            assert len(projected_normed_base_shape) == obs_info.n_bins, (
                projected_normed_base_shape.shape,
                obs_info.n_bins,
            )

            if sys_shape_info.plot_subsets:
                for subset_id in range(sys_shape_info.relative_shape_error.shape[2]):
                    yield SpecificShapePlotInfoContainer(
                        name=sys_shape_info.name,
                        latex_str=sys_shape_info.latex_str,
                        subset_index=subset_id,
                        normed_base_shape=projected_normed_base_shape,
                        shape_error=projected_shape_error_splits[subset_id][obs_index],
                        stat_error=projected_stat_error[obs_index],
                        observable=obs_info,
                        reco_ch_info=reco_ch_info,
                        component_info=component_info,
                        scale_factor=sys_shape_info.scale_factor,
                        add_statistical_uncertainty=add_statistical_uncertainty,
                    )

            yield SpecificShapePlotInfoContainer(
                name=sys_shape_info.name,
                latex_str=sys_shape_info.latex_str,
                subset_index=-1,
                normed_base_shape=projected_normed_base_shape,
                shape_error=projected_shape_error_combination[obs_index],
                stat_error=projected_stat_error[obs_index],
                observable=obs_info,
                reco_ch_info=reco_ch_info,
                component_info=component_info,
                scale_factor=sys_shape_info.scale_factor,
                add_statistical_uncertainty=add_statistical_uncertainty,
            )

    def create_systematics_shape_plots(
        self,
        sys_shape_info: ShapePlotInfoContainer,
        add_statistical_uncertainties: bool = True,
    ) -> Generator[SpecificShapePlotInfoContainer, None, None]:
        target_dir_path: PathType = os.path.join(self.output_dir_path, sys_shape_info.name)
        os.makedirs(target_dir_path, exist_ok=True)

        for component_index, comp_info in enumerate(self.fit_components):
            if sys_shape_info.components_to_consider is not None:
                if comp_info.name not in sys_shape_info.components_to_consider:
                    continue

            # Select by component by index in 2nd dimension
            _normed_shape_for_comp: np.ndarray = sys_shape_info.normed_base_shape[:, component_index]
            _rel_shape_error_for_comp: np.ndarray = sys_shape_info.relative_shape_error[:, component_index, :]
            _rel_stat_error_for_comp: np.ndarray = sys_shape_info.relative_stat_error[:, component_index]

            assert len(_normed_shape_for_comp.shape) == 1, _normed_shape_for_comp.shape
            assert len(_rel_shape_error_for_comp.shape) == 2, _rel_shape_error_for_comp.shape

            # Split 1st dimension by reco channel
            _normed_shape_reco_ch_splits: List[np.ndarray] = np.split(
                _normed_shape_for_comp,
                self._fit_binning_info.reco_ch_split_ids,
                axis=0,
            )
            _rel_shape_error_shape_reco_ch_splits: List[np.ndarray] = np.split(
                _rel_shape_error_for_comp,
                self._fit_binning_info.reco_ch_split_ids,
                axis=0,
            )

            assert len(_normed_shape_reco_ch_splits) == len(self.fit_reco_channels), (
                len(_normed_shape_reco_ch_splits),
                len(self.fit_reco_channels),
                len(_rel_shape_error_shape_reco_ch_splits),
            )
            assert len(_rel_shape_error_shape_reco_ch_splits) == len(self.fit_reco_channels), (
                len(_rel_shape_error_shape_reco_ch_splits),
                len(self.fit_reco_channels),
            )

            _rel_shape_error_stat_reco_ch_splits: List[np.ndarray] = np.split(
                _rel_stat_error_for_comp,
                self._fit_binning_info.reco_ch_split_ids,
                axis=0,
            )

            for reco_ch_index, reco_ch_info in enumerate(self.fit_reco_channels):
                if comp_info not in reco_ch_info.components:
                    continue

                for _this_shape_info in self.calculate_specific_shape_info_for(
                    sys_shape_info=sys_shape_info,
                    selected_normed_base_shape=_normed_shape_reco_ch_splits[reco_ch_index],
                    selected_relative_shape_error=_rel_shape_error_shape_reco_ch_splits[reco_ch_index],
                    selected_relative_stat_error=_rel_shape_error_stat_reco_ch_splits[reco_ch_index],
                    reco_ch_info=reco_ch_info,
                    component_info=comp_info,
                    add_statistical_uncertainty=add_statistical_uncertainties,
                ):
                    self.plot_systematics_shape_effect_for(
                        spec_sys_shape_info=_this_shape_info,
                        target_dir_path=target_dir_path,
                    )
                    yield _this_shape_info

    def plot_systematics_shape_effect_for(
        self,
        spec_sys_shape_info: SpecificShapePlotInfoContainer,
        target_dir_path: PathType,
    ) -> None:
        set_matplotlibrc_params()
        final_target_dir_path: PathType = os.path.join(target_dir_path, spec_sys_shape_info.observable.col_name)
        os.makedirs(final_target_dir_path, exist_ok=True)

        fn_subset_str: str = (
            "All" if spec_sys_shape_info.subset_index == -1 else f"SubSet{spec_sys_shape_info.subset_index}"
        )

        plot_file_name: str = f"{self.plot_name_prefix}"
        plot_file_name += f"_{spec_sys_shape_info.observable.col_name}"
        plot_file_name += f"_comp{spec_sys_shape_info.component_info.name}_ch{spec_sys_shape_info.reco_ch_info.name}"
        plot_file_name += f"_{spec_sys_shape_info.name}_{fn_subset_str}"

        fig, axes = plt.subplots(
            nrows=2,
            ncols=1,
            figsize=self.fig_size,
            dpi=300,
            sharex="all",
            gridspec_kw={"height_ratios": [self.height_ratio[0], self.height_ratio[1]]},
        )  # type: FigureType, Sequence[AxesType]

        assert len(axes) == 2, len(axes)
        ax1: AxesType = axes[0]
        ax2: AxesType = axes[1]

        legend_loc: int = 1
        y_axis_scale: float = 1.2

        if spec_sys_shape_info.scale_factor is not None:
            scale_factor: float = spec_sys_shape_info.scale_factor
            scale_info_text: str = r" ($\times " + f"{scale_factor}$)"
        else:
            scale_factor = 1.0
            scale_info_text = ""

        bin_edges: np.ndarray = np.array(spec_sys_shape_info.observable.bin_edges)
        bin_mids: np.ndarray = (bin_edges[1:] + bin_edges[:-1]) / 2.0
        bin_widths: np.ndarray = bin_edges[1:] - bin_edges[:-1]

        norm_bin_count: np.ndarray = spec_sys_shape_info.normed_base_shape
        u_bin_shape_uncert: np.ndarray = spec_sys_shape_info.shape_error
        d_bin_shape_uncert: np.ndarray = -1.0 * spec_sys_shape_info.shape_error

        bin_stat_uncert: np.ndarray = spec_sys_shape_info.stat_error

        assert np.all(u_bin_shape_uncert * d_bin_shape_uncert <= 0.0), (
            np.sum(u_bin_shape_uncert * d_bin_shape_uncert > 0.0),
            u_bin_shape_uncert * d_bin_shape_uncert,
        )

        # bin_shape_uncert_height: np.ndarray = np.abs(u_bin_shape_uncert) + np.abs(d_bin_shape_uncert)
        bin_shape_uncert_lower_value: np.ndarray = np.minimum(u_bin_shape_uncert, d_bin_shape_uncert)
        assert np.all(bin_shape_uncert_lower_value <= 0.0), (
            np.sum(bin_shape_uncert_lower_value > 0.0),
            bin_shape_uncert_lower_value,
        )

        ax1.hist(
            x=bin_mids,
            bins=bin_edges,
            weights=norm_bin_count,
            stacked=True,
            edgecolor=KITColors.kit_black,
            lw=0.3,
            color=spec_sys_shape_info.component_info.color,
            label=spec_sys_shape_info.component_info.latex_label,
            histtype="stepfilled",
        )

        # ax1.bar(
        #     x=bin_mids,
        #     height=scale_factor * bin_shape_uncert_height,
        #     width=bin_widths,
        #     bottom=norm_bin_count + scale_factor * bin_shape_uncert_lower_value,
        #     color="black",
        #     hatch="///////",
        #     fill=False,
        #     lw=0,
        #     label=f"sys. shape unc. {scale_info_text}",
        # )

        ax1.bar(
            x=bin_mids,
            height=scale_factor * np.abs(u_bin_shape_uncert),
            width=bin_widths,
            bottom=norm_bin_count
            + scale_factor * np.where(u_bin_shape_uncert < 0.0, u_bin_shape_uncert, np.zeros_like(u_bin_shape_uncert)),
            color=KITColors.kit_black,
            hatch="//////",
            fill=False,
            lw=0,
            label=f"up sys.{scale_info_text}",
        )
        ax1.bar(
            x=bin_mids,
            height=scale_factor * np.abs(d_bin_shape_uncert),
            width=bin_widths,
            bottom=norm_bin_count
            + scale_factor * np.where(d_bin_shape_uncert < 0.0, d_bin_shape_uncert, np.zeros_like(d_bin_shape_uncert)),
            color=KITColors.kit_black,
            hatch="\\\\\\\\\\\\",
            fill=False,
            lw=0,
            label=f"down sys. {scale_info_text}",
        )

        # Pull Plot
        if spec_sys_shape_info.add_statistical_uncertainty:
            ax2.bar(
                x=bin_mids,
                height=2 * bin_stat_uncert,
                width=bin_widths,
                bottom=-1.0 * bin_stat_uncert,
                color=KITColors.grey,
                alpha=0.4,
                fill=True,
                lw=0,
                label="stat.",
            )
        ax2.hist(
            x=[bin_mids, bin_mids],
            bins=bin_edges,
            weights=[u_bin_shape_uncert, d_bin_shape_uncert],
            stacked=False,
            color=[KITColors.kit_blue, KITColors.kit_orange],
            lw=1.3,
            label=["up", "down"],  # type: ignore  # The type here is correct!
            histtype="step",
        )

        ax1.set_xlim(bin_edges[0], bin_edges[-1])
        ax2.set_xlim(bin_edges[0], bin_edges[-1])

        y_limits = ax1.get_ylim()
        ax1.set_ylim(bottom=y_limits[0], top=y_axis_scale * y_limits[1])

        title_subset_str: str = "all" if spec_sys_shape_info.subset_index == -1 else f"{spec_sys_shape_info.subset_index}"
        plot_title: str = (
            spec_sys_shape_info.latex_str
            + f" ({title_subset_str}) Shape Sys. in "
            + spec_sys_shape_info.reco_ch_info.latex_label
        )

        ax1.set_title(plot_title, fontsize=20)

        ax1.set_ylabel(r"Normalized Bin Count", fontsize=18, **ylabel_pos)
        ax2.set_xlabel(spec_sys_shape_info.observable.hist_var.x_label, fontsize=18, **xlabel_pos)

        ax1.legend(frameon=False, loc=legend_loc, ncol=1, fontsize=14)
        ax2.legend(frameon=False, loc=legend_loc, ncol=1, fontsize=6)

        export(
            fig=fig,
            filename=plot_file_name,
            target_dir=final_target_dir_path,
            close_figure=False,
        )


# endregion
