"""
R(D) and R(D*) result significance ellipse plotter.

Felix Metzner 2023
"""


import os
import numpy as np
import matplotlib.pyplot as plt
import matplotlib.transforms as transforms

from uncertainties import ufloat
from dataclasses import dataclass
from matplotlib.patches import Ellipse
from typing import Tuple, Optional, ClassVar

from templatefitter.plotter.plot_utilities import export, AxesType
from templatefitter.plotter.plot_style import KITColors, set_matplotlibrc_params, xlabel_pos, ylabel_pos

from rdstar.utilities import PathType
from rdstar.utilities.logging import log_to_default

from rdstar.branching_fractions import constants_and_measurements as bf_calc_consts

from rdstar.offline_analysis.fitting.dedicated_fit_approach.fit_result_container import RDStarFitResultContainer


__all__ = [
    "RDStarMoneyPlotter",
]


logger = log_to_default(name=__name__)


# region money plot info container


@dataclass(frozen=True)
class MeasurementValue:
    ratio_value: float
    stat_error: float
    full_error: float
    fixed_sys_error: Optional[float] = None

    @property
    def sys_error(self) -> float:
        if self.fixed_sys_error is not None:
            return self.fixed_sys_error
        else:
            return np.sqrt(np.square(self.full_error) - np.square(self.stat_error))


@dataclass(frozen=True)
class MoneyPlotInfoContainer:
    label: str
    color: str
    rd_value: MeasurementValue
    rdstar_value: MeasurementValue
    rd_rdstar_stat_rho: float
    rd_rdstar_full_rho: float
    fixed_rd_rdstar_sys_rho: Optional[float] = None

    @property
    def rd_rdstar_sys_rho(self) -> float:
        if self.fixed_rd_rdstar_sys_rho is None:
            sys_cov: np.ndarray = self.full_cov - self.stat_cov
            return sys_cov[0][1] / np.sqrt(sys_cov[0][0] * sys_cov[1][1])
        else:
            return self.fixed_rd_rdstar_sys_rho

    @property
    def full_cov(self) -> np.ndarray:
        a: float = self.rd_value.full_error
        b: float = self.rdstar_value.full_error
        ab: float = a * b * self.rd_rdstar_full_rho
        return np.array([[a ** 2, ab], [ab, b ** 2]])

    @property
    def stat_cov(self) -> np.ndarray:
        a: float = self.rd_value.stat_error
        b: float = self.rdstar_value.stat_error
        ab: float = a * b * self.rd_rdstar_stat_rho
        return np.array([[a ** 2, ab], [ab, b ** 2]])

    @property
    def sys_cov(self) -> np.ndarray:
        a: float = self.rd_value.sys_error
        b: float = self.rdstar_value.sys_error
        ab: float = a * b * self.rd_rdstar_sys_rho
        return np.array([[a ** 2, ab], [ab, b ** 2]])

    @staticmethod
    def from_fit_result(
        label: str,
        color: str,
        full_fit_result: RDStarFitResultContainer,
        stat_fit_result: RDStarFitResultContainer,
    ) -> "MoneyPlotInfoContainer":
        return MoneyPlotInfoContainer(
            label=label,
            color=color,
            rd_value=MeasurementValue(
                ratio_value=full_fit_result.param_values[0],
                stat_error=stat_fit_result.errors[0],
                full_error=full_fit_result.errors[0],
            ),
            rdstar_value=MeasurementValue(
                ratio_value=full_fit_result.param_values[1],
                stat_error=stat_fit_result.errors[1],
                full_error=full_fit_result.errors[1],
            ),
            rd_rdstar_stat_rho=stat_fit_result.corr_mat[0][1],
            rd_rdstar_full_rho=full_fit_result.corr_mat[0][1],
        )


# endregion


belle_2015_result: MoneyPlotInfoContainer = MoneyPlotInfoContainer(
    label="Belle 2015",
    color=KITColors.kit_blue,
    rd_value=MeasurementValue(
        ratio_value=0.375,
        stat_error=0.064,
        full_error=0.0690797,
        fixed_sys_error=0.026,
    ),
    rdstar_value=MeasurementValue(
        ratio_value=0.293,
        stat_error=0.038,
        full_error=0.0408534,
        fixed_sys_error=0.015,
    ),
    rd_rdstar_stat_rho=-0.56,
    rd_rdstar_full_rho=-0.50,
    fixed_rd_rdstar_sys_rho=-0.11,
)


# region RDStar fit plotter class
class RDStarMoneyPlotter:

    sm_rd_value: ClassVar[ufloat] = bf_calc_consts.sm_r_d
    sm_rd_star_value: ClassVar[ufloat] = bf_calc_consts.sm_r_d_st

    default_this_sm_label: ClassVar[str] = "This work (SM)"
    default_this_np_label: ClassVar[str] = "This work (NP)"
    default_this_color: ClassVar[str] = KITColors.kit_orange

    output_dir_name: ClassVar[str] = "RD_RDs_MoneyPlot"
    plot_file_name: ClassVar[str] = "rd_vs_rdstar_money_plot"

    def __init__(
        self,
        evaluation_name: str,
        base_output_path: PathType,
        full_rd_rdstar_sm_result: RDStarFitResultContainer,
        stat_rd_rdstar_sm_result: RDStarFitResultContainer,
        full_rd_rdstar_np_result: RDStarFitResultContainer,
        stat_rd_rdstar_np_result: RDStarFitResultContainer,
        this_sm_label: Optional[str] = None,
        this_np_label: Optional[str] = None,
    ) -> None:

        self.evaluation_name: str = evaluation_name
        self.base_output_path: PathType = base_output_path

        self.output_path: PathType = os.path.join(self.base_output_path, self.output_dir_name, self.evaluation_name)

        self.this_sm_label: str = self.default_this_sm_label if this_sm_label is None else this_sm_label
        self.this_np_label: str = self.default_this_np_label if this_np_label is None else this_np_label

        self.this_sm_full_result: RDStarFitResultContainer = full_rd_rdstar_sm_result
        self.this_sm_stat_result: RDStarFitResultContainer = stat_rd_rdstar_sm_result
        self.this_np_full_result: RDStarFitResultContainer = full_rd_rdstar_np_result
        self.this_np_stat_result: RDStarFitResultContainer = stat_rd_rdstar_np_result

        self.this_sm_info: MoneyPlotInfoContainer = MoneyPlotInfoContainer.from_fit_result(
            label=self.this_sm_label,
            color=self.default_this_color,
            full_fit_result=self.this_sm_full_result,
            stat_fit_result=self.this_sm_stat_result,
        )
        self.this_np_info: MoneyPlotInfoContainer = MoneyPlotInfoContainer.from_fit_result(
            label=self.this_np_label,
            color=self.default_this_color,
            full_fit_result=self.this_np_full_result,
            stat_fit_result=self.this_np_stat_result,
        )

        self.all_measurements: Tuple[MoneyPlotInfoContainer, ...] = (
            belle_2015_result,
            self.this_np_info,
            self.this_sm_info,
        )

    @staticmethod
    def confidence_ellipse(
        x: float,
        dx: float,
        y: float,
        dy: float,
        rho: float,
        ax: AxesType,
        n_std: float = 1.0,
        alpha: float = 0.3,
        face_color: Optional[str] = None,
        **kwargs,
    ):

        cov: np.ndarray = np.array([[dx ** 2, dx * dy * rho], [dx * dy * rho, dy ** 2]])

        scale_x: float = np.sqrt(cov[0, 0]) * n_std
        mean_x: float = x

        scale_y: float = np.sqrt(cov[1, 1]) * n_std
        mean_y: float = y

        ell_radius_x: float = np.sqrt(1 + rho)
        ell_radius_y: float = np.sqrt(1 - rho)

        ellipse: Ellipse = Ellipse(
            xy=(0, 0),
            width=ell_radius_x * 2,
            height=ell_radius_y * 2,
            facecolor="none" if face_color is None else face_color,
            **kwargs,
        )

        if face_color is not None:
            ellipse.set_facecolor(face_color)
            ellipse.set_alpha(alpha)

        transform = transforms.Affine2D().rotate_deg(45).scale(scale_x, scale_y).translate(mean_x, mean_y)

        ellipse.set_transform(transform + ax.transData)

        return ax.add_patch(ellipse)

    def produce_significance_plot(
        self,
    ) -> None:
        set_matplotlibrc_params()

        os.makedirs(self.output_path, exist_ok=True)

        fig, ax = plt.subplots(figsize=(6, 6), dpi=300)

        ax.scatter(
            self.sm_rd_value.nominal_value,
            self.sm_rd_star_value.nominal_value,
            c=KITColors.grey,
            s=60,
            marker="+",
            label="Theory Expectation",
        )

        ax.set_xlim(0.2, 0.5)
        ax.set_ylim(0.2, 0.38)

        for measurement_info in self.all_measurements:

            x: float = measurement_info.rd_value.ratio_value
            y: float = measurement_info.rdstar_value.ratio_value

            stat_dx: float = measurement_info.rd_value.stat_error
            stat_dy: float = measurement_info.rdstar_value.stat_error

            full_dx: float = measurement_info.rd_value.full_error
            full_dy: float = measurement_info.rdstar_value.full_error

            stat_rho: float = measurement_info.rd_rdstar_stat_rho
            full_rho: float = measurement_info.rd_rdstar_full_rho

            if measurement_info.color is not None:
                if full_dx is None:
                    _label: str = measurement_info.label + " (stat. only)"
                else:
                    _label = measurement_info.label + " (stat. and dashed sys.)"

                self.confidence_ellipse(
                    x=x,
                    dx=stat_dx,
                    y=y,
                    dy=stat_dy,
                    rho=stat_rho,
                    ax=ax,
                    n_std=1.0,
                    alpha=0.75,
                    label=_label,
                    face_color=measurement_info.color,
                    edgecolor=measurement_info.color,
                )

                ax.scatter(x, y, color=measurement_info.color, s=6)

                if full_dx is not None and full_dy is not None:
                    self.confidence_ellipse(
                        x=x,
                        dx=full_dx,
                        y=y,
                        dy=full_dy,
                        rho=full_rho,
                        ax=ax,
                        n_std=1.0,
                        alpha=1.0,
                        label="",
                        face_color=None,
                        edgecolor=measurement_info.color,
                        linestyle="--",
                    )
            else:
                self.confidence_ellipse(
                    x=x,
                    dx=stat_dx,
                    y=y,
                    dy=stat_dy,
                    rho=stat_rho,
                    ax=ax,
                    n_std=1.0,
                    alpha=0.6,
                    label=measurement_info.label,
                    face_color=KITColors.grey,
                    edgecolor=KITColors.kit_black,
                )
                ax.scatter(x, y, color=KITColors.dark_grey, s=5)

        plt.xlabel(r"$\mathcal{R}(\mathrm{D})$", fontsize=18, **xlabel_pos)
        plt.ylabel(r"$\mathcal{R}(\mathrm{D}^{*})$", fontsize=18, **ylabel_pos)

        ax.legend()

        handles, labels = ax.get_legend_handles_labels()
        ax.legend(handles[::-1], labels[::-1], loc=2)

        export(
            fig=fig,
            filename=self.plot_file_name,
            target_dir=self.output_path,
            close_figure=False,
        )


# endregion
