"""
Definition of fit routine class which executes a fit.

Felix Metzner 2021
"""

import os
import json
import pandas as pd

from dataclasses import dataclass
from typing import Optional, Tuple, List, Dict, NamedTuple, Any

from rdstar.utilities.logging import log_to_default
from rdstar.utilities import PathType, merge_plot_pdf_files

from rdstar.offline_analysis.fitting.fit_case import FitCase
from rdstar.offline_analysis.fitting.mva_cut_info import MVACutInfo
from rdstar.offline_analysis.observables.common_observables import Cols
from rdstar.offline_analysis.fitting.fit_model_creator import FitModelCreator
from rdstar.offline_analysis.mc_matching.sig_id_costomizer_tools import SigIDAndGenIDCombiner
from rdstar.offline_analysis.result_calculation.ratio_calculator import RatioCalculator, UpDownUValue
from rdstar.offline_analysis.corrections.efficiency_descrepancy_corrector import ReconstructionModeEfficiencyCorrector
from rdstar.offline_analysis.efficiency_calculation.efficiency_systematics_calculator import EfficiencySysInfoContainer
from rdstar.offline_analysis.efficiency_calculation.simple_efficiency_calculator import (
    EfficiencyNumbersContainer,
    SimpleEfficiencyCalculator,
)

from templatefitter.fitter import TemplateFitter
from templatefitter.minimizer import MinimizeResult
from templatefitter.plotter.fit_result_plots import FitResultPlotter
from templatefitter.plotter.fit_template_plots import FitTemplatesPlotter


__all__ = [
    "FitRoutine",
    "FitResultContainer",
]


logger = log_to_default(name=__name__)


class FitOutputInfo(NamedTuple):
    key: str
    name: str
    path: PathType
    tmp_path: PathType

    def get_output_dict_info(self) -> Dict[str, PathType]:
        return {self.key: self.path}


@dataclass
class MutableFitResultContainer:
    fit_result: MinimizeResult
    file_output_dict: Dict[str, PathType]
    efficiency_info: Optional[EfficiencyNumbersContainer] = None
    efficiency_sys_info: Optional[EfficiencySysInfoContainer] = None
    rd_ratio_result: Optional[UpDownUValue] = None
    rdstar_ratio_result: Optional[UpDownUValue] = None
    rddstar_ratio_result: Optional[UpDownUValue] = None

    def updated_output_dict(self, new_items: Dict[str, PathType]) -> None:
        assert isinstance(new_items, dict), type(new_items).__name__
        assert all(k not in self.file_output_dict.keys() for k in new_items.keys()), [
            k for k in new_items.keys() if k in self.file_output_dict.keys()
        ]
        assert all(v not in self.file_output_dict.values() for v in new_items.values()), [
            v for v in new_items.values() if v in self.file_output_dict.values()
        ]
        initial_len: int = len(self.file_output_dict)
        self.file_output_dict.update(new_items)
        assert len(self.file_output_dict) == initial_len + len(new_items), (
            len(self.file_output_dict),
            initial_len,
            len(new_items),
        )


class FitResultContainer(NamedTuple):
    fit_result: MinimizeResult
    file_output_dict: Dict[str, PathType]
    efficiency_info: Optional[EfficiencyNumbersContainer] = None
    efficiency_sys_info: Optional[EfficiencySysInfoContainer] = None
    rd_ratio_result: Optional[UpDownUValue] = None
    rdstar_ratio_result: Optional[UpDownUValue] = None
    rddstar_ratio_result: Optional[UpDownUValue] = None

    @staticmethod
    def from_mutable(mutable_fit_result_container: MutableFitResultContainer) -> "FitResultContainer":
        return FitResultContainer(
            fit_result=mutable_fit_result_container.fit_result,
            file_output_dict=mutable_fit_result_container.file_output_dict,
            efficiency_info=mutable_fit_result_container.efficiency_info,
            efficiency_sys_info=mutable_fit_result_container.efficiency_sys_info,
            rd_ratio_result=mutable_fit_result_container.rd_ratio_result,
            rdstar_ratio_result=mutable_fit_result_container.rdstar_ratio_result,
            rddstar_ratio_result=mutable_fit_result_container.rddstar_ratio_result,
        )

    def as_dict(self) -> Dict[str, Any]:
        return {
            "fit_result": self.fit_result.as_dict,
            "file_output_dict": self.file_output_dict,
            "efficiency_info": None if self.efficiency_info is None else self.efficiency_info.as_dict(),
            "efficiency_sys_info": None if self.efficiency_sys_info is None else self.efficiency_sys_info.as_dict(),
            "rd_ratio_result": None if self.rd_ratio_result is None else self.rd_ratio_result.as_dict,
            "rdstar_ratio_result": None if self.rdstar_ratio_result is None else self.rdstar_ratio_result.as_dict,
            "rddstar_ratio_result": None if self.rddstar_ratio_result is None else self.rddstar_ratio_result.as_dict,
        }

    @staticmethod
    def from_dict(input_dict: Dict[str, Any]) -> "FitResultContainer":
        eff_dict: Optional[EfficiencyNumbersContainer] = None
        if input_dict["efficiency_info"] is not None:
            eff_dict = EfficiencyNumbersContainer()
            eff_dict.fill_from_dict(dict_input=input_dict["efficiency_info"], convert_keys=True)

        return FitResultContainer(
            fit_result=MinimizeResult.from_dict(result_dict=input_dict["fit_result"]),
            file_output_dict=input_dict["file_output_dict"],
            efficiency_info=eff_dict,
            efficiency_sys_info=None
            if input_dict["efficiency_sys_info"] is None
            else EfficiencySysInfoContainer.from_dict(input_dict=input_dict["efficiency_sys_info"]),
            rd_ratio_result=None
            if input_dict["rd_ratio_result"] is None
            else UpDownUValue.from_dict(input_dict=input_dict["rd_ratio_result"]),
            rdstar_ratio_result=None
            if input_dict["rdstar_ratio_result"] is None
            else UpDownUValue.from_dict(input_dict=input_dict["rdstar_ratio_result"]),
            rddstar_ratio_result=None
            if input_dict["rddstar_ratio_result"] is None
            else UpDownUValue.from_dict(input_dict=input_dict["rddstar_ratio_result"]),
        )

    def write_to_file(self, file_path: PathType) -> None:
        with open(file_path, "w") as json_file:
            json.dump(self.as_dict(), json_file, indent=2)

    @staticmethod
    def from_file(file_path: PathType) -> "FitResultContainer":
        with open(file_path) as f:
            input_dict = json.load(f)
        assert isinstance(input_dict, dict), type(input_dict).__name__
        return FitResultContainer.from_dict(input_dict=input_dict)


class FitRoutine:
    fit_result_output_key: str = "FitResult"
    fit_significance_output_key: str = "FitSignificance"
    efficiency_info_key: str = "EfficiencyInfo"
    efficiency_sys_info_key: str = "EfficiencySystematicsInfo"
    pre_fit_plot_output_key: str = "PreFitPlot"
    post_fit_plot_output_key: str = "PostFitPlot"
    result_info_key: str = "ResultContainer"

    output_info: Tuple[Tuple[str, str], ...] = (
        (fit_result_output_key, "data"),
        (fit_significance_output_key, "data"),
        (efficiency_info_key, "data"),
        (efficiency_sys_info_key, "data"),
        (pre_fit_plot_output_key, "plot"),
        (post_fit_plot_output_key, "plot"),
        (result_info_key, "data"),
    )

    output_type_ext_map: Dict[str, str] = {
        "data": "json",
        "plot": "pdf",
    }

    def __init__(
        self,
        fit_case: FitCase,
        base_output_dir_path: PathType,
        file_dict_dir_path: PathType,
        mva_cut_info: Optional[MVACutInfo] = None,
    ) -> None:
        self._fit_case: FitCase = fit_case

        self.mva_cut_info: Optional[MVACutInfo] = mva_cut_info

        self.base_path: PathType = base_output_dir_path
        self.file_dict_dir_path: PathType = file_dict_dir_path

    def run_fit_routine(
        self,
        mc_sample: pd.DataFrame,
        data_sample: Optional[pd.DataFrame] = None,
    ) -> FitResultContainer:
        if self.fit_case.asimov_data:
            assert data_sample is None
            selected_data_sample: Optional[pd.DataFrame] = None
        else:
            assert data_sample is not None
            selected_data_sample = self.apply_scope_cuts(df=data_sample)

        sig_id_customizer: Optional[SigIDAndGenIDCombiner] = self.fit_case.get_additional_setting_with_default(
            setting_key="sig_id_customizer",
            default_setting=None,
        )
        if sig_id_customizer is not None:
            assert isinstance(sig_id_customizer, SigIDAndGenIDCombiner), type(sig_id_customizer).__name__
            sig_id_customizer.customize_sig_id(df=mc_sample)

        selected_mc_sample: pd.DataFrame = self.apply_scope_cuts(df=mc_sample)

        result_container: MutableFitResultContainer = self._run_fit(
            mc_sample=selected_mc_sample,
            data_sample=selected_data_sample,
        )

        eff_info = SimpleEfficiencyCalculator(
            df=selected_mc_sample,
            sig_id_column_name=self.fit_case.sig_id_column_name,
            sig_id_data_type_mapping=self.fit_case.sig_id_to_data_type_mapping,
            file_dict_base_path=self.file_dict_dir_path,
            use_new_fei_calibration_weights=self.fit_case.add_new_fei_calibration_factors,
            additional_scaling_weight_columns=self.get_additional_efficiency_scaling_columns(),
        )
        eff_info_output_info: FitOutputInfo = self.get_output_info(output_key=self.efficiency_info_key)

        eff_info.write_eff_dict_to_file(file_path=eff_info_output_info.tmp_path)
        os.rename(src=eff_info_output_info.tmp_path, dst=eff_info_output_info.path)

        result_container.efficiency_info = eff_info.get_efficiency_info_container()
        result_container.updated_output_dict(new_items=eff_info_output_info.get_output_dict_info())

        # Efficiency Systematics
        eff_systematics_info_container: EfficiencySysInfoContainer = self.create_efficiency_sys_info_container()
        eff_systematics_info_container.fill(df=selected_mc_sample)
        eff_sys_info_output_info: FitOutputInfo = self.get_output_info(output_key=self.efficiency_sys_info_key)
        eff_systematics_info_container.write_to_file(target_file_path=eff_sys_info_output_info.path)
        result_container.efficiency_sys_info = eff_systematics_info_container
        result_container.updated_output_dict(new_items=eff_sys_info_output_info.get_output_dict_info())

        # Ratio calculation
        if self.fit_case.sample_key in ("q2_sideband", "mbc_sideband"):
            rd_dst_calculator = RatioCalculator(
                fit_case=self.fit_case,
                fit_result=result_container.fit_result,
                numerator_component_key="Dl",
                denominator_component_key="DSl",
                efficiency_dict=eff_info.eff_numbers_container,
                efficiencies_sys_container=eff_systematics_info_container,
            )
            result_container.rddstar_ratio_result = rd_dst_calculator.efficiency_corrected_ratio
        elif self.fit_case.sample_key == "":
            rd_calculator = RatioCalculator(
                fit_case=self.fit_case,
                fit_result=result_container.fit_result,
                numerator_component_key="Dtau",
                denominator_component_key="Dl",
                efficiency_dict=eff_info.eff_numbers_container,
                efficiencies_sys_container=eff_systematics_info_container,
                factor=2.0,
            )
            rdst_calculator = RatioCalculator(
                fit_case=self.fit_case,
                fit_result=result_container.fit_result,
                numerator_component_key="DStau",
                denominator_component_key="DSl",
                efficiency_dict=eff_info.eff_numbers_container,
                efficiencies_sys_container=eff_systematics_info_container,
                factor=2.0,
            )
            result_container.rd_ratio_result = rd_calculator.efficiency_corrected_ratio
            result_container.rdstar_ratio_result = rdst_calculator.efficiency_corrected_ratio
        else:
            raise KeyError(f"No ratio calculation defined for fit case with sample key '{self.fit_case.sample_key}'!")

        result_output_info: FitOutputInfo = self.get_output_info(output_key=self.result_info_key)
        result_container.updated_output_dict(new_items=result_output_info.get_output_dict_info())

        self.check_output_dict(output_dict=result_container.file_output_dict)
        results = FitResultContainer.from_mutable(mutable_fit_result_container=result_container)

        results.write_to_file(file_path=result_output_info.tmp_path)
        os.rename(src=result_output_info.tmp_path, dst=result_output_info.path)

        return results

    def apply_scope_cuts(self, df: pd.DataFrame) -> pd.DataFrame:
        query_str_list: List[str] = [
            self.fit_case.get_cut_query(reco_mode_column_str=Cols.decay_mode_column_name),
            #
        ]

        mva_cut_info: Optional[MVACutInfo] = self.fit_case.get_additional_setting_with_default(
            setting_key=MVACutInfo.fit_case_setting_key,
            default_setting=None,
        )

        if mva_cut_info is None:
            mva_cut_info = self.mva_cut_info

        if mva_cut_info is not None:
            assert isinstance(mva_cut_info, MVACutInfo), type(mva_cut_info)
            mva_cut_info.check_columns(df=df)
            mva_cut_info.check_reco_modes(fit_case_reco_mode_id_map=self.fit_case.get_reco_mode_id_map())

            mva_cut_str = mva_cut_info.get_cut_query_str(fit_case_reco_mode_id_map=self.fit_case.get_reco_mode_id_map())
            query_str_list.append(mva_cut_str)

        query_str: str = " & ".join(query_str_list)
        return df.query(query_str, inplace=False).reset_index(drop=True, inplace=False)

    def _run_fit(
        self,
        mc_sample: pd.DataFrame,
        data_sample: Optional[pd.DataFrame] = None,
    ) -> MutableFitResultContainer:
        output_dict: Dict[str, PathType] = dict()

        fit_model_creator = FitModelCreator(fit_case=self.fit_case_name)

        fit_model_creator.setup_model(
            mc_df=mc_sample,
            data_df=data_sample,
            use_asimov_data=self.fit_case.asimov_data,
            round_asimov_data_bin_counts=self.fit_case.round_asimov_data,
        )
        fit_model = fit_model_creator.fit_model

        if self.fit_case.create_template_plots or self.fit_case.create_2d_template_plots:
            template_plotter = FitTemplatesPlotter(
                variables_by_channel=fit_model_creator.hist_variables_by_mode_dict,
                fit_model=fit_model,
                channel_label_dict=self.fit_case.reco_mode_labels_dict,
            )

            if self.fit_case.create_template_plots:
                template_plotter.plot_projected_templates(
                    use_initial_values=True,
                    output_dir_path=self.base_path,
                    output_name_tag=self.fit_case.output_tag,
                )

            if self.fit_case.create_2d_template_plots and self.fit_case.has_channels_with_multiple_dimensions:
                template_plotter.plot_2d_templates(
                    use_initial_values=True,
                    output_dir_path=self.base_path,
                    output_name_tag=self.fit_case.output_tag,
                    base_color=None,
                    alternative_temp_color=self.fit_case.alternative_template_colors,
                )

        fit_result_plotter, pre_fit_plot_paths = None, {}  # type: Optional[FitResultPlotter], Dict[str, List[PathType]]
        if self.fit_case.create_plots or self.fit_case.create_projection_plots:
            fit_result_plotter = FitResultPlotter(
                variables_by_channel=fit_model_creator.hist_variables_by_mode_dict,
                fit_model=fit_model,
                reference_dimension=0,
                channel_label_dict=self.fit_case.reco_mode_labels_dict,
                data_label=self.fit_case.plot_info_text,
            )

            if self.fit_case.create_plots:
                pre_fit_plot_paths = fit_result_plotter.plot_fit_result(
                    use_initial_values=True,
                    output_dir_path=self.base_path,
                    output_name_tag=self.fit_case.output_tag,
                )
                assert len(pre_fit_plot_paths["pdf"]) == len(pre_fit_plot_paths["png"]), (
                    len(pre_fit_plot_paths["pdf"]),
                    len(pre_fit_plot_paths["png"]),
                )
                pre_fit_plot_o_info: FitOutputInfo = self.get_output_info(output_key=self.pre_fit_plot_output_key)
                self.add_to_output_dict(output_dict=output_dict, output_info=pre_fit_plot_o_info)
                merge_plot_pdf_files(input_paths=pre_fit_plot_paths["pdf"], output_path=pre_fit_plot_o_info.tmp_path)
                os.rename(src=pre_fit_plot_o_info.tmp_path, dst=pre_fit_plot_o_info.path)

            if self.fit_case.create_projection_plots:
                for fit_dim in range(self.fit_case.max_dimensions):
                    fit_result_plotter.plot_fit_result_projections(
                        project_to=fit_dim,
                        use_initial_values=True,
                        output_dir_path=self.base_path,
                        output_name_tag=self.fit_case.output_tag,
                    )

        fitter = TemplateFitter(
            fit_model=fit_model,
            minimizer_id=self.fit_case.minimizer_id,
        )
        fit_result: MinimizeResult = fitter.do_fit(
            update_templates=True,
            get_hesse=True,
            verbose=True,
            fix_nui_params=self.fit_case.fix_nuisance_parameters,
        )

        fit_result_output_info: FitOutputInfo = self.get_output_info(output_key=self.fit_result_output_key)
        self.add_to_output_dict(output_dict=output_dict, output_info=fit_result_output_info)
        fit_result.save_to(path=fit_result_output_info.tmp_path, overwrite=True)
        os.rename(src=fit_result_output_info.tmp_path, dst=fit_result_output_info.path)

        if self.fit_case.create_plots:
            assert fit_result_plotter is not None
            post_fit_plot_paths = fit_result_plotter.plot_fit_result(
                output_dir_path=self.base_path,
                output_name_tag=self.fit_case.output_tag,
            )
            assert len(post_fit_plot_paths["pdf"]) == len(post_fit_plot_paths["png"]), (
                len(post_fit_plot_paths["pdf"]),
                len(post_fit_plot_paths["png"]),
            )
            assert len(post_fit_plot_paths["pdf"]) == len(pre_fit_plot_paths["pdf"]), (
                len(post_fit_plot_paths["pdf"]),
                len(pre_fit_plot_paths["pdf"]),
            )
            post_fit_plot_o_info: FitOutputInfo = self.get_output_info(output_key=self.post_fit_plot_output_key)
            self.add_to_output_dict(output_dict=output_dict, output_info=post_fit_plot_o_info)
            merge_plot_pdf_files(input_paths=post_fit_plot_paths["pdf"], output_path=post_fit_plot_o_info.tmp_path)
            os.rename(src=post_fit_plot_o_info.tmp_path, dst=post_fit_plot_o_info.path)

        if self.fit_case.create_projection_plots:
            assert fit_result_plotter is not None
            for fit_dim in range(self.fit_case.max_dimensions):
                fit_result_plotter.plot_fit_result_projections(
                    project_to=fit_dim,
                    use_initial_values=False,
                    output_dir_path=self.base_path,
                    output_name_tag=self.fit_case.output_tag,
                )

        if self.fit_case.run_significance_evaluation:
            significance_dict: Dict[str, float] = dict()
            _fixed_component_yields: List[str] = [
                FitModelCreator.get_yield_parameter_name(component_name=c) for c in self.fit_case.fix_component_yields
            ]

            for yield_parameter in fit_model.get_yield_parameter_names():
                if yield_parameter in _fixed_component_yields:
                    continue
                try:
                    significance: float = fitter.get_significance(
                        yield_parameter=yield_parameter,
                        fix_nui_params=True,
                        catch_exception=True,
                    )
                except IndexError as ie:
                    logger.warning(
                        f"Catching IndexError in significance calculation for yield parameter "
                        f"'{yield_parameter}'. The error was:\n{str(ie)}"
                    )
                    significance = -1.0

                significance_dict.update({yield_parameter: significance})

            fit_sig_output_info: FitOutputInfo = self.get_output_info(output_key=self.fit_significance_output_key)
            self.add_to_output_dict(output_dict=output_dict, output_info=fit_sig_output_info)
            with open(fit_sig_output_info.tmp_path, "w") as fp_sig:
                json.dump(obj=significance_dict, fp=fp_sig, indent=2)
            os.rename(src=fit_sig_output_info.tmp_path, dst=fit_sig_output_info.path)

        return MutableFitResultContainer(
            fit_result=fit_result,
            file_output_dict=output_dict,
        )

    def create_efficiency_sys_info_container(self) -> EfficiencySysInfoContainer:
        relevant_component_names: List[str] = ["Dl", "DSl"]  # Normalization Mode in Sideband Case
        if self.fit_case.sample_key == "":
            relevant_component_names.extend(["Dtau", "DStau"])  # R(D) and R(D*) case

        relevant_sig_ids: List[Tuple[int, ...]] = [
            self.fit_case.get_sig_id_for_component(component_name=comp_name) for comp_name in relevant_component_names
        ]
        relevant_sys_info = EfficiencySysInfoContainer.get_relevant_sys_infos(sys_cols=self.fit_case.systematics_cols)

        return EfficiencySysInfoContainer(
            sig_ids=relevant_sig_ids,
            sys_infos=relevant_sys_info,
            weight_col_name=None,  # None means the same column as for the efficiency calculation is used! -> Good!
            sig_id_column_name=self.fit_case.sig_id_column_name,
            use_only_given_sig_ids=False,
            use_new_fei_calibration_weights=self.fit_case.add_new_fei_calibration_factors,
            additional_scaling_weight_columns=self.get_additional_efficiency_scaling_columns(),
        )

    def get_additional_efficiency_scaling_columns(self) -> Optional[Tuple[str, ...]]:
        col_names: List[str] = []
        if self.fit_case.add_reco_mode_efficiency_correction:
            col_names.append(ReconstructionModeEfficiencyCorrector.reco_eff_correction_weight_column_name)

        return tuple(col_names) if col_names else None

    @staticmethod
    def add_to_output_dict(output_dict: Dict[str, PathType], output_info: FitOutputInfo) -> None:
        assert output_info.key not in output_dict.keys(), (output_info.key, output_info.path)
        assert output_info.path not in output_dict.values(), (output_info.key, output_info.path)
        output_dict.update({output_info.key: output_info.path})

    def get_fit_routine_output(self) -> Dict[str, PathType]:
        output_dict: Dict[str, PathType] = dict()

        assert len(self.output_info) == len(set([k for k, _ in self.output_info])), self.output_info
        assert all(v in ("data", "plot") for _, v in self.output_info), self.output_info

        fit_result_plot_keys: List[str] = [self.pre_fit_plot_output_key, self.post_fit_plot_output_key]

        for output_key, output_type in self.output_info:
            if not self.fit_case.create_plots and output_key in fit_result_plot_keys:
                continue

            if not self.fit_case.run_significance_evaluation and output_key == self.fit_significance_output_key:
                continue

            o_info: FitOutputInfo = self.get_output_info(output_key=output_key)
            self.add_to_output_dict(output_dict=output_dict, output_info=o_info)
        return output_dict

    def get_output_info(self, output_key: str) -> FitOutputInfo:
        assert len(self.output_info) == len(set([k for k, _ in self.output_info])), self.output_info
        assert all(v in self.output_type_ext_map.keys() for _, v in self.output_info), self.output_info
        assert output_key in [k for k, _ in self.output_info], (output_key, self.output_info)

        output_mapping: Dict[str, str] = dict(self.output_info)

        o_key: str = self.get_task_output_key(fit_case=self.fit_case, output_key=output_key)
        file_name = f"{o_key}.{self.output_type_ext_map[output_mapping[output_key]]}"
        o_path = os.path.join(self.base_path, file_name)
        tmp_path = os.path.join(self.base_path, f"tmp_{file_name}")

        return FitOutputInfo(
            key=o_key,
            name=file_name,
            path=o_path,
            tmp_path=tmp_path,
        )

    @staticmethod
    def get_task_output_key(fit_case: FitCase, output_key: str) -> str:
        return f"{fit_case.output_tag}_{output_key}"

    def check_output_dict(self, output_dict: Dict[str, PathType]) -> None:
        od: Dict[str, PathType] = self.get_fit_routine_output()
        if not output_dict == od:
            assert len(output_dict) == len(od), set(output_dict.keys()) ^ set(od.keys())
            assert len(set(output_dict.keys()) ^ set(od.keys())) == 0, set(output_dict.keys()) ^ set(od.keys())
            assert len(set(output_dict.values()) ^ set(od.values())) == 0, set(output_dict.values()) ^ set(od.values())
            assert len(set(output_dict.items()) ^ set(od.items())) == 0, set(output_dict.items()) ^ set(od.items())

    @property
    def fit_case(self) -> FitCase:
        return self._fit_case

    @property
    def fit_case_name(self) -> str:
        return self.fit_case.name
