"""
Definitions for Simple Signal Setup with Decoupled D* Modes

Felix Metzner 2021
"""

import copy

from typing import Dict, Tuple

from templatefitter.plotter.plot_style import KITColors

from rdstar.latex_labels.latex_process_labels import LatexDecayProcesses

from rdstar.offline_analysis.reco_modes.reco_mode_collections import BSigRecoModes

from rdstar.offline_analysis.mc_matching import default_sig_id_col_name
from rdstar.offline_analysis.mc_matching.old_mc_matching_info import OldMCMatchingIDsQueryInfo
from rdstar.offline_analysis.mc_matching.mc_matching_utilities import MCSigIDNameToQueryMappingType
from rdstar.offline_analysis.mc_matching.sig_id_costomizer_tools import (
    SigIDAndGenIDCombiner,
    SigIDAndGenIDCombinerInputType,
)

from rdstar.offline_analysis.fitting.fit_case import TemplateSetup
from rdstar.offline_analysis.fitting.rdstar_fit_variables import RecoModeNames
from rdstar.offline_analysis.fitting.special_fit_template_setups.special_template_setup_info_base import (
    SpecialTemplateSetupInfo,
)


__all__ = [
    "Q2SidebandSplitBkgTemplateSetupInfo",
    "Q2SidebandSplitBkgTemplateSetupInfoUsingDF",
]

main_reco_mode_keys: Tuple[str, ...] = tuple(v.name for v in BSigRecoModes.main_modes.reco_mode_infos)
_sig_id_mapping: MCSigIDNameToQueryMappingType = copy.copy(OldMCMatchingIDsQueryInfo.get_sig_id_mapping())


_base_q2_sideband_split_d_star_template_setup: Dict[str, Tuple[str, ...]] = {
    "continuum": ("CBKG",),
    "bbBkg": ("GBKG", "HBKG", "PiLNu", "DSSlbkg", "DSSl", "DStau", "Dtau"),
    "DSlDownFeed": ("DSlDownFeed",),
    "DSl": ("DSl",),
    "Dl": ("Dl",),
}

reco_mode_to_component_map: Dict[str, Tuple[str, ...]] = {
    RecoModeNames.DM: ("DSlDownFeed",),
    RecoModeNames.DZ: ("DSlDownFeed",),
    RecoModeNames.DSM: ("DSl",),
    RecoModeNames.DSZ: ("DSl",),
}

q2_sideband_split_d_star_custom_sig_id_map: SigIDAndGenIDCombinerInputType = {
    ("DSl", 4): {RecoModeNames.DSZ: (_sig_id_mapping["DSl"].id,), RecoModeNames.DSM: (_sig_id_mapping["DSl"].id,)},
    ("DSlDownFeed", 1002): {
        RecoModeNames.DZ: (_sig_id_mapping["DSl"].id,),
        RecoModeNames.DM: (_sig_id_mapping["DSl"].id,),
    },
}

q2_sideband_split_d_star_sig_customizer = SigIDAndGenIDCombiner(
    customizing_dict=q2_sideband_split_d_star_custom_sig_id_map,
    sig_id_column_name=default_sig_id_col_name,
    decay_mode_dict=copy.copy(BSigRecoModes.main_modes.main_reco_mode_infos_mapping),
)
q2_sideband_split_d_star_custom_sig_queries: MCSigIDNameToQueryMappingType = (
    q2_sideband_split_d_star_sig_customizer.get_updated_component_dict(initial_sig_query_dict=copy.copy(_sig_id_mapping))
)

_q2_sideband_split_d_star_template_setup: Dict[str, TemplateSetup] = {
    dec_mode_name: TemplateSetup(
        template_setup={
            comp_name: sub_comps
            for comp_name, sub_comps in _base_q2_sideband_split_d_star_template_setup.items()
            if comp_name in reco_mode_to_component_map[dec_mode_name]
            or comp_name not in [vv for v in reco_mode_to_component_map.values() for vv in v]
        }
    )
    for dec_mode_name in main_reco_mode_keys
}

_q2_sideband_split_d_star_plot_colors: Dict[str, str] = {
    "Dl": KITColors.kit_blue,
    "DSl": KITColors.kit_cyan,
    "DSlDownFeed": KITColors.kit_green,
    "bbBkg": KITColors.grey,
    "continuum": KITColors.light_grey,
}
_q2_sideband_split_d_star_plot_labels: Dict[str, str] = {
    "Dl": LatexDecayProcesses.to_d_l_nu,
    "DSl": LatexDecayProcesses.to_d_st_l_nu,
    "DSlDownFeed": LatexDecayProcesses.to_d_st_l_nu_down_feed,
    "bbBkg": LatexDecayProcesses.to_b_bbar_bkg,
    "continuum": LatexDecayProcesses.to_continuum_txt,
}


Q2SidebandSplitBkgTemplateSetupInfo = SpecialTemplateSetupInfo(
    template_setup=_q2_sideband_split_d_star_template_setup,
    plot_color_dict=_q2_sideband_split_d_star_plot_colors,
    plot_labels_dict=_q2_sideband_split_d_star_plot_labels,
    custom_sig_queries=q2_sideband_split_d_star_custom_sig_queries,
    sig_id_customizer=q2_sideband_split_d_star_sig_customizer,
)


#############################################
# Setup Using the down feed for calculation #
#############################################

_base_q2_sideband_split_d_star_inverted_template_setup: Dict[str, Tuple[str, ...]] = {
    "continuum": ("CBKG",),
    "bbBkg": ("GBKG", "HBKG", "PiLNu", "DSSlbkg", "DSSl", "DStau", "Dtau"),
    "DSlRegular": ("DSlRegular",),
    "DSl": ("DSl",),
    "Dl": ("Dl",),
}

inverted_reco_mode_to_component_map: Dict[str, Tuple[str, ...]] = {
    RecoModeNames.DM: ("DSl",),
    RecoModeNames.DZ: ("DSl",),
    RecoModeNames.DSM: ("DSlRegular",),
    RecoModeNames.DSZ: ("DSlRegular",),
}

q2_sideband_split_d_star_inverted_custom_sig_id_map: SigIDAndGenIDCombinerInputType = {
    ("DSlRegular", 1001): {
        RecoModeNames.DSZ: (_sig_id_mapping["DSl"].id,),
        RecoModeNames.DSM: (_sig_id_mapping["DSl"].id,),
    },
    ("DSl", 4): {
        RecoModeNames.DZ: (_sig_id_mapping["DSl"].id,),
        RecoModeNames.DM: (_sig_id_mapping["DSl"].id,),
    },
}

q2_sideband_split_d_star_inverted_sig_customizer = SigIDAndGenIDCombiner(
    customizing_dict=q2_sideband_split_d_star_inverted_custom_sig_id_map,
    sig_id_column_name=default_sig_id_col_name,
    decay_mode_dict=copy.copy(BSigRecoModes.main_modes.main_reco_mode_infos_mapping),
)
q2_sideband_split_d_star_inverted_custom_sig_queries: MCSigIDNameToQueryMappingType = (
    q2_sideband_split_d_star_inverted_sig_customizer.get_updated_component_dict(
        initial_sig_query_dict=copy.copy(_sig_id_mapping)
    )
)

_q2_sideband_split_d_star_inverted_template_setup: Dict[str, TemplateSetup] = {
    dec_mode_name: TemplateSetup(
        template_setup={
            comp_name: sub_comps
            for comp_name, sub_comps in _base_q2_sideband_split_d_star_inverted_template_setup.items()
            if comp_name in inverted_reco_mode_to_component_map[dec_mode_name]
            or comp_name not in [vv for v in inverted_reco_mode_to_component_map.values() for vv in v]
        }
    )
    for dec_mode_name in main_reco_mode_keys
}


_q2_sideband_split_d_star_inverted_plot_colors: Dict[str, str] = {
    "Dl": KITColors.kit_blue,
    "DSlRegular": KITColors.kit_cyan,
    "DSl": KITColors.kit_green,
    "bbBkg": KITColors.grey,
    "continuum": KITColors.light_grey,
}
_q2_sideband_split_d_star_inverted_plot_labels: Dict[str, str] = {
    "Dl": LatexDecayProcesses.to_d_l_nu,
    "DSlRegular": LatexDecayProcesses.to_d_st_l_nu,
    "DSl": LatexDecayProcesses.to_d_st_l_nu_down_feed,
    "bbBkg": LatexDecayProcesses.to_b_bbar_bkg,
    "continuum": LatexDecayProcesses.to_continuum_txt,
}


Q2SidebandSplitBkgTemplateSetupInfoUsingDF = SpecialTemplateSetupInfo(
    template_setup=_q2_sideband_split_d_star_inverted_template_setup,
    plot_color_dict=_q2_sideband_split_d_star_inverted_plot_colors,
    plot_labels_dict=_q2_sideband_split_d_star_inverted_plot_labels,
    custom_sig_queries=q2_sideband_split_d_star_inverted_custom_sig_queries,
    sig_id_customizer=q2_sideband_split_d_star_inverted_sig_customizer,
)
