"""
Queries for generator level MC matching.

Felix Metzner 2023
"""


import pandas as pd

from typing import Tuple, Callable

from rdstar.offline_analysis.observables.common_observables import Cols
from rdstar.offline_analysis.pdg_code_collections import PDGCodeCollection as PcC


__all__ = [
    "GenMCMatchingQueries",
]


class GenMCMatchingQueries:

    required_columns: Tuple[str, ...] = (
        Cols.is_continuum_evt,
        #
        Cols.mc_b1_1st_degree_d_meson_pdg,
        Cols.mc_b2_1st_degree_d_meson_pdg,
        #
        Cols.mc_b1_lepton_pdg,
        Cols.mc_b2_lepton_pdg,
        Cols.mc_b1_lepton_first_daughter_pdg,
        Cols.mc_b2_lepton_first_daughter_pdg,
        #
        Cols.mc_b1_4th_daughter_pdg,
        Cols.mc_b2_4th_daughter_pdg,
    )

    # Observable Helpers
    @staticmethod
    def get_first_b_meson_daughter_pdg_colum_for(b_index: int) -> str:
        return f"genUpsilon4S__bomcDaughter__bo{b_index}__cm__spmcDaughter__bo0__cmPDG__bc__bc__bc"

    @staticmethod
    def get_second_b_meson_daughter_pdg_colum_for(b_index: int) -> str:
        return f"genUpsilon4S__bomcDaughter__bo{b_index}__cm__spmcDaughter__bo1__cmPDG__bc__bc__bc"

    @staticmethod
    def get_first_daughter_of_second_b_meson_daughter_pdg_colum_for(b_index: int) -> str:
        return f"genUpsilon4S__bomcDaughter__bo{b_index}__cm__spmcDaughter__bo1__cmmcDaughter__bo0__cmPDG__bc__bc__bc__bc"

    @staticmethod
    def get_fourth_b_meson_daughter_pdg_colum_for(b_index: int) -> str:
        return f"genUpsilon4S__bomcDaughter__bo{b_index}__cm__spmcDaughter__bo3__cmPDG__bc__bc__bc"

    #####################
    # Low Level Queries #
    #####################

    @staticmethod
    def is_continuum_event(df: pd.DataFrame) -> pd.Series:
        return df[Cols.is_continuum_evt] == 1.0

    @staticmethod
    def is_not_continuum_event(df: pd.DataFrame) -> pd.Series:
        return df[Cols.is_continuum_evt] == 0.0

    @staticmethod
    def second_daughter_is_lepton(df: pd.DataFrame, b_index: int) -> pd.Series:
        lepton_col_name: str = GenMCMatchingQueries.get_second_b_meson_daughter_pdg_colum_for(b_index=b_index)
        return abs(df[lepton_col_name]).isin(PcC.light_lepton_pdg_ids)

    @staticmethod
    def second_daughter_is_tau(df: pd.DataFrame, b_index: int) -> pd.Series:
        lepton_col_name: str = GenMCMatchingQueries.get_second_b_meson_daughter_pdg_colum_for(b_index=b_index)
        return abs(df[lepton_col_name]) == PcC.tau_pdg_id

    @staticmethod
    def second_daughter_is_tau_to_l(df: pd.DataFrame, b_index: int) -> pd.Series:
        lep_from_tau_col: str = GenMCMatchingQueries.get_first_daughter_of_second_b_meson_daughter_pdg_colum_for(
            b_index=b_index
        )
        is_b_to_tau: pd.Series = GenMCMatchingQueries.second_daughter_is_tau(df=df, b_index=b_index)
        is_tau_from_lepton: pd.Series = abs(df[lep_from_tau_col]).isin(PcC.light_lepton_pdg_ids)

        return is_b_to_tau & is_tau_from_lepton

    @staticmethod
    def second_daughter_is_tau_to_had(df: pd.DataFrame, b_index: int) -> pd.Series:
        lep_from_tau_col: str = GenMCMatchingQueries.get_first_daughter_of_second_b_meson_daughter_pdg_colum_for(
            b_index=b_index
        )
        is_b_to_tau: pd.Series = GenMCMatchingQueries.second_daughter_is_tau(df=df, b_index=b_index)
        is_not_tau_from_lepton: pd.Series = ~abs(df[lep_from_tau_col]).isin(PcC.light_lepton_pdg_ids)

        return is_b_to_tau & is_not_tau_from_lepton

    @staticmethod
    def first_daughter_is_d_meson(df: pd.DataFrame, b_index: int) -> pd.Series:
        d_col_name: str = GenMCMatchingQueries.get_first_b_meson_daughter_pdg_colum_for(b_index=b_index)
        return abs(df[d_col_name]).isin(PcC.d_pdg_ids)

    @staticmethod
    def first_daughter_is_dst_meson(df: pd.DataFrame, b_index: int) -> pd.Series:
        dst_col_name: str = GenMCMatchingQueries.get_first_b_meson_daughter_pdg_colum_for(b_index=b_index)
        return abs(df[dst_col_name]).isin(PcC.dst_pdg_ids)

    @staticmethod
    def first_daughter_is_pi(df: pd.DataFrame, b_index: int) -> pd.Series:
        pi_col_name: str = GenMCMatchingQueries.get_first_b_meson_daughter_pdg_colum_for(b_index=b_index)
        return abs(df[pi_col_name]).isin(PcC.pion_pdg_ids)

    @staticmethod
    def first_daughter_is_d_one_dss_meson(df: pd.DataFrame, b_index: int) -> pd.Series:
        dss_col_name: str = GenMCMatchingQueries.get_first_b_meson_daughter_pdg_colum_for(b_index=b_index)
        return abs(df[dss_col_name]).isin(PcC.done_pdg_ids)

    # TODO
    @staticmethod
    def first_daughter_is_d_twostar_dss_meson(df: pd.DataFrame, b_index: int) -> pd.Series:
        dss_col_name: str = GenMCMatchingQueries.get_first_b_meson_daughter_pdg_colum_for(b_index=b_index)
        return abs(df[dss_col_name]).isin(PcC.dtwostar_pdg_ids)

    @staticmethod
    def first_daughter_is_d_primeone_dss_meson(df: pd.DataFrame, b_index: int) -> pd.Series:
        dss_col_name: str = GenMCMatchingQueries.get_first_b_meson_daughter_pdg_colum_for(b_index=b_index)
        return abs(df[dss_col_name]).isin(PcC.dprimeone_pdg_ids)

    @staticmethod
    def first_daughter_is_d_zerostar_dss_meson(df: pd.DataFrame, b_index: int) -> pd.Series:
        dss_col_name: str = GenMCMatchingQueries.get_first_b_meson_daughter_pdg_colum_for(b_index=b_index)
        return abs(df[dss_col_name]).isin(PcC.dzerostar_pdg_ids)

    @staticmethod
    def first_daughter_is_any_d_meson(df: pd.DataFrame, b_index: int) -> pd.Series:
        dss_col_name: str = GenMCMatchingQueries.get_first_b_meson_daughter_pdg_colum_for(b_index=b_index)
        return abs(df[dss_col_name]).isin(PcC.all_d_pdg_ids)

    @staticmethod
    def first_daughter_is_not_any_d_meson(df: pd.DataFrame, b_index: int) -> pd.Series:
        return ~GenMCMatchingQueries.first_daughter_is_any_d_meson(df=df, b_index=b_index)

    @staticmethod
    def b_has_three_daughters(df: pd.DataFrame, b_index: int) -> pd.Series:
        fourth_daughter_col: str = GenMCMatchingQueries.get_fourth_b_meson_daughter_pdg_colum_for(b_index=b_index)
        # Allow radiative photons also as fourth daughter
        return df[fourth_daughter_col].isin(PcC.valid_forth_daughter)

    ###################
    # B Decay Queries #
    ###################

    @staticmethod
    def b_to_d_tau(df: pd.DataFrame, b_index: int) -> pd.Series:
        not_continuum: pd.Series = GenMCMatchingQueries.is_not_continuum_event(df=df)
        has_d: pd.Series = GenMCMatchingQueries.first_daughter_is_d_meson(df=df, b_index=b_index)
        has_tau_to_l: pd.Series = GenMCMatchingQueries.second_daughter_is_tau_to_l(df=df, b_index=b_index)
        has_three_daughters: pd.Series = GenMCMatchingQueries.b_has_three_daughters(df=df, b_index=b_index)

        return not_continuum & has_d & has_tau_to_l & has_three_daughters

    @staticmethod
    def b_to_dst_tau(df: pd.DataFrame, b_index: int) -> pd.Series:
        not_continuum: pd.Series = GenMCMatchingQueries.is_not_continuum_event(df=df)
        has_dst: pd.Series = GenMCMatchingQueries.first_daughter_is_dst_meson(df=df, b_index=b_index)
        has_tau_to_l: pd.Series = GenMCMatchingQueries.second_daughter_is_tau_to_l(df=df, b_index=b_index)
        has_three_daughters: pd.Series = GenMCMatchingQueries.b_has_three_daughters(df=df, b_index=b_index)

        return not_continuum & has_dst & has_tau_to_l & has_three_daughters

    @staticmethod
    def b_to_d_l(df: pd.DataFrame, b_index: int) -> pd.Series:
        not_continuum: pd.Series = GenMCMatchingQueries.is_not_continuum_event(df=df)
        has_d: pd.Series = GenMCMatchingQueries.first_daughter_is_d_meson(df=df, b_index=b_index)
        has_l: pd.Series = GenMCMatchingQueries.second_daughter_is_lepton(df=df, b_index=b_index)
        has_three_daughters: pd.Series = GenMCMatchingQueries.b_has_three_daughters(df=df, b_index=b_index)

        return not_continuum & has_d & has_l & has_three_daughters

    @staticmethod
    def b_to_d_st_l(df: pd.DataFrame, b_index: int) -> pd.Series:
        not_continuum: pd.Series = GenMCMatchingQueries.is_not_continuum_event(df=df)
        has_dst: pd.Series = GenMCMatchingQueries.first_daughter_is_dst_meson(df=df, b_index=b_index)
        has_l: pd.Series = GenMCMatchingQueries.second_daughter_is_lepton(df=df, b_index=b_index)
        has_three_daughters: pd.Series = GenMCMatchingQueries.b_has_three_daughters(df=df, b_index=b_index)

        return not_continuum & has_dst & has_l & has_three_daughters

    @staticmethod
    def b_to_d_one_l(df: pd.DataFrame, b_index: int) -> pd.Series:
        not_continuum: pd.Series = GenMCMatchingQueries.is_not_continuum_event(df=df)
        has_done: pd.Series = GenMCMatchingQueries.first_daughter_is_d_one_dss_meson(df=df, b_index=b_index)
        has_l: pd.Series = GenMCMatchingQueries.second_daughter_is_lepton(df=df, b_index=b_index)
        has_three_daughters: pd.Series = GenMCMatchingQueries.b_has_three_daughters(df=df, b_index=b_index)

        return not_continuum & has_done & has_l & has_three_daughters

    @staticmethod
    def b_to_d_twostar_l(df: pd.DataFrame, b_index: int) -> pd.Series:
        not_continuum: pd.Series = GenMCMatchingQueries.is_not_continuum_event(df=df)
        has_dtwostar: pd.Series = GenMCMatchingQueries.first_daughter_is_d_twostar_dss_meson(df=df, b_index=b_index)
        has_l: pd.Series = GenMCMatchingQueries.second_daughter_is_lepton(df=df, b_index=b_index)
        has_three_daughters: pd.Series = GenMCMatchingQueries.b_has_three_daughters(df=df, b_index=b_index)

        return not_continuum & has_dtwostar & has_l & has_three_daughters

    @staticmethod
    def b_to_d_primeone_l(df: pd.DataFrame, b_index: int) -> pd.Series:
        not_continuum: pd.Series = GenMCMatchingQueries.is_not_continuum_event(df=df)
        has_dprimeone: pd.Series = GenMCMatchingQueries.first_daughter_is_d_primeone_dss_meson(df=df, b_index=b_index)
        has_l: pd.Series = GenMCMatchingQueries.second_daughter_is_lepton(df=df, b_index=b_index)
        has_three_daughters: pd.Series = GenMCMatchingQueries.b_has_three_daughters(df=df, b_index=b_index)

        return not_continuum & has_dprimeone & has_l & has_three_daughters

    @staticmethod
    def b_to_d_zerostar_l(df: pd.DataFrame, b_index: int) -> pd.Series:
        not_continuum: pd.Series = GenMCMatchingQueries.is_not_continuum_event(df=df)
        has_dzerostar: pd.Series = GenMCMatchingQueries.first_daughter_is_d_zerostar_dss_meson(df=df, b_index=b_index)
        has_l: pd.Series = GenMCMatchingQueries.second_daughter_is_lepton(df=df, b_index=b_index)
        has_three_daughters: pd.Series = GenMCMatchingQueries.b_has_three_daughters(df=df, b_index=b_index)

        return not_continuum & has_dzerostar & has_l & has_three_daughters

    @staticmethod
    def b_to_d_one_tau(df: pd.DataFrame, b_index: int) -> pd.Series:
        not_continuum: pd.Series = GenMCMatchingQueries.is_not_continuum_event(df=df)
        has_done: pd.Series = GenMCMatchingQueries.first_daughter_is_d_one_dss_meson(df=df, b_index=b_index)
        has_tau: pd.Series = GenMCMatchingQueries.second_daughter_is_tau_to_l(df=df, b_index=b_index)
        has_three_daughters: pd.Series = GenMCMatchingQueries.b_has_three_daughters(df=df, b_index=b_index)

        return not_continuum & has_done & has_tau & has_three_daughters

    @staticmethod
    def b_to_d_twostar_tau(df: pd.DataFrame, b_index: int) -> pd.Series:
        not_continuum: pd.Series = GenMCMatchingQueries.is_not_continuum_event(df=df)
        has_dtwostar: pd.Series = GenMCMatchingQueries.first_daughter_is_d_twostar_dss_meson(df=df, b_index=b_index)
        has_tau: pd.Series = GenMCMatchingQueries.second_daughter_is_tau_to_l(df=df, b_index=b_index)
        has_three_daughters: pd.Series = GenMCMatchingQueries.b_has_three_daughters(df=df, b_index=b_index)

        return not_continuum & has_dtwostar & has_tau & has_three_daughters

    @staticmethod
    def b_to_d_primeone_tau(df: pd.DataFrame, b_index: int) -> pd.Series:
        not_continuum: pd.Series = GenMCMatchingQueries.is_not_continuum_event(df=df)
        has_dprimeone: pd.Series = GenMCMatchingQueries.first_daughter_is_d_primeone_dss_meson(df=df, b_index=b_index)
        has_tau: pd.Series = GenMCMatchingQueries.second_daughter_is_tau_to_l(df=df, b_index=b_index)
        has_three_daughters: pd.Series = GenMCMatchingQueries.b_has_three_daughters(df=df, b_index=b_index)

        return not_continuum & has_dprimeone & has_tau & has_three_daughters

    @staticmethod
    def b_to_d_zerostar_tau(df: pd.DataFrame, b_index: int) -> pd.Series:
        not_continuum: pd.Series = GenMCMatchingQueries.is_not_continuum_event(df=df)
        has_dzerostar: pd.Series = GenMCMatchingQueries.first_daughter_is_d_zerostar_dss_meson(df=df, b_index=b_index)
        has_tau: pd.Series = GenMCMatchingQueries.second_daughter_is_tau_to_l(df=df, b_index=b_index)
        has_three_daughters: pd.Series = GenMCMatchingQueries.b_has_three_daughters(df=df, b_index=b_index)

        return not_continuum & has_dzerostar & has_tau & has_three_daughters

    @staticmethod
    def hadronic_tau(df: pd.DataFrame, b_index: int) -> pd.Series:
        not_continuum: pd.Series = GenMCMatchingQueries.is_not_continuum_event(df=df)
        has_any_d: pd.Series = GenMCMatchingQueries.first_daughter_is_any_d_meson(df=df, b_index=b_index)
        has_tau_to_had: pd.Series = GenMCMatchingQueries.second_daughter_is_tau_to_had(df=df, b_index=b_index)
        has_three_daughters: pd.Series = GenMCMatchingQueries.b_has_three_daughters(df=df, b_index=b_index)

        return not_continuum & has_any_d & has_tau_to_had & has_three_daughters

    @staticmethod
    def b_to_pi_l(df: pd.DataFrame, b_index: int) -> pd.Series:
        not_continuum: pd.Series = GenMCMatchingQueries.is_not_continuum_event(df=df)
        has_pi: pd.Series = GenMCMatchingQueries.first_daughter_is_pi(df=df, b_index=b_index)
        has_l: pd.Series = GenMCMatchingQueries.second_daughter_is_lepton(df=df, b_index=b_index)
        has_three_daughters: pd.Series = GenMCMatchingQueries.b_has_three_daughters(df=df, b_index=b_index)

        return not_continuum & has_pi & has_l & has_three_daughters

    @staticmethod
    def other(df: pd.DataFrame, b_index: int) -> pd.Series:
        not_continuum: pd.Series = GenMCMatchingQueries.is_not_continuum_event(df=df)
        not_b_to_d_l: pd.Series = ~GenMCMatchingQueries.b_to_d_l(df=df, b_index=b_index)
        not_b_to_dst_l: pd.Series = ~GenMCMatchingQueries.b_to_d_st_l(df=df, b_index=b_index)
        not_b_to_d_tau: pd.Series = ~GenMCMatchingQueries.b_to_d_tau(df=df, b_index=b_index)
        not_b_to_dst_tau: pd.Series = ~GenMCMatchingQueries.b_to_dst_tau(df=df, b_index=b_index)
        not_b_to_dss_d_one_l: pd.Series = ~GenMCMatchingQueries.b_to_d_one_l(df=df, b_index=b_index)
        not_b_to_dss_d_one_tau: pd.Series = ~GenMCMatchingQueries.b_to_d_one_tau(df=df, b_index=b_index)
        not_b_to_dss_d_twostar_l: pd.Series = ~GenMCMatchingQueries.b_to_d_twostar_l(df=df, b_index=b_index)
        not_b_to_dss_d_twostar_tau: pd.Series = ~GenMCMatchingQueries.b_to_d_twostar_tau(df=df, b_index=b_index)
        not_b_to_dss_d_primeone_l: pd.Series = ~GenMCMatchingQueries.b_to_d_primeone_l(df=df, b_index=b_index)
        not_b_to_dss_d_primeone_tau: pd.Series = ~GenMCMatchingQueries.b_to_d_primeone_tau(df=df, b_index=b_index)
        not_b_to_dss_d_zerostar_l: pd.Series = ~GenMCMatchingQueries.b_to_d_zerostar_l(df=df, b_index=b_index)
        not_b_to_dss_d_zerostar_tau: pd.Series = ~GenMCMatchingQueries.b_to_d_zerostar_tau(df=df, b_index=b_index)
        not_hadronic_tau: pd.Series = ~GenMCMatchingQueries.hadronic_tau(df=df, b_index=b_index)
        not_b_to_pi_l: pd.Series = ~GenMCMatchingQueries.b_to_pi_l(df=df, b_index=b_index)

        return (
            not_continuum
            & not_hadronic_tau
            & not_b_to_d_l
            & not_b_to_dst_l
            & not_b_to_d_tau
            & not_b_to_dst_tau
            & not_b_to_dss_d_one_l
            & not_b_to_dss_d_one_tau
            & not_b_to_dss_d_twostar_l
            & not_b_to_dss_d_twostar_tau
            & not_b_to_dss_d_primeone_l
            & not_b_to_dss_d_primeone_tau
            & not_b_to_dss_d_zerostar_l
            & not_b_to_dss_d_zerostar_tau
            & not_b_to_pi_l
        )

    @staticmethod
    def continuum(df: pd.DataFrame, b_index: int) -> pd.Series:
        return GenMCMatchingQueries.is_continuum_event(df=df)


if __name__ == "__main__":

    observable_helper_functions: Tuple[Callable, ...] = (
        GenMCMatchingQueries.get_first_b_meson_daughter_pdg_colum_for,
        GenMCMatchingQueries.get_second_b_meson_daughter_pdg_colum_for,
        GenMCMatchingQueries.get_first_daughter_of_second_b_meson_daughter_pdg_colum_for,
        GenMCMatchingQueries.get_fourth_b_meson_daughter_pdg_colum_for,
    )

    required_obs_column_names: Tuple[str, ...] = tuple(
        obs_fcn(b_index=i) for obs_fcn in observable_helper_functions for i in (0, 1)
    )

    assert all(c in GenMCMatchingQueries.required_columns for c in required_obs_column_names), (
        c for c in required_obs_column_names if c not in GenMCMatchingQueries.required_columns
    )
