"""
Queries for extended mc matching information.

Felix Metzner 2023
"""

import numpy as np
import pandas as pd

from typing import Tuple, Dict, Union, Callable

from rdstar.offline_analysis.observables.common_observables import Cols
from rdstar.offline_analysis.pdg_code_collections import PDGCodeCollection as PcC


__all__ = [
    "MCSigQueries",
]


class MCSigQueries:

    required_columns: Tuple[str, ...] = (
        Cols.is_continuum_bkg_col,
        #
        Cols.data_type,
        #
        Cols.mdst_index_via_lepton_col,
        Cols.mdst_index_via_tau_col,
        #
        Cols.reco_lepton_truth_matched_pdg,
        #
        Cols.mc_b1_lepton_pdg,
        Cols.mc_b2_lepton_pdg,
        Cols.mc_b1_lepton_first_daughter_pdg,
        Cols.mc_b2_lepton_first_daughter_pdg,
        #
        Cols.mc_b1_3rd_daughter_pdg,
        Cols.mc_b2_3rd_daughter_pdg,
        Cols.mc_b1_4th_daughter_pdg,
        Cols.mc_b2_4th_daughter_pdg,
        Cols.mc_b1_5th_daughter_pdg,
        Cols.mc_b2_5th_daughter_pdg,
        #
        Cols.mc_b1_1st_degree_d_meson_second_daughter_pdg,
        Cols.mc_b2_1st_degree_d_meson_second_daughter_pdg,
        Cols.mc_b1_1st_degree_d_meson_third_daughter_pdg,
        Cols.mc_b2_1st_degree_d_meson_third_daughter_pdg,
        Cols.mc_b1_1st_degree_d_meson_fourth_daughter_pdg,
        Cols.mc_b2_1st_degree_d_meson_fourth_daughter_pdg,
        #
        Cols.mc_b1_1st_degree_d_meson_pdg,
        Cols.mc_b2_1st_degree_d_meson_pdg,
        Cols.mc_b1_2nd_degree_d_meson_pdg,
        Cols.mc_b2_2nd_degree_d_meson_pdg,
        Cols.mc_b1_3rd_degree_d_meson_pdg,
        Cols.mc_b2_3rd_degree_d_meson_pdg,
        #
        Cols.mc_b1_meson_pdg,
        Cols.mc_b2_meson_pdg,
        #
        Cols.mc_b1_meson_charge,
        Cols.mc_b2_meson_charge,
        #
        Cols.mdst_index_via_d_meson_daughter_col,
        Cols.mdst_index_via_d_meson_daughter_2_col,
        Cols.mdst_index_via_d_meson_daughter_indirect_1_col,
        Cols.mdst_index_via_d_meson_daughter_indirect_2_col,
        Cols.mdst_index_via_d_meson_indirect_1_col,
        Cols.mdst_index_via_d_meson_indirect_2_col,
    )

    sig_b_mdst_indices: Tuple[int, ...] = (1, 2)
    no_sig_b_match_value: int = -999
    no_main_sig_d_match_value: int = -9991
    no_light_lepton_match_value: int = -9992

    charged_b_meson_flavour_values: Tuple[int, ...] = (-1, 1)
    neutral_b_meson_flavour_values: int = 0
    no_b_meson_flavour_values: int = -999

    @staticmethod
    def get_b_meson_mdst_index_from_reco_lepton(df: pd.DataFrame) -> np.ndarray:
        sig_b_mdst_id: np.array = df[Cols.mdst_index_via_lepton_col].values
        sig_b_via_tau_mdst_id: np.array = df[Cols.mdst_index_via_tau_col].values

        is_sig_b: np.array = np.isin(element=sig_b_mdst_id, test_elements=MCSigQueries.sig_b_mdst_indices)
        is_sig_b_via_tau: np.array = np.isin(element=sig_b_via_tau_mdst_id, test_elements=MCSigQueries.sig_b_mdst_indices)

        assert not np.any(is_sig_b & is_sig_b_via_tau), np.sum(is_sig_b & is_sig_b_via_tau)

        b_mdst_index_array: np.array = np.select(
            condlist=[is_sig_b, is_sig_b_via_tau],
            choicelist=[sig_b_mdst_id, sig_b_via_tau_mdst_id],
            default=MCSigQueries.no_sig_b_match_value,
        )

        _unique_indices: np.array = np.unique(b_mdst_index_array)
        _valid_indices: Tuple[int, ...] = MCSigQueries.sig_b_mdst_indices + (MCSigQueries.no_sig_b_match_value,)
        assert np.all(np.isin(element=_unique_indices, test_elements=_valid_indices)), (_unique_indices, _valid_indices)
        return b_mdst_index_array

    @staticmethod
    def get_b_meson_mdst_index_from_d_meson(df: pd.DataFrame) -> np.ndarray:
        mdst_id_via_d_daughter: np.ndarray = df[Cols.mdst_index_via_d_meson_daughter_col].values
        mdst_id_via_d_daughter_2: np.ndarray = df[Cols.mdst_index_via_d_meson_daughter_2_col].values
        mdst_id_via_d_daughter_indirect_1: np.ndarray = df[Cols.mdst_index_via_d_meson_daughter_indirect_1_col].values
        mdst_id_via_d_daughter_indirect_2: np.ndarray = df[Cols.mdst_index_via_d_meson_daughter_indirect_2_col].values
        mdst_id_via_d_indirect_1: np.ndarray = df[Cols.mdst_index_via_d_meson_indirect_1_col].values
        mdst_id_via_d_indirect_2: np.ndarray = df[Cols.mdst_index_via_d_meson_indirect_2_col].values

        # fmt: off
        is_via_d_daughter: np.array = np.isin(element=mdst_id_via_d_daughter, test_elements=MCSigQueries.sig_b_mdst_indices)
        is_via_d_daughter_2: np.array = np.isin(element=mdst_id_via_d_daughter_2, test_elements=MCSigQueries.sig_b_mdst_indices)
        is_via_d_daughter_indirect_1: np.array = np.isin(element=mdst_id_via_d_daughter_indirect_1, test_elements=MCSigQueries.sig_b_mdst_indices)
        is_via_d_daughter_indirect_2: np.array = np.isin(element=mdst_id_via_d_daughter_indirect_2, test_elements=MCSigQueries.sig_b_mdst_indices)
        is_via_d_indirect_1: np.array = np.isin(element=mdst_id_via_d_indirect_1, test_elements=MCSigQueries.sig_b_mdst_indices)
        is_via_d_indirect_2: np.array = np.isin(element=mdst_id_via_d_indirect_2, test_elements=MCSigQueries.sig_b_mdst_indices)

        is_primary_via_d_daughter_2: np.array = is_via_d_daughter_2 & np.logical_not(is_via_d_daughter)
        is_primary_via_d_daughter_indirect_1: np.array = is_via_d_daughter_indirect_1 & np.logical_not(is_primary_via_d_daughter_2)
        is_primary_via_d_daughter_indirect_2: np.array = is_via_d_daughter_indirect_2 & np.logical_not(is_primary_via_d_daughter_indirect_1)
        is_primary_via_d_indirect_1: np.array = is_via_d_indirect_1 & np.logical_not(is_primary_via_d_daughter_indirect_2)
        is_primary_via_d_indirect_2: np.array = is_via_d_indirect_2 & np.logical_not(is_primary_via_d_indirect_1)
        # fmt: on

        b_mdst_index_array: np.array = np.select(
            condlist=[
                is_via_d_daughter,
                is_primary_via_d_daughter_2,
                is_primary_via_d_daughter_indirect_1,
                is_primary_via_d_daughter_indirect_2,
                is_primary_via_d_indirect_1,
                is_primary_via_d_indirect_2,
            ],
            choicelist=[
                mdst_id_via_d_daughter,
                mdst_id_via_d_daughter_2,
                mdst_id_via_d_daughter_indirect_1,
                mdst_id_via_d_daughter_indirect_2,
                mdst_id_via_d_indirect_1,
                mdst_id_via_d_indirect_2,
            ],
            default=MCSigQueries.no_sig_b_match_value,
        )

        _unique_indices: np.array = np.unique(b_mdst_index_array)
        _valid_indices: Tuple[int, ...] = MCSigQueries.sig_b_mdst_indices + (MCSigQueries.no_sig_b_match_value,)
        assert np.all(np.isin(element=_unique_indices, test_elements=_valid_indices)), (_unique_indices, _valid_indices)
        return b_mdst_index_array

    @staticmethod
    def get_observable_for_sig_b(df: pd.DataFrame, obs_col_name: str) -> np.ndarray:
        b1_obs_start: str = "genUpsilon4S__bomcDaughter__bo0"
        b2_obs_start: str = "genUpsilon4S__bomcDaughter__bo1"
        assert any(obs_col_name.startswith(c) for c in (b1_obs_start, b2_obs_start)), obs_col_name
        assert obs_col_name in df.columns, obs_col_name

        b1_obs_col: str = b1_obs_start + obs_col_name[len(b1_obs_start) :]
        b2_obs_col: str = b2_obs_start + obs_col_name[len(b2_obs_start) :]
        assert any(c == obs_col_name for c in (b1_obs_col, b2_obs_col)), (obs_col_name, (b1_obs_col, b2_obs_col))
        assert b1_obs_col in df.columns, b1_obs_col
        assert b2_obs_col in df.columns, b2_obs_col

        b_mdst_index_array = MCSigQueries.get_b_meson_mdst_index_from_reco_lepton(df=df)

        b1_obs_values: np.array = df[b1_obs_col].values
        b2_obs_values: np.array = df[b2_obs_col].values

        return np.select(
            condlist=[b_mdst_index_array == 1, b_mdst_index_array == 2],
            choicelist=[b1_obs_values, b2_obs_values],
            default=MCSigQueries.no_sig_b_match_value,
        )

    @staticmethod
    def get_observable_for_any_b(df: pd.DataFrame, obs_col_name: str) -> np.ndarray:
        b1_obs_start: str = "genUpsilon4S__bomcDaughter__bo0"
        b2_obs_start: str = "genUpsilon4S__bomcDaughter__bo1"
        assert any(obs_col_name.startswith(c) for c in (b1_obs_start, b2_obs_start)), obs_col_name
        assert obs_col_name in df.columns, obs_col_name

        b1_obs_col: str = b1_obs_start + obs_col_name[len(b1_obs_start) :]
        b2_obs_col: str = b2_obs_start + obs_col_name[len(b2_obs_start) :]
        assert any(c == obs_col_name for c in (b1_obs_col, b2_obs_col)), (obs_col_name, (b1_obs_col, b2_obs_col))
        assert b1_obs_col in df.columns, b1_obs_col
        assert b2_obs_col in df.columns, b2_obs_col

        b_mdst_index_array = MCSigQueries.get_b_meson_mdst_index_from_d_meson(df=df)

        b1_obs_values: np.array = df[b1_obs_col].values
        b2_obs_values: np.array = df[b2_obs_col].values

        return np.select(
            condlist=[b_mdst_index_array == 1, b_mdst_index_array == 2],
            choicelist=[b1_obs_values, b2_obs_values],
            default=MCSigQueries.no_sig_b_match_value,
        )

    @staticmethod
    def get_true_b_meson_pdg(df: pd.DataFrame) -> np.ndarray:
        return MCSigQueries.get_observable_for_sig_b(
            df=df,
            obs_col_name=Cols.mc_b1_meson_pdg,
        )

    @staticmethod
    def get_true_b_meson_flavour(df: pd.DataFrame) -> np.ndarray:
        # Returns values in
        #   1, -1: charged,
        #   0: neutral,
        #   -999: no B meson!
        return MCSigQueries.get_observable_for_sig_b(
            df=df,
            obs_col_name=Cols.mc_b1_meson_charge,
        )

    @staticmethod
    def is_charged1_or_neutral0_b_meson_or_neither(df: pd.DataFrame) -> np.ndarray:
        flavour_vals: np.ndarray = MCSigQueries.get_true_b_meson_flavour(df=df)
        flavour_vals[abs(flavour_vals) == 1] = 1

        _expected_flavour_vals: Tuple[int, ...] = (0, 1, MCSigQueries.no_sig_b_match_value)
        _unique_test_flavour_vals: Tuple[int, ...] = tuple(np.unique(flavour_vals))

        assert all(v in _expected_flavour_vals for v in _unique_test_flavour_vals), (
            _unique_test_flavour_vals,
            list(set(v for v in _unique_test_flavour_vals if v not in _expected_flavour_vals)),
        )
        num_no_flavour_with_int = np.sum(flavour_vals == MCSigQueries.no_sig_b_match_value)

        is_no_b_mask: np.ndarray = flavour_vals == MCSigQueries.no_sig_b_match_value
        flavour_vals[is_no_b_mask] = np.nan

        num_no_flavour_with_nan = np.sum(np.isnan(flavour_vals))
        assert num_no_flavour_with_nan == num_no_flavour_with_int, (num_no_flavour_with_nan, num_no_flavour_with_int)

        return flavour_vals

    @staticmethod
    def is_continuum_event(df: pd.DataFrame) -> np.ndarray:
        is_continuum_event_info: np.ndarray = df[Cols.is_continuum_bkg_col].values
        assert all(v in [0, 1] for v in np.unique(is_continuum_event_info)), np.unique(is_continuum_event_info)

        is_cont_event: np.ndarray = is_continuum_event_info == 1
        return is_cont_event

    @staticmethod
    def is_not_continuum_event(df: pd.DataFrame) -> np.ndarray:
        is_continuum_event_info: np.ndarray = df[Cols.is_continuum_bkg_col].values
        is_not_cont_event: np.ndarray = is_continuum_event_info == 0
        return is_not_cont_event

    @staticmethod
    def is_event_without_lepton_to_b_meson_relation(df: pd.DataFrame) -> np.ndarray:
        b_meson_flavour: np.ndarray = MCSigQueries.is_charged1_or_neutral0_b_meson_or_neither(df=df)
        is_not_from_b_meson: np.ndarray = np.isnan(b_meson_flavour)
        return is_not_from_b_meson

    @staticmethod
    def is_event_with_lepton_to_b_meson_relation(df: pd.DataFrame) -> np.ndarray:
        b_meson_flavour: np.ndarray = MCSigQueries.is_charged1_or_neutral0_b_meson_or_neither(df=df)
        is_from_b_meson: np.ndarray = np.logical_not(np.isnan(b_meson_flavour))
        return is_from_b_meson

    @staticmethod
    def is_charged_b_meson(df: pd.DataFrame) -> np.ndarray:
        flavour_vals = MCSigQueries.get_true_b_meson_flavour(df=df)
        return abs(flavour_vals) == 1

    @staticmethod
    def is_neutral_b_meson(df: pd.DataFrame) -> np.ndarray:
        flavour_vals = MCSigQueries.get_true_b_meson_flavour(df=df)
        return abs(flavour_vals) == 0

    @staticmethod
    def get_true_lepton_pdg(df: pd.DataFrame) -> np.ndarray:
        return MCSigQueries.get_observable_for_sig_b(
            df=df,
            obs_col_name=Cols.mc_b1_lepton_pdg,
        )

    @staticmethod
    def get_true_lepton_daughter_pdg(df: pd.DataFrame) -> np.ndarray:
        return MCSigQueries.get_observable_for_sig_b(
            df=df,
            obs_col_name=Cols.mc_b1_lepton_first_daughter_pdg,
        )

    @staticmethod
    def get_mc_matched_reco_lepton_pdg(df: pd.DataFrame) -> np.ndarray:
        return df[Cols.reco_lepton_truth_matched_pdg]

    @staticmethod
    def get_primary_light_lepton_pdg(df: pd.DataFrame) -> np.ndarray:
        lepton_pdg: np.ndarray = MCSigQueries.get_mc_matched_reco_lepton_pdg(df=df)
        is_light_lepton: np.ndarray = np.isin(element=abs(lepton_pdg), test_elements=PcC.light_lepton_pdg_ids)
        lepton_pdg[np.logical_not(is_light_lepton)] = MCSigQueries.no_light_lepton_match_value
        return lepton_pdg

    @staticmethod
    def get_true_d_meson_pdg(df: pd.DataFrame) -> np.ndarray:
        return MCSigQueries.get_observable_for_sig_b(
            df=df,
            obs_col_name=Cols.mc_b1_1st_degree_d_meson_pdg,
        )

    @staticmethod
    def get_true_d_meson_first_daughter_pdg(df: pd.DataFrame) -> np.ndarray:
        return MCSigQueries.get_observable_for_sig_b(
            df=df,
            obs_col_name=Cols.mc_b1_1st_degree_d_meson_first_daughter_pdg,
        )

    @staticmethod
    def get_true_d_meson_second_daughter_pdg(df: pd.DataFrame) -> np.ndarray:
        return MCSigQueries.get_observable_for_sig_b(
            df=df,
            obs_col_name=Cols.mc_b1_1st_degree_d_meson_second_daughter_pdg,
        )

    @staticmethod
    def get_true_d_meson_third_daughter_pdg(df: pd.DataFrame) -> np.ndarray:
        return MCSigQueries.get_observable_for_sig_b(
            df=df,
            obs_col_name=Cols.mc_b1_1st_degree_d_meson_third_daughter_pdg,
        )

    @staticmethod
    def get_true_d_meson_fourth_daughter_pdg(df: pd.DataFrame) -> np.ndarray:
        return MCSigQueries.get_observable_for_sig_b(
            df=df,
            obs_col_name=Cols.mc_b1_1st_degree_d_meson_fourth_daughter_pdg,
        )

    @staticmethod
    def get_main_d_meson_pdg(df: pd.DataFrame) -> np.ndarray:
        primary_d_pdg: np.array = MCSigQueries.get_observable_for_sig_b(
            df=df,
            obs_col_name=Cols.mc_b1_1st_degree_d_meson_pdg,
        )
        secondary_d_pdg: np.array = MCSigQueries.get_observable_for_sig_b(
            df=df,
            obs_col_name=Cols.mc_b1_2nd_degree_d_meson_pdg,
        )
        tertiary_d_pdg: np.array = MCSigQueries.get_observable_for_sig_b(
            df=df,
            obs_col_name=Cols.mc_b1_3rd_degree_d_meson_pdg,
        )

        primary_abs_d_pdg: np.array = abs(primary_d_pdg)
        secondary_abs_d_pdg: np.array = abs(secondary_d_pdg)
        tertiary_abs_d_pdg: np.array = abs(tertiary_d_pdg)

        primary_d_is_main: np.array = np.isin(element=primary_abs_d_pdg, test_elements=PcC.d_wwo_st_pdg_ids)
        secondary_d_could_be_main: np.array = np.isin(element=secondary_abs_d_pdg, test_elements=PcC.d_wwo_st_pdg_ids)
        secondary_d_is_main: np.array = secondary_d_could_be_main & np.logical_not(primary_d_is_main)
        tertiary_d_could_be_main: np.array = np.isin(element=tertiary_abs_d_pdg, test_elements=PcC.d_wwo_st_pdg_ids)
        tertiary_d_is_main: np.array = tertiary_d_could_be_main & np.logical_not(secondary_d_is_main)

        return np.select(
            condlist=[primary_d_is_main, secondary_d_is_main, tertiary_d_is_main],
            choicelist=[primary_d_pdg, secondary_d_pdg, tertiary_d_pdg],
            default=MCSigQueries.no_main_sig_d_match_value,  # This is the case, if no true d meson is available
        )

    @staticmethod
    def reco_lepton_is_light_lepton(df: pd.DataFrame) -> np.ndarray:
        lepton_pdg: np.ndarray = MCSigQueries.get_mc_matched_reco_lepton_pdg(df=df)
        is_light_lepton: np.ndarray = np.isin(element=abs(lepton_pdg), test_elements=PcC.light_lepton_pdg_ids)
        return is_light_lepton

    @staticmethod
    def reco_lepton_is_electron(df: pd.DataFrame) -> np.ndarray:
        lepton_pdg: np.ndarray = MCSigQueries.get_mc_matched_reco_lepton_pdg(df=df)
        is_electron: np.ndarray = abs(lepton_pdg) == PcC.electron_pdg_id
        return is_electron

    @staticmethod
    def reco_lepton_is_muon(df: pd.DataFrame) -> np.ndarray:
        lepton_pdg: np.ndarray = MCSigQueries.get_mc_matched_reco_lepton_pdg(df=df)
        is_muon: np.ndarray = abs(lepton_pdg) == PcC.muon_pdg_id
        return is_muon

    @staticmethod
    def get_true_xth_daughter_pdg(df: pd.DataFrame, x: int) -> np.ndarray:
        assert x in list(np.arange(1, 6)), (x, list(np.arange(1, 6)))

        col_name_mapping: Dict[int, str] = {
            1: Cols.mc_b1_lepton_pdg,
            2: Cols.mc_b1_1st_degree_d_meson_pdg,
            3: Cols.mc_b1_3rd_daughter_pdg,
            4: Cols.mc_b1_4th_daughter_pdg,
            5: Cols.mc_b1_5th_daughter_pdg,
        }

        return MCSigQueries.get_observable_for_sig_b(df=df, obs_col_name=col_name_mapping[x])

    #
    # Lepton Matching
    @staticmethod
    def is_event_with_lepton_pdgs(df: pd.DataFrame, pdgs: Union[int, Tuple[int, ...]]) -> np.ndarray:
        true_lepton_pdg: np.ndarray = MCSigQueries.get_true_lepton_pdg(df=df)
        if isinstance(pdgs, int):
            return np.abs(true_lepton_pdg) == pdgs
        elif isinstance(pdgs, tuple):
            assert all(isinstance(pdg, int) for pdg in pdgs), pdgs
            return np.isin(element=np.abs(true_lepton_pdg), test_elements=pdgs)

    @staticmethod
    def is_true_tau_event(df: pd.DataFrame) -> np.ndarray:
        return MCSigQueries.is_event_with_lepton_pdgs(df=df, pdgs=PcC.tau_pdg_id)

    @staticmethod
    def is_true_tau_via_leptons_event(df: pd.DataFrame) -> np.ndarray:
        is_true_tau_event: np.ndarray = MCSigQueries.is_true_tau_event(df=df)

        lepton_dauther_pdg: np.ndarray = MCSigQueries.get_true_lepton_daughter_pdg(df=df)
        is_tau_to_lepton: np.ndarray = np.isin(element=np.abs(lepton_dauther_pdg), test_elements=PcC.light_lepton_pdg_ids)

        reco_lepton_is_light_lepton: np.ndarray = MCSigQueries.reco_lepton_is_light_lepton(df=df)

        return is_true_tau_event & is_tau_to_lepton & reco_lepton_is_light_lepton

    @staticmethod
    def is_true_tau_via_electron_event(df: pd.DataFrame) -> np.ndarray:
        is_true_tau_event: np.ndarray = MCSigQueries.is_true_tau_event(df=df)

        lepton_dauther_pdg: np.ndarray = MCSigQueries.get_true_lepton_daughter_pdg(df=df)
        is_tau_to_e: np.ndarray = np.abs(lepton_dauther_pdg) == PcC.electron_pdg_id

        reco_lepton_is_electron: np.ndarray = MCSigQueries.reco_lepton_is_electron(df=df)

        return is_true_tau_event & is_tau_to_e & reco_lepton_is_electron

    @staticmethod
    def is_true_tau_via_muon_event(df: pd.DataFrame) -> np.ndarray:
        is_true_tau_event: np.ndarray = MCSigQueries.is_true_tau_event(df=df)

        lepton_dauther_pdg: np.ndarray = MCSigQueries.get_true_lepton_daughter_pdg(df=df)
        is_tau_to_mu: np.ndarray = np.abs(lepton_dauther_pdg) == PcC.muon_pdg_id

        reco_lepton_is_muon: np.ndarray = MCSigQueries.reco_lepton_is_muon(df=df)

        return is_true_tau_event & is_tau_to_mu & reco_lepton_is_muon

    @staticmethod
    def is_true_tau_via_others_event(df: pd.DataFrame) -> np.ndarray:
        is_true_tau_event: np.ndarray = MCSigQueries.is_true_tau_event(df=df)

        lepton_daughter_pdg: np.ndarray = MCSigQueries.get_true_lepton_daughter_pdg(df=df)
        is_tau_to_other: np.ndarray = np.isin(element=np.abs(lepton_daughter_pdg), test_elements=PcC.light_lepton_pdg_ids)

        return is_true_tau_event & np.logical_not(is_tau_to_other)

    @staticmethod
    def is_true_electron_event(df: pd.DataFrame) -> np.ndarray:
        is_true_electron_event: np.ndarray = MCSigQueries.is_event_with_lepton_pdgs(df=df, pdgs=PcC.electron_pdg_id)
        reco_lepton_is_electron: np.ndarray = MCSigQueries.reco_lepton_is_electron(df=df)
        return is_true_electron_event & reco_lepton_is_electron

    @staticmethod
    def is_true_muon_event(df: pd.DataFrame) -> np.ndarray:
        is_true_muon_event: np.ndarray = MCSigQueries.is_event_with_lepton_pdgs(df=df, pdgs=PcC.muon_pdg_id)
        reco_lepton_is_muon: np.ndarray = MCSigQueries.reco_lepton_is_muon(df=df)
        return is_true_muon_event & reco_lepton_is_muon

    @staticmethod
    def is_true_light_lepton_event(df: pd.DataFrame) -> np.ndarray:
        is_true_light_lep_evt: np.ndarray = MCSigQueries.is_event_with_lepton_pdgs(df=df, pdgs=PcC.light_lepton_pdg_ids)
        reco_lepton_is_light_lepton: np.ndarray = MCSigQueries.reco_lepton_is_light_lepton(df=df)
        return is_true_light_lep_evt & reco_lepton_is_light_lepton

    @staticmethod
    def is_any_true_lepton_event(df: pd.DataFrame) -> np.ndarray:
        is_true_lepton_evt: np.ndarray = MCSigQueries.is_event_with_lepton_pdgs(df=df, pdgs=PcC.all_lepton_pdg_ids)
        reco_lepton_is_light_lepton: np.ndarray = MCSigQueries.reco_lepton_is_light_lepton(df=df)
        return is_true_lepton_evt & reco_lepton_is_light_lepton

    @staticmethod
    def is_any_true_lepton_including_hadronic_tau_event(df: pd.DataFrame) -> np.ndarray:
        is_true_lepton_evt: np.ndarray = MCSigQueries.is_event_with_lepton_pdgs(df=df, pdgs=PcC.all_lepton_pdg_ids)

        reco_lepton_is_light_lepton: np.ndarray = MCSigQueries.reco_lepton_is_light_lepton(df=df)
        is_hadronic_tau: np.ndarray = MCSigQueries.is_true_tau_via_others_event(df=df)

        return is_true_lepton_evt & (reco_lepton_is_light_lepton | is_hadronic_tau)

    @staticmethod
    def is_truth_level_fake_lepton_event(df: pd.DataFrame) -> np.ndarray:
        return np.logical_not(MCSigQueries.is_event_with_lepton_pdgs(df=df, pdgs=PcC.all_lepton_pdg_ids))

    @staticmethod
    def is_lepton_cross_feed_event(
        df: pd.DataFrame,
        reco_pdgs: Union[int, Tuple[int, ...]],
        truth_pdgs: Union[int, Tuple[int, ...]],
        truth_daughter_pdgs: Union[None, int, Tuple[int, ...]],
    ) -> np.ndarray:

        reco_lepton_pdg: np.ndarray = MCSigQueries.get_mc_matched_reco_lepton_pdg(df=df)
        if isinstance(reco_pdgs, int):
            is_reco_lepton_match: np.ndarray = np.abs(reco_lepton_pdg) == reco_pdgs
        elif isinstance(reco_pdgs, tuple):
            assert all(isinstance(pdg, int) for pdg in reco_pdgs), reco_pdgs
            is_reco_lepton_match = np.isin(element=np.abs(reco_lepton_pdg), test_elements=reco_pdgs)
        else:
            raise ValueError(reco_pdgs, type(reco_pdgs))

        is_true_lepton_match: np.ndarray = MCSigQueries.is_event_with_lepton_pdgs(df=df, pdgs=truth_pdgs)

        truth_lepton_daughter_pdg: np.ndarray = MCSigQueries.get_true_lepton_daughter_pdg(df=df)
        if truth_daughter_pdgs is None:
            is_true_daughter_match: np.ndarray = np.full(shape=len(is_true_lepton_match), fill_value=True)
        elif isinstance(truth_daughter_pdgs, int):
            is_true_daughter_match = np.abs(truth_lepton_daughter_pdg) == truth_daughter_pdgs
        elif isinstance(truth_daughter_pdgs, tuple):
            assert all(isinstance(pdg, int) for pdg in truth_daughter_pdgs), truth_daughter_pdgs
            is_true_daughter_match = np.isin(element=np.abs(truth_lepton_daughter_pdg), test_elements=truth_daughter_pdgs)
        else:
            raise ValueError(truth_daughter_pdgs, type(truth_daughter_pdgs))

        return is_reco_lepton_match & is_true_lepton_match & is_true_daughter_match

    #
    # D Meson Matching
    @staticmethod
    def is_event_with_d_meson_pdgs(df: pd.DataFrame, pdgs: Union[int, Tuple[int, ...]]) -> np.ndarray:
        true_d_pdg: np.ndarray = MCSigQueries.get_true_d_meson_pdg(df=df)
        if isinstance(pdgs, int):
            return np.abs(true_d_pdg) == pdgs
        elif isinstance(pdgs, tuple):
            assert all(isinstance(pdg, int) for pdg in pdgs), pdgs
            return np.isin(element=np.abs(true_d_pdg), test_elements=pdgs)

    @staticmethod
    def is_any_d_event(df: pd.DataFrame) -> np.ndarray:
        return MCSigQueries.is_event_with_d_meson_pdgs(df=df, pdgs=PcC.all_d_pdg_ids)

    @staticmethod
    def is_not_any_d_event(df: pd.DataFrame) -> np.ndarray:
        return np.logical_not(MCSigQueries.is_any_d_event(df=df))

    @staticmethod
    def is_d_wwo_st_event(df: pd.DataFrame) -> np.ndarray:
        return MCSigQueries.is_event_with_d_meson_pdgs(df=df, pdgs=PcC.d_wwo_st_pdg_ids)

    @staticmethod
    def is_any_neutral_d_event(df: pd.DataFrame) -> np.ndarray:
        return MCSigQueries.is_event_with_d_meson_pdgs(df=df, pdgs=PcC.all_neutral_d_pdg_ids)

    @staticmethod
    def is_any_charged_d_event(df: pd.DataFrame) -> np.ndarray:
        return MCSigQueries.is_event_with_d_meson_pdgs(df=df, pdgs=PcC.all_charged_d_pdg_ids)

    @staticmethod
    def is_d_event(df: pd.DataFrame) -> np.ndarray:
        return MCSigQueries.is_event_with_d_meson_pdgs(df=df, pdgs=PcC.d_pdg_ids)

    @staticmethod
    def is_neutral_d_event(df: pd.DataFrame) -> np.ndarray:
        return MCSigQueries.is_event_with_d_meson_pdgs(df=df, pdgs=PcC.d_neutral_pdg_id)

    @staticmethod
    def is_charged_d_event(df: pd.DataFrame) -> np.ndarray:
        return MCSigQueries.is_event_with_d_meson_pdgs(df=df, pdgs=PcC.d_charged_pdg_id)

    @staticmethod
    def is_dst_event(df: pd.DataFrame) -> np.ndarray:
        return MCSigQueries.is_event_with_d_meson_pdgs(df=df, pdgs=PcC.dst_pdg_ids)

    @staticmethod
    def is_neutral_dst_event(df: pd.DataFrame) -> np.ndarray:
        return MCSigQueries.is_event_with_d_meson_pdgs(df=df, pdgs=PcC.dst_neutral_pdg_id)

    @staticmethod
    def is_charged_dst_event(df: pd.DataFrame) -> np.ndarray:
        return MCSigQueries.is_event_with_d_meson_pdgs(df=df, pdgs=PcC.dst_charged_pdg_id)

    @staticmethod
    def is_dstst_event(df: pd.DataFrame) -> np.ndarray:
        return MCSigQueries.is_event_with_d_meson_pdgs(df=df, pdgs=PcC.dstst_pdg_ids)

    @staticmethod
    def is_neutral_dstst_event(df: pd.DataFrame) -> np.ndarray:
        return MCSigQueries.is_event_with_d_meson_pdgs(df=df, pdgs=PcC.neutral_dstst_pdg_ids)

    @staticmethod
    def is_charged_dstst_event(df: pd.DataFrame) -> np.ndarray:
        return MCSigQueries.is_event_with_d_meson_pdgs(df=df, pdgs=PcC.charged_dstst_pdg_ids)

    @staticmethod
    def is_charged_dstst_done_event(df: pd.DataFrame) -> np.ndarray:
        return MCSigQueries.is_event_with_d_meson_pdgs(df=df, pdgs=PcC.done_pdg_ids)

    @staticmethod
    def is_charged_dstst_dtwostar_event(df: pd.DataFrame) -> np.ndarray:
        return MCSigQueries.is_event_with_d_meson_pdgs(df=df, pdgs=PcC.dtwostar_pdg_ids)

    @staticmethod
    def is_charged_dstst_dzerostar_event(df: pd.DataFrame) -> np.ndarray:
        return MCSigQueries.is_event_with_d_meson_pdgs(df=df, pdgs=PcC.dzerostar_pdg_ids)

    @staticmethod
    def is_charged_dstst_dprimeone_event(df: pd.DataFrame) -> np.ndarray:
        return MCSigQueries.is_event_with_d_meson_pdgs(df=df, pdgs=PcC.dprimeone_pdg_ids)

    @staticmethod
    def is_neutral_dstst_via_pi0_event(df: pd.DataFrame) -> np.ndarray:
        is_neutral_dstst: np.ndarray = MCSigQueries.is_event_with_d_meson_pdgs(df=df, pdgs=PcC.neutral_dstst_pdg_ids)
        is_dstst_to_x_neutral_pion: np.ndarray = MCSigQueries.second_primary_d_meson_daughter_is_neutral_pion(df=df)
        is_dstst_with_no_3rd_daughter: np.ndarray = MCSigQueries.third_primary_d_meson_daughter_is_photon_or_nothing(
            df=df
        )

        return is_neutral_dstst & is_dstst_to_x_neutral_pion & is_dstst_with_no_3rd_daughter

    @staticmethod
    def is_neutral_dstst_via_pi_plus_event(df: pd.DataFrame) -> np.ndarray:
        is_neutral_dstst: np.ndarray = MCSigQueries.is_event_with_d_meson_pdgs(df=df, pdgs=PcC.neutral_dstst_pdg_ids)
        is_dstst_to_x_charged_pion: np.ndarray = MCSigQueries.second_primary_d_meson_daughter_is_charged_pion(df=df)
        is_dstst_with_no_3rd_daughter: np.ndarray = MCSigQueries.third_primary_d_meson_daughter_is_photon_or_nothing(
            df=df
        )

        return is_neutral_dstst & is_dstst_to_x_charged_pion & is_dstst_with_no_3rd_daughter

    @staticmethod
    def is_charged_dstst_via_pi0_event(df: pd.DataFrame) -> np.ndarray:
        is_charged_dstst: np.ndarray = MCSigQueries.is_event_with_d_meson_pdgs(df=df, pdgs=PcC.charged_dstst_pdg_ids)
        is_dstst_to_x_neutral_pion: np.ndarray = MCSigQueries.second_primary_d_meson_daughter_is_neutral_pion(df=df)
        is_dstst_with_no_3rd_daughter: np.ndarray = MCSigQueries.third_primary_d_meson_daughter_is_photon_or_nothing(
            df=df
        )

        return is_charged_dstst & is_dstst_to_x_neutral_pion & is_dstst_with_no_3rd_daughter

    @staticmethod
    def is_charged_dstst_via_pi_plus_event(df: pd.DataFrame) -> np.ndarray:
        is_charged_dstst: np.ndarray = MCSigQueries.is_event_with_d_meson_pdgs(df=df, pdgs=PcC.charged_dstst_pdg_ids)
        is_dstst_to_x_charged_pion: np.ndarray = MCSigQueries.second_primary_d_meson_daughter_is_charged_pion(df=df)
        is_dstst_with_no_3rd_daughter: np.ndarray = MCSigQueries.third_primary_d_meson_daughter_is_photon_or_nothing(
            df=df
        )

        return is_charged_dstst & is_dstst_to_x_charged_pion & is_dstst_with_no_3rd_daughter

    @staticmethod
    def get_dstst_query_with_dstst_requirement(df: pd.DataFrame, dstst: str, other_query: Callable) -> np.ndarray:
        d_st_st_mapping: Dict[str, Tuple[int, ...]] = {
            "Done": PcC.done_pdg_ids,
            "Dtwostar": PcC.dtwostar_pdg_ids,
            "Dzerostar": PcC.dzerostar_pdg_ids,
            "Dprimeone": PcC.dprimeone_pdg_ids,
        }
        assert dstst in d_st_st_mapping, (dstst, list(d_st_st_mapping.keys()))

        initial_query_result: np.ndarray = other_query(df=df)
        is_dstst_meson_type: np.ndarray = MCSigQueries.is_event_with_d_meson_pdgs(df=df, pdgs=d_st_st_mapping[dstst])

        return initial_query_result & is_dstst_meson_type

    @staticmethod
    def is_dstst_with_more_than_two_daughters_event(df: pd.DataFrame) -> np.ndarray:
        is_dstst: np.ndarray = MCSigQueries.is_dstst_event(df=df)
        is_dstst_wo_3rd_daughter: np.ndarray = MCSigQueries.third_primary_d_meson_daughter_is_photon_or_nothing(df=df)
        is_dstst_with_3rd_daughter: np.ndarray = np.logical_not(is_dstst_wo_3rd_daughter)

        return is_dstst & is_dstst_with_3rd_daughter

    @staticmethod
    def is_dstst_with_wrong_daughters_event(df: pd.DataFrame) -> np.ndarray:
        is_dstst: np.ndarray = MCSigQueries.is_dstst_event(df=df)
        first_dstst_daughter_pdg: np.ndarray = MCSigQueries.get_true_d_meson_first_daughter_pdg(df=df)
        second_dstst_daughter_pdg: np.ndarray = MCSigQueries.get_true_d_meson_second_daughter_pdg(df=df)

        wrong_first_daughter: np.ndarray = np.logical_not(np.isin(np.abs(first_dstst_daughter_pdg), PcC.d_wwo_st_pdg_ids))
        wrong_second_daughter: np.ndarray = np.logical_not(np.isin(np.abs(second_dstst_daughter_pdg), PcC.pion_pdg_ids))

        return is_dstst & (wrong_first_daughter | wrong_second_daughter)

    @staticmethod
    def first_primary_d_meson_daughter_is_d(df: pd.DataFrame) -> np.ndarray:
        d_1st_daughter_pdg = MCSigQueries.get_true_d_meson_first_daughter_pdg(df=df)
        return np.isin(elements=np.abs(d_1st_daughter_pdg), test_elements=PcC.d_pdg_ids)

    @staticmethod
    def first_primary_d_meson_daughter_is_dst(df: pd.DataFrame) -> np.ndarray:
        d_1st_daughter_pdg = MCSigQueries.get_true_d_meson_first_daughter_pdg(df=df)
        return np.isin(elements=np.abs(d_1st_daughter_pdg), test_elements=PcC.dst_pdg_ids)

    @staticmethod
    def first_primary_d_meson_daughter_is_charged_d(df: pd.DataFrame) -> np.ndarray:
        d_1st_daughter_pdg = MCSigQueries.get_true_d_meson_first_daughter_pdg(df=df)
        return np.abs(d_1st_daughter_pdg) == PcC.d_charged_pdg_id

    @staticmethod
    def first_primary_d_meson_daughter_is_neutral_d(df: pd.DataFrame) -> np.ndarray:
        d_1st_daughter_pdg = MCSigQueries.get_true_d_meson_first_daughter_pdg(df=df)
        return np.abs(d_1st_daughter_pdg) == PcC.d_neutral_pdg_id

    @staticmethod
    def first_primary_d_meson_daughter_is_charged_dst(df: pd.DataFrame) -> np.ndarray:
        d_1st_daughter_pdg = MCSigQueries.get_true_d_meson_first_daughter_pdg(df=df)
        return np.abs(d_1st_daughter_pdg) == PcC.dst_charged_pdg_id

    @staticmethod
    def first_primary_d_meson_daughter_is_neutral_dst(df: pd.DataFrame) -> np.ndarray:
        d_1st_daughter_pdg = MCSigQueries.get_true_d_meson_first_daughter_pdg(df=df)
        return np.abs(d_1st_daughter_pdg) == PcC.dst_neutral_pdg_id

    @staticmethod
    def second_primary_d_meson_daughter_is_charged_pion(df: pd.DataFrame) -> np.ndarray:
        d_2nd_daughter_pdg = MCSigQueries.get_true_d_meson_second_daughter_pdg(df=df)
        return np.abs(d_2nd_daughter_pdg) == PcC.pi_charged_pdg_id

    @staticmethod
    def second_primary_d_meson_daughter_is_neutral_pion(df: pd.DataFrame) -> np.ndarray:
        d_2nd_daughter_pdg = MCSigQueries.get_true_d_meson_second_daughter_pdg(df=df)
        return np.abs(d_2nd_daughter_pdg) == PcC.pi_neutral_pdg_id

    @staticmethod
    def second_primary_d_meson_daughter_is_any_pion(df: pd.DataFrame) -> np.ndarray:
        d_2nd_daughter_pdg = MCSigQueries.get_true_d_meson_second_daughter_pdg(df=df)
        return np.isin(element=np.abs(d_2nd_daughter_pdg), test_elements=PcC.pion_pdg_ids)

    @staticmethod
    def second_primary_d_meson_daughter_is_photon(df: pd.DataFrame) -> np.ndarray:
        d_2nd_daughter_pdg = MCSigQueries.get_true_d_meson_second_daughter_pdg(df=df)
        return np.abs(d_2nd_daughter_pdg) == PcC.photon_pdg_id

    @staticmethod
    def third_primary_d_meson_daughter_is_photon_or_nothing(df: pd.DataFrame) -> np.ndarray:
        third_d_daughter_pdg: np.ndarray = MCSigQueries.get_true_d_meson_third_daughter_pdg(df=df)
        return np.isin(element=third_d_daughter_pdg, test_elements=PcC.photon_or_nothing_pdg_ids)

    # Neutrino and Remaining Particles Matching

    @staticmethod
    def xth_daughter_is_any_neutrino(df: pd.DataFrame, x: int) -> np.ndarray:
        xth_daughter_pdg: np.ndarray = MCSigQueries.get_true_xth_daughter_pdg(df=df, x=x)
        return np.isin(element=np.abs(xth_daughter_pdg), test_elements=PcC.all_neutrino_pdg_ids)

    @staticmethod
    def xth_daughter_is_light_lepton_neutrino(df: pd.DataFrame, x: int) -> np.ndarray:
        xth_daughter_pdg: np.ndarray = MCSigQueries.get_true_xth_daughter_pdg(df=df, x=x)
        return np.isin(element=np.abs(xth_daughter_pdg), test_elements=PcC.nu_light_lepton_pdg_ids)

    @staticmethod
    def xth_daughter_is_y_neutrino(df: pd.DataFrame, x: int, y: int) -> np.ndarray:
        assert abs(y) in PcC.lepton_to_nu_pdg_mapping.keys() or y in PcC.lepton_to_nu_pdg_mapping.values(), y
        if abs(y) in PcC.lepton_to_nu_pdg_mapping.keys():
            nu_pdg: int = PcC.lepton_to_nu_pdg_mapping[abs(y)]
        else:
            nu_pdg = abs(y)
        xth_daughter_pdg: np.ndarray = MCSigQueries.get_true_xth_daughter_pdg(df=df, x=x)
        return np.abs(xth_daughter_pdg) == nu_pdg

    @staticmethod
    def xth_daughter_is_photon_or_nothing(df: pd.DataFrame, x: int) -> np.ndarray:
        xth_daughter_pdg: np.ndarray = MCSigQueries.get_true_xth_daughter_pdg(df=df, x=x)
        return np.isin(element=xth_daughter_pdg, test_elements=PcC.photon_or_nothing_pdg_ids)

    # Functions for BKG identification
    @staticmethod
    def is_lepton_fake(df: pd.DataFrame) -> np.ndarray:
        mc_matched_lepton_pdg: np.ndarray = MCSigQueries.get_mc_matched_reco_lepton_pdg(df=df)
        is_fake_reco_lepton: np.ndarray = np.logical_not(
            np.isin(element=np.abs(mc_matched_lepton_pdg), test_elements=PcC.all_lepton_pdg_ids)
        )

        is_truth_level_lepton_fake: np.ndarray = MCSigQueries.is_truth_level_fake_lepton_event(df=df)
        is_not_cont_event: np.ndarray = MCSigQueries.is_not_continuum_event(df=df)

        return is_fake_reco_lepton & is_truth_level_lepton_fake & is_not_cont_event

    @staticmethod
    def is_lepton_reco_fake(df: pd.DataFrame) -> np.ndarray:
        mc_matched_lepton_pdg: np.ndarray = MCSigQueries.get_mc_matched_reco_lepton_pdg(df=df)
        is_fake_reco_lepton: np.ndarray = np.logical_not(
            np.isin(element=np.abs(mc_matched_lepton_pdg), test_elements=PcC.all_lepton_pdg_ids)
        )

        is_truth_level_lepton: np.ndarray = MCSigQueries.is_event_with_lepton_pdgs(df=df, pdgs=PcC.all_lepton_pdg_ids)
        is_not_turth_level_hadronic_tau: np.ndarray = np.logical_not(MCSigQueries.is_true_tau_via_others_event(df=df))

        is_not_cont_event: np.ndarray = MCSigQueries.is_not_continuum_event(df=df)

        return is_fake_reco_lepton & is_truth_level_lepton & is_not_turth_level_hadronic_tau & is_not_cont_event

    ##################################
    # B to lepton candidate matching #
    ##################################

    @staticmethod
    def is_b_to_x_tau_nu_with_leptonic_tau(df: pd.DataFrame) -> np.ndarray:
        is_tau_via_lepton: np.ndarray = MCSigQueries.is_true_tau_via_leptons_event(df=df)
        is_tau_neutrino: np.ndarray = MCSigQueries.xth_daughter_is_y_neutrino(df=df, x=3, y=PcC.nu_tau_pdg_id)
        no_more_daughters: np.ndarray = MCSigQueries.xth_daughter_is_photon_or_nothing(df=df, x=4)

        return is_tau_via_lepton & is_tau_neutrino & no_more_daughters

    @staticmethod
    def is_b_to_x_tau_nu_with_leptonic_tau_via_e(df: pd.DataFrame) -> np.ndarray:
        is_tau_via_e: np.ndarray = MCSigQueries.is_true_tau_via_electron_event(df=df)
        is_tau_neutrino: np.ndarray = MCSigQueries.xth_daughter_is_y_neutrino(df=df, x=3, y=PcC.nu_tau_pdg_id)
        no_more_daughters: np.ndarray = MCSigQueries.xth_daughter_is_photon_or_nothing(df=df, x=4)

        return is_tau_via_e & is_tau_neutrino & no_more_daughters

    @staticmethod
    def is_b_to_x_tau_nu_with_leptonic_tau_via_mu(df: pd.DataFrame) -> np.ndarray:
        is_tau_via_mu: np.ndarray = MCSigQueries.is_true_tau_via_muon_event(df=df)
        is_tau_neutrino: np.ndarray = MCSigQueries.xth_daughter_is_y_neutrino(df=df, x=3, y=PcC.nu_tau_pdg_id)
        no_more_daughters: np.ndarray = MCSigQueries.xth_daughter_is_photon_or_nothing(df=df, x=4)

        return is_tau_via_mu & is_tau_neutrino & no_more_daughters

    @staticmethod
    def is_b_to_x_tau_nu_with_hadronic_tau(df: pd.DataFrame) -> np.ndarray:
        is_tau_via_other: np.ndarray = MCSigQueries.is_true_tau_via_others_event(df=df)
        is_tau_neutrino: np.ndarray = MCSigQueries.xth_daughter_is_y_neutrino(df=df, x=3, y=PcC.nu_tau_pdg_id)
        no_more_daughters: np.ndarray = MCSigQueries.xth_daughter_is_photon_or_nothing(df=df, x=4)

        return is_tau_via_other & is_tau_neutrino & no_more_daughters

    @staticmethod
    def is_b_to_x_tau_nu_with_any_tau(df: pd.DataFrame) -> np.ndarray:
        is_tau_via_any: np.ndarray = MCSigQueries.is_true_tau_event(df=df)
        is_tau_neutrino: np.ndarray = MCSigQueries.xth_daughter_is_y_neutrino(df=df, x=3, y=PcC.nu_tau_pdg_id)
        no_more_daughters: np.ndarray = MCSigQueries.xth_daughter_is_photon_or_nothing(df=df, x=4)

        return is_tau_via_any & is_tau_neutrino & no_more_daughters

    @staticmethod
    def is_b_to_x_electron_nu(df: pd.DataFrame) -> np.ndarray:
        is_electron: np.ndarray = MCSigQueries.is_true_electron_event(df=df)
        is_electron_neutrino: np.ndarray = MCSigQueries.xth_daughter_is_y_neutrino(df=df, x=3, y=PcC.nu_electron_pdg_id)
        no_more_daughters: np.ndarray = MCSigQueries.xth_daughter_is_photon_or_nothing(df=df, x=4)

        return is_electron & is_electron_neutrino & no_more_daughters

    @staticmethod
    def is_b_to_x_muon_nu(df: pd.DataFrame) -> np.ndarray:
        is_muon: np.ndarray = MCSigQueries.is_true_muon_event(df=df)
        is_muon_neutrino: np.ndarray = MCSigQueries.xth_daughter_is_y_neutrino(df=df, x=3, y=PcC.nu_muon_pdg_id)
        no_more_daughters: np.ndarray = MCSigQueries.xth_daughter_is_photon_or_nothing(df=df, x=4)

        return is_muon & is_muon_neutrino & no_more_daughters

    @staticmethod
    def is_b_to_x_any_lepton_nu(df: pd.DataFrame) -> np.ndarray:
        is_any_lepton: np.ndarray = MCSigQueries.is_any_true_lepton_event(df=df)
        is_any_neutrino: np.ndarray = MCSigQueries.xth_daughter_is_any_neutrino(df=df, x=3)
        no_more_daughters: np.ndarray = MCSigQueries.xth_daughter_is_photon_or_nothing(df=df, x=4)

        return is_any_lepton & is_any_neutrino & no_more_daughters

    @staticmethod
    def is_b_to_x_any_lepton_nu_including_hadronic_tau(df: pd.DataFrame) -> np.ndarray:
        is_any_lepton_including_had_tau: np.ndarray = MCSigQueries.is_any_true_lepton_including_hadronic_tau_event(df=df)
        is_any_neutrino: np.ndarray = MCSigQueries.xth_daughter_is_any_neutrino(df=df, x=3)
        no_more_daughters: np.ndarray = MCSigQueries.xth_daughter_is_photon_or_nothing(df=df, x=4)

        return is_any_lepton_including_had_tau & is_any_neutrino & no_more_daughters

    @staticmethod
    def is_b_to_x_light_lepton_nu(df: pd.DataFrame) -> np.ndarray:
        is_light_lepton: np.ndarray = MCSigQueries.is_true_light_lepton_event(df=df)
        is_light_lepton_neutrino: np.ndarray = MCSigQueries.xth_daughter_is_light_lepton_neutrino(df=df, x=3)
        no_more_daughters: np.ndarray = MCSigQueries.xth_daughter_is_photon_or_nothing(df=df, x=4)

        return is_light_lepton & is_light_lepton_neutrino & no_more_daughters

    #######################################
    # Some B meson queries for direct use #
    #######################################

    @staticmethod
    def is_lepton_but_not_from_b_meson(df: pd.DataFrame) -> np.ndarray:
        mc_matched_lepton_pdg: np.ndarray = MCSigQueries.get_mc_matched_reco_lepton_pdg(df=df)
        is_lepton_pdg: np.ndarray = np.isin(element=np.abs(mc_matched_lepton_pdg), test_elements=PcC.all_lepton_pdg_ids)

        is_not_from_b_meson: np.ndarray = MCSigQueries.is_event_without_lepton_to_b_meson_relation(df=df)
        is_not_cont_event: np.ndarray = MCSigQueries.is_not_continuum_event(df=df)

        return is_lepton_pdg & is_not_from_b_meson & is_not_cont_event

    @staticmethod
    def is_lepton_but_no_related_d_meson(df: pd.DataFrame) -> np.ndarray:
        is_b_to_x_any_lepton_nu_event: np.ndarray = MCSigQueries.is_b_to_x_any_lepton_nu_including_hadronic_tau(df=df)

        is_not_d_meson_event: np.ndarray = MCSigQueries.is_not_any_d_event(df=df)

        return is_b_to_x_any_lepton_nu_event & is_not_d_meson_event

    @staticmethod
    def is_dstst_meson_l_nu_with_dstst_to_three_event(df: pd.DataFrame) -> np.ndarray:
        is_b_to_x_any_lepton_nu_event: np.ndarray = MCSigQueries.is_b_to_x_any_lepton_nu_including_hadronic_tau(df=df)
        is_dstst_to_three_event: np.ndarray = MCSigQueries.is_dstst_with_more_than_two_daughters_event(df=df)

        return is_b_to_x_any_lepton_nu_event & is_dstst_to_three_event

    @staticmethod
    def is_b_to_pi_l_nu_event(df: pd.DataFrame) -> np.ndarray:
        is_b_to_x_any_lepton_nu_event: np.ndarray = MCSigQueries.is_b_to_x_any_lepton_nu_including_hadronic_tau(df=df)
        is_b_to_pi_event: np.ndarray = MCSigQueries.is_event_with_d_meson_pdgs(df=df, pdgs=PcC.pion_pdg_ids)

        return is_b_to_x_any_lepton_nu_event & is_b_to_pi_event

    @staticmethod
    def is_b_to_neutral_pi_l_nu_event(df: pd.DataFrame) -> np.ndarray:
        is_b_to_x_light_lepton_nu_event: np.ndarray = MCSigQueries.is_b_to_x_light_lepton_nu(df=df)
        is_b_to_neutral_pi_event: np.ndarray = MCSigQueries.is_event_with_d_meson_pdgs(df=df, pdgs=PcC.pi_neutral_pdg_id)

        return is_b_to_x_light_lepton_nu_event & is_b_to_neutral_pi_event

    @staticmethod
    def is_b_to_charged_pi_l_nu_event(df: pd.DataFrame) -> np.ndarray:
        is_b_to_x_light_lepton_nu_event: np.ndarray = MCSigQueries.is_b_to_x_light_lepton_nu(df=df)
        is_b_to_charged_pi_event: np.ndarray = MCSigQueries.is_event_with_d_meson_pdgs(df=df, pdgs=PcC.pi_charged_pdg_id)

        return is_b_to_x_light_lepton_nu_event & is_b_to_charged_pi_event

    @staticmethod
    def is_b_to_neutral_pi_tau_nu_event(df: pd.DataFrame) -> np.ndarray:
        is_b_to_x_tau_nu_event: np.ndarray = MCSigQueries.is_b_to_x_tau_nu_with_any_tau(df=df)
        is_b_to_neutral_pi_event: np.ndarray = MCSigQueries.is_event_with_d_meson_pdgs(df=df, pdgs=PcC.pi_neutral_pdg_id)

        return is_b_to_x_tau_nu_event & is_b_to_neutral_pi_event

    @staticmethod
    def is_b_to_charged_pi_tau_nu_event(df: pd.DataFrame) -> np.ndarray:
        is_b_to_x_tau_nu_event: np.ndarray = MCSigQueries.is_b_to_x_tau_nu_with_any_tau(df=df)
        is_b_to_charged_pi_event: np.ndarray = MCSigQueries.is_event_with_d_meson_pdgs(df=df, pdgs=PcC.pi_charged_pdg_id)

        return is_b_to_x_tau_nu_event & is_b_to_charged_pi_event

    ##########################################
    # Extended signal queries for direct use #
    ##########################################

    # B -> D(*) leptonic tau
    @staticmethod
    def cB_to_nD_tau_nu_query(df: pd.DataFrame) -> np.ndarray:
        is_tau: np.ndarray = MCSigQueries.is_b_to_x_tau_nu_with_leptonic_tau(df=df)

        is_neutral_d: np.ndarray = MCSigQueries.is_neutral_d_event(df=df)
        is_charged_b: np.ndarray = MCSigQueries.is_charged_b_meson(df=df)

        return is_tau & is_neutral_d & is_charged_b

    @staticmethod
    def cB_to_nD_tau_via_e_nu_query(df: pd.DataFrame) -> np.ndarray:
        is_tau: np.ndarray = MCSigQueries.is_b_to_x_tau_nu_with_leptonic_tau_via_e(df=df)

        is_neutral_d: np.ndarray = MCSigQueries.is_neutral_d_event(df=df)
        is_charged_b: np.ndarray = MCSigQueries.is_charged_b_meson(df=df)

        return is_tau & is_neutral_d & is_charged_b

    @staticmethod
    def cB_to_nD_tau_via_mu_nu_query(df: pd.DataFrame) -> np.ndarray:
        is_tau: np.ndarray = MCSigQueries.is_b_to_x_tau_nu_with_leptonic_tau_via_mu(df=df)

        is_neutral_d: np.ndarray = MCSigQueries.is_neutral_d_event(df=df)
        is_charged_b: np.ndarray = MCSigQueries.is_charged_b_meson(df=df)

        return is_tau & is_neutral_d & is_charged_b

    @staticmethod
    def nB_to_cD_tau_nu_query(df: pd.DataFrame) -> np.ndarray:
        is_tau: np.ndarray = MCSigQueries.is_b_to_x_tau_nu_with_leptonic_tau(df=df)

        is_charged_d: np.ndarray = MCSigQueries.is_charged_d_event(df=df)
        is_neutral_b: np.ndarray = MCSigQueries.is_neutral_b_meson(df=df)

        return is_tau & is_charged_d & is_neutral_b

    @staticmethod
    def nB_to_cD_tau_via_e_nu_query(df: pd.DataFrame) -> np.ndarray:
        is_tau: np.ndarray = MCSigQueries.is_b_to_x_tau_nu_with_leptonic_tau_via_e(df=df)

        is_charged_d: np.ndarray = MCSigQueries.is_charged_d_event(df=df)
        is_neutral_b: np.ndarray = MCSigQueries.is_neutral_b_meson(df=df)

        return is_tau & is_charged_d & is_neutral_b

    @staticmethod
    def nB_to_cD_tau_via_mu_nu_query(df: pd.DataFrame) -> np.ndarray:
        is_tau: np.ndarray = MCSigQueries.is_b_to_x_tau_nu_with_leptonic_tau_via_mu(df=df)

        is_charged_d: np.ndarray = MCSigQueries.is_charged_d_event(df=df)
        is_neutral_b: np.ndarray = MCSigQueries.is_neutral_b_meson(df=df)

        return is_tau & is_charged_d & is_neutral_b

    @staticmethod
    def cB_to_nDst_tau_nu_query(df: pd.DataFrame) -> np.ndarray:
        is_tau: np.ndarray = MCSigQueries.is_b_to_x_tau_nu_with_leptonic_tau(df=df)

        is_neutral_dst: np.ndarray = MCSigQueries.is_neutral_dst_event(df=df)
        is_charged_b: np.ndarray = MCSigQueries.is_charged_b_meson(df=df)

        return is_tau & is_neutral_dst & is_charged_b

    @staticmethod
    def cB_to_nDst_tau_via_e_nu_query(df: pd.DataFrame) -> np.ndarray:
        is_tau: np.ndarray = MCSigQueries.is_b_to_x_tau_nu_with_leptonic_tau_via_e(df=df)

        is_neutral_dst: np.ndarray = MCSigQueries.is_neutral_dst_event(df=df)
        is_charged_b: np.ndarray = MCSigQueries.is_charged_b_meson(df=df)

        return is_tau & is_neutral_dst & is_charged_b

    @staticmethod
    def cB_to_nDst_tau_via_mu_nu_query(df: pd.DataFrame) -> np.ndarray:
        is_tau: np.ndarray = MCSigQueries.is_b_to_x_tau_nu_with_leptonic_tau_via_mu(df=df)

        is_neutral_dst: np.ndarray = MCSigQueries.is_neutral_dst_event(df=df)
        is_charged_b: np.ndarray = MCSigQueries.is_charged_b_meson(df=df)

        return is_tau & is_neutral_dst & is_charged_b

    @staticmethod
    def nB_to_cDst_tau_nu_query(df: pd.DataFrame) -> np.ndarray:
        is_tau: np.ndarray = MCSigQueries.is_b_to_x_tau_nu_with_leptonic_tau(df=df)

        is_charged_dst: np.ndarray = MCSigQueries.is_charged_dst_event(df=df)
        is_neutral_b: np.ndarray = MCSigQueries.is_neutral_b_meson(df=df)

        return is_tau & is_charged_dst & is_neutral_b

    @staticmethod
    def nB_to_cDst_tau_via_e_nu_query(df: pd.DataFrame) -> np.ndarray:
        is_tau: np.ndarray = MCSigQueries.is_b_to_x_tau_nu_with_leptonic_tau_via_e(df=df)

        is_charged_dst: np.ndarray = MCSigQueries.is_charged_dst_event(df=df)
        is_neutral_b: np.ndarray = MCSigQueries.is_neutral_b_meson(df=df)

        return is_tau & is_charged_dst & is_neutral_b

    @staticmethod
    def nB_to_cDst_tau_via_mu_nu_query(df: pd.DataFrame) -> np.ndarray:
        is_tau: np.ndarray = MCSigQueries.is_b_to_x_tau_nu_with_leptonic_tau_via_mu(df=df)

        is_charged_dst: np.ndarray = MCSigQueries.is_charged_dst_event(df=df)
        is_neutral_b: np.ndarray = MCSigQueries.is_neutral_b_meson(df=df)

        return is_tau & is_charged_dst & is_neutral_b

    @staticmethod
    def cB_to_nDst_tau_nu_with_Dst_to_pi_query(df: pd.DataFrame) -> np.ndarray:
        is_tau: np.ndarray = MCSigQueries.is_b_to_x_tau_nu_with_leptonic_tau(df=df)

        is_neutral_dst: np.ndarray = MCSigQueries.is_neutral_dst_event(df=df)
        is_dst_to_pi: np.ndarray = MCSigQueries.second_primary_d_meson_daughter_is_any_pion(df=df)
        is_charged_b: np.ndarray = MCSigQueries.is_charged_b_meson(df=df)

        return is_tau & is_neutral_dst & is_dst_to_pi & is_charged_b

    @staticmethod
    def cB_to_nDst_tau_via_e_nu_with_Dst_to_pi_query(df: pd.DataFrame) -> np.ndarray:
        is_tau: np.ndarray = MCSigQueries.is_b_to_x_tau_nu_with_leptonic_tau_via_e(df=df)

        is_neutral_dst: np.ndarray = MCSigQueries.is_neutral_dst_event(df=df)
        is_dst_to_pi: np.ndarray = MCSigQueries.second_primary_d_meson_daughter_is_any_pion(df=df)
        is_charged_b: np.ndarray = MCSigQueries.is_charged_b_meson(df=df)

        return is_tau & is_neutral_dst & is_dst_to_pi & is_charged_b

    @staticmethod
    def cB_to_nDst_tau_via_mu_nu_with_Dst_to_pi_query(df: pd.DataFrame) -> np.ndarray:
        is_tau: np.ndarray = MCSigQueries.is_b_to_x_tau_nu_with_leptonic_tau_via_mu(df=df)

        is_neutral_dst: np.ndarray = MCSigQueries.is_neutral_dst_event(df=df)
        is_dst_to_pi: np.ndarray = MCSigQueries.second_primary_d_meson_daughter_is_any_pion(df=df)
        is_charged_b: np.ndarray = MCSigQueries.is_charged_b_meson(df=df)

        return is_tau & is_neutral_dst & is_dst_to_pi & is_charged_b

    @staticmethod
    def nB_to_cDst_tau_nu_with_Dst_to_pi_query(df: pd.DataFrame) -> np.ndarray:
        is_tau: np.ndarray = MCSigQueries.is_b_to_x_tau_nu_with_leptonic_tau(df=df)

        is_charged_dst: np.ndarray = MCSigQueries.is_charged_dst_event(df=df)
        is_dst_to_pi: np.ndarray = MCSigQueries.second_primary_d_meson_daughter_is_any_pion(df=df)
        is_neutral_b: np.ndarray = MCSigQueries.is_neutral_b_meson(df=df)

        return is_tau & is_charged_dst & is_dst_to_pi & is_neutral_b

    @staticmethod
    def nB_to_cDst_tau_via_e_nu_with_Dst_to_pi_query(df: pd.DataFrame) -> np.ndarray:
        is_tau: np.ndarray = MCSigQueries.is_b_to_x_tau_nu_with_leptonic_tau_via_e(df=df)

        is_charged_dst: np.ndarray = MCSigQueries.is_charged_dst_event(df=df)
        is_dst_to_pi: np.ndarray = MCSigQueries.second_primary_d_meson_daughter_is_any_pion(df=df)
        is_neutral_b: np.ndarray = MCSigQueries.is_neutral_b_meson(df=df)

        return is_tau & is_charged_dst & is_dst_to_pi & is_neutral_b

    @staticmethod
    def nB_to_cDst_tau_via_mu_nu_with_Dst_to_pi_query(df: pd.DataFrame) -> np.ndarray:
        is_tau: np.ndarray = MCSigQueries.is_b_to_x_tau_nu_with_leptonic_tau_via_mu(df=df)

        is_charged_dst: np.ndarray = MCSigQueries.is_charged_dst_event(df=df)
        is_dst_to_pi: np.ndarray = MCSigQueries.second_primary_d_meson_daughter_is_any_pion(df=df)
        is_neutral_b: np.ndarray = MCSigQueries.is_neutral_b_meson(df=df)

        return is_tau & is_charged_dst & is_dst_to_pi & is_neutral_b

    @staticmethod
    def cB_to_nDst_tau_nu_with_Dst_to_photon_query(df: pd.DataFrame) -> np.ndarray:
        is_tau: np.ndarray = MCSigQueries.is_b_to_x_tau_nu_with_leptonic_tau(df=df)

        is_neutral_dst: np.ndarray = MCSigQueries.is_neutral_dst_event(df=df)
        is_dst_to_photon: np.ndarray = MCSigQueries.second_primary_d_meson_daughter_is_photon(df=df)
        is_charged_b: np.ndarray = MCSigQueries.is_charged_b_meson(df=df)

        return is_tau & is_neutral_dst & is_dst_to_photon & is_charged_b

    @staticmethod
    def cB_to_nDst_tau_via_e_nu_with_Dst_to_photon_query(df: pd.DataFrame) -> np.ndarray:
        is_tau: np.ndarray = MCSigQueries.is_b_to_x_tau_nu_with_leptonic_tau_via_e(df=df)

        is_neutral_dst: np.ndarray = MCSigQueries.is_neutral_dst_event(df=df)
        is_dst_to_photon: np.ndarray = MCSigQueries.second_primary_d_meson_daughter_is_photon(df=df)
        is_charged_b: np.ndarray = MCSigQueries.is_charged_b_meson(df=df)

        return is_tau & is_neutral_dst & is_dst_to_photon & is_charged_b

    @staticmethod
    def cB_to_nDst_tau_via_mu_nu_with_Dst_to_photon_query(df: pd.DataFrame) -> np.ndarray:
        is_tau: np.ndarray = MCSigQueries.is_b_to_x_tau_nu_with_leptonic_tau_via_mu(df=df)

        is_neutral_dst: np.ndarray = MCSigQueries.is_neutral_dst_event(df=df)
        is_dst_to_photon: np.ndarray = MCSigQueries.second_primary_d_meson_daughter_is_photon(df=df)
        is_charged_b: np.ndarray = MCSigQueries.is_charged_b_meson(df=df)

        return is_tau & is_neutral_dst & is_dst_to_photon & is_charged_b

    @staticmethod
    def nB_to_cDst_tau_nu_with_Dst_to_photon_query(df: pd.DataFrame) -> np.ndarray:
        is_tau: np.ndarray = MCSigQueries.is_b_to_x_tau_nu_with_leptonic_tau(df=df)

        is_charged_dst: np.ndarray = MCSigQueries.is_charged_dst_event(df=df)
        is_dst_to_photon: np.ndarray = MCSigQueries.second_primary_d_meson_daughter_is_photon(df=df)
        is_neutral_b: np.ndarray = MCSigQueries.is_neutral_b_meson(df=df)

        return is_tau & is_charged_dst & is_dst_to_photon & is_neutral_b

    @staticmethod
    def nB_to_cDst_tau_via_e_nu_with_Dst_to_photon_query(df: pd.DataFrame) -> np.ndarray:
        is_tau: np.ndarray = MCSigQueries.is_b_to_x_tau_nu_with_leptonic_tau_via_e(df=df)

        is_charged_dst: np.ndarray = MCSigQueries.is_charged_dst_event(df=df)
        is_dst_to_photon: np.ndarray = MCSigQueries.second_primary_d_meson_daughter_is_photon(df=df)
        is_neutral_b: np.ndarray = MCSigQueries.is_neutral_b_meson(df=df)

        return is_tau & is_charged_dst & is_dst_to_photon & is_neutral_b

    @staticmethod
    def nB_to_cDst_tau_via_mu_nu_with_Dst_to_photon_query(df: pd.DataFrame) -> np.ndarray:
        is_tau: np.ndarray = MCSigQueries.is_b_to_x_tau_nu_with_leptonic_tau_via_mu(df=df)

        is_charged_dst: np.ndarray = MCSigQueries.is_charged_dst_event(df=df)
        is_dst_to_photon: np.ndarray = MCSigQueries.second_primary_d_meson_daughter_is_photon(df=df)
        is_neutral_b: np.ndarray = MCSigQueries.is_neutral_b_meson(df=df)

        return is_tau & is_charged_dst & is_dst_to_photon & is_neutral_b

    # B -> D(*) hadronic tau
    @staticmethod
    def cB_to_nD_hadronic_tau_nu_query(df: pd.DataFrame) -> np.ndarray:
        is_hadronic_tau: np.ndarray = MCSigQueries.is_b_to_x_tau_nu_with_hadronic_tau(df=df)

        is_neutral_d: np.ndarray = MCSigQueries.is_neutral_d_event(df=df)
        is_charged_b: np.ndarray = MCSigQueries.is_charged_b_meson(df=df)

        return is_hadronic_tau & is_neutral_d & is_charged_b

    @staticmethod
    def nB_to_cD_hadronic_tau_nu_query(df: pd.DataFrame) -> np.ndarray:
        is_hadronic_tau: np.ndarray = MCSigQueries.is_b_to_x_tau_nu_with_hadronic_tau(df=df)

        is_charged_d: np.ndarray = MCSigQueries.is_charged_d_event(df=df)
        is_neutral_b: np.ndarray = MCSigQueries.is_neutral_b_meson(df=df)

        return is_hadronic_tau & is_charged_d & is_neutral_b

    @staticmethod
    def cB_to_nDst_hadronic_tau_nu_query(df: pd.DataFrame) -> np.ndarray:
        is_hadronic_tau: np.ndarray = MCSigQueries.is_b_to_x_tau_nu_with_hadronic_tau(df=df)

        is_neutral_dst: np.ndarray = MCSigQueries.is_neutral_dst_event(df=df)
        is_charged_b: np.ndarray = MCSigQueries.is_charged_b_meson(df=df)

        return is_hadronic_tau & is_neutral_dst & is_charged_b

    @staticmethod
    def nB_to_cDst_hadronic_tau_nu_query(df: pd.DataFrame) -> np.ndarray:
        is_hadronic_tau: np.ndarray = MCSigQueries.is_b_to_x_tau_nu_with_hadronic_tau(df=df)

        is_charged_dst: np.ndarray = MCSigQueries.is_charged_dst_event(df=df)
        is_neutral_b: np.ndarray = MCSigQueries.is_neutral_b_meson(df=df)

        return is_hadronic_tau & is_charged_dst & is_neutral_b

    # B -> D(*) light lepton
    @staticmethod
    def cB_to_nD_e_nu_query(df: pd.DataFrame) -> np.ndarray:
        is_electron: np.ndarray = MCSigQueries.is_b_to_x_electron_nu(df=df)

        is_neutral_d: np.ndarray = MCSigQueries.is_neutral_d_event(df=df)
        is_charged_b: np.ndarray = MCSigQueries.is_charged_b_meson(df=df)

        return is_electron & is_neutral_d & is_charged_b

    @staticmethod
    def nB_to_cD_e_nu_query(df: pd.DataFrame) -> np.ndarray:
        is_electron: np.ndarray = MCSigQueries.is_b_to_x_electron_nu(df=df)

        is_charged_d: np.ndarray = MCSigQueries.is_charged_d_event(df=df)
        is_neutral_b: np.ndarray = MCSigQueries.is_neutral_b_meson(df=df)

        return is_electron & is_charged_d & is_neutral_b

    @staticmethod
    def cB_to_nD_mu_nu_query(df: pd.DataFrame) -> np.ndarray:
        is_muon: np.ndarray = MCSigQueries.is_b_to_x_muon_nu(df=df)

        is_neutral_d: np.ndarray = MCSigQueries.is_neutral_d_event(df=df)
        is_charged_b: np.ndarray = MCSigQueries.is_charged_b_meson(df=df)

        return is_muon & is_neutral_d & is_charged_b

    @staticmethod
    def nB_to_cD_mu_nu_query(df: pd.DataFrame) -> np.ndarray:
        is_muon: np.ndarray = MCSigQueries.is_b_to_x_muon_nu(df=df)

        is_charged_d: np.ndarray = MCSigQueries.is_charged_d_event(df=df)
        is_neutral_b: np.ndarray = MCSigQueries.is_neutral_b_meson(df=df)

        return is_muon & is_charged_d & is_neutral_b

    @staticmethod
    def cB_to_nDst_e_nu_query(df: pd.DataFrame) -> np.ndarray:
        is_electron: np.ndarray = MCSigQueries.is_b_to_x_electron_nu(df=df)

        is_neutral_dst: np.ndarray = MCSigQueries.is_neutral_dst_event(df=df)
        is_charged_b: np.ndarray = MCSigQueries.is_charged_b_meson(df=df)

        return is_electron & is_neutral_dst & is_charged_b

    @staticmethod
    def nB_to_cDst_e_nu_query(df: pd.DataFrame) -> np.ndarray:
        is_electron: np.ndarray = MCSigQueries.is_b_to_x_electron_nu(df=df)

        is_charged_dst: np.ndarray = MCSigQueries.is_charged_dst_event(df=df)
        is_neutral_b: np.ndarray = MCSigQueries.is_neutral_b_meson(df=df)

        return is_electron & is_charged_dst & is_neutral_b

    @staticmethod
    def cB_to_nDst_e_nu_with_Dst_to_pi_query(df: pd.DataFrame) -> np.ndarray:
        is_electron: np.ndarray = MCSigQueries.is_b_to_x_electron_nu(df=df)

        is_neutral_dst: np.ndarray = MCSigQueries.is_neutral_dst_event(df=df)
        is_dst_to_pi: np.ndarray = MCSigQueries.second_primary_d_meson_daughter_is_any_pion(df=df)
        is_charged_b: np.ndarray = MCSigQueries.is_charged_b_meson(df=df)

        return is_electron & is_neutral_dst & is_dst_to_pi & is_charged_b

    @staticmethod
    def nB_to_cDst_e_nu_with_Dst_to_pi_query(df: pd.DataFrame) -> np.ndarray:
        is_electron: np.ndarray = MCSigQueries.is_b_to_x_electron_nu(df=df)

        is_charged_dst: np.ndarray = MCSigQueries.is_charged_dst_event(df=df)
        is_dst_to_pi: np.ndarray = MCSigQueries.second_primary_d_meson_daughter_is_any_pion(df=df)
        is_neutral_b: np.ndarray = MCSigQueries.is_neutral_b_meson(df=df)

        return is_electron & is_charged_dst & is_dst_to_pi & is_neutral_b

    @staticmethod
    def cB_to_nDst_e_nu_with_Dst_to_photon_query(df: pd.DataFrame) -> np.ndarray:
        is_electron: np.ndarray = MCSigQueries.is_b_to_x_electron_nu(df=df)

        is_neutral_dst: np.ndarray = MCSigQueries.is_neutral_dst_event(df=df)
        is_dst_to_photon: np.ndarray = MCSigQueries.second_primary_d_meson_daughter_is_photon(df=df)
        is_charged_b: np.ndarray = MCSigQueries.is_charged_b_meson(df=df)

        return is_electron & is_neutral_dst & is_dst_to_photon & is_charged_b

    @staticmethod
    def nB_to_cDst_e_nu_with_Dst_to_photon_query(df: pd.DataFrame) -> np.ndarray:
        is_electron: np.ndarray = MCSigQueries.is_b_to_x_electron_nu(df=df)

        is_charged_dst: np.ndarray = MCSigQueries.is_charged_dst_event(df=df)
        is_dst_to_photon: np.ndarray = MCSigQueries.second_primary_d_meson_daughter_is_photon(df=df)
        is_neutral_b: np.ndarray = MCSigQueries.is_neutral_b_meson(df=df)

        return is_electron & is_charged_dst & is_dst_to_photon & is_neutral_b

    @staticmethod
    def cB_to_nDst_mu_nu_query(df: pd.DataFrame) -> np.ndarray:
        is_muon: np.ndarray = MCSigQueries.is_b_to_x_muon_nu(df=df)

        is_neutral_dst: np.ndarray = MCSigQueries.is_neutral_dst_event(df=df)
        is_charged_b: np.ndarray = MCSigQueries.is_charged_b_meson(df=df)

        return is_muon & is_neutral_dst & is_charged_b

    @staticmethod
    def nB_to_cDst_mu_nu_query(df: pd.DataFrame) -> np.ndarray:
        is_muon: np.ndarray = MCSigQueries.is_b_to_x_muon_nu(df=df)

        is_charged_dst: np.ndarray = MCSigQueries.is_charged_dst_event(df=df)
        is_neutral_b: np.ndarray = MCSigQueries.is_neutral_b_meson(df=df)

        return is_muon & is_charged_dst & is_neutral_b

    @staticmethod
    def cB_to_nDst_mu_nu_with_Dst_to_pi_query(df: pd.DataFrame) -> np.ndarray:
        is_muon: np.ndarray = MCSigQueries.is_b_to_x_muon_nu(df=df)

        is_neutral_dst: np.ndarray = MCSigQueries.is_neutral_dst_event(df=df)
        is_dst_to_pi: np.ndarray = MCSigQueries.second_primary_d_meson_daughter_is_any_pion(df=df)
        is_charged_b: np.ndarray = MCSigQueries.is_charged_b_meson(df=df)

        return is_muon & is_neutral_dst & is_dst_to_pi & is_charged_b

    @staticmethod
    def nB_to_cDst_mu_nu_with_Dst_to_pi_query(df: pd.DataFrame) -> np.ndarray:
        is_muon: np.ndarray = MCSigQueries.is_b_to_x_muon_nu(df=df)

        is_charged_dst: np.ndarray = MCSigQueries.is_charged_dst_event(df=df)
        is_dst_to_pi: np.ndarray = MCSigQueries.second_primary_d_meson_daughter_is_any_pion(df=df)
        is_neutral_b: np.ndarray = MCSigQueries.is_neutral_b_meson(df=df)

        return is_muon & is_charged_dst & is_dst_to_pi & is_neutral_b

    @staticmethod
    def cB_to_nDst_mu_nu_with_Dst_to_photon_query(df: pd.DataFrame) -> np.ndarray:
        is_muon: np.ndarray = MCSigQueries.is_b_to_x_muon_nu(df=df)

        is_neutral_dst: np.ndarray = MCSigQueries.is_neutral_dst_event(df=df)
        is_dst_to_photon: np.ndarray = MCSigQueries.second_primary_d_meson_daughter_is_photon(df=df)
        is_charged_b: np.ndarray = MCSigQueries.is_charged_b_meson(df=df)

        return is_muon & is_neutral_dst & is_dst_to_photon & is_charged_b

    @staticmethod
    def nB_to_cDst_mu_nu_with_Dst_to_photon_query(df: pd.DataFrame) -> np.ndarray:
        is_muon: np.ndarray = MCSigQueries.is_b_to_x_muon_nu(df=df)

        is_charged_dst: np.ndarray = MCSigQueries.is_charged_dst_event(df=df)
        is_dst_to_photon: np.ndarray = MCSigQueries.second_primary_d_meson_daughter_is_photon(df=df)
        is_neutral_b: np.ndarray = MCSigQueries.is_neutral_b_meson(df=df)

        return is_muon & is_charged_dst & is_dst_to_photon & is_neutral_b

    # B -> D** lepton
    @staticmethod
    def cB_to_nDstst_via_pi0_e_nu_query(df: pd.DataFrame) -> np.ndarray:
        is_electron: np.ndarray = MCSigQueries.is_b_to_x_electron_nu(df=df)

        is_neutral_dstst_via_pi_zero: np.ndarray = MCSigQueries.is_neutral_dstst_via_pi0_event(df=df)
        is_charged_b: np.ndarray = MCSigQueries.is_charged_b_meson(df=df)

        return is_electron & is_neutral_dstst_via_pi_zero & is_charged_b

    @staticmethod
    def nB_to_cDstst_via_pi0_e_nu_query(df: pd.DataFrame) -> np.ndarray:
        is_electron: np.ndarray = MCSigQueries.is_b_to_x_electron_nu(df=df)

        is_charged_dstst_via_pi_zero: np.ndarray = MCSigQueries.is_charged_dstst_via_pi0_event(df=df)
        is_neutral_b: np.ndarray = MCSigQueries.is_neutral_b_meson(df=df)

        return is_electron & is_charged_dstst_via_pi_zero & is_neutral_b

    @staticmethod
    def cB_to_nDstst_via_pi0_mu_nu_query(df: pd.DataFrame) -> np.ndarray:
        is_muon: np.ndarray = MCSigQueries.is_b_to_x_muon_nu(df=df)

        is_neutral_dstst_via_pi_zero: np.ndarray = MCSigQueries.is_neutral_dstst_via_pi0_event(df=df)
        is_charged_b: np.ndarray = MCSigQueries.is_charged_b_meson(df=df)

        return is_muon & is_neutral_dstst_via_pi_zero & is_charged_b

    @staticmethod
    def nB_to_cDstst_via_pi0_mu_nu_query(df: pd.DataFrame) -> np.ndarray:
        is_muon: np.ndarray = MCSigQueries.is_b_to_x_muon_nu(df=df)

        is_charged_dstst_via_pi_zero: np.ndarray = MCSigQueries.is_charged_dstst_via_pi0_event(df=df)
        is_neutral_b: np.ndarray = MCSigQueries.is_neutral_b_meson(df=df)

        return is_muon & is_charged_dstst_via_pi_zero & is_neutral_b

    @staticmethod
    def cB_to_nDstst_via_pi0_tau_nu_query(df: pd.DataFrame) -> np.ndarray:
        is_tau: np.ndarray = MCSigQueries.is_b_to_x_tau_nu_with_leptonic_tau(df=df)

        is_neutral_dstst_via_pi_zero: np.ndarray = MCSigQueries.is_neutral_dstst_via_pi0_event(df=df)
        is_charged_b: np.ndarray = MCSigQueries.is_charged_b_meson(df=df)

        return is_tau & is_neutral_dstst_via_pi_zero & is_charged_b

    @staticmethod
    def nB_to_cDstst_via_pi0_tau_nu_query(df: pd.DataFrame) -> np.ndarray:
        is_tau: np.ndarray = MCSigQueries.is_b_to_x_tau_nu_with_leptonic_tau(df=df)

        is_charged_dstst_via_pi_zero: np.ndarray = MCSigQueries.is_charged_dstst_via_pi0_event(df=df)
        is_neutral_b: np.ndarray = MCSigQueries.is_neutral_b_meson(df=df)

        return is_tau & is_charged_dstst_via_pi_zero & is_neutral_b

    @staticmethod
    def cB_to_nDstst_via_pip_e_nu_query(df: pd.DataFrame) -> np.ndarray:
        is_electron: np.ndarray = MCSigQueries.is_b_to_x_electron_nu(df=df)

        is_neutral_dstst_via_pi_plus: np.ndarray = MCSigQueries.is_neutral_dstst_via_pi_plus_event(df=df)
        is_charged_b: np.ndarray = MCSigQueries.is_charged_b_meson(df=df)

        return is_electron & is_neutral_dstst_via_pi_plus & is_charged_b

    @staticmethod
    def nB_to_cDstst_via_pip_e_nu_query(df: pd.DataFrame) -> np.ndarray:
        is_electron: np.ndarray = MCSigQueries.is_b_to_x_electron_nu(df=df)

        is_charged_dstst_via_pi_plus: np.ndarray = MCSigQueries.is_charged_dstst_via_pi_plus_event(df=df)
        is_neutral_b: np.ndarray = MCSigQueries.is_neutral_b_meson(df=df)

        return is_electron & is_charged_dstst_via_pi_plus & is_neutral_b

    @staticmethod
    def cB_to_nDstst_via_pip_mu_nu_query(df: pd.DataFrame) -> np.ndarray:
        is_muon: np.ndarray = MCSigQueries.is_b_to_x_muon_nu(df=df)

        is_neutral_dstst_via_pi_plus: np.ndarray = MCSigQueries.is_neutral_dstst_via_pi_plus_event(df=df)
        is_charged_b: np.ndarray = MCSigQueries.is_charged_b_meson(df=df)

        return is_muon & is_neutral_dstst_via_pi_plus & is_charged_b

    @staticmethod
    def nB_to_cDstst_via_pip_mu_nu_query(df: pd.DataFrame) -> np.ndarray:
        is_muon: np.ndarray = MCSigQueries.is_b_to_x_muon_nu(df=df)

        is_charged_dstst_via_pi_plus: np.ndarray = MCSigQueries.is_charged_dstst_via_pi_plus_event(df=df)
        is_neutral_b: np.ndarray = MCSigQueries.is_neutral_b_meson(df=df)

        return is_muon & is_charged_dstst_via_pi_plus & is_neutral_b

    @staticmethod
    def cB_to_nDstst_via_pip_tau_nu_query(df: pd.DataFrame) -> np.ndarray:
        is_tau: np.ndarray = MCSigQueries.is_b_to_x_tau_nu_with_leptonic_tau(df=df)

        is_neutral_dstst_via_pi_plus: np.ndarray = MCSigQueries.is_neutral_dstst_via_pi_plus_event(df=df)
        is_charged_b: np.ndarray = MCSigQueries.is_charged_b_meson(df=df)

        return is_tau & is_neutral_dstst_via_pi_plus & is_charged_b

    @staticmethod
    def nB_to_cDstst_via_pip_tau_nu_query(df: pd.DataFrame) -> np.ndarray:
        is_tau: np.ndarray = MCSigQueries.is_b_to_x_tau_nu_with_leptonic_tau(df=df)

        is_charged_dstst_via_pi_plus: np.ndarray = MCSigQueries.is_charged_dstst_via_pi_plus_event(df=df)
        is_neutral_b: np.ndarray = MCSigQueries.is_neutral_b_meson(df=df)

        return is_tau & is_charged_dstst_via_pi_plus & is_neutral_b

    @staticmethod
    def cB_to_nDone_via_pi0_e_nu_query(df: pd.DataFrame) -> np.ndarray:
        return MCSigQueries.get_dstst_query_with_dstst_requirement(
            df=df,
            dstst="Done",
            other_query=MCSigQueries.cB_to_nDstst_via_pi0_e_nu_query,
        )

    @staticmethod
    def cB_to_nDtwostar_via_pi0_e_nu_query(df: pd.DataFrame) -> np.ndarray:
        return MCSigQueries.get_dstst_query_with_dstst_requirement(
            df=df,
            dstst="Dtwostar",
            other_query=MCSigQueries.cB_to_nDstst_via_pi0_e_nu_query,
        )

    @staticmethod
    def cB_to_nDzerostar_via_pi0_e_nu_query(df: pd.DataFrame) -> np.ndarray:
        return MCSigQueries.get_dstst_query_with_dstst_requirement(
            df=df,
            dstst="Dzerostar",
            other_query=MCSigQueries.cB_to_nDstst_via_pi0_e_nu_query,
        )

    @staticmethod
    def cB_to_nDprimeone_via_pi0_e_nu_query(df: pd.DataFrame) -> np.ndarray:
        return MCSigQueries.get_dstst_query_with_dstst_requirement(
            df=df,
            dstst="Dprimeone",
            other_query=MCSigQueries.cB_to_nDstst_via_pi0_e_nu_query,
        )

    @staticmethod
    def nB_to_cDone_via_pi0_e_nu_query(df: pd.DataFrame) -> np.ndarray:
        return MCSigQueries.get_dstst_query_with_dstst_requirement(
            df=df,
            dstst="Done",
            other_query=MCSigQueries.nB_to_cDstst_via_pi0_e_nu_query,
        )

    @staticmethod
    def nB_to_cDtwostar_via_pi0_e_nu_query(df: pd.DataFrame) -> np.ndarray:
        return MCSigQueries.get_dstst_query_with_dstst_requirement(
            df=df,
            dstst="Dtwostar",
            other_query=MCSigQueries.nB_to_cDstst_via_pi0_e_nu_query,
        )

    @staticmethod
    def nB_to_cDzerostar_via_pi0_e_nu_query(df: pd.DataFrame) -> np.ndarray:
        return MCSigQueries.get_dstst_query_with_dstst_requirement(
            df=df,
            dstst="Dzerostar",
            other_query=MCSigQueries.nB_to_cDstst_via_pi0_e_nu_query,
        )

    @staticmethod
    def nB_to_cDprimeone_via_pi0_e_nu_query(df: pd.DataFrame) -> np.ndarray:
        return MCSigQueries.get_dstst_query_with_dstst_requirement(
            df=df,
            dstst="Dprimeone",
            other_query=MCSigQueries.nB_to_cDstst_via_pi0_e_nu_query,
        )

    @staticmethod
    def cB_to_nDone_via_pi0_mu_nu_query(df: pd.DataFrame) -> np.ndarray:
        return MCSigQueries.get_dstst_query_with_dstst_requirement(
            df=df,
            dstst="Done",
            other_query=MCSigQueries.cB_to_nDstst_via_pi0_mu_nu_query,
        )

    @staticmethod
    def cB_to_nDtwostar_via_pi0_mu_nu_query(df: pd.DataFrame) -> np.ndarray:
        return MCSigQueries.get_dstst_query_with_dstst_requirement(
            df=df,
            dstst="Dtwostar",
            other_query=MCSigQueries.cB_to_nDstst_via_pi0_mu_nu_query,
        )

    @staticmethod
    def cB_to_nDzerostar_via_pi0_mu_nu_query(df: pd.DataFrame) -> np.ndarray:
        return MCSigQueries.get_dstst_query_with_dstst_requirement(
            df=df,
            dstst="Dzerostar",
            other_query=MCSigQueries.cB_to_nDstst_via_pi0_mu_nu_query,
        )

    @staticmethod
    def cB_to_nDprimeone_via_pi0_mu_nu_query(df: pd.DataFrame) -> np.ndarray:
        return MCSigQueries.get_dstst_query_with_dstst_requirement(
            df=df,
            dstst="Dprimeone",
            other_query=MCSigQueries.cB_to_nDstst_via_pi0_mu_nu_query,
        )

    @staticmethod
    def nB_to_cDone_via_pi0_mu_nu_query(df: pd.DataFrame) -> np.ndarray:
        return MCSigQueries.get_dstst_query_with_dstst_requirement(
            df=df,
            dstst="Done",
            other_query=MCSigQueries.nB_to_cDstst_via_pi0_mu_nu_query,
        )

    @staticmethod
    def nB_to_cDtwostar_via_pi0_mu_nu_query(df: pd.DataFrame) -> np.ndarray:
        return MCSigQueries.get_dstst_query_with_dstst_requirement(
            df=df,
            dstst="Dtwostar",
            other_query=MCSigQueries.nB_to_cDstst_via_pi0_mu_nu_query,
        )

    @staticmethod
    def nB_to_cDzerostar_via_pi0_mu_nu_query(df: pd.DataFrame) -> np.ndarray:
        return MCSigQueries.get_dstst_query_with_dstst_requirement(
            df=df,
            dstst="Dzerostar",
            other_query=MCSigQueries.nB_to_cDstst_via_pi0_mu_nu_query,
        )

    @staticmethod
    def nB_to_cDprimeone_via_pi0_mu_nu_query(df: pd.DataFrame) -> np.ndarray:
        return MCSigQueries.get_dstst_query_with_dstst_requirement(
            df=df,
            dstst="Dprimeone",
            other_query=MCSigQueries.nB_to_cDstst_via_pi0_mu_nu_query,
        )

    @staticmethod
    def cB_to_nDone_via_pi0_tau_nu_query(df: pd.DataFrame) -> np.ndarray:
        return MCSigQueries.get_dstst_query_with_dstst_requirement(
            df=df,
            dstst="Done",
            other_query=MCSigQueries.cB_to_nDstst_via_pi0_tau_nu_query,
        )

    @staticmethod
    def cB_to_nDtwostar_via_pi0_tau_nu_query(df: pd.DataFrame) -> np.ndarray:
        return MCSigQueries.get_dstst_query_with_dstst_requirement(
            df=df,
            dstst="Dtwostar",
            other_query=MCSigQueries.cB_to_nDstst_via_pi0_tau_nu_query,
        )

    @staticmethod
    def cB_to_nDzerostar_via_pi0_tau_nu_query(df: pd.DataFrame) -> np.ndarray:
        return MCSigQueries.get_dstst_query_with_dstst_requirement(
            df=df,
            dstst="Dzerostar",
            other_query=MCSigQueries.cB_to_nDstst_via_pi0_tau_nu_query,
        )

    @staticmethod
    def cB_to_nDprimeone_via_pi0_tau_nu_query(df: pd.DataFrame) -> np.ndarray:
        return MCSigQueries.get_dstst_query_with_dstst_requirement(
            df=df,
            dstst="Dprimeone",
            other_query=MCSigQueries.cB_to_nDstst_via_pi0_tau_nu_query,
        )

    @staticmethod
    def nB_to_cDone_via_pi0_tau_nu_query(df: pd.DataFrame) -> np.ndarray:
        return MCSigQueries.get_dstst_query_with_dstst_requirement(
            df=df,
            dstst="Done",
            other_query=MCSigQueries.nB_to_cDstst_via_pi0_tau_nu_query,
        )

    @staticmethod
    def nB_to_cDtwostar_via_pi0_tau_nu_query(df: pd.DataFrame) -> np.ndarray:
        return MCSigQueries.get_dstst_query_with_dstst_requirement(
            df=df,
            dstst="Dtwostar",
            other_query=MCSigQueries.nB_to_cDstst_via_pi0_tau_nu_query,
        )

    @staticmethod
    def nB_to_cDzerostar_via_pi0_tau_nu_query(df: pd.DataFrame) -> np.ndarray:
        return MCSigQueries.get_dstst_query_with_dstst_requirement(
            df=df,
            dstst="Dzerostar",
            other_query=MCSigQueries.nB_to_cDstst_via_pi0_tau_nu_query,
        )

    @staticmethod
    def nB_to_cDprimeone_via_pi0_tau_nu_query(df: pd.DataFrame) -> np.ndarray:
        return MCSigQueries.get_dstst_query_with_dstst_requirement(
            df=df,
            dstst="Dprimeone",
            other_query=MCSigQueries.nB_to_cDstst_via_pi0_tau_nu_query,
        )

    @staticmethod
    def cB_to_nDone_via_pip_e_nu_query(df: pd.DataFrame) -> np.ndarray:
        return MCSigQueries.get_dstst_query_with_dstst_requirement(
            df=df,
            dstst="Done",
            other_query=MCSigQueries.cB_to_nDstst_via_pip_e_nu_query,
        )

    @staticmethod
    def cB_to_nDtwostar_via_pip_e_nu_query(df: pd.DataFrame) -> np.ndarray:
        return MCSigQueries.get_dstst_query_with_dstst_requirement(
            df=df,
            dstst="Dtwostar",
            other_query=MCSigQueries.cB_to_nDstst_via_pip_e_nu_query,
        )

    @staticmethod
    def cB_to_nDzerostar_via_pip_e_nu_query(df: pd.DataFrame) -> np.ndarray:
        return MCSigQueries.get_dstst_query_with_dstst_requirement(
            df=df,
            dstst="Dzerostar",
            other_query=MCSigQueries.cB_to_nDstst_via_pip_e_nu_query,
        )

    @staticmethod
    def cB_to_nDprimeone_via_pip_e_nu_query(df: pd.DataFrame) -> np.ndarray:
        return MCSigQueries.get_dstst_query_with_dstst_requirement(
            df=df,
            dstst="Dprimeone",
            other_query=MCSigQueries.cB_to_nDstst_via_pip_e_nu_query,
        )

    @staticmethod
    def nB_to_cDone_via_pip_e_nu_query(df: pd.DataFrame) -> np.ndarray:
        return MCSigQueries.get_dstst_query_with_dstst_requirement(
            df=df,
            dstst="Done",
            other_query=MCSigQueries.nB_to_cDstst_via_pip_e_nu_query,
        )

    @staticmethod
    def nB_to_cDtwostar_via_pip_e_nu_query(df: pd.DataFrame) -> np.ndarray:
        return MCSigQueries.get_dstst_query_with_dstst_requirement(
            df=df,
            dstst="Dtwostar",
            other_query=MCSigQueries.nB_to_cDstst_via_pip_e_nu_query,
        )

    @staticmethod
    def nB_to_cDzerostar_via_pip_e_nu_query(df: pd.DataFrame) -> np.ndarray:
        return MCSigQueries.get_dstst_query_with_dstst_requirement(
            df=df,
            dstst="Dzerostar",
            other_query=MCSigQueries.nB_to_cDstst_via_pip_e_nu_query,
        )

    @staticmethod
    def nB_to_cDprimeone_via_pip_e_nu_query(df: pd.DataFrame) -> np.ndarray:
        return MCSigQueries.get_dstst_query_with_dstst_requirement(
            df=df,
            dstst="Dprimeone",
            other_query=MCSigQueries.nB_to_cDstst_via_pip_e_nu_query,
        )

    @staticmethod
    def cB_to_nDone_via_pip_mu_nu_query(df: pd.DataFrame) -> np.ndarray:
        return MCSigQueries.get_dstst_query_with_dstst_requirement(
            df=df,
            dstst="Done",
            other_query=MCSigQueries.cB_to_nDstst_via_pip_mu_nu_query,
        )

    @staticmethod
    def cB_to_nDtwostar_via_pip_mu_nu_query(df: pd.DataFrame) -> np.ndarray:
        return MCSigQueries.get_dstst_query_with_dstst_requirement(
            df=df,
            dstst="Dtwostar",
            other_query=MCSigQueries.cB_to_nDstst_via_pip_mu_nu_query,
        )

    @staticmethod
    def cB_to_nDzerostar_via_pip_mu_nu_query(df: pd.DataFrame) -> np.ndarray:
        return MCSigQueries.get_dstst_query_with_dstst_requirement(
            df=df,
            dstst="Dzerostar",
            other_query=MCSigQueries.cB_to_nDstst_via_pip_mu_nu_query,
        )

    @staticmethod
    def cB_to_nDprimeone_via_pip_mu_nu_query(df: pd.DataFrame) -> np.ndarray:
        return MCSigQueries.get_dstst_query_with_dstst_requirement(
            df=df,
            dstst="Dprimeone",
            other_query=MCSigQueries.cB_to_nDstst_via_pip_mu_nu_query,
        )

    @staticmethod
    def nB_to_cDone_via_pip_mu_nu_query(df: pd.DataFrame) -> np.ndarray:
        return MCSigQueries.get_dstst_query_with_dstst_requirement(
            df=df,
            dstst="Done",
            other_query=MCSigQueries.nB_to_cDstst_via_pip_mu_nu_query,
        )

    @staticmethod
    def nB_to_cDtwostar_via_pip_mu_nu_query(df: pd.DataFrame) -> np.ndarray:
        return MCSigQueries.get_dstst_query_with_dstst_requirement(
            df=df,
            dstst="Dtwostar",
            other_query=MCSigQueries.nB_to_cDstst_via_pip_mu_nu_query,
        )

    @staticmethod
    def nB_to_cDzerostar_via_pip_mu_nu_query(df: pd.DataFrame) -> np.ndarray:
        return MCSigQueries.get_dstst_query_with_dstst_requirement(
            df=df,
            dstst="Dzerostar",
            other_query=MCSigQueries.nB_to_cDstst_via_pip_mu_nu_query,
        )

    @staticmethod
    def nB_to_cDprimeone_via_pip_mu_nu_query(df: pd.DataFrame) -> np.ndarray:
        return MCSigQueries.get_dstst_query_with_dstst_requirement(
            df=df,
            dstst="Dprimeone",
            other_query=MCSigQueries.nB_to_cDstst_via_pip_mu_nu_query,
        )

    @staticmethod
    def cB_to_nDone_via_pip_tau_nu_query(df: pd.DataFrame) -> np.ndarray:
        return MCSigQueries.get_dstst_query_with_dstst_requirement(
            df=df,
            dstst="Done",
            other_query=MCSigQueries.cB_to_nDstst_via_pip_tau_nu_query,
        )

    @staticmethod
    def cB_to_nDtwostar_via_pip_tau_nu_query(df: pd.DataFrame) -> np.ndarray:
        return MCSigQueries.get_dstst_query_with_dstst_requirement(
            df=df,
            dstst="Dtwostar",
            other_query=MCSigQueries.cB_to_nDstst_via_pip_tau_nu_query,
        )

    @staticmethod
    def cB_to_nDzerostar_via_pip_tau_nu_query(df: pd.DataFrame) -> np.ndarray:
        return MCSigQueries.get_dstst_query_with_dstst_requirement(
            df=df,
            dstst="Dzerostar",
            other_query=MCSigQueries.cB_to_nDstst_via_pip_tau_nu_query,
        )

    @staticmethod
    def cB_to_nDprimeone_via_pip_tau_nu_query(df: pd.DataFrame) -> np.ndarray:
        return MCSigQueries.get_dstst_query_with_dstst_requirement(
            df=df,
            dstst="Dprimeone",
            other_query=MCSigQueries.cB_to_nDstst_via_pip_tau_nu_query,
        )

    @staticmethod
    def nB_to_cDone_via_pip_tau_nu_query(df: pd.DataFrame) -> np.ndarray:
        return MCSigQueries.get_dstst_query_with_dstst_requirement(
            df=df,
            dstst="Done",
            other_query=MCSigQueries.nB_to_cDstst_via_pip_tau_nu_query,
        )

    @staticmethod
    def nB_to_cDtwostar_via_pip_tau_nu_query(df: pd.DataFrame) -> np.ndarray:
        return MCSigQueries.get_dstst_query_with_dstst_requirement(
            df=df,
            dstst="Dtwostar",
            other_query=MCSigQueries.nB_to_cDstst_via_pip_tau_nu_query,
        )

    @staticmethod
    def nB_to_cDzerostar_via_pip_tau_nu_query(df: pd.DataFrame) -> np.ndarray:
        return MCSigQueries.get_dstst_query_with_dstst_requirement(
            df=df,
            dstst="Dzerostar",
            other_query=MCSigQueries.nB_to_cDstst_via_pip_tau_nu_query,
        )

    @staticmethod
    def nB_to_cDprimeone_via_pip_tau_nu_query(df: pd.DataFrame) -> np.ndarray:
        return MCSigQueries.get_dstst_query_with_dstst_requirement(
            df=df,
            dstst="Dprimeone",
            other_query=MCSigQueries.nB_to_cDstst_via_pip_tau_nu_query,
        )

    # B -> D** Hadronic Tau
    @staticmethod
    def cB_to_nDstst_via_pi0_HADRONIC_tau_nu_query(df: pd.DataFrame) -> np.ndarray:
        is_hadronic_tau: np.ndarray = MCSigQueries.is_b_to_x_tau_nu_with_hadronic_tau(df=df)

        is_neutral_dstst_via_pi_zero: np.ndarray = MCSigQueries.is_neutral_dstst_via_pi0_event(df=df)
        is_charged_b: np.ndarray = MCSigQueries.is_charged_b_meson(df=df)

        return is_hadronic_tau & is_neutral_dstst_via_pi_zero & is_charged_b

    @staticmethod
    def nB_to_cDstst_via_pi0_HADRONIC_tau_nu_query(df: pd.DataFrame) -> np.ndarray:
        is_hadronic_tau: np.ndarray = MCSigQueries.is_b_to_x_tau_nu_with_hadronic_tau(df=df)

        is_charged_dstst_via_pi_zero: np.ndarray = MCSigQueries.is_charged_dstst_via_pi0_event(df=df)
        is_neutral_b: np.ndarray = MCSigQueries.is_neutral_b_meson(df=df)

        return is_hadronic_tau & is_charged_dstst_via_pi_zero & is_neutral_b

    @staticmethod
    def cB_to_nDstst_via_pip_HADRONIC_tau_nu_query(df: pd.DataFrame) -> np.ndarray:
        is_hadronic_tau: np.ndarray = MCSigQueries.is_b_to_x_tau_nu_with_hadronic_tau(df=df)

        is_neutral_dstst_via_pi_plus: np.ndarray = MCSigQueries.is_neutral_dstst_via_pi_plus_event(df=df)
        is_charged_b: np.ndarray = MCSigQueries.is_charged_b_meson(df=df)

        return is_hadronic_tau & is_neutral_dstst_via_pi_plus & is_charged_b

    @staticmethod
    def nB_to_cDstst_via_pip_HADRONIC_tau_nu_query(df: pd.DataFrame) -> np.ndarray:
        is_hadronic_tau: np.ndarray = MCSigQueries.is_b_to_x_tau_nu_with_hadronic_tau(df=df)

        is_charged_dstst_via_pi_plus: np.ndarray = MCSigQueries.is_charged_dstst_via_pi_plus_event(df=df)
        is_neutral_b: np.ndarray = MCSigQueries.is_neutral_b_meson(df=df)

        return is_hadronic_tau & is_charged_dstst_via_pi_plus & is_neutral_b

    @staticmethod
    def cB_to_nDone_via_pi0_HADRONIC_tau_nu_query(df: pd.DataFrame) -> np.ndarray:
        return MCSigQueries.get_dstst_query_with_dstst_requirement(
            df=df,
            dstst="Done",
            other_query=MCSigQueries.cB_to_nDstst_via_pi0_HADRONIC_tau_nu_query,
        )

    @staticmethod
    def cB_to_nDtwostar_via_pi0_HADRONIC_tau_nu_query(df: pd.DataFrame) -> np.ndarray:
        return MCSigQueries.get_dstst_query_with_dstst_requirement(
            df=df,
            dstst="Dtwostar",
            other_query=MCSigQueries.cB_to_nDstst_via_pi0_HADRONIC_tau_nu_query,
        )

    @staticmethod
    def cB_to_nDzerostar_via_pi0_HADRONIC_tau_nu_query(df: pd.DataFrame) -> np.ndarray:
        return MCSigQueries.get_dstst_query_with_dstst_requirement(
            df=df,
            dstst="Dzerostar",
            other_query=MCSigQueries.cB_to_nDstst_via_pi0_HADRONIC_tau_nu_query,
        )

    @staticmethod
    def cB_to_nDprimeone_via_pi0_HADRONIC_tau_nu_query(df: pd.DataFrame) -> np.ndarray:
        return MCSigQueries.get_dstst_query_with_dstst_requirement(
            df=df,
            dstst="Dprimeone",
            other_query=MCSigQueries.cB_to_nDstst_via_pi0_HADRONIC_tau_nu_query,
        )

    @staticmethod
    def nB_to_cDone_via_pi0_HADRONIC_tau_nu_query(df: pd.DataFrame) -> np.ndarray:
        return MCSigQueries.get_dstst_query_with_dstst_requirement(
            df=df,
            dstst="Done",
            other_query=MCSigQueries.nB_to_cDstst_via_pi0_HADRONIC_tau_nu_query,
        )

    @staticmethod
    def nB_to_cDtwostar_via_pi0_HADRONIC_tau_nu_query(df: pd.DataFrame) -> np.ndarray:
        return MCSigQueries.get_dstst_query_with_dstst_requirement(
            df=df,
            dstst="Dtwostar",
            other_query=MCSigQueries.nB_to_cDstst_via_pi0_HADRONIC_tau_nu_query,
        )

    @staticmethod
    def nB_to_cDzerostar_via_pi0_HADRONIC_tau_nu_query(df: pd.DataFrame) -> np.ndarray:
        return MCSigQueries.get_dstst_query_with_dstst_requirement(
            df=df,
            dstst="Dzerostar",
            other_query=MCSigQueries.nB_to_cDstst_via_pi0_HADRONIC_tau_nu_query,
        )

    @staticmethod
    def nB_to_cDprimeone_via_pi0_HADRONIC_tau_nu_query(df: pd.DataFrame) -> np.ndarray:
        return MCSigQueries.get_dstst_query_with_dstst_requirement(
            df=df,
            dstst="Dprimeone",
            other_query=MCSigQueries.nB_to_cDstst_via_pi0_HADRONIC_tau_nu_query,
        )

    @staticmethod
    def cB_to_nDone_via_pip_HADRONIC_tau_nu_query(df: pd.DataFrame) -> np.ndarray:
        return MCSigQueries.get_dstst_query_with_dstst_requirement(
            df=df,
            dstst="Done",
            other_query=MCSigQueries.cB_to_nDstst_via_pip_HADRONIC_tau_nu_query,
        )

    @staticmethod
    def cB_to_nDtwostar_via_pip_HADRONIC_tau_nu_query(df: pd.DataFrame) -> np.ndarray:
        return MCSigQueries.get_dstst_query_with_dstst_requirement(
            df=df,
            dstst="Dtwostar",
            other_query=MCSigQueries.cB_to_nDstst_via_pip_HADRONIC_tau_nu_query,
        )

    @staticmethod
    def cB_to_nDzerostar_via_pip_HADRONIC_tau_nu_query(df: pd.DataFrame) -> np.ndarray:
        return MCSigQueries.get_dstst_query_with_dstst_requirement(
            df=df,
            dstst="Dzerostar",
            other_query=MCSigQueries.cB_to_nDstst_via_pip_HADRONIC_tau_nu_query,
        )

    @staticmethod
    def cB_to_nDprimeone_via_pip_HADRONIC_tau_nu_query(df: pd.DataFrame) -> np.ndarray:
        return MCSigQueries.get_dstst_query_with_dstst_requirement(
            df=df,
            dstst="Dprimeone",
            other_query=MCSigQueries.cB_to_nDstst_via_pip_HADRONIC_tau_nu_query,
        )

    @staticmethod
    def nB_to_cDone_via_pip_HADRONIC_tau_nu_query(df: pd.DataFrame) -> np.ndarray:
        return MCSigQueries.get_dstst_query_with_dstst_requirement(
            df=df,
            dstst="Done",
            other_query=MCSigQueries.nB_to_cDstst_via_pip_HADRONIC_tau_nu_query,
        )

    @staticmethod
    def nB_to_cDtwostar_via_pip_HADRONIC_tau_nu_query(df: pd.DataFrame) -> np.ndarray:
        return MCSigQueries.get_dstst_query_with_dstst_requirement(
            df=df,
            dstst="Dtwostar",
            other_query=MCSigQueries.nB_to_cDstst_via_pip_HADRONIC_tau_nu_query,
        )

    @staticmethod
    def nB_to_cDzerostar_via_pip_HADRONIC_tau_nu_query(df: pd.DataFrame) -> np.ndarray:
        return MCSigQueries.get_dstst_query_with_dstst_requirement(
            df=df,
            dstst="Dzerostar",
            other_query=MCSigQueries.nB_to_cDstst_via_pip_HADRONIC_tau_nu_query,
        )

    @staticmethod
    def nB_to_cDprimeone_via_pip_HADRONIC_tau_nu_query(df: pd.DataFrame) -> np.ndarray:
        return MCSigQueries.get_dstst_query_with_dstst_requirement(
            df=df,
            dstst="Dprimeone",
            other_query=MCSigQueries.nB_to_cDstst_via_pip_HADRONIC_tau_nu_query,
        )

    # Lepton Cross Feed Backgrounds
    @staticmethod
    def lepton_tau_e_to_mu_crossfeed_query(df: pd.DataFrame) -> np.ndarray:
        is_b_to_x_l_nu_event: np.ndarray = MCSigQueries.is_b_to_x_any_lepton_nu(df=df)
        is_any_d_meson_event: np.ndarray = MCSigQueries.is_any_d_event(df=df)
        is_corss_feed: np.ndarray = MCSigQueries.is_lepton_cross_feed_event(
            df=df,
            reco_pdgs=PcC.electron_pdg_id,
            truth_pdgs=PcC.tau_pdg_id,
            truth_daughter_pdgs=PcC.muon_pdg_id,
        )
        return is_b_to_x_l_nu_event & is_any_d_meson_event & is_corss_feed

    @staticmethod
    def lepton_tau_mu_to_e_crossfeed_query(df: pd.DataFrame) -> np.ndarray:
        is_b_to_x_l_nu_event: np.ndarray = MCSigQueries.is_b_to_x_any_lepton_nu(df=df)
        is_any_d_meson_event: np.ndarray = MCSigQueries.is_any_d_event(df=df)
        is_corss_feed: np.ndarray = MCSigQueries.is_lepton_cross_feed_event(
            df=df,
            reco_pdgs=PcC.muon_pdg_id,
            truth_pdgs=PcC.tau_pdg_id,
            truth_daughter_pdgs=PcC.electron_pdg_id,
        )
        return is_b_to_x_l_nu_event & is_any_d_meson_event & is_corss_feed

    @staticmethod
    def lepton_e_to_mu_crossfeed_query(df: pd.DataFrame) -> np.ndarray:
        is_b_to_x_l_nu_event: np.ndarray = MCSigQueries.is_b_to_x_any_lepton_nu(df=df)
        is_any_d_meson_event: np.ndarray = MCSigQueries.is_any_d_event(df=df)
        is_corss_feed: np.ndarray = MCSigQueries.is_lepton_cross_feed_event(
            df=df,
            reco_pdgs=PcC.electron_pdg_id,
            truth_pdgs=PcC.muon_pdg_id,
            truth_daughter_pdgs=None,
        )
        return is_b_to_x_l_nu_event & is_any_d_meson_event & is_corss_feed

    @staticmethod
    def lepton_mu_to_e_crossfeed_query(df: pd.DataFrame) -> np.ndarray:
        is_b_to_x_l_nu_event: np.ndarray = MCSigQueries.is_b_to_x_any_lepton_nu(df=df)
        is_any_d_meson_event: np.ndarray = MCSigQueries.is_any_d_event(df=df)
        is_corss_feed: np.ndarray = MCSigQueries.is_lepton_cross_feed_event(
            df=df,
            reco_pdgs=PcC.muon_pdg_id,
            truth_pdgs=PcC.electron_pdg_id,
            truth_daughter_pdgs=None,
        )
        return is_b_to_x_l_nu_event & is_any_d_meson_event & is_corss_feed

    @staticmethod
    def is_any_lepton_crossfeed_event(df: pd.DataFrame) -> np.ndarray:
        is_b_to_x_l_nu_event: np.ndarray = MCSigQueries.is_b_to_x_any_lepton_nu(df=df)
        is_any_d_meson_event: np.ndarray = MCSigQueries.is_any_d_event(df=df)

        is_tem_cross_feed: np.ndarray = MCSigQueries.is_lepton_cross_feed_event(
            df=df,
            reco_pdgs=PcC.electron_pdg_id,
            truth_pdgs=PcC.tau_pdg_id,
            truth_daughter_pdgs=PcC.muon_pdg_id,
        )
        is_tme_cross_feed: np.ndarray = MCSigQueries.is_lepton_cross_feed_event(
            df=df,
            reco_pdgs=PcC.muon_pdg_id,
            truth_pdgs=PcC.tau_pdg_id,
            truth_daughter_pdgs=PcC.electron_pdg_id,
        )
        is_em_cross_feed: np.ndarray = MCSigQueries.is_lepton_cross_feed_event(
            df=df,
            reco_pdgs=PcC.electron_pdg_id,
            truth_pdgs=PcC.muon_pdg_id,
            truth_daughter_pdgs=None,
        )
        is_me_cross_feed: np.ndarray = MCSigQueries.is_lepton_cross_feed_event(
            df=df,
            reco_pdgs=PcC.muon_pdg_id,
            truth_pdgs=PcC.electron_pdg_id,
            truth_daughter_pdgs=None,
        )
        is_any_cross_feed: np.ndarray = is_tem_cross_feed | is_tme_cross_feed | is_em_cross_feed | is_me_cross_feed

        return is_b_to_x_l_nu_event & is_any_d_meson_event & is_any_cross_feed

    # D** Daughter Mismatch Backgrounds
    @staticmethod
    def dss_daughter_mismatch_nb_cdss_e_query(df: pd.DataFrame) -> np.ndarray:
        is_b_to_x_e_nu_event: np.ndarray = MCSigQueries.is_b_to_x_electron_nu(df=df)
        is_dstst_daughter_mismatch: np.ndarray = MCSigQueries.is_dstst_with_wrong_daughters_event(df=df)

        is_charged_dstst: np.ndarray = MCSigQueries.is_charged_dstst_event(df=df)
        is_neutral_b: np.ndarray = MCSigQueries.is_neutral_b_meson(df=df)

        return is_b_to_x_e_nu_event & is_dstst_daughter_mismatch & is_charged_dstst & is_neutral_b

    @staticmethod
    def dss_daughter_mismatch_nb_cdss_mu_query(df: pd.DataFrame) -> np.ndarray:
        is_b_to_x_mu_nu_event: np.ndarray = MCSigQueries.is_b_to_x_muon_nu(df=df)
        is_dstst_daughter_mismatch: np.ndarray = MCSigQueries.is_dstst_with_wrong_daughters_event(df=df)

        is_charged_dstst: np.ndarray = MCSigQueries.is_charged_dstst_event(df=df)
        is_neutral_b: np.ndarray = MCSigQueries.is_neutral_b_meson(df=df)

        return is_b_to_x_mu_nu_event & is_dstst_daughter_mismatch & is_charged_dstst & is_neutral_b

    @staticmethod
    def dss_daughter_mismatch_nb_cdss_tau_query(df: pd.DataFrame) -> np.ndarray:
        is_b_to_x_tau_nu_event: np.ndarray = MCSigQueries.is_b_to_x_tau_nu_with_any_tau(df=df)
        is_dstst_daughter_mismatch: np.ndarray = MCSigQueries.is_dstst_with_wrong_daughters_event(df=df)

        is_charged_dstst: np.ndarray = MCSigQueries.is_charged_dstst_event(df=df)
        is_neutral_b: np.ndarray = MCSigQueries.is_neutral_b_meson(df=df)

        return is_b_to_x_tau_nu_event & is_dstst_daughter_mismatch & is_charged_dstst & is_neutral_b

    @staticmethod
    def dss_daughter_mismatch_cb_cdss_e_query(df: pd.DataFrame) -> np.ndarray:
        is_b_to_x_e_nu_event: np.ndarray = MCSigQueries.is_b_to_x_electron_nu(df=df)
        is_dstst_daughter_mismatch: np.ndarray = MCSigQueries.is_dstst_with_wrong_daughters_event(df=df)

        is_charged_dstst: np.ndarray = MCSigQueries.is_charged_dstst_event(df=df)
        is_charged_b: np.ndarray = MCSigQueries.is_charged_b_meson(df=df)

        return is_b_to_x_e_nu_event & is_dstst_daughter_mismatch & is_charged_dstst & is_charged_b

    @staticmethod
    def dss_daughter_mismatch_cb_cdss_mu_query(df: pd.DataFrame) -> np.ndarray:
        is_b_to_x_mu_nu_event: np.ndarray = MCSigQueries.is_b_to_x_muon_nu(df=df)
        is_dstst_daughter_mismatch: np.ndarray = MCSigQueries.is_dstst_with_wrong_daughters_event(df=df)

        is_charged_dstst: np.ndarray = MCSigQueries.is_charged_dstst_event(df=df)
        is_charged_b: np.ndarray = MCSigQueries.is_charged_b_meson(df=df)

        return is_b_to_x_mu_nu_event & is_dstst_daughter_mismatch & is_charged_dstst & is_charged_b

    @staticmethod
    def dss_daughter_mismatch_cb_cdss_tau_query(df: pd.DataFrame) -> np.ndarray:
        is_b_to_x_tau_nu_event: np.ndarray = MCSigQueries.is_b_to_x_tau_nu_with_any_tau(df=df)
        is_dstst_daughter_mismatch: np.ndarray = MCSigQueries.is_dstst_with_wrong_daughters_event(df=df)

        is_charged_dstst: np.ndarray = MCSigQueries.is_charged_dstst_event(df=df)
        is_charged_b: np.ndarray = MCSigQueries.is_charged_b_meson(df=df)

        return is_b_to_x_tau_nu_event & is_dstst_daughter_mismatch & is_charged_dstst & is_charged_b

    @staticmethod
    def dss_daughter_mismatch_nb_ndss_e_query(df: pd.DataFrame) -> np.ndarray:
        is_b_to_x_e_nu_event: np.ndarray = MCSigQueries.is_b_to_x_electron_nu(df=df)
        is_dstst_daughter_mismatch: np.ndarray = MCSigQueries.is_dstst_with_wrong_daughters_event(df=df)

        is_neutral_dstst: np.ndarray = MCSigQueries.is_neutral_dstst_event(df=df)
        is_neutral_b: np.ndarray = MCSigQueries.is_neutral_b_meson(df=df)

        return is_b_to_x_e_nu_event & is_dstst_daughter_mismatch & is_neutral_dstst & is_neutral_b

    @staticmethod
    def dss_daughter_mismatch_nb_ndss_mu_query(df: pd.DataFrame) -> np.ndarray:
        is_b_to_x_mu_nu_event: np.ndarray = MCSigQueries.is_b_to_x_muon_nu(df=df)
        is_dstst_daughter_mismatch: np.ndarray = MCSigQueries.is_dstst_with_wrong_daughters_event(df=df)

        is_neutral_dstst: np.ndarray = MCSigQueries.is_neutral_dstst_event(df=df)
        is_neutral_b: np.ndarray = MCSigQueries.is_neutral_b_meson(df=df)

        return is_b_to_x_mu_nu_event & is_dstst_daughter_mismatch & is_neutral_dstst & is_neutral_b

    @staticmethod
    def dss_daughter_mismatch_nb_ndss_tau_query(df: pd.DataFrame) -> np.ndarray:
        is_b_to_x_tau_nu_event: np.ndarray = MCSigQueries.is_b_to_x_tau_nu_with_any_tau(df=df)
        is_dstst_daughter_mismatch: np.ndarray = MCSigQueries.is_dstst_with_wrong_daughters_event(df=df)

        is_neutral_dstst: np.ndarray = MCSigQueries.is_neutral_dstst_event(df=df)
        is_neutral_b: np.ndarray = MCSigQueries.is_neutral_b_meson(df=df)

        return is_b_to_x_tau_nu_event & is_dstst_daughter_mismatch & is_neutral_dstst & is_neutral_b

    @staticmethod
    def dss_daughter_mismatch_cb_ndss_e_query(df: pd.DataFrame) -> np.ndarray:
        is_b_to_x_e_nu_event: np.ndarray = MCSigQueries.is_b_to_x_electron_nu(df=df)
        is_dstst_daughter_mismatch: np.ndarray = MCSigQueries.is_dstst_with_wrong_daughters_event(df=df)

        is_neutral_dstst: np.ndarray = MCSigQueries.is_neutral_dstst_event(df=df)
        is_charged_b: np.ndarray = MCSigQueries.is_charged_b_meson(df=df)

        return is_b_to_x_e_nu_event & is_dstst_daughter_mismatch & is_neutral_dstst & is_charged_b

    @staticmethod
    def dss_daughter_mismatch_cb_ndss_mu_query(df: pd.DataFrame) -> np.ndarray:
        is_b_to_x_mu_nu_event: np.ndarray = MCSigQueries.is_b_to_x_muon_nu(df=df)
        is_dstst_daughter_mismatch: np.ndarray = MCSigQueries.is_dstst_with_wrong_daughters_event(df=df)

        is_neutral_dstst: np.ndarray = MCSigQueries.is_neutral_dstst_event(df=df)
        is_charged_b: np.ndarray = MCSigQueries.is_charged_b_meson(df=df)

        return is_b_to_x_mu_nu_event & is_dstst_daughter_mismatch & is_neutral_dstst & is_charged_b

    @staticmethod
    def dss_daughter_mismatch_cb_ndss_tau_query(df: pd.DataFrame) -> np.ndarray:
        is_b_to_x_tau_nu_event: np.ndarray = MCSigQueries.is_b_to_x_tau_nu_with_any_tau(df=df)
        is_dstst_daughter_mismatch: np.ndarray = MCSigQueries.is_dstst_with_wrong_daughters_event(df=df)

        is_neutral_dstst: np.ndarray = MCSigQueries.is_neutral_dstst_event(df=df)
        is_charged_b: np.ndarray = MCSigQueries.is_charged_b_meson(df=df)

        return is_b_to_x_tau_nu_event & is_dstst_daughter_mismatch & is_neutral_dstst & is_charged_b

    @staticmethod
    def is_any_dss_daughter_mismatch_event(df: pd.DataFrame) -> np.ndarray:
        is_b_to_x_l_nu_event: np.ndarray = MCSigQueries.is_b_to_x_any_lepton_nu_including_hadronic_tau(df=df)
        is_dstst_daughter_mismatch: np.ndarray = MCSigQueries.is_dstst_with_wrong_daughters_event(df=df)
        return is_b_to_x_l_nu_event & is_dstst_daughter_mismatch

    # Hadronic Lepton Backgrounds
    @staticmethod
    def lepton_fake_h_bkg_query(df: pd.DataFrame) -> np.ndarray:
        return MCSigQueries.is_lepton_fake(df=df)

    @staticmethod
    def lepton_reco_fake_h_bkg_query(df: pd.DataFrame) -> np.ndarray:
        return MCSigQueries.is_lepton_reco_fake(df=df)

    # Other Backgrounds

    @staticmethod
    def lepton_but_not_from_b_g_bkg_query(df: pd.DataFrame) -> np.ndarray:
        return MCSigQueries.is_lepton_but_not_from_b_meson(df=df)

    @staticmethod
    def b_to_d_xx_l_xx_g_bkg_query(df: pd.DataFrame) -> np.ndarray:
        is_lepton_truth_fake_event: np.ndarray = MCSigQueries.is_lepton_fake(df=df)
        is_lepton_reco_fake_event: np.ndarray = MCSigQueries.is_lepton_reco_fake(df=df)
        is_not_lep_fake: np.ndarray = np.logical_not(is_lepton_truth_fake_event | is_lepton_reco_fake_event)

        is_lepton_with_b_relation: np.ndarray = MCSigQueries.is_event_with_lepton_to_b_meson_relation(df=df)

        is_not_b_to_x_l_nu_event: np.ndarray = ~MCSigQueries.is_b_to_x_any_lepton_nu_including_hadronic_tau(df=df)
        is_any_d_event: np.ndarray = MCSigQueries.is_any_d_event(df=df)

        return is_not_lep_fake & is_lepton_with_b_relation & is_not_b_to_x_l_nu_event & is_any_d_event

    @staticmethod
    def lepton_from_other_b_decay_g_bkg_query(df: pd.DataFrame) -> np.ndarray:
        is_not_b_to_x_any_lep_nu_event: np.ndarray = ~MCSigQueries.is_b_to_x_any_lepton_nu_including_hadronic_tau(df=df)

        is_lepton_truth_fake_event: np.ndarray = MCSigQueries.is_lepton_fake(df=df)
        is_lepton_reco_fake_event: np.ndarray = MCSigQueries.is_lepton_reco_fake(df=df)
        is_not_lep_fake: np.ndarray = np.logical_not(is_lepton_truth_fake_event | is_lepton_reco_fake_event)

        is_not_b_to_d_xx_l_nu_g_bkg_event = np.logical_not(MCSigQueries.b_to_d_xx_l_xx_g_bkg_query(df=df))

        is_lepton_from_b: np.ndarray = MCSigQueries.is_event_with_lepton_to_b_meson_relation(df=df)

        return is_not_b_to_x_any_lep_nu_event & is_not_lep_fake & is_not_b_to_d_xx_l_nu_g_bkg_event & is_lepton_from_b

    @staticmethod
    def d_meson_fake_g_bkg_query(df: pd.DataFrame) -> np.ndarray:
        is_no_d_meson_match_event: np.ndarray = MCSigQueries.is_lepton_but_no_related_d_meson(df=df)
        is_not_pi_l_nu_event: np.ndarray = np.logical_not(MCSigQueries.is_b_to_pi_l_nu_event(df=df))
        return is_no_d_meson_match_event & is_not_pi_l_nu_event

    @staticmethod
    def dstst_to_three_g_bkg_query(df: pd.DataFrame) -> np.ndarray:
        return MCSigQueries.is_dstst_meson_l_nu_with_dstst_to_three_event(df=df)

    @staticmethod
    def GBKG_query(df: pd.DataFrame) -> np.ndarray:
        # The following covers all B -> D(*(*)) l nu events and B -> pi l nu events, as well as D meson fake events
        is_b_to_x_any_lepton_nu_event: np.ndarray = MCSigQueries.is_b_to_x_any_lepton_nu_including_hadronic_tau(df=df)

        # The following covers lepton fakes
        is_lepton_truth_fake_event: np.ndarray = MCSigQueries.is_lepton_fake(df=df)
        is_lepton_reco_fake_event: np.ndarray = MCSigQueries.is_lepton_reco_fake(df=df)
        is_lepton_fake_event: np.ndarray = is_lepton_truth_fake_event | is_lepton_reco_fake_event

        # The following covers D** daughter mismatch and lepton cross-feed events
        is_lepton_crossfeed: np.ndarray = MCSigQueries.is_any_lepton_crossfeed_event(df=df)
        is_dss_daughter_mismatch: np.ndarray = MCSigQueries.is_any_dss_daughter_mismatch_event(df=df)
        is_any_mismatch_event: np.ndarray = is_lepton_crossfeed | is_dss_daughter_mismatch

        # The following covers lepton that do not originate directly from B decays and also not from continuum
        is_lepton_no_from_b_evt: np.ndarray = MCSigQueries.lepton_but_not_from_b_g_bkg_query(df=df)
        # The following covers B to any D xx l nu events
        is_b_to_d_xx_l_xx_evt: np.ndarray = MCSigQueries.b_to_d_xx_l_xx_g_bkg_query(df=df)
        # The following covers leptons from other B decays
        is_l_from_other_b_decays: np.ndarray = MCSigQueries.lepton_from_other_b_decay_g_bkg_query(df=df)

        is_other_lepton_event: np.ndarray = is_lepton_no_from_b_evt | is_b_to_d_xx_l_xx_evt | is_l_from_other_b_decays

        # The following covers continuum
        is_continuum_event: np.ndarray = MCSigQueries.general_CBKG_query(df=df)

        return ~(
            is_b_to_x_any_lepton_nu_event
            | is_lepton_fake_event
            | is_any_mismatch_event
            | is_other_lepton_event
            | is_continuum_event
        )

    @staticmethod
    def charm_CBKG_query(df: pd.DataFrame) -> np.ndarray:
        is_continuum: np.ndarray = MCSigQueries.is_continuum_event(df=df)
        is_charm: np.ndarray = df[Cols.data_type] == "Charm"
        is_off_res_charm: np.ndarray = df[Cols.data_type] == "OffResonanceCharm"
        return is_continuum & (is_charm | is_off_res_charm)

    @staticmethod
    def uds_CBKG_query(df: pd.DataFrame) -> np.ndarray:
        is_continuum: np.ndarray = MCSigQueries.is_continuum_event(df=df)
        is_uds: np.ndarray = df[Cols.data_type] == "UDS"
        is_off_res_uds: np.ndarray = df[Cols.data_type] == "OffResonanceUDS"
        return is_continuum & (is_uds | is_off_res_uds)

    @staticmethod
    def general_CBKG_query(df: pd.DataFrame) -> np.ndarray:
        is_continuum: np.ndarray = MCSigQueries.is_continuum_event(df=df)
        return is_continuum

    @staticmethod
    def PiLNu_query(df: pd.DataFrame) -> np.ndarray:
        return MCSigQueries.is_b_to_pi_l_nu_event(df=df)

    @staticmethod
    def neutral_pi_light_lepton_nu_query(df: pd.DataFrame) -> np.ndarray:
        return MCSigQueries.is_b_to_neutral_pi_l_nu_event(df=df)

    @staticmethod
    def charged_pi_light_lepton_nu_query(df: pd.DataFrame) -> np.ndarray:
        return MCSigQueries.is_b_to_charged_pi_l_nu_event(df=df)

    @staticmethod
    def neutral_pi_tau_nu_query(df: pd.DataFrame) -> np.ndarray:
        return MCSigQueries.is_b_to_neutral_pi_tau_nu_event(df=df)

    @staticmethod
    def charged_pi_tau_nu_query(df: pd.DataFrame) -> np.ndarray:
        return MCSigQueries.is_b_to_charged_pi_tau_nu_event(df=df)
