"""
Setups for fit of a possible shift in the squared missing mass distributions.

Felix Metzner 2021
"""

from typing import Optional, Tuple, Dict, Any, NamedTuple

from templatefitter.plotter.histogram_variable import HistVariable

from rdstar.latex_labels.latex_units import LatexUnits
from rdstar.latex_labels.latex_observable_names import LatexObsNames
from rdstar.offline_analysis.observables.common_observables import Cols
from rdstar.offline_analysis.weight_management.weight_information import WeightManager
from rdstar.offline_analysis.mms_resolution_correction.resolution_correction import MissingMassResolutionCorrector
from rdstar.offline_analysis.mms_resolution_correction.resolution_correction_setups import (
    missing_mass_hist_var as resolution_evaluation_base_missing_mass_hist_var,
)


__all__ = [
    "ShiftEvaluationInfo",
    "mm2_shift_eval_infos_tuple",
    "mm2_shift_eval_infos_dict",
    "missing_mass_like_for_res_corr_eval_setup",
    "missing_mass_with_fei_calib_like_for_res_corr_eval_setup",
]


class ShiftEvaluationInfo(NamedTuple):
    name: str
    hist_var: HistVariable
    new_col_name_appendix: str
    fei_weight_col: Optional[str]

    @property
    def as_dict(self) -> Dict[str, Any]:
        res_dict = self._asdict()
        res_dict.update({"hist_var": self.hist_var_as_dict})
        return res_dict

    @staticmethod
    def from_dict(input_dict: Dict[str, Any]) -> "ShiftEvaluationInfo":
        return ShiftEvaluationInfo(
            **{
                k: ShiftEvaluationInfo.hist_var_from_dict(input_dict=v) if k == "hist_var" else v
                for k, v in input_dict.items()
            }
        )

    @property
    def hist_var_as_dict(self) -> Dict[str, Any]:
        return {
            "df_label": self.hist_var.df_label,
            "n_bins": self.hist_var.n_bins,
            "scope": self.hist_var.scope,
            "var_name": self.hist_var.variable_name,
            "unit": self.hist_var.unit,
            "use_log_scale": self.hist_var.use_log_scale,
        }

    @staticmethod
    def hist_var_from_dict(input_dict: Dict[str, Any]) -> HistVariable:
        return HistVariable(
            df_label=input_dict["df_label"],
            n_bins=input_dict["n_bins"],
            scope=input_dict["scope"],
            var_name=input_dict["var_name"] if input_dict["var_name"] else None,
            unit=input_dict["unit"] if input_dict["unit"] else None,
            use_log_scale=input_dict["use_log_scale"],
        )

    def create_copy_with_other_base_column(self, new_base_column_name: str) -> "ShiftEvaluationInfo":
        new_hist_var = self.get_new_hist_var_for(new_base_column_name=new_base_column_name)  # type: HistVariable
        return self.create_copy_with_other_hist_var(new_hist_var=new_hist_var)

    def create_copy_with_other_hist_var(self, new_hist_var: HistVariable) -> "ShiftEvaluationInfo":
        assert new_hist_var != self.hist_var, (new_hist_var, self.hist_var)

        return ShiftEvaluationInfo(
            name=self.name,
            hist_var=new_hist_var,
            new_col_name_appendix=self.new_col_name_appendix,
            fei_weight_col=self.fei_weight_col,
        )

    def create_copy_with_extended_new_col_name_appendix(self, new_appendix: str) -> "ShiftEvaluationInfo":
        return ShiftEvaluationInfo(
            name=self.name,
            hist_var=self.hist_var,
            new_col_name_appendix=f"{self.new_col_name_appendix}_{new_appendix}",
            fei_weight_col=self.fei_weight_col,
        )

    def create_copy_with_new_in_and_output_cols(
        self,
        new_base_column_name: str,
        new_appendix: str,
    ) -> "ShiftEvaluationInfo":
        new_hist_var = self.get_new_hist_var_for(new_base_column_name=new_base_column_name)  # type: HistVariable
        return ShiftEvaluationInfo(
            name=self.name,
            hist_var=new_hist_var,
            new_col_name_appendix=new_appendix,
            fei_weight_col=self.fei_weight_col,
        )

    def get_new_hist_var_for(self, new_base_column_name: str) -> HistVariable:
        assert new_base_column_name != self.hist_var.df_label, (new_base_column_name, self.hist_var.df_label)
        return HistVariable(
            df_label=new_base_column_name,
            n_bins=self.hist_var.n_bins,
            scope=self.hist_var.scope,
            var_name=self.hist_var.variable_name,
            unit=self.hist_var.unit if self.hist_var.unit else None,
            use_log_scale=self.hist_var.use_log_scale,
        )

    def get_new_col_name(self, base_col_name: str, shift_case_name: str) -> str:
        appendix = f"_{self.new_col_name_appendix}" if self.new_col_name_appendix else ""  # type: str
        return f"{base_col_name}_{shift_case_name}{appendix}"


# Definition of some required variables:
missing_mass = HistVariable(  # Should be identical to same variable in histogram_variables.py
    df_label=Cols.original_missing_mass,
    n_bins=30,
    scope=(-2.0, 6.0),
    var_name=LatexObsNames.fitting_mm2,
    unit=LatexUnits.gev_sq,
)

_old_resolution_corrected_mms_col_name = MissingMassResolutionCorrector.get_main_single_stage_correction_column_name(
    use_laplace_smeared_version=False
)  # type: str

missing_mass_res_corr_extreme_fine_binning = HistVariable(
    df_label=_old_resolution_corrected_mms_col_name,  # Should be identical to same variable in histogram_variables.py
    n_bins=100,
    scope=(-0.5, 3.0),
    var_name=LatexObsNames.fitting_mm2,
    unit=LatexUnits.gev_sq,
)


# Setups:
missing_mass_shift_eval_setup = ShiftEvaluationInfo(
    name="missing_mass",
    hist_var=missing_mass,
    new_col_name_appendix="MSSSv1",
    fei_weight_col=None,
)  # type: ShiftEvaluationInfo

missing_mass_like_for_res_corr_eval_setup = ShiftEvaluationInfo(
    name="missing_mass_like_for_res_corr_eval_setup",
    hist_var=resolution_evaluation_base_missing_mass_hist_var,
    new_col_name_appendix="MSSSv2",
    fei_weight_col=None,
)  # type: ShiftEvaluationInfo

missing_mass_res_corr_extreme_fine_binning_shift_eval_setup = ShiftEvaluationInfo(
    name="missing_mass_res_corr_extreme_fine_binning",
    hist_var=missing_mass_res_corr_extreme_fine_binning,
    new_col_name_appendix="MSSSv8",
    fei_weight_col=None,
)  # type: ShiftEvaluationInfo

missing_mass_with_fei_calib_shift_eval_setup = ShiftEvaluationInfo(
    name="missing_mass_with_fei_calib",
    hist_var=missing_mass,
    new_col_name_appendix="MSSSv1wFC",
    fei_weight_col=WeightManager.used_fei_calib_col,
)  # type: ShiftEvaluationInfo

missing_mass_with_fei_calib_like_for_res_corr_eval_setup = ShiftEvaluationInfo(
    name="missing_mass_with_fei_calib_like_for_res_corr_eval_setup",
    hist_var=resolution_evaluation_base_missing_mass_hist_var,
    new_col_name_appendix="MSSSv2wFC",
    fei_weight_col=WeightManager.used_fei_calib_col,
)  # type: ShiftEvaluationInfo

missing_mass_with_fei_calib_res_corr_extreme_fine_binning_shift_eval_setup = ShiftEvaluationInfo(
    name="missing_mass_with_fei_calib_res_corr_extreme_fine_binning",
    hist_var=missing_mass_res_corr_extreme_fine_binning,
    new_col_name_appendix="MSSSv8wFC",
    fei_weight_col=WeightManager.used_fei_calib_col,
)  # type: ShiftEvaluationInfo


_mm2_shift_eval_infos_tuple = (
    missing_mass_shift_eval_setup,
    missing_mass_like_for_res_corr_eval_setup,
    missing_mass_res_corr_extreme_fine_binning_shift_eval_setup,
    missing_mass_with_fei_calib_shift_eval_setup,
    missing_mass_with_fei_calib_like_for_res_corr_eval_setup,
    missing_mass_with_fei_calib_res_corr_extreme_fine_binning_shift_eval_setup,
)  # type: Tuple[ShiftEvaluationInfo, ...]

assert len(set([v.name for v in _mm2_shift_eval_infos_tuple])) == len(_mm2_shift_eval_infos_tuple)

mm2_shift_eval_infos_tuple = _mm2_shift_eval_infos_tuple  # type: Tuple[ShiftEvaluationInfo, ...]
mm2_shift_eval_infos_dict = {v.name: v for v in mm2_shift_eval_infos_tuple}  # type: Dict[str, ShiftEvaluationInfo]
