"""
Logistics Tasks
    - Copying Data
    - ROOT to hdf5 conversion

Felix Metzner 2019
"""

import os
import b2luigi as luigi

from luigi.util import requires
from typing import Any, Tuple, Dict, List

from analysistools.dataframe_tools import hdf5_utils
from analysistools.batch_tools.tasks import FileDictArchiveTask
from analysistools.batch_tools.file_handler import load_json_config
from analysistools.dataframe_tools.dataframe_tools import (
    copy_online_analysis_output_from_remote,
    copy_file_dict_archive_from_remote,
)

from rdstar.utilities import PathType
from rdstar.offline_analysis.tasks import utilities


__all__ = [
    "DataCopyTask",
    "DataInitialConversionTask",
]


class DataCopyTask(utilities.BaseOfflineTask):
    """
    Task copies n-tuples produced on the KEKcc to ETP servers. Should be run on desktop
    machines, as only the IP addresses of those have direct access to the KEKcc servers.

    The files that will be copied are specified via a production tag and a list of config
    files, which should be stored in the rdstar project in rdstar/configs/<production_tag>/...
    """

    offline_config_path = luigi.Parameter(positional=False)  # type: PathType

    file_dict_dir_name = "file_dicts"  # type: str
    file_dict_base_dir_name = "FileDict"  # type: str

    @property
    def priority(self) -> int:
        return 1000

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

        offline_config = utilities.load_yaml_config(yaml_config_path=self.offline_config_path)
        self.working_dir = offline_config["working_dir"]  # type: PathType

        self.production_tag = offline_config["production_tag"]
        self.batch_configs = offline_config["batch_configs"]

        self.remote_host = offline_config["file_transfer"]["remote_host"]
        self.remote_source_dir = offline_config["file_transfer"]["remote_source_dir"]

        self.batch_config_list = self.get_configs()

    @staticmethod
    def get_local_file_dict_path(offline_config_path: PathType) -> PathType:
        offline_config = utilities.load_yaml_config(yaml_config_path=offline_config_path)
        main_work_dir_path = offline_config["working_dir"]  # type: PathType
        assert os.path.isdir(main_work_dir_path), main_work_dir_path

        fd_path = os.path.join(main_work_dir_path, DataCopyTask.file_dict_base_dir_name, DataCopyTask.file_dict_dir_name)
        assert os.path.isdir(fd_path), fd_path
        return fd_path

    @staticmethod
    def get_local_file_dict_path_from_main_dir(main_work_dir_path: PathType) -> PathType:
        return os.path.join(main_work_dir_path, DataCopyTask.file_dict_base_dir_name, DataCopyTask.file_dict_dir_name)

    def get_configs(self) -> List[Dict[Any, Any]]:
        config_dir = utilities.get_config_dir_path_for(production_tag=self.production_tag)  # type: PathType
        config_files = [os.path.join(config_dir, c) for c in self.batch_configs]  # type: List[PathType]
        return [load_json_config(c) for c in config_files]

    def get_remote_file_dict_source_path(self) -> PathType:
        first_batch_config = self.get_configs()[0]
        return FileDictArchiveTask.get_output_path(
            batch_config=first_batch_config,
            local_config_dir_path=utilities.get_config_directory(),
        )

    def copy_process(
        self,
        dry_run: bool = False,
    ) -> List[Tuple[PathType, PathType]]:
        output_tuple_list = copy_online_analysis_output_from_remote(
            remote_host=self.remote_host,
            source_dir=self.remote_source_dir,
            target_dir=self.working_dir,
            configs=self.batch_config_list,
            overwrite=False,
            dry_run=dry_run,
        )  # type: List[Tuple[PathType, PathType]]

        file_dict_output_tuple_entry = copy_file_dict_archive_from_remote(
            remote_host=self.remote_host,
            source_path=self.get_remote_file_dict_source_path(),
            target_dir=os.path.join(self.working_dir, self.file_dict_base_dir_name),
            file_dict_dir_name=self.file_dict_dir_name,
            overwrite=False,
            dry_run=dry_run,
        )  # type: Tuple[PathType, PathType]

        output_tuple_list.append(file_dict_output_tuple_entry)

        return output_tuple_list

    def run(self) -> None:
        utilities.host_check(hosts=utilities.GlobalParameters().copy_host_list)
        os.makedirs(self.working_dir, exist_ok=True)
        self.copy_process()

    def output(self) -> Dict[str, luigi.LocalTarget]:
        source_target_list = self.copy_process(dry_run=True)

        output_dict = dict()
        for _, target_path in source_target_list:
            t_path = target_path
            output_dict.update({os.path.splitext(os.path.basename(t_path))[0]: luigi.LocalTarget(t_path)})

        return output_dict


@requires(DataCopyTask)
class DataInitialConversionTask(utilities.BaseOfflineTask):
    data_type: str = luigi.Parameter(positional=False)
    stream: int = luigi.IntParameter(positional=False)
    tuple_type: str = luigi.Parameter(positional=False)
    reduced_production_tag: str = luigi.Parameter(positional=False)

    @property
    def resources(self) -> Dict[str, int]:
        if self.data_type in ["Charged", "Charm", "UDS"]:
            return {utilities.offline_memory_resource_name: 8}
        elif self.data_type in ["Mixed"]:
            return {utilities.offline_memory_resource_name: 5}
        else:
            return {utilities.offline_memory_resource_name: 1}

    @property
    def priority(self) -> int:
        return super().priority * 10

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

        offline_config = utilities.load_yaml_config(yaml_config_path=self.offline_config_path)
        self.working_dir = offline_config["working_dir"]  # type: PathType
        self.base_dir = os.path.join(self.working_dir, str(self.data_type))

        self.stream_tag = f"_{self.stream}"
        if self.stream == -999:
            self.stream_tag = ""

        self.file_name = f"{self.data_type}{self.stream_tag}_{self.tuple_type}"

        self.file_tag = "base"

    def conversion_process(
        self,
        dry_run: bool = False,
    ) -> Tuple[PathType, PathType]:
        source = os.path.join(self.base_dir, f"{self.file_name}.root")
        target = os.path.join(self.base_dir, f"{self.file_name}_{self.file_tag}.h5")

        if not dry_run:
            assert os.path.isfile(source), source

            temp_target = os.path.join(self.base_dir, f"tmp_{self.file_name}_{self.file_tag}.h5")

            hdf5_utils.convert_root_to_hdf5(
                root_file_path=source,
                hdf5_file_path=temp_target,
                hdf5_key=self.file_name,
                data_type=str(self.data_type),
                stream_tag=self.stream_tag,
                add_unique_event_id=False,
            )

            os.rename(src=temp_target, dst=target)

        return source, target

    def run(self) -> None:
        utilities.host_check(hosts=utilities.GlobalParameters().work_host_list)
        self.conversion_process()

    def output(self) -> Dict[str, luigi.LocalTarget]:
        _, target_path = self.conversion_process(dry_run=True)
        return {self.file_name: luigi.LocalTarget(target_path)}
