"""
Validation of pre-selection sub-samples to evaluate quality of observables used in the selections and in mva trainings.

Felix Metzner 2021
"""

import os
import pandas as pd

import b2luigi as luigi

from luigi.util import inherits
from typing import Tuple, List, Dict, Generator, Optional

from rdstar.utilities import PathType
from rdstar.offline_analysis.selection.selection import Selection
from rdstar.offline_analysis.selection.selection_setups import SelectionSetups
from rdstar.offline_analysis.weight_management.weight_information import WeightManager
from rdstar.offline_analysis.weight_management.weight_finalization import WeightFinalizer
from rdstar.offline_analysis.selection.selection_info_containers import SelectionInfo, SelectionSpecifics

from rdstar.offline_analysis.tasks import utilities
from rdstar.offline_analysis.tasks.logistics_tasks import DataCopyTask
from rdstar.offline_analysis.tasks.weight_tasks import WeightInfoGathererTask
from rdstar.offline_analysis.tasks.combiner_tasks import EvaluationSampleCombinerTask
from rdstar.offline_analysis.validation.pre_selection_validator import PreSelectionValidator
from rdstar.offline_analysis.data_handling.sample_setup_definitions import SampleSetupHandler


__all__ = [
    "PreSelectionValidationManagerTask",
]


@inherits(EvaluationSampleCombinerTask)
class PreSelectionSamplePreparerTask(utilities.BaseOfflineTask):
    config_combinations: utilities.ConfigCombinationsType = luigi.ListParameter(positional=False, hashed=True)

    base_dir_name: str = "CombinedEvaluationSamplesWithWeights"
    validation_output_tag: str = "evaluation_sample_with_weights"

    @property
    def resources(self) -> Dict[str, int]:
        return {utilities.offline_memory_resource_name: 2}

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

        self.base_path: PathType = os.path.join(self.working_dir, self.base_dir_name)

        self.cs_tag: str = self.config_params[0]["cs_tag"]
        self.main_selection_name: str = SelectionSetups.used_selection_name

        self._input_dict: Optional[Dict[str, PathType]] = None

    @property
    def input_dict(self) -> Dict[str, PathType]:
        if self._input_dict is None:
            self._input_dict = {k: v.path for d in self.input() if isinstance(d, dict) for k, v in d.items()}
        return self._input_dict

    def run_preparation(
        self,
        input_path: PathType,
        output_key: str,
        tmp_output_path: PathType,
        output_path: PathType,
    ) -> None:
        _flag_info_dict_path: PathType = WeightManager.get_flag_info_dict_path_from_task_input_dict(
            input_dict=self.input_dict,
            selection_name=self.main_selection_name,
        )

        df = pd.read_hdf(input_path)
        assert isinstance(df, pd.DataFrame), (type(df).__name__, input_path)

        WeightFinalizer.apply_weight_finalization_to_df(
            df=df,
            flag_info_dict_path=_flag_info_dict_path,
            file_dict_dir_path=DataCopyTask.get_local_file_dict_path_from_main_dir(main_work_dir_path=self.working_dir),
            combo_name=self.combo_name,
            key_hint=None,
        )
        df.to_hdf(tmp_output_path, key=output_key, format="fixed")
        os.rename(tmp_output_path, output_path)

    def get_input_key(self) -> str:
        return f"{self.combo_name}_{EvaluationSampleCombinerTask.validation_output_tag}"

    def prepare_validation_samples(self, dry_run: bool = False) -> Tuple[str, PathType]:
        output_key: str = f"{self.combo_name}_{self.validation_output_tag}"
        output_path: PathType = os.path.join(self.base_path, f"{output_key}.h5")
        temp_output_path: PathType = os.path.join(self.base_path, f"tmp_{output_key}.h5")

        if not dry_run:
            os.makedirs(self.base_path, exist_ok=True)

            input_path: PathType = self.input_dict[self.get_input_key()]
            assert os.path.exists(input_path), (self.combo_name, input_path)

            if self.combo_name.startswith("Data"):
                if os.path.exists(output_path):
                    assert os.path.islink(output_path), output_path
                    os.remove(output_path)
                os.symlink(src=input_path, dst=output_path)
            else:
                self.run_preparation(
                    input_path=input_path,
                    output_key=output_key,
                    tmp_output_path=temp_output_path,
                    output_path=output_path,
                )

        return output_key, output_path

    def run(self) -> None:
        utilities.host_check(hosts=utilities.GlobalParameters().work_host_list)
        self.prepare_validation_samples()

    def output(self) -> Dict[str, luigi.LocalTarget]:
        output_key, target_path = self.prepare_validation_samples(dry_run=True)  # type: str, PathType
        return {output_key: luigi.LocalTarget(target_path)}

    def requires(self) -> Generator[utilities.TaskType, None, None]:
        yield self.clone(
            EvaluationSampleCombinerTask,
            combo_name=self.combo_name,
            working_dir=self.working_dir,
            config_params=self.config_params,
        )

        yield self.clone(
            WeightInfoGathererTask,
            selection_name=self.main_selection_name,
            cs_tag=self.cs_tag,
            config_combinations=self.config_combinations,
            weight_key=WeightManager.pid_weight_info_key,
        )


class PreSelectionValidationTask(utilities.BaseOfflineTask):
    sample_setup_key: str = luigi.Parameter(positional=False)
    apply_subset_of_selection: bool = luigi.BoolParameter(positional=False)
    only_bcs_sub_selection: bool = luigi.BoolParameter(positional=False)
    apply_fei_correction_weight: bool = luigi.BoolParameter(positional=False)

    config_combinations: utilities.ConfigCombinationsType = luigi.ListParameter(positional=False, hashed=True)
    working_dir: PathType = luigi.Parameter(positional=False)

    base_dir_name: str = "PreSelectionEvaluation"

    selection_bcs_random_seed: int = 1337

    # Column name information:
    u_eid_col: str = utilities.UniqueEventIDColumnInfos.get_unique_event_id_col_name()

    @property
    def resources(self) -> Dict[str, int]:
        return {utilities.offline_memory_resource_name: 8}

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

        if self.only_bcs_sub_selection:
            assert (
                self.apply_subset_of_selection
            ), "If only_bcs_sub_selection can only be set to True, if apply_subset_of_selection is also True!"

        settings_tag: str = "with_bf_corr_weights"
        if self.apply_fei_correction_weight:
            settings_tag = "with_bf_and_fei_corr_weights"
        if self.apply_subset_of_selection:
            if self.only_bcs_sub_selection:
                settings_tag += "_with_only_bcs_subset_selection"
            else:
                settings_tag += "_with_subset_selection"

        self.base_path: PathType = os.path.join(
            self.working_dir,
            self.base_dir_name,
            self.sample_setup_key,
            settings_tag,
        )

        self.data_sample_key: str = SampleSetupHandler.get_data_sample_key(sample_name=self.sample_setup_key)
        self.mc_sample_keys: Tuple[str, ...] = SampleSetupHandler.get_mc_sample_keys(sample_name=self.sample_setup_key)

        self.cs_tag: str = self.config_combinations[0][2][0]["cs_tag"]
        self.all_required_combo_names: List[str] = [self.data_sample_key, *self.mc_sample_keys]

        if self.apply_fei_correction_weight:
            self.weight_column: str = WeightManager.final_weight_with_additions_col
        else:
            self.weight_column = WeightManager.final_weight_with_bf_corrections_col

        self._input_dict: Optional[Dict[str, PathType]] = None

    @property
    def input_dict(self) -> Dict[str, PathType]:
        if self._input_dict is None:
            self._input_dict = {k: v.path for d in self.input() if isinstance(d, dict) for k, v in d.items()}
        return self._input_dict

    def load_mc_sample(self) -> pd.DataFrame:
        mc_paths: List[PathType] = [
            p
            for mc_key in self.mc_sample_keys
            for k, p in self.input_dict.items()
            if (
                k == f"{mc_key}_{PreSelectionSamplePreparerTask.validation_output_tag}"
                and os.path.splitext(p)[1] == ".h5"
            )
        ]
        assert len(mc_paths) == len(self.mc_sample_keys), (len(mc_paths), mc_paths, self.mc_sample_keys)

        if len(mc_paths) > 1:
            mc_data_dir: PathType = os.path.dirname(mc_paths[0])
            assert os.path.isdir(mc_data_dir), mc_data_dir
            mc_combined_file_name: PathType = f"combined_mc_samples_for_{self.sample_setup_key}.h5"
            mc_combined_file_path: PathType = os.path.join(mc_data_dir, mc_combined_file_name)

            if os.path.isfile(mc_combined_file_path):
                mc_sample = pd.read_hdf(mc_combined_file_path)
            else:
                combined_mc_key: str = f"combined_mc_samples_for_{self.sample_setup_key}"
                tmp_mc_combined_file_name: PathType = f"tmp_{mc_combined_file_name}"
                tmp_mc_combined_file_path: PathType = os.path.join(mc_data_dir, tmp_mc_combined_file_name)

                mc_sample = pd.concat([pd.read_hdf(mc_path) for mc_path in mc_paths], ignore_index=True, sort=False)
                mc_sample.to_hdf(tmp_mc_combined_file_path, key=combined_mc_key, format="fixed")
                os.rename(tmp_mc_combined_file_path, mc_combined_file_path)
        else:
            mc_sample = pd.read_hdf(mc_paths[0])

        return mc_sample

    def load_recorded_data_sample(self) -> pd.DataFrame:
        data_paths: List[PathType] = [
            p
            for k, p in self.input_dict.items()
            if (
                k == f"{self.data_sample_key}_{PreSelectionSamplePreparerTask.validation_output_tag}"
                and os.path.splitext(p)[1] == ".h5"
            )
        ]
        assert len(data_paths) == 1, (len(data_paths), data_paths)
        data_path = data_paths[0]
        assert os.path.exists(data_path), data_path
        data_df = pd.read_hdf(data_path)
        assert isinstance(data_df, pd.DataFrame)
        return data_df

    def apply_selection_subset(
        self,
        selection_name: str,
        mc_df: pd.DataFrame,
        data_df: pd.DataFrame,
        subset_check_cut_stings: Tuple[str, ...],
    ) -> Tuple[pd.DataFrame, pd.DataFrame]:
        subset_selection_info: SelectionInfo = self.get_subset_selection_info(
            selection_name=selection_name,
            subset_check_cut_stings=subset_check_cut_stings,
        )

        if self.only_bcs_sub_selection:
            sel_mc_df: pd.DataFrame = Selection.apply_bcs(
                df=mc_df,
                primary_var=subset_selection_info.selection.primary_bcs_var,
                secondary_var=subset_selection_info.selection.secondary_bcs_var,
                bcs_random_seed=self.selection_bcs_random_seed,
                unique_event_id_column_name=self.u_eid_col,
            )
            sel_data_df: pd.DataFrame = Selection.apply_bcs(
                df=data_df,
                primary_var=subset_selection_info.selection.primary_bcs_var,
                secondary_var=subset_selection_info.selection.secondary_bcs_var,
                bcs_random_seed=self.selection_bcs_random_seed,
                unique_event_id_column_name=self.u_eid_col,
            )
        else:
            event_selection = Selection(
                selection_name=selection_name,
                unique_event_id_column_name=self.u_eid_col,
                bcs_random_seed=self.selection_bcs_random_seed,
            )

            sel_mc_df, _ = event_selection.apply_selection(
                df=mc_df,
                selection_info=subset_selection_info,
                evaluate_selection=False,
            )
            sel_data_df, _ = event_selection.apply_selection(
                df=data_df,
                selection_info=subset_selection_info,
                evaluate_selection=False,
            )

        return sel_mc_df, sel_data_df

    def get_subset_selection_info(
        self,
        selection_name: str,
        subset_check_cut_stings: Tuple[str, ...],
    ) -> SelectionInfo:
        main_selection_tag: str = SelectionSetups.get_main_selection_tag(
            selection_name=selection_name,
            cs_tag=self.cs_tag,
        )
        main_selection_specifics: SelectionSpecifics = SelectionSetups.get_selection_infos(
            selection_name=selection_name,
            cs_tag=self.cs_tag,
        )[main_selection_tag]

        cut_infos_subset = tuple(ci for ci in main_selection_specifics.cuts if ci.cut_str in subset_check_cut_stings)
        assert len(cut_infos_subset) == len(subset_check_cut_stings), (
            len(cut_infos_subset),
            len(subset_check_cut_stings),
        )

        subset_of_selection_specifics = SelectionSpecifics(
            name=main_selection_tag,
            cuts=cut_infos_subset,
            post_bcs_cuts=main_selection_specifics.post_bcs_cuts,
            primary_bcs_var=main_selection_specifics.primary_bcs_var,
            secondary_bcs_var=main_selection_specifics.secondary_bcs_var,
            check_cuts=tuple(),
            check_bcs=False,
            apply_post_mva_selection_cut=False,
            run_evaluation=False,
            mva_setup_name=None,
            post_mva_cut_info=None,
        )

        return SelectionInfo(
            selection_tag=main_selection_tag,
            selection=subset_of_selection_specifics,
            target_path="none",
            output_key="none",
        )

    def create_pre_selection_validation_plots(
        self,
        dry_run: bool = False,
    ) -> Dict[str, PathType]:
        pre_selection_validator = PreSelectionValidator(
            weight_column=self.weight_column,
            export_path=self.base_path,
            adaptive_binning=True,
            fig_size=None,
        )

        if dry_run:
            mc_sample_df: pd.DataFrame = pd.DataFrame()
            data_sample_df: pd.DataFrame = pd.DataFrame()
        else:
            os.makedirs(self.base_path, exist_ok=True)
            mc_sample_df = self.load_mc_sample()
            data_sample_df = self.load_recorded_data_sample()

            if self.apply_subset_of_selection:
                sel_mc_df, sel_data_df = self.apply_selection_subset(
                    selection_name=pre_selection_validator.subset_selection_name,
                    mc_df=mc_sample_df,
                    data_df=data_sample_df,
                    subset_check_cut_stings=pre_selection_validator.subset_check_cut_stings,
                )
                del (mc_sample_df,)
                del (data_sample_df,)
                mc_sample_df = sel_mc_df
                data_sample_df = sel_data_df

        return pre_selection_validator.create_plots(
            mc_df=mc_sample_df,
            data_df=data_sample_df,
            dry_run=dry_run,
        )

    def run(self) -> None:
        utilities.host_check(hosts=utilities.GlobalParameters().work_host_list)
        self.create_pre_selection_validation_plots()

    def output(self) -> Dict[str, luigi.LocalTarget]:
        output_dict: Dict[str, PathType] = self.create_pre_selection_validation_plots(dry_run=True)
        return {output_key: luigi.LocalTarget(target_path) for output_key, target_path in output_dict.items()}

    def requires(self) -> Generator[utilities.TaskType, None, None]:
        for combo_name, _, relevant_configs in self.config_combinations:
            if combo_name in self.all_required_combo_names:
                yield self.clone(
                    PreSelectionSamplePreparerTask,
                    combo_name=combo_name,
                    config_params=relevant_configs,
                )


class PreSelectionValidationManagerTask(luigi.WrapperTask):
    config_combinations: utilities.ConfigCombinationsType = luigi.ListParameter(positional=False, hashed=True)
    working_dir: PathType = luigi.Parameter(positional=False)

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

    def get_log_file_dir(self) -> PathType:
        return self.working_dir

    def get_task_file_dir(self) -> PathType:
        return self.working_dir

    def requires(self) -> Generator[utilities.TaskType, None, None]:
        for sample_setup_key in SampleSetupHandler.on_resonance_sample_setup_names:
            for only_bcs_sub_selection in [True, False]:
                yield self.clone(
                    PreSelectionValidationTask,
                    sample_setup_key=sample_setup_key,
                    only_bcs_sub_selection=only_bcs_sub_selection,
                    apply_subset_of_selection=True,
                    apply_fei_correction_weight=False,
                )
