"""
Analysis Flow Validation -- Multiplicity Evaluation

Felix Metzner 2019
"""

import os
import copy
import numpy as np
import pandas as pd

from matplotlib import pyplot as plt
from typing import Any, NamedTuple, Union, Optional, Tuple, List, Dict, Callable

from templatefitter.plotter.plot_style import KITColors
from templatefitter.plotter.plot_utilities import export
from templatefitter.plotter.plot_utilities import color_fader, AxesType, FigureType

from analysistools.plotting.latex_tables import df_to_split_latex_table

from rdstar.utilities import PathType

from rdstar.offline_analysis.reco_modes.reco_mode_infos import RecoModeInfos
from rdstar.offline_analysis.mc_matching.old_mc_matching_info import OldMCMatchingIDsQueryInfo

from rdstar.offline_analysis.selection.selection_evaluation_utilities import ValueCounter, ValuePerOtherCounter
from rdstar.offline_analysis.selection.selection_evaluation_info_container import (
    SelectionEvaluationContainer,
    SelectionEvaluationContainerEntry,
)
from rdstar.offline_analysis.validation.analysis_flow_validation import (
    get_cut_flow_table_col_name,
    update_table_fraction_output_dict,
)


__all__ = [
    "multiplicity_info_collection",
    "evaluate_multiplicity_info",
]


class MultiplicityInfo(NamedTuple):
    key: str
    info_type: str
    column_name_dict: Union[None, Dict[Union[str, int, float], str], Dict[int, str]] = None
    table_fractions: Optional[List[int]] = None

    @property
    def key_and_type(self) -> str:
        return f"key: {self.key}, info_type: {self.info_type}"

    @property
    def multiplicity_info_size(self) -> int:
        sub_info = SelectionEvaluationContainerEntry.get_sub_info_class_by_key(key=self.info_type)
        entry_type = sub_info.get_entry_type(entry_key=self.key)
        if entry_type == ValueCounter:
            return 1
        elif entry_type == ValuePerOtherCounter:
            return 2
        else:
            raise TypeError(f"Unsupported entry type {entry_type} for {self.info_type}.{self.key}!")


main_reco_mode_id_to_latex_label_mapping: Dict[int, str] = copy.copy(RecoModeInfos.get_reco_id_to_latex_label())

multiplicity_info_collection: List[MultiplicityInfo] = [
    MultiplicityInfo(
        key="event_multiplicities", info_type="multiplicity_info", column_name_dict=None, table_fractions=[1]
    ),
    MultiplicityInfo(
        key="event_multiplicities_per_data_type",
        info_type="multiplicity_info",
        column_name_dict=None,
        table_fractions=[3, 6],
    ),
    MultiplicityInfo(
        key="event_multiplicities_per_decay_id",
        info_type="multiplicity_info",
        column_name_dict=main_reco_mode_id_to_latex_label_mapping,
        table_fractions=[2, 4],
    ),
    MultiplicityInfo(
        key="event_multiplicities_per_sig_id",
        info_type="multiplicity_mc_info",
        column_name_dict=OldMCMatchingIDsQueryInfo.get_sig_id_to_latex_label_without_pilnu(),
        table_fractions=[1, 3],
    ),
    MultiplicityInfo(
        key="reco_sig_or_norm_match_event_multiplicities_per_decay_id",
        info_type="multiplicity_mc_info",
        column_name_dict=main_reco_mode_id_to_latex_label_mapping,
        table_fractions=[2, 4],
    ),
    MultiplicityInfo(
        key="u4s_match_reco_sig_or_norm_match_event_multiplicities_per_decay_id",
        info_type="multiplicity_mc_info",
        column_name_dict=main_reco_mode_id_to_latex_label_mapping,
        table_fractions=[2, 4],
    ),
    MultiplicityInfo(
        key="b_sig_match_reco_sig_or_norm_match_event_multiplicities_per_decay_id",
        info_type="multiplicity_mc_info",
        column_name_dict=main_reco_mode_id_to_latex_label_mapping,
        table_fractions=[2, 4],
    ),
]


data_multiplicity_info_collection: List[MultiplicityInfo] = [
    MultiplicityInfo(
        key="event_multiplicities", info_type="multiplicity_info", column_name_dict=None, table_fractions=[1]
    ),
    MultiplicityInfo(
        key="event_multiplicities_per_decay_id",
        info_type="multiplicity_info",
        column_name_dict=main_reco_mode_id_to_latex_label_mapping,
        table_fractions=[2, 4],
    ),
]


def evaluate_multiplicity_info(
    info_container: Optional[SelectionEvaluationContainer],
    multiplicity_info: MultiplicityInfo,
    selection_name: str,
    selection_tag: str,
    base_target_dir_path: PathType,
    for_data: bool,
    create_table_pdfs: bool,
    dry_run: bool = False,
) -> Dict[str, PathType]:
    output_dict: Dict[str, PathType] = {}
    data_dir_suffix: str = "OnData" if for_data else ""
    tables_dir_path: PathType = os.path.join(base_target_dir_path, f"EventMultiplicityTables{data_dir_suffix}")
    plots_dir_path: PathType = os.path.join(base_target_dir_path, f"EventMultiplicityPlots{data_dir_suffix}")

    em_table_output_dict = create_multiplicity_table(
        info_container=info_container,
        multiplicity_info=multiplicity_info,
        selection_name=selection_name,
        selection_tag=selection_tag,
        for_data=for_data,
        target_dir_path=tables_dir_path,
        create_pdfs=create_table_pdfs,
        dry_run=dry_run,
    )
    assert all(k not in output_dict for k in em_table_output_dict), (em_table_output_dict, output_dict)
    output_dict.update(em_table_output_dict)

    em_plot_output_dict = create_multiplicity_plots(
        info_container=info_container,
        multiplicity_info=multiplicity_info,
        selection_name=selection_name,
        selection_tag=selection_tag,
        target_dir_path=plots_dir_path,
        for_data=for_data,
        max_cands=8,
        explicit_overflow_label=True,
        dry_run=dry_run,
    )
    assert all(k not in output_dict for k in em_plot_output_dict), (em_plot_output_dict, output_dict)
    output_dict.update(em_plot_output_dict)

    return output_dict


def create_multiplicity_plots(
    info_container: Optional[SelectionEvaluationContainer],
    multiplicity_info: MultiplicityInfo,
    selection_name: str,
    selection_tag: str,
    target_dir_path: PathType,
    dry_run: bool = False,
    max_cands: int = 8,
    explicit_overflow_label: bool = False,
    log_y_scale: bool = False,
    for_data: bool = False,
) -> Dict[str, PathType]:
    output_dict: Dict[str, PathType] = {}

    file_name_tag: str = ""
    if for_data:
        assert multiplicity_info.info_type == "multiplicity_info", multiplicity_info.key_and_type
        file_name_tag = "data_"

    file_name: str = f"{selection_tag}_emp_{file_name_tag}{multiplicity_info.key}_plot"
    assert file_name not in output_dict, file_name
    output_dict.update({file_name: os.path.join(target_dir_path, f"{file_name}.pdf")})

    if not dry_run:
        assert info_container is not None
        os.makedirs(target_dir_path, exist_ok=True)

        fig, _ = create_multiplicity_plot(
            multiplicity_info=multiplicity_info,
            info_container=info_container,
            cut_list=info_container.full_cut_list,
            selection_name=selection_name,
            max_cands=max_cands,
            explicit_overflow_label=explicit_overflow_label,
            log_y_scale=log_y_scale,
        )
        export(fig=fig, filename=file_name, target_dir=target_dir_path)
        plt.close(fig=fig)

    return output_dict


def create_multiplicity_plot(
    multiplicity_info: MultiplicityInfo,
    info_container: SelectionEvaluationContainer,
    cut_list: Tuple[str, ...],
    selection_name: str,
    max_cands: int = 8,
    explicit_overflow_label: bool = False,
    log_y_scale: bool = False,
) -> Tuple[FigureType, AxesType]:

    if explicit_overflow_label:
        overflow_name: str = rf"$\geqq {max_cands + 1}$"
    else:
        overflow_name = "OF"
        if max_cands <= 20:
            overflow_name = "Overflow"

    if multiplicity_info.multiplicity_info_size == 1:
        emp_dict = {
            i: copy.copy(
                info_container[k]
                .get_valid_sub_info_by_key(key=multiplicity_info.info_type)
                .get_value(key=multiplicity_info.key)
            )
            for i, k in info_container.level_to_entry_id_map.items()
        }
    elif multiplicity_info.multiplicity_info_size == 2:
        tmp_dict = {
            i: copy.copy(
                info_container[k]
                .get_valid_sub_info_by_key(key=multiplicity_info.info_type)
                .get_value(key=multiplicity_info.key)
            )
            for i, k in info_container.level_to_entry_id_map.items()
        }
        assert all(all(isinstance(v, dict) for v in sub_di.values()) for sub_di in tmp_dict.values())

        emp_series_dict: Dict[int, Any] = {k: ValuePerOtherCounter(v).as_series for k, v in tmp_dict.items()}
        # TODO: Here multiple plots could be created if we do not reduce the information:
        emp_dict = {k: ValueCounter(v.groupby(level=[1]).sum()).value_counts for k, v in emp_series_dict.items()}
    else:
        raise RuntimeError(
            f"No implementation of multiplicity plot for info {multiplicity_info.key_and_type}"
            f"with info size {multiplicity_info.multiplicity_info_size}"
        )

    assert all(isinstance(k, int) for k in emp_dict.keys()), list(emp_dict.keys())
    assert all(all(isinstance(k, int) for k in sub_di.keys()) for sub_di in emp_dict.values())
    assert all(all(isinstance(v, int) for v in sub_di.values()) for sub_di in emp_dict.values())

    # Remove events with multiplicity above overflow value
    reduced_emp_dict = {k: {kk: vv for kk, vv in v.items() if kk <= max_cands + 1} for k, v in emp_dict.items()}

    # Add overflow bin:
    of_bin_emp_dict = {k: sum([vv for kk, vv in v.items() if kk > max_cands + 1]) for k, v in emp_dict.items()}
    for k in reduced_emp_dict.keys():
        _current_count: int = reduced_emp_dict[k].get(max_cands + 1, 0)
        reduced_emp_dict[k][max_cands + 1] = _current_count + of_bin_emp_dict[k]

    width_elements = 1.0 / len(reduced_emp_dict.keys())
    width = width_elements * 0.8
    width_space = width_elements * 0.15

    fig = plt.figure(num=None, figsize=(6.0, 3.0), dpi=200, facecolor="w", edgecolor="k")
    ax = plt.gca()

    for i, (key, data) in enumerate(reduced_emp_dict.items()):
        assert i == key, (i, key)
        label = get_cut_flow_table_col_name(
            index=i,
            cut_list=cut_list,
            selection_name=selection_name,
        )
        if i < len(reduced_emp_dict) - 1:
            color = color_fader(KITColors.kit_green, KITColors.kit_black, i / (len(reduced_emp_dict) - 1))
        else:
            color = KITColors.kit_red
        x_pos = np.array(list(data.keys())) + i * (width + width_space)
        ax.bar(x_pos, data.values(), width, label=label, color=color)

    ax.set_xlabel("Event Multiplicity")
    ax.set_xlim(0.6, max_cands + 2)
    ax.set_xticks(np.arange(1, max_cands + 2))
    ax.set_xticks([], minor=True)
    x_tick_labels = [x if x <= max_cands else overflow_name for x in list(np.arange(1, max_cands + 2))]
    ax.set_xticklabels(x_tick_labels, ha="left")

    ax.set_ylabel("Number of Events")
    if log_y_scale:
        ax.set_yscale("log")

    ax.legend(loc=1, ncol=2)

    return fig, ax


def get_multiplicity_data(multiplicity_dict: Dict[int, int]) -> np.array:
    if len(multiplicity_dict) == 0:
        return np.array([], dtype=np.int32)

    try:
        return np.repeat(
            a=np.array(list(multiplicity_dict.keys())),
            repeats=np.array(list(multiplicity_dict.values())),
        )
    except TypeError:
        return np.repeat(
            a=np.array(list(multiplicity_dict.keys())),
            repeats=np.array(list(map(int, multiplicity_dict.values()))),
        )


def create_multiplicity_table(
    info_container: Optional[SelectionEvaluationContainer],
    multiplicity_info: MultiplicityInfo,
    selection_name: str,
    selection_tag: str,
    target_dir_path: PathType,
    for_data: bool,
    create_pdfs: bool,
    dry_run: bool = False,
) -> Dict[str, PathType]:
    assert multiplicity_info.table_fractions is not None
    if for_data:
        assert multiplicity_info.info_type == "multiplicity_info", multiplicity_info.key_and_type

    output_dict: Dict[str, PathType] = {}
    data_tag: str = "for_data_" if for_data else ""
    base_name: str = f"{selection_tag}_emt_{data_tag}{multiplicity_info.key}"

    em_table_functions: Dict[str, Callable] = {
        "median": np.median,
        "max": np.max,
        "min": np.min,
        "mean": np.mean,
    }

    all_key: int = -1
    all_col_name: str = "All"
    single_col_name: str = "Multiplicity"

    em_data_dict: Dict[int, Dict[int, np.array]] = {}

    if not dry_run:
        assert info_container is not None
        info_k: str = multiplicity_info.key
        info_t: str = multiplicity_info.info_type

        sub_containers = {
            k: info_container[e_id].get_valid_sub_info_by_key(key=info_t).get_value(key=info_k)
            for k, e_id in info_container.level_to_entry_id_map.items()
        }  # my type hint: Dict[int, Dict[int, Union[int, Dict[int, int]]]]

        if multiplicity_info.multiplicity_info_size == 1:
            assert all(isinstance(k, int) and isinstance(v, int) for vv in sub_containers.values() for k, v in vv.items())
            _s_containers: Dict[int, Dict[int, int]] = sub_containers
            em_data_dict = {all_key: {k: get_multiplicity_data(multiplicity_dict=v) for k, v in _s_containers.items()}}
        elif multiplicity_info.multiplicity_info_size == 2:
            assert all(isinstance(k, (int, str)) for vv in sub_containers.values() for k in vv.keys())
            assert all(isinstance(v, dict) for vv in sub_containers.values() for v in vv.values())
            assert all(
                isinstance(k, int) and isinstance(v, int)
                for vvv in sub_containers.values()
                for vv in vvv.values()
                for k, v in vv.items()
            )
            _sub_containers: Dict[int, Dict[Union[int, str], Dict[int, int]]] = sub_containers

            sub_keys: List[int] = list(set([kk for k in info_container.levels for kk in sub_containers[k].keys()]))
            _combined = {
                k: ValueCounter(ValuePerOtherCounter(v).as_series.groupby(level=[1]).sum()).value_counts
                for k, v in _sub_containers.items()
            }  # my type hint:  Dict[int, Dict[Union[int, float, str], Union[int, float]]]
            assert all(isinstance(k, int) and isinstance(v, int) for vv in _combined.values() for k, v in vv.items())
            combined_containers: Dict[int, Dict[int, int]] = {
                k: {int(kk): int(vv) for kk, vv in v.items()} for k, v in _combined.items()
            }

            em_data_dict = {
                kk: {
                    cut_k: get_multiplicity_data(multiplicity_dict=sub_container[kk])
                    if kk in sub_container.keys()
                    else np.array([])
                    for cut_k, sub_container in _sub_containers.items()
                }
                for kk in sub_keys
            }

            assert all_key not in em_data_dict.keys(), (all_key, em_data_dict.keys())
            em_all = {all_key: {k: get_multiplicity_data(multiplicity_dict=v) for k, v in combined_containers.items()}}
            em_data_dict.update(em_all)
        else:
            raise RuntimeError(
                f"No implementation of multiplicity plot for info {multiplicity_info.key_and_type}"
                f"with info size {multiplicity_info.multiplicity_info_size}"
            )

    for func_name, func in em_table_functions.items():
        name_prefix: str = f"{base_name}_{func_name}"

        update_table_fraction_output_dict(
            output_dict=output_dict,
            name_prefix=name_prefix,
            base_path=target_dir_path,
            create_pdfs=create_pdfs,
            table_fractions=multiplicity_info.table_fractions,
        )

        if dry_run:
            continue

        assert info_container is not None
        os.makedirs(target_dir_path, exist_ok=True)

        em_dict = {
            d_id: {
                k: func(em_data[k]) if isinstance(em_data[k], np.ndarray) and len(em_data[k]) > 0 else np.nan
                for k in info_container.levels
            }
            for d_id, em_data in em_data_dict.items()
        }
        em_df = pd.DataFrame.from_dict(em_dict, orient="index")
        em_df.fillna("---", inplace=True)

        cut_names = {
            c: get_cut_flow_table_col_name(
                index=c,
                cut_list=info_container.full_cut_list,
                selection_name=selection_name,
            )
            for c in em_df.columns
        }
        em_df.rename(columns=cut_names, inplace=True)
        em_df = em_df.T

        if multiplicity_info.column_name_dict is None:
            if len(em_df.columns) == 1 and all_key in em_df.columns:
                col_dict: Union[Dict[Union[str, int, float], str], Dict[int, str]] = {all_key: single_col_name}
                em_df.rename(columns={k: rf"{k}_{v}" for k, v in col_dict.items()}, inplace=True)
            else:
                em_df.rename(columns={c: rf"{i}_{c}" for i, c in enumerate(list(em_df.columns))}, inplace=True)
        else:
            col_dict = copy.copy(multiplicity_info.column_name_dict)
            col_dict.update({all_key: all_col_name})
            assert all(c in col_dict.keys() for c in em_df.columns), (em_df.columns, col_dict.keys())
            em_df.rename(columns={k: rf"{k}_{v}" for k, v in col_dict.items()}, inplace=True)

        df_to_split_latex_table(
            df=em_df,
            table_fractions=multiplicity_info.table_fractions,
            name_prefix=name_prefix,
            output_lookup_dict=output_dict,
            placeholder_element="---",
            remove_column_name_index=True,
            allow_smaller_sub_dicts=True,
            create_pdf=create_pdfs,
            float_format=lambda s: f"{s:0.2f}" if isinstance(s, float) else "! --- !",
        )

    return output_dict
