"""
Steering file for the n-tuple file generation for the B -> D(*) tau nu analysis.
Felix Metzner 2018
"""

import basf2

from typing import List

from modularAnalysis import setAnalysisConfigParams, statistics

from analysistools.processing import job_interface
from analysistools.processing import path_utilities
from analysistools.processing import skim_cuts_utils

from rdstar.online_analysis import b2bii_utils
from rdstar.online_analysis.decay_utilities import Reconstructor
from rdstar.online_analysis import decay_utilities, skim_cuts, decay_channels


cmd_args = job_interface.get_commandline_args()  # type: job_interface.BatchJobCMDLineArgs

path_utilities.setup_environment_vars()
path_utilities.setup_databases(config=cmd_args.config)
path_utilities.fei_backward_compatibility()

main_path = basf2.create_path()  # type: basf2.Path

b2bii_utils.b2bii_read_and_convert(
    path=main_path,
    input_files=cmd_args.input_files,
    input_sequences=cmd_args.input_sequences,
)

setAnalysisConfigParams(
    configParametersAndValues={"mcMatchingVersion": "Belle"},
    path=main_path,
)

main_path.add_module(module="CurlerCloneMarker")

decay_utilities.fsp.fsp_collection.get_final_state_particles(
    path=main_path,
    use_mc_info=cmd_args.config["use_mc_info"],
)

skim_cuts_utils.apply_skim_from_dict(
    path=main_path,
    skim_dict=skim_cuts.pre_skim_cuts,
)

reconstructor = Reconstructor(
    path=main_path,
    config=cmd_args.config,
    decay_channels_module=decay_channels,
    do_gen_level_reco=True,
)

###########
# D meson #
###########
reconstructor.reconstruct_D()

general_vars = [
    "PDG",
    "matchedMC(PDG)",
    #
    "mcMother(PDG)",
    "mcMother(mcMother(PDG))",
    "mcMother(mcMother(mcMother(PDG)))",
    #
    "mcMother(mcDaughter(0, PDG))",
    "mcMother(mcDaughter(1, PDG))",
    #
    "mcMother(mcMother(mcDaughter(0, PDG)))",
    "mcMother(mcMother(mcDaughter(1, PDG)))",
    #
    "mcMother(mcMother(mcMother(mcDaughter(0, PDG))))",
    "mcMother(mcMother(mcMother(mcDaughter(1, PDG))))",
    #
    "mcDaughter(0, PDG)",
    "mcDaughter(1, PDG)",
    #
    "isSignal",
    "extraInfo(decayModeID)",
    "mcDaughter(0, PDG)",
    "mcDaughter(1, PDG)",
    "matchedMC(E)",
    #
    "dM",
    "abs(dM)",
    "M",
    "InvM",
    "abs(SigMBF)",
    "SigMBF",
    #
    "E",
    "Q",
    "dQ",
    "useCMSFrame(p)",
    "useCMSFrame(pt)",
    "p",
    "pt",
    "px",
    "py",
    "pz",
    #
    "isContinuumEvent",
    #
    "useCMSFrame(daughter(0, p))",
    "useCMSFrame(daughter(1, p))",
    "useCMSFrame(daughter(2, p))",
    "useCMSFrame(daughter(3, p))",
    "useCMSFrame(daughter(4, p))",
    "useRestFrame(daughter(0, p))",
    "useRestFrame(daughter(1, p))",
    "useRestFrame(daughter(2, p))",
    "useRestFrame(daughter(3, p))",
    "useRestFrame(daughter(4, p))",
    "daughter(0, InvM)",
    "daughter(1, InvM)",
    "daughter(2, InvM)",
    "daughter(3, InvM)",
    "daughter(4, InvM)",
    #
    "daughter(0, trackCurlerCloneIndicator)",
    "daughter(1, trackCurlerCloneIndicator)",
    "daughter(2, trackCurlerCloneIndicator)",
    "daughter(3, trackCurlerCloneIndicator)",
    "daughter(4, trackCurlerCloneIndicator)",
]  # type: List[str]

main_path.add_module(
    module="VariablesToNtuple",
    fileName=cmd_args.output_dict["dzero_ntuple"],
    particleList="D0:all",
    variables=general_vars,
)

main_path.add_module(
    module="VariablesToNtuple",
    fileName=cmd_args.output_dict["dcharged_ntuple"],
    particleList="D+:all",
    variables=general_vars,
)

main_path.add_module(
    module="VariablesToNtuple",
    fileName=cmd_args.output_dict["dzero_gen_ntuple"],
    particleList="D0:all_gen",
    variables=general_vars,
)

main_path.add_module(
    module="VariablesToNtuple",
    fileName=cmd_args.output_dict["dcharged_gen_ntuple"],
    particleList="D+:all_gen",
    variables=general_vars,
)

############
# D* meson #
############
reconstructor.reconstruct_Dstar()

dst_vars = [
    "daughterMotherDiffOf(0, M)",
    "daughterMotherDiffOf(0, InvM)",
    "daughterInvariantMass(0, 1)",
    "abs(massDifferenceSignificance(0))",
    "daughter(1, p)",
    "daughter(1, InvM)",
    "daughter(1, SigMBF)",
    "daughter(1, M)",
    "daughter(1, E)",
    "useCMSFrame(daughter(1, p))",
    "useRestFrame(daughter(1, p))",
    "daughter(0, p)",
    "daughter(0, InvM)",
    "daughter(0, SigMBF)",
    "daughter(0, M)",
    "daughter(0, E)",
    "useCMSFrame(daughter(0, p))",
    "useRestFrame(daughter(0, p))",
    "useLabFrame(daughterAngleInBetween(0, 1))",
    "daughter(1, trackCurlerCloneIndicator)",
]  # type: List[str]

main_path.add_module(
    module="VariablesToNtuple",
    fileName=cmd_args.output_dict["dstzero_ntuple"],
    particleList="D*0:all",
    variables=general_vars + dst_vars,
)

main_path.add_module(
    module="VariablesToNtuple",
    fileName=cmd_args.output_dict["dstcharged_ntuple"],
    particleList="D*+:all",
    variables=general_vars + dst_vars,
)

main_path.add_module(
    module="VariablesToNtuple",
    fileName=cmd_args.output_dict["dstzero_gen_ntuple"],
    particleList="D*0:all_gen",
    variables=general_vars + dst_vars,
)

main_path.add_module(
    module="VariablesToNtuple",
    fileName=cmd_args.output_dict["dstcharged_gen_ntuple"],
    particleList="D*+:all_gen",
    variables=general_vars + dst_vars,
)

#############
# D** meson #
#############
#
# reconstructor.reconstruct_D2star()
#
#
# main_path.add_module(
#     "VariablesToNtuple",
#     fileName=cmd_args.output_dict["donezero_ntuple"],
#     particleList="D_10:all",
#     variables=general_vars + dst_vars,
# )
#
# main_path.add_module(
#     "VariablesToNtuple",
#     fileName=cmd_args.output_dict["donecharged_ntuple"],
#     particleList="D_1+:all",
#     variables=general_vars + dst_vars,
# )
#
#
# main_path.add_module(
#     "VariablesToNtuple",
#     fileName=cmd_args.output_dict["dtwostzero_ntuple"],
#     particleList="D_2*0:all",
#     variables=general_vars + dst_vars,
# )
#
# main_path.add_module(
#     "VariablesToNtuple",
#     fileName=cmd_args.output_dict["dtwostcharged_ntuple"],
#     particleList="D_2*+:all",
#     variables=general_vars + dst_vars,
# )
#
#
# main_path.add_module(
#     "VariablesToNtuple",
#     fileName=cmd_args.output_dict["dprimeonezero_ntuple"],
#     particleList="D"_10:all",
#     variables=general_vars + dst_vars,
# )
#
# main_path.add_module(
#     "VariablesToNtuple",
#     fileName=cmd_args.output_dict["dprimeonecharged_ntuple"],
#     particleList="D"_1+:all",
#     variables=general_vars + dst_vars,
# )
#
#
# main_path.add_module(
#     "VariablesToNtuple",
#     fileName=cmd_args.output_dict["dzerostzero_ntuple"],
#     particleList="D_0*0:all",
#     variables=general_vars + dst_vars,
# )
#
# main_path.add_module(
#     "VariablesToNtuple",
#     fileName=cmd_args.output_dict["dzerostcharged_ntuple"],
#     particleList="D_0*+:all",
#     variables=general_vars + dst_vars,
# )
#


basf2.print_path(path=main_path)
basf2.process(path=main_path)
print(statistics)
