"""
Provides observable lists for low level studies.

Patrick Ecker, Felix Metzner 2021
"""

import copy
from typing import List


__all__ = [
    "LowLevelStudyObservables",
]


class LowLevelStudyObservables:
    basic_vars = [
        "PDG",
        "expNum",
        "runNum",
        "evtNum",
        "nTracks",
        "Ecms",
        "beamE",
        "beamPx",
        "beamPy",
        "beamPz",
        "nCleanedTracks(dr < 2 and abs(dz) < 4)",
        "nCleanedTracks(dr < 2 and abs(dz) < 4 and pt > 0.2)",
        "nCleanedECLClusters(goodBelleGamma == 1 and clusterTrackMatch == 0)",
        "nECLClusters",
        "belleECLEnergy",
        "totalEnergyOfParticlesInList(gamma:fsp)",
        "nKLMClusters",
        "KLMEnergy",
    ]  # type: List[str]

    momentum_variables = ["p", "pt", "px", "py", "pz", "theta", "cosTheta", "phi"]  # type: List[str]

    energy_variables = ["E", "Q", "dQ"]  # type: List[str]

    mass_variables = ["M", "InvM", "dM", "abs(dM)", "SigMBF", "abs(SigMBF)"]  # type: List[str]

    impact_param_variables = ["dr", "dz", "abs(dz)"]  # type: List[str]

    track_variables = ["pValue", "nCDCHits", "nSVDHits", "trackCurlerCloneIndicator"]  # type: List[str]

    fsp_mc_vars = [
        "isSignal",
        "matchedMC(PDG)",
        "mcMother(PDG)",
        "mcMother(mcMother(PDG))",
        "mcMother(mcMother(mcMother(PDG)))",
        "mcMother(mcMother(mcMother(mcMother(PDG))))",
        "mcMother(mcDaughter(0, PDG))",
        "mcMother(mcMother(mcDaughter(0, PDG)))",
        "productionIdentifier",
        "isContinuumEvent",
    ]  # type: List[str]

    cluster_variables = [
        "goodBelleGamma",
        "clusterE1E9",
        "clusterE9E21",
        "clusterE9E25",
        "clusterNHits",
        "clusterHighestE",
        "E",
    ]  # type: List[str]

    particle_id_variables = [
        "eIDBelle",
        "muIDBelle",
        "muIDBelleQuality",
        "atcPIDBelle(2,3)",
        "atcPIDBelle(3,2)",
        "extraInfo(ksnbStandard)",
        "extraInfo(goodKs)",
    ]  # type: List[str]

    def __init__(self, use_mc_info: bool = True) -> None:
        self._use_mc_info = use_mc_info  # type: bool

    @staticmethod
    def get_for_daughter(daughter_index: int, variable_list: List[str]) -> List[str]:
        assert 0 <= daughter_index <= 20, daughter_index
        return [f"daughter({daughter_index}, {var})" for var in variable_list]

    @staticmethod
    def get_in_cms_frame(variable_list: List[str]) -> List[str]:
        return [f"useCMSFrame({var})" for var in variable_list]

    @staticmethod
    def get_in_rest_frame(variable_list: List[str]) -> List[str]:
        return [f"useRestFrame({var})" for var in variable_list]

    @staticmethod
    def get_mc_matched(variable_list: List[str]) -> List[str]:
        return [f"matchedMC({var})" for var in variable_list]

    @staticmethod
    def get_for_mc_daughter(daughter_index: int, variable_list: List[str]) -> List[str]:
        assert 0 <= daughter_index <= 20, daughter_index
        return [f"mcDaughter({daughter_index}, {var})" for var in variable_list]

    def get_neutral_pion_observables(self) -> List[str]:
        pi0_var_list = [
            "daughterInvariantMass(0, 1)",
            "daughterAngle(0, 1)",
            "useRestFrame(daughterAngle(0, 1))",
            "daughterAngleInBetween(0, 1)",
            "useRestFrame(daughterAngleInBetween(0, 1))",
            "daughterSumOf(E)",
            "daughterDiffOf(0, 1, E)",
            "formula(daughterDiffOf(0, 1, E) / daughterSumOf(E))",
            "daughterDiffOf(0, 1, phi)",
            "useCMSFrame(daughterDiffOf(0, 1, phi))",
            "daughterDiffOf(0, 1, theta)",
            "useCMSFrame(daughterDiffOf(0, 1, theta))",
            "daughterDiffOf(0, 1, p)",
            "useCMSFrame(daughterDiffOf(0, 1, p))",
            "daughterDiffOf(0, 1, pt)",
            "useCMSFrame(daughterDiffOf(0, 1, pt))",
            "daughterDiffOf(0, 1, pz)",
            "useCMSFrame(daughterDiffOf(0, 1, pz))",
        ]  # type: List[str]

        pi0_var_list.extend(copy.copy(self.basic_vars))
        pi0_var_list.extend(copy.copy(self.momentum_variables))
        pi0_var_list.extend(self.get_in_cms_frame(variable_list=self.momentum_variables))
        pi0_var_list.extend(copy.copy(self.energy_variables))
        pi0_var_list.extend(copy.copy(self.mass_variables))
        pi0_var_list.extend(copy.copy(self.particle_id_variables))
        for d_index in [0, 1]:
            pi0_var_list.extend(self.get_for_daughter(daughter_index=d_index, variable_list=self.cluster_variables))

        if self._use_mc_info:
            pi0_var_list.extend(copy.copy(self.fsp_mc_vars))

        return sorted(list(set(pi0_var_list)))

    def get_charged_pion_observables(self) -> List[str]:
        pi_var_list = []  # type: List[str]

        pi_var_list.extend(copy.copy(self.basic_vars))
        pi_var_list.extend(copy.copy(self.momentum_variables))
        pi_var_list.extend(self.get_in_cms_frame(variable_list=self.momentum_variables))
        pi_var_list.extend(copy.copy(self.track_variables))
        pi_var_list.extend(copy.copy(self.impact_param_variables))
        pi_var_list.extend(copy.copy(self.energy_variables))
        pi_var_list.extend(copy.copy(self.mass_variables))
        pi_var_list.extend(copy.copy(self.particle_id_variables))
        if self._use_mc_info:
            pi_var_list.extend(copy.copy(self.fsp_mc_vars))

        return sorted(list(set(pi_var_list)))

    def get_lepton_observables(self, include_fsr_vars: bool = True) -> List[str]:
        lepton_var_list = []  # type: List[str]

        fsr_correction_validation_vars = [
            "daughterInvariantMass(0, 1)",  # for fsr corrected e
            "extraInfo(fsrCorrectedE)",  # for fsr corrected e
            "extraInfo(fsrCorrectedPx)",  # for fsr corrected e
            "extraInfo(fsrCorrectedPy)",  # for fsr corrected e
            "extraInfo(fsrCorrectedPz)",  # for fsr corrected e
            "extraInfo(fsrCorrectedP)",  # for fsr corrected e
            "extraInfo(fsrGammaE)",  # for fsr corrected e
            "extraInfo(fsrGammaPx)",  # for fsr corrected e
            "extraInfo(fsrGammaPy)",  # for fsr corrected e
            "extraInfo(fsrGammaPz)",  # for fsr corrected e
            "extraInfo(fsrGammaP)",  # for fsr corrected e
            "extraInfo(fsrCorrected)",  # for fsr corrected e
            "extraInfo(hasFSRCorrection)",  # for fsr corrected e
        ]  # type: List[str]

        if include_fsr_vars:
            # for fsr corrected e:
            lepton_var_list.extend(fsr_correction_validation_vars)

        track_vars = copy.copy(self.track_variables) + copy.copy(self.impact_param_variables)  # type: List[str]
        lepton_var_list.extend(track_vars)

        if include_fsr_vars:
            # for fsr corrected e:
            lepton_var_list.extend(self.get_for_daughter(daughter_index=0, variable_list=track_vars))

        lepton_var_list.extend(copy.copy(self.basic_vars))
        lepton_var_list.extend(copy.copy(self.momentum_variables))
        lepton_var_list.extend(self.get_in_cms_frame(variable_list=self.momentum_variables))
        lepton_var_list.extend(copy.copy(self.energy_variables))
        lepton_var_list.extend(copy.copy(self.mass_variables))
        lepton_var_list.extend(copy.copy(self.particle_id_variables))

        if self._use_mc_info:
            lepton_var_list.extend(copy.copy(self.fsp_mc_vars))

            if include_fsr_vars:
                # for fsr corrected e:
                additional_vars_for_mc_match = copy.copy(self.momentum_variables)
                additional_vars_for_mc_match.extend(copy.copy(self.impact_param_variables))
                additional_vars_for_mc_match.append("E")
                mc_match_vars = self.get_mc_matched(variable_list=additional_vars_for_mc_match)
                lepton_var_list.extend(mc_match_vars)
                lepton_var_list.extend(self.get_for_daughter(daughter_index=0, variable_list=mc_match_vars))

        return sorted(list(set(lepton_var_list)))

    def get_charged_kaon_observables(self) -> List[str]:
        kaon_var_list = []  # type: List[str]

        kaon_var_list.extend(copy.copy(self.basic_vars))
        kaon_var_list.extend(copy.copy(self.track_variables))
        kaon_var_list.extend(copy.copy(self.impact_param_variables))
        kaon_var_list.extend(copy.copy(self.momentum_variables))
        kaon_var_list.extend(self.get_in_cms_frame(variable_list=self.momentum_variables))
        kaon_var_list.extend(copy.copy(self.energy_variables))
        kaon_var_list.extend(copy.copy(self.mass_variables))
        kaon_var_list.extend(copy.copy(self.particle_id_variables))

        if self._use_mc_info:
            kaon_var_list.extend(copy.copy(self.fsp_mc_vars))

        return sorted(list(set(kaon_var_list)))

    def get_kaon_short_observables(self) -> List[str]:
        ks_var_list = []  # type: List[str]

        ks_var_list.extend(copy.copy(self.basic_vars))
        ks_var_list.extend(copy.copy(self.momentum_variables))
        ks_var_list.extend(self.get_in_cms_frame(variable_list=self.momentum_variables))
        ks_var_list.extend(copy.copy(self.impact_param_variables))
        ks_var_list.extend(copy.copy(self.energy_variables))
        ks_var_list.extend(copy.copy(self.mass_variables))
        ks_var_list.extend(copy.copy(self.particle_id_variables))

        if self._use_mc_info:
            ks_var_list.extend(copy.copy(self.fsp_mc_vars))

        return sorted(list(set(ks_var_list)))

    def get_gamma_observables(self) -> List[str]:
        gamma_var_list = [
            "extraInfo(hasFSRCorrection)"  # for fsr corrected e
        ]  # type: List[str]

        gamma_var_list.extend(copy.copy(self.basic_vars))
        gamma_var_list.extend(copy.copy(self.cluster_variables))
        gamma_var_list.extend(copy.copy(self.energy_variables))
        gamma_var_list.extend(copy.copy(self.mass_variables))
        gamma_var_list.extend(copy.copy(self.momentum_variables))
        gamma_var_list.extend(self.get_in_cms_frame(variable_list=self.momentum_variables))
        gamma_var_list.extend(copy.copy(self.impact_param_variables))

        if self._use_mc_info:
            gamma_var_list.extend(copy.copy(self.fsp_mc_vars))
            # for fsr corrected e:
            additional_vars_for_mc_match = copy.copy(self.momentum_variables)
            additional_vars_for_mc_match.append("E")
            mc_match_vars = self.get_mc_matched(variable_list=additional_vars_for_mc_match)
            gamma_var_list.extend(mc_match_vars)

        return sorted(list(set(gamma_var_list)))

    def get_d_meson_observables(self) -> List[str]:

        d_meson_var_list = [
            "extraInfo(decayModeID)",
            "useCMSFrame(daughter(0, p))",
            "useCMSFrame(daughter(1, p))",
            "useCMSFrame(daughter(2, p))",
            "useCMSFrame(daughter(3, p))",
            "useCMSFrame(daughter(4, p))",
            "useRestFrame(daughter(0, p))",
            "useRestFrame(daughter(1, p))",
            "useRestFrame(daughter(2, p))",
            "useRestFrame(daughter(3, p))",
            "useRestFrame(daughter(4, p))",
            "daughter(0, InvM)",
            "daughter(1, InvM)",
            "daughter(2, InvM)",
            "daughter(3, InvM)",
            "daughter(4, InvM)",
            "daughter(0, trackCurlerCloneIndicator)",
            "daughter(1, trackCurlerCloneIndicator)",
            "daughter(2, trackCurlerCloneIndicator)",
            "daughter(3, trackCurlerCloneIndicator)",
            "daughter(4, trackCurlerCloneIndicator)",
        ]  # type: List[str]

        d_meson_var_list.extend(copy.copy(self.basic_vars))
        d_meson_var_list.extend(copy.copy(self.momentum_variables))
        d_meson_var_list.extend(self.get_in_cms_frame(variable_list=self.momentum_variables))
        d_meson_var_list.extend(copy.copy(self.energy_variables))
        d_meson_var_list.extend(copy.copy(self.mass_variables))

        if self._use_mc_info:
            d_meson_var_list.extend(copy.copy(self.fsp_mc_vars))
            extra_mc_vars = [
                "isContinuumEvent",
                "matchedMC(E)",
                "mcMother(mcDaughter(1, PDG))",
                "mcMother(mcMother(mcDaughter(1, PDG)))",
                "mcMother(mcMother(mcMother(mcDaughter(0, PDG))))",
                "mcMother(mcMother(mcMother(mcDaughter(1, PDG))))",
            ]  # type: List[str]

            for d_index in range(7):
                d_meson_var_list.extend(self.get_for_mc_daughter(daughter_index=d_index, variable_list=["PDG"]))
                d_meson_var_list.extend(self.get_for_daughter(daughter_index=d_index, variable_list=["matchedMC(PDG)"]))
            d_meson_var_list.extend(extra_mc_vars)

        return sorted(list(set(d_meson_var_list)))

    def get_d_star_meson_observables(self) -> List[str]:
        d_star_meson_var_list = [
            "daughter(0, extraInfo(decayModeID))",
            "daughterMotherDiffOf(0, M)",
            "daughterMotherDiffOf(0, InvM)",
            "daughterInvariantMass(0, 1)",
            "abs(massDifferenceSignificance(0))",
            "useLabFrame(daughterAngleInBetween(0, 1))",
            "daughter(1, trackCurlerCloneIndicator)",
            "useCMSFrame(daughter(1, p))",
            "useCMSFrame(daughter(0, p))",
            "useRestFrame(daughter(1, p))",
            "useRestFrame(daughter(0, p))",
        ]  # type: List[str]

        d_daughter_vars = [
            "p",
            "E",
        ]  # type: List[str]
        d_daughter_vars.extend(copy.copy(self.mass_variables))

        for d_index in range(2):
            d_star_meson_var_list.extend(self.get_for_daughter(daughter_index=d_index, variable_list=d_daughter_vars))

        d_star_meson_var_list.extend(
            self.get_for_daughter(
                daughter_index=1,
                variable_list=[v for v in self.get_neutral_pion_observables() if v not in self.basic_vars],
            )
        )

        d_star_meson_var_list.extend(self.get_d_meson_observables())

        return sorted(list(set(d_star_meson_var_list)))

    @staticmethod
    def print_var_list(var_list: List[str]) -> None:
        print("\n\t- " + "\n\t- ".join(var_list) + "\n\n")


if __name__ == "__main__":
    print("Testing Low Level Study Observables:")

    vars_w_mc = LowLevelStudyObservables(use_mc_info=True)
    vars_wo_mc = LowLevelStudyObservables(use_mc_info=False)

    print("Lepton Vars with MC Info and FSR Info")
    LowLevelStudyObservables.print_var_list(var_list=vars_w_mc.get_lepton_observables(include_fsr_vars=True))

    print("Lepton Vars with out MC Info and without FSR Info")
    LowLevelStudyObservables.print_var_list(var_list=vars_wo_mc.get_lepton_observables(include_fsr_vars=False))

    print("Charged Pion Vars without MC Info")
    LowLevelStudyObservables.print_var_list(var_list=vars_wo_mc.get_charged_pion_observables())

    print("Neutral Pion Vars without MC Info")
    LowLevelStudyObservables.print_var_list(var_list=vars_wo_mc.get_neutral_pion_observables())

    print("Charged Kaon Vars without MC Info")
    LowLevelStudyObservables.print_var_list(var_list=vars_wo_mc.get_charged_kaon_observables())

    print("KShort Vars without MC Info")
    LowLevelStudyObservables.print_var_list(var_list=vars_wo_mc.get_kaon_short_observables())

    print("Gamma Vars without MC Info")
    LowLevelStudyObservables.print_var_list(var_list=vars_wo_mc.get_gamma_observables())

    print("D Meson Vars without MC Info")
    LowLevelStudyObservables.print_var_list(var_list=vars_wo_mc.get_d_meson_observables())

    print("D* Meson Vars without MC Info")
    LowLevelStudyObservables.print_var_list(var_list=vars_wo_mc.get_d_star_meson_observables())
